//
//  Project   : Navier-Stokes
//  File      : convdiff.cc
//  Author    : Ronald Kriemann
//  Purpose   : geophys example
//
// arch-tag: f7f64cee-0ed0-4ac2-9757-bea17fba074d
//

extern "C" {
#include <stdio.h>
#include <stdlib.h>
}

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/timer.hh"
#include "misc/cmdline.hh"
#include "misc/TRandom.hh"
#include "misc/matlab.hh"

#include "net/T3DFrontier.hh"
#include "net/T3DCellEdge.hh"
#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"
#include "net/TDomain.hh"
#include "net/T3DMeshGenerator.hh"

#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "discr/TConvDiffHierarchy.hh"
#include "discr/TCdrEqn.hh"

#include "matrix/blockvector.hh"
#include "matrix/TCycRedMatrixBlock.hh"

#include "solver/secondNF.hh"
#include "solver/fastGaussSeidel.hh"
#include "solver/bicgStab.hh"
#include "solver/cgSolver.hh"
#include "solver/TSOR.hh"
#include "solver/TCyclicReduction.hh"

#include "graph/T3DGraph.hh"
#include "graph/T3DPermBuilder.hh"

#include "visual/drawapp.hh"
#include "visual/matrixview.hh"

class TZero : public TFunction<3>
{
public:
    TZero () {}
    
    TFloat eval ( TVector <3> * )
    {
        return 0;
    }
};

class TMyCdrEqn : public TCdrEqn<3>
{
public:
    TMyCdrEqn ( TFloat aEpsilon = 1.0, 
                TBlockVector<TFloat> *aConvVector = NULL )
            : TCdrEqn<3>( aEpsilon, aConvVector )
    {}

    void evalDiff ( TVector <3> * v, TMatrix<3> * D ) const
    {
        TFloat  z = (*v)[Z];
        
        if (z >= -5.5)
        {
            (*D)[0][0]=10; (*D)[0][1]=0;  (*D)[0][2]=0;
            (*D)[1][0]=0;  (*D)[1][1]=10; (*D)[1][2]=0;
            (*D)[2][0]=0;  (*D)[2][1]=0;  (*D)[2][2]=10;
        }// if
        else if (z >= -25)
        {
            (*D)[0][0]=100; (*D)[0][1]=0;   (*D)[0][2]=0;
            (*D)[1][0]=0;   (*D)[1][1]=100; (*D)[1][2]=0;
            (*D)[2][0]=0;   (*D)[2][1]=0;   (*D)[2][2]=100;
        }// else if
        else
        {
            (*D)[0][0]=1; (*D)[0][1]=0; (*D)[0][2]=0;
            (*D)[1][0]=0; (*D)[1][1]=1; (*D)[1][2]=0;
            (*D)[2][0]=0; (*D)[2][1]=0; (*D)[2][2]=1;
        }// else
    }
            
    TFloat evalReac ( TVector <3> * ) const
    {
        return 0;
    }
   
    void evalConv ( TVector<3> *, TVector<3> * b ) const
    {
        (*b)[X] = 0;
        (*b)[Y] = 0;
        (*b)[Z] = 0;
    }

    TFloat evalRhs ( TVector <3> * v, uint i = 0 ) const
    {
#define EPS  1
        
        TFloat x, y, z;

        x = (*v)[X];
        y = (*v)[Y];
        z = (*v)[Z];

        if ((z >= -EPS) && (z <= EPS) &&
            (y >= -EPS) && (y <= EPS))
        {
            if ((x >= (-7.0-EPS)) &&
                (x <= (-7.0+EPS)))
                return -1;
            else if ((x >= (7.0-EPS)) &&
                     (x <= (7.0+EPS)))
                return 1;
            else
                return 0;
        }// if
        else
            return 0;
    }
};

class TMyDomain : public TDomain
{
protected:
    // direct access to all boundaries
    TBoundary   * arrBoundary[6];

    // direct access to min and max points
    TFloat  minX, minY, minZ;
    TFloat  maxX, maxY, maxZ;
    
public:
    TMyDomain ( const TVector<3> & min,
                const TVector<3> & max )
            : TDomain ()
    {
        setMin( (TVector<3>*) & min );
        setMax( (TVector<3>*) & max );

        minX = min[X]; minY = min[Y]; minZ = min[Z];
        maxX = max[X]; maxY = max[Y]; maxZ = max[Z];

        arrBoundary[0] = new TBoundary( "b0" );
        arrBoundary[1] = new TBoundary( "b1" );
        arrBoundary[2] = new TBoundary( "b2" );
        arrBoundary[3] = new TBoundary( "b3" );
        arrBoundary[4] = new TBoundary( "b4" );
        arrBoundary[5] = new TBoundary( "b5" );

        for (uint i = 0; i < 6; i++ )
            addBoundary( arrBoundary[i] );
    }

    TBoundary * boundary ( TFloat x, TFloat y, TFloat z ) const
    {
        if (z == minZ)
            return arrBoundary[0];
        else if (z == maxZ)
            return arrBoundary[2];
        else
        {
            if (x == minX)
                return arrBoundary[3];
            else if (x == maxX)
                return arrBoundary[1];
            else
            {
                if (y == minY)
                    return arrBoundary[4];
                else if (y == maxY)
                    return arrBoundary[5];
            }// else
        }// else

        return INTERIOR;
    }
};

//
// mainfunction
//
int
main ( int argc, char **argv ) 
{
    uint    i;
    TTimer  timer;

    //
    // create domain and define net-points
    //
    pGlobalDomain = new TMyDomain( TVector<3>( -5500, -5500, -5500 ),
                                   TVector<3>( 5500, 5500, 0 ) );

/*
    TFloat  arrCoordX[73] = { -5500, -4250, -3250, -2500, -2000, -1500, -1200, -900,
                              -700, -550, -425, -325, -250, -200, -150, -120, -90,
                              -70, -55, -42.5, -32.5, -25, -20, -15, -9, -7, -5.5,
                              -4.25, -3.25, -2.5, -2, -1.5, -1.2, -1, -0.8, 0, 0.8,
                              1, 1.2, 1.5, 2, 2.5, 3.25, 4.25, 5.5, 7, 9, 12, 15, 20,
                              25, 32.5, 42.5, 55, 70, 90, 120, 150, 200, 250, 325,
                              425, 550, 700, 900, 1200, 1500, 2000, 2500, 3250, 4250,
                              5500 };
    TFloat  arrCoordZ[39] = { -5500, -4250, -3250, -2500, -2000, -1500, -1200, -900,
                              -700, -550, -425, -325, -250, -200, -150, -120, -90,
                              -70, -55, -42.5, -32.5, -25, -20, -15, -9, -7, -5.5,
                              -4.25, -3.25, -2.5, -2, -1.5, -1.25, -1.0,
                              -0.75, -0.5, -0.2, 0 };
*/
    TFloat  arrCoordX[] = { -5500, -3250, -2000, -1200,
                             -700, -425, -250, -150, -90,
                             -70, -42.5, -25, -15, -7, -5.5,
                             -4.25, -2.5, -1.5, -1, 0,
                             1, 1.5, 2.5, 4.25, 5.5, 7, 15,
                             25, 42.5, 70, 90, 150, 250,
                             425, 700, 1200, 2000, 3250, 5500 };
    TFloat  arrCoordZ[] = { -5500, -3250, -2000, -1200,
                            -700, -425, -250, -150, -90,
                            -70, -42.5, -25, -15, -7, -5.5,
                            -4.25, -2.5, -1.5, -1.0,
                            -0.75, -0.2, 0 };
                            
    //
    // parse command line
    //
    param.convDirName = "circle";
    param.noOfNets    = 4;
    param.criterion   = 3;
    param.xi          = 0.2;
    param.epsilon     = 1.0;
    param.steps       = 1000;
    param.setStartTri( "grids/cube.tri" );

    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl << endl;

    //
    // init net and discr functions
    //
    T3DMeshGenerator  * pMeshGen = new T3DMeshGenerator;
    TNet<3>           * pNet;

    cout << "main (...) : used memory           = " << TMemory::used() << endl;
    timer.start();
    pNet = pMeshGen->buildNet( sizeof(arrCoordX)/sizeof(TFloat), arrCoordX,
                               sizeof(arrCoordX)/sizeof(TFloat), arrCoordX,
                               sizeof(arrCoordZ)/sizeof(TFloat), arrCoordZ,
                               pGlobalDomain );
    timer.stop();
    cout << "main (...) : time for building net = " << timer << endl;
    cout << "main (...) : used memory           = " << TMemory::used() << endl;
    
    delete pMeshGen;
    cout << "main (...) : used memory           = " << TMemory::used() << endl;

    //
    // init discr functions
    //
    TCdrEqn<3>          * pCdrEqn;
    TFunction<3>        * pZero;

    pCdrEqn = new TMyCdrEqn;

    pCdrEqn->isLaplace(  FALSE );
    pCdrEqn->Diffusion(  TRUE  );
    pCdrEqn->Convection( FALSE  );
    pCdrEqn->setEpsilon( param.epsilon );
  
    pZero = new TZero;

    //
    // supply boundary Descriptons
    //
    TListIterator< TBoundary >  * pIterBound;
    TCdrBoundaryDesc<3>         * pBoundDesc[6];
    TBoundary                   * pBoundary;

    for ( i = 0; i < 5; i++ )
    {
        pBoundDesc[i] = new TCdrBoundaryDesc<3>( 1 );
        pBoundDesc[i]->setType( 0, DIR_TYPE );
        pBoundDesc[i]->setFunc( 0, pZero );
        pBoundDesc[i]->setDoFInfo();
    }// for

    pBoundDesc[5] = new TCdrBoundaryDesc<3>( 1 );
    pBoundDesc[5]->setType( 0, NEU_TYPE );
    pBoundDesc[5]->setFunc( 0, pZero );
    pBoundDesc[5]->setDoFInfo();
    
    pIterBound = new TListIterator< TBoundary >;
    pIterBound->setList( pGlobalDomain->getBoundaries() );
    
    pIterBound->first();
    while ((pBoundary = (*pIterBound)++) != NULL)
    {
        if (pBoundary->getName() == "b0" )
            pBoundary->setBoundaryDesc( pBoundDesc[0] );
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( pBoundDesc[1] );
        if (pBoundary->getName() == "b2" )
            pBoundary->setBoundaryDesc( pBoundDesc[2] );
        if (pBoundary->getName() == "b3" )
            pBoundary->setBoundaryDesc( pBoundDesc[3] );
        if (pBoundary->getName() == "b4" )
            pBoundary->setBoundaryDesc( pBoundDesc[4] );
        if (pBoundary->getName() == "b5" )
            pBoundary->setBoundaryDesc( pBoundDesc[5] );
    }// while

    //
    // discretize
    //
    TConvDiffDoFManager<3>  * DoFManager;
    TBlockMatrix            * pA;
    ulong                     dim;
    TBlockVector<TFloat>    * pRHS;
    TBlockVector<TFloat>    * pX;
    TConvDiffFE<3>          * discretiser;
    
    DoFManager  = new TConvDiffDoFManager<3>();
    discretiser = new TConvDiffFE<3>( Tabata );
    pRHS        = new TBlockVector<TFloat>;
    pX          = new TBlockVector<TFloat>;

    cout << "main (...) : initDofDesc" << endl;

    DoFManager->initDoFDesc( pNet );
        
    cout << "main (...) : number of DoFs        = " << DoFManager->getNoOfDoFs() << endl;
    cout << "main (...) : build stiffnessmatrix" << endl;

    pA  = DoFManager->buildStiffnessMatrix( pNet, pNet );
    dim = (*pA)[0][0].getNoOfRows();
    
    cout << "main (...) : stiffnessmatrix built" << endl;

    pRHS->setNoOfBlocks( 1 );
    (*pRHS)[0].setSize( dim );
    
    pX->setNoOfBlocks( 1 );
    (*pX)[0].setSize( dim );
    
    (*pRHS) = 0;
    
    cout << "main (...) : used memory           = " << TMemory::used() << " kB" << endl;

    //
    // Diskretisierung
    //
    timer.start();
    discretiser->discretize( pCdrEqn, pNet, pA,  pRHS );
    timer.stop();
    cout << "main (...) : time for discretizing = " << timer << endl << endl;
    
    if (param.draw)
    {
//        matWriteMatrixBlock( "files/matrixblock.mat",
//                             (*pA)[0][0], (*pA)[0][0].getNoOfRows(), "A" );
        TDrawApp::drawMatrix( & (*pA)[0][0], param.xi, "matrixblock" );
    }// if

    //
    // build coarse net permutation
    //
    
    timer.start();
    
    TCycRedMatrixBlock*   pCycRedMat = new TCycRedMatrixBlock( (*pA)[0][0] );
    uint                  MSIZE=17;
    TCoarsener            coarser( param.xi, param.criterion, SEMI );
    uint                  maxLevel;
    
    maxLevel = pCycRedMat->decompose( coarser, 1, MSIZE,
                                      TRUE, 100, param.maxLevel );

    cout << "main (...) : maxlevel = " << maxLevel << endl;
    timer.stop();
    
    cout << "main (...) : time for decompostion = " << timer << endl;
    
    (*(pA))[0].change( 0, pCycRedMat );
    (*pRHS)[0].permute( pCycRedMat->getInvPerm() );
    
    //
    // solve with original matrix
    //
    TSecondNF        * pPrec;
    TSolver          * pSolver;
    TSolver          * pInnerSolver;
    TSolver          * pCoarseSolver;
    TCyclicReduction * pCycRedSolver;

//     TFloat      alt, neu, beta, omega;
    
//     (*pX) = 1;
    
//     pSolver = new TJacobi;

//     pSolver->setMode( 2 );
//     pSolver->setAccuracy( 1e-13 );
//     pSolver->setMaxit( 20 );
//     pSolver->setVerbosity( 0 );

//     pSolver->solve( pA, pRHS, pX );
//     alt = pX->euclidNorm();
    
//     pSolver->solve( pA, pRHS, pX );
//     neu = pX->euclidNorm();

//     beta = exp( 0.05 * log( neu/alt ) );
//     omega = 2.0 / (1.0 + sqrt( 1.0 - (beta * beta)));

//     cout << "main (...) : omega = " << omega << endl;
    
    pPrec = new TSymmetricGaussSeidel;
    pPrec->setMode( 0 );
    pPrec->setMaxit( 1 );
    pPrec->setVerbosity( 0 );
//     ((TSOR*)pPrec)->setOmega( 1.5 );
//     ((TSOR*)pPrec)->setMethod( SOR_SYMMETRIC );

    pInnerSolver = new TFastSymmetricGaussSeidel;
    pInnerSolver->setMode( MODE_REDUCEBY );
    pInnerSolver->setAccuracy( 1e-10 );
    pInnerSolver->setMaxit( 20 );
    pInnerSolver->setVerbosity( 0 );
//     ((TSOR*)pInnerSolver)->setOmega( 1.5 );
//     ((TSOR*)pInnerSolver)->setMethod( SOR_SYMMETRIC );

    pCoarseSolver = new TCGSolver( pPrec );
    pCoarseSolver->setMode( MODE_REDUCEBY );
    pCoarseSolver->setAccuracy( 1e-10 );
    pCoarseSolver->setMaxit( 40 );
    pCoarseSolver->setVerbosity( 0 );

    pCycRedSolver = new TCyclicReduction( pA, pRHS, pX,
                                          param.startLevel, 
                                          maxLevel, 
                                          param.cycle,
                                          pInnerSolver,
                                          pCoarseSolver, 
                                          param.steps );
    pCycRedSolver->setMaxit( 1 );
    pCycRedSolver->setAccuracy( 1e-13 );
    pCycRedSolver->setVerbosity( 0 );
    
    pSolver = new TCGSolver( pCycRedSolver );
//    pSolver = new TSOR;
    pSolver->setMode( 0 );
    pSolver->setAccuracy( 1e-13 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 0xff );
//    ((TSOR*)pSolver)->setOmega( omega );
//    ((TSOR*)pSolver)->setMethod( SOR_SYMMETRIC );

    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "main (...) : time for solving = " << timer << endl;

    //
    // clean up
    //
    
    delete pCdrEqn;
    delete discretiser;
    delete pRHS;
    delete pX;
    delete pA;
    delete DoFManager;
    delete pGlobalDomain;

    delete pPrec;
    delete pSolver;
  
    TMemory::statistics();
}
