//
//  Project  : Navier-Stokes
//  File     : convDiff.cc
//  Author   : Ronald Kriemann / Sabine Gutsch Thomas Probst
//  Purpose  : test for discretization and multigrid solution
//             of a convection-diffusion problem; 
//             comparison with exact solution
//
// arch-tag: db1914a0-73eb-410c-ae2c-edac08495e95
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/fastGaussSeidel.hh"
#include "solver/TGMRES.hh"
#include "solver/TSOR.hh"

#include "discr/convDiffDoFManager.hh"
#include "discr/convDiffFE.hh"
#include "discr/convDiffHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/graphtool.hh"
#include "graph/rcm.hh"
#include "graph/fvs.hh"

// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y
class TMyConvDiffEqn : public TBasisEqn<2>
{
public:
    TMyConvDiffEqn () 
            : TBasisEqn<2> () {}
    
    virtual TFloat f0 (const TVector<2> & v )
    {
        return ( -4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] ) ); 
    }
    virtual TFloat dirichlet0 ( const TVector<2> & v )
    { 
        return ( v[0] * v[0] + v[1] * v[1] );
    }
};

int 
main( int argc, char **argv ) 
{
    TTimer  timer;
    uint    i = 0;
   
    param.steps      = 1000;
    param.noOfNets   = 6;
    param.startLevel = 1;
    param.ordering   = CONCENTRIC_MIN;
    param.xi         = 1.0/3.0;
    param.mgSort     = MGSORT_INDEP; 
    param.criterion  = 3;
    param.epsilon    = 1e-3;
    param.draw       = FALSE;
    param.convDirName = "circle";
    param.solver     = "bicg";
    strcpy(param.starttri, "grids/quad1.tri" );

    param.evalCmdLine( argc, argv );

    param.endLevel = param.noOfNets-1;
    cout << "chosen parameter: " << endl << param << endl;

    //
    // build and refine net
    //

    TNetHierarchy<2> * pNetHier = new TNetHierarchy<2>( param.starttri );
    TNet<2>          * pNet;

    timer.start();
    
    for (i = 0; i < param.noOfNets-1; i++)
        pNet = pNetHier->buildNet();
    
    timer.stop();
    cout << "main (...) : time to refine = " << timer << endl;
    cout << "main (...) : used memory = " << TMemory::used() << " kB" << endl;
    
    //
    // build stiffness matrix sparsity structure
    //

    TBlockMatrix          * pA;
    TBlockVector<TFloat>  * pRHS;
    TBlockVector<TFloat>  * pX;
    ulong                   dim;
    T2DConfDiffDoFManager * pDoFManager;
    TMyConvDiffEqn        * pEqn;
    T2DConvDiffFE         * pFE;
    
    pDoFManager = new T2DConfDiffDoFManager();
    pEqn        = new TMyConvDiffEqn();

    // set epsilon
    pEqn->setEpsilon( param.epsilon );
    
    // set reaction term
    if (param.reactionTerm)
        pEqn->setReacFunction( oneReac ); 
    else 
        pEqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (param.convDirName == "none")
    {
        pEqn->setConvFunctions( noConv0, noConv1 );
    }// if
    else if (param.convDirName == "xline")
    {
        pEqn->setConvFunctions( xline0, xline1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/10.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convDirName == "curve")
    {
        pEqn->setConvFunctions( curve0, curve1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convDirName == "circle")
    {
        pEqn->setConvFunctions( circle0, circle1 ); 
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convDirName == "square") 
    {
        pEqn->setConvFunctions( square0, square1 );    
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convDirName == "foursquares") 
    {
        pEqn->setConvFunctions( foursquares0, foursquares1 ); 
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convDirName == "4circles") 
    {
        pEqn->setConvFunctions( fourcircles0, fourcircles1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else
    {
        cout << "main (...) : convdir unsupported" << endl;
        exit( 1 );
    }// else
    
    //
    // Diskretisierung
    //
    pFE  = new T2DConvDiffFE( Tabata );
    pRHS = new TBlockVector<TFloat>;
    pX   = new TBlockVector<TFloat>;

    pA  = pDoFManager->buildStiffnessMatrix( pNet, pNet );
    dim = (*pA)[0][0].getNoOfRows();
    cout << "main (...) : numer of DoF's = " << dim << endl;
    
    pRHS->setNoOfBlocks( 1 );
    (*pRHS)[0].setSize( dim );
    
    pX->setNoOfBlocks( 1 );
    (*pX)[0].setSize( dim );
    
    (*pRHS) = 0;
    
    timer.start();
    
    pFE->discrDiff( pEqn, pNet, pA, pRHS );
    pFE->discrConv( pEqn, pNet, &(*pA)[0][0], pRHS );
    if ( pEqn->getReacc() !=  noReac)
    {
        TBlockVector<TFloat> * pM = new TBlockVector<TFloat>( 0.0, *pX );
        // lumped mass matrix in vector pM
        pFE->discrMass( pEqn, pNet, &(*pM)[0] ); 
        for ( i = 0; i < dim; i++)
            (*pA)[0][0][i][i] += (*pM)[0][i];
        delete pM;
    }// if

    timer.stop();
    cout << "main (...) : time for disretization = " << timer << endl;
    cout << "main (...) : used memory = " << TMemory::used() << " kB" << endl;

    timer.start();
    
    ///////////////////////////////////////////////////
    //
    // build graph from matrix and work with it
    //

    TDynArray< TVector<2> * > * arrVectors = new TDynArray< TVector<2> * >;
    pFE->getCoord( pNet, *arrVectors );

    if (param.draw && (verbosity >= VERBOSE_RESULT))
        TDrawApp::draw( *pNet, "net" );

    timer.start();
    TBlockInfo  * pBlockinfo = new TBlockInfo;
    
    pBlockinfo->setSortMethod( SORT_OFD );

    //
    // build data structures for each net
    //
    TVectorBlock<ulong>  * vecPerm;
    TVectorBlock<ulong>  * cyclicBlocks;
    ulong                  fvsSize;
 
    if ( param.ordering != NONE )
    {
        vecPerm      = new TVectorBlock<ulong>;
        cyclicBlocks = new TVectorBlock<ulong>;
    }// if

    //
    // now build permutation vectors
    //
    timer.start();
    if ( param.ordering & CONCENTRIC )
    {
        pBlockinfo->buildCYCPerm( * vecPerm, * cyclicBlocks,
                                  * arrVectors,
                                  (*pA)[0][0], 
                                  param.xi, 
                                  param.criterion,
                                  param.ordering );
        cout << "main (...) : number of cycles = " << cyclicBlocks->getSize()/2 << endl;
    }// if
    
    if (( param.ordering == HEURISTIC_FVS ) ||
        ( param.ordering == PLANAR_FVS ))
    {
        fvsSize = 0;
        pBlockinfo->buildFVSPerm( * vecPerm,
                                  fvsSize,
                                  * arrVectors,
                                  (*pA)[0][0], 
                                  param.xi,
                                  param.criterion,
                                  SEARCH_DFS,
                                  param.ordering );
        cout << "main (...) : number of fvs-nodes = " << fvsSize << endl;
    }// if

    if ( param.ordering == RCM )
        pBlockinfo->buildRCMPerm( * vecPerm, (*pA)[0][0] );

    timer.stop();
    cout << "main (...) : time to build permutation = " << timer << endl;
    
    //
    // build the inverse permutations
    //
    TVectorBlock<ulong> * vecInvPerm;
    
    if ( param.ordering != NONE )
    {
        vecInvPerm = new TVectorBlock<ulong>( 0, vecPerm->getSize() );
        invertPerm( (*vecInvPerm), *vecPerm );
    }// if

    // reorder
    if (param.ordering != NONE )
    {
        timer.start();
        (*pRHS)[0].permute( vecInvPerm );
        (*pA)[0][0].permute( vecInvPerm, vecInvPerm );
        timer.stop();
        cout << "main (...) : time for permute = "<< timer << endl;
    }// if

    // draw matrix
    if (param.draw)
        TDrawApp::drawMatrix( pA, param.xi, "pmatrix");

    /////////////////////////////////////
    //
    // solver
    //

    TSolver * pPrec;
    TSolver * pSolver;
    
    if (( param.ordering == HEURISTIC_FVS ) || ( param.ordering == PLANAR_FVS ))
        pPrec = new TBGS;
    else 
        pPrec = new TSGS;

    if (param.solver == "bicg")        pSolver = new TBicgStab( pPrec );
    else if (param.solver == "cg")     pSolver = new TCGSolver( pPrec );
    else if (param.solver == "gmres")  pSolver = new TGMRES( pPrec );
    else if (param.solver == "sgs")    pSolver = new TSGS;
    else if (param.solver == "bgs")    pSolver = new TBGS;
    else if (param.solver == "fgs")    pSolver = new TFGS;
    else
    {
        cout << " main : unsupported solver (" << param.solver << ")" << endl;
        exit( 1 );
    }// else
        
    pPrec->setAccuracy( 1e-10 );
    pPrec->setMaxit( param.steps );
    pPrec->setVerbosity( 0 );

    pSolver->setAccuracy( 1e-10 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 0xff );
    
    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "main (...) : time to solve = " << timer << endl;

    //
    // compare with exact solution (in function dirichlet1)
    //
    TBlockVector<TFloat> *exact = new TBlockVector<TFloat>( *pRHS );
    
    for ( i = 0; i < (*exact)[0].getSize(); i++) 
        (*exact)[0][i] = pEqn->dirichlet0( *(*arrVectors)[i] );
    
    if ( param.ordering != NONE )
        (*exact)[0].permute( vecInvPerm );

    TBlockVector<TFloat> *error = new TBlockVector<TFloat>( *exact );
    *error -= *pX;
    
    cout << "main (...) : error = " << error->weightedEuclidNorm() << endl;

    /////////////////////////////////////
    //
    // clean up
    //

    pNetHier->free();
    
    delete pDoFManager;
    delete pEqn;
    delete pFE;

    delete pX;
    delete pRHS;
    delete pA;

    delete arrVectors;

    delete pPrec;
    delete pSolver;
    
    delete exact;
    delete error;

    delete pBlockinfo;

    if ( param.ordering != NONE )
    {
        delete vecInvPerm;
        delete vecPerm;
        delete cyclicBlocks;
    }// if
}
