//
//  Project   : Navier-Stokes
//  File      : stokes.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D conv-diff
//
// arch-tag: b78acaa2-80e6-442e-bc76-b3d5dd8b6415
//

extern "C" {
#include <stdio.h>
#include <stdlib.h>
}

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/timer.hh"
#include "misc/cmdline.hh"
#include "misc/TRandom.hh"

#include "net/T3DFrontier.hh"
#include "net/T3DCellEdge.hh"
#include "net/net.hh"
#include "net/TStokesBoundaryDesc.hh"
#include "net/TDomain.hh"

#include "discr/TStokesDoFManager.hh"
#include "discr/TStokesFE.hh"
#include "discr/flowdef.hh"

#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/schur.hh"
#include "solver/bicgStab.hh"

#include "graph/T3DGraph.hh"
#include "graph/T3DPermBuilder.hh"

#include "visual/drawapp.hh"

void
disturbNet ( TNet<3> * pNet )
{
    uint                       i, j;
    TListIterator< TCell<3> >  iterCells( pNet->getList() );
    TCell<3>                  *pCell;
    T3DFrontier               *pFrontier;
    T3DCellEdge               *pEdge;
    TFloat                     length, delta;
    uint                       noOfFrontiers, noOfEdges;
    TVector<3>                 v0, v1;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFrontiers = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = (T3DFrontier*) pCell->getFrontier( i );

            noOfEdges = pFrontier->getNoOfEdges();
            for (j = 0; j < noOfEdges; j++)
            {
                pEdge = pFrontier->getEdge( j );

                v0 = * pEdge->getNode( 0 );
                v1 = * pEdge->getNode( 2 );

                length = (v1 - v0).euclidNorm();

                if (length == 0)
                    continue;

                delta = length / 150;

                v0[X] += TRandom::random(-delta, delta);
                v0[Y] += TRandom::random(-delta, delta);
                v0[Z] += TRandom::random(-delta, delta);

                v1[X] += TRandom::random(-delta, delta);
                v1[Y] += TRandom::random(-delta, delta);
                v1[Z] += TRandom::random(-delta, delta);

                if (pEdge->getNode(0)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(0))[X] = v0[X];
                    (*pEdge->getNode(0))[Y] = v0[Y];
                    (*pEdge->getNode(0))[Z] = v0[Z];
                }// if

                if (pEdge->getNode(2)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(2))[X] = v1[X];
                    (*pEdge->getNode(2))[Y] = v1[Y];
                    (*pEdge->getNode(2))[Z] = v1[Z];
                }// if
            }// for
        }// for
    }// while
}

class TDirichletFun0 : public TFunction<3>
{
public:
    TFloat eval(TVector<3> *v )
	{ 
        return ( sin((*v)[0]) * sin((*v)[1]) );
	}
};

class TDirichletFun1 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return  ( cos((*v)[0]) * cos((*v)[1]) );
	}
};

class TDirichletFun2 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return (0.0);
	}
};

class TPressure : public TFunction<3>
{
public:
    TFloat eval ( TVector <3> *v )
	{
        return (2.0 * cos((*v)[0]) * sin((*v)[1]) -
                2.0 * sin(1.0) * (1.0 - cos(1.0)));
	}
};

//
// mainfunction
//
int
main ( int argc, char **argv ) 
{
    uint    i;
    TTimer  timer;

    //
    // parse command line
    //
    param.convDirName = "xline";
    param.noOfNets    = 4;
    param.criterion   = 7;
    param.xi          = 1.0;
    param.steps       = 40;
    param.epsilon      = 1.0;
    param.setStartTri( "grids/cube.tri" );

    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl << endl;

    //
    // init net and discr functions
    //
    TNetHierarchy<3>  * pHierarchy = new TNetHierarchy<3>( param.starttri );
    TNet<3>           * pCNet, * pFNet;

    //
    // init discr functions
    //
    TStokesBasisEqn<3>  * pStokesEqn;
    TFunction<3>        * pDirichlet[3];
    TFunction<3>        * pPressure;

    if (param.convDirName == "none")
    {
        pStokesEqn = new TNoConvStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.2, 0.1, 1.0/3.0, "flow" );
    }// if
    if (param.convDirName == "xline")
    {
        pStokesEqn = new TXLineStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.2, 0.1, 1.0/3.0, "flow" );
    }// if
    if (param.convDirName == "curve")
    {
        pStokesEqn = new TCurveStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 1.0/3.0, 0.1, 0.1, "flow" );
    }// if
    if (param.convDirName == "circle")
    {
        pStokesEqn = new TCircleStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 1.0/3.0, 0.1, "flow" );
    }// if
    if (param.convDirName == "4circles")
    {
        pStokesEqn = new T4CirclesStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 1.0/3.0, 0.1, "flow" );
    }// if
    if (param.convDirName == "diagcircle")
    {
        pStokesEqn = new TDiagCircleStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    if (param.convDirName == "ball")
    {
        pStokesEqn = new TBallStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 0.1, 0.1, "flow" );
    }// if

//     pStokesEqn->isLaplace(  TRUE );
//     pStokesEqn->Diffusion(  TRUE  );
//     pStokesEqn->Convection( TRUE  );
    pStokesEqn->setEpsilon( param.epsilon );
  
    pDirichlet[0] = new TDirichletFun0();
    pDirichlet[1] = new TDirichletFun1();
    pDirichlet[2] = new TDirichletFun2();
    pPressure     = new TPressure();

    //
    // supply boundary Descriptons
    //
    TListIterator< TBoundary >  * pIterBound;
    TStokesBoundaryDesc<3>      * pBoundDesc;
    TBoundary                   * pBoundary;
    
    pBoundDesc = new TStokesBoundaryDesc<3>( 4 );

    for ( i = 0; i < 3; i++ )
    {
        pBoundDesc->setType( i, DIR_TYPE);
        pBoundDesc->setFunc( i, pDirichlet[i] );
        pBoundDesc->setDoFInfo( i, FALSE);
    }// for
    
    pBoundDesc->setType( 3, NEU_TYPE);
    pBoundDesc->setDoFInfo( 3, TRUE);

    pIterBound = new TListIterator< TBoundary >;
    pIterBound->setList( pDomain->getBoundaries() );
    
    pIterBound->first();
    while ((pBoundary = (*pIterBound)++) != NULL)
    {
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( pBoundDesc );
    }// while

    //
    // discretize
    //
    TStokesDoFManager<3>    * DoFManager;
    TBlockMatrix            * pA;
    ulong                     dim, noOfBlocks;
    TBlockVector<TFloat>    * pRHS;
    TBlockVector<TFloat>    * pX;
    TBlockVector<TFloat>    * pexact;
    TBlockVector<TFloat>    * presidual;
    TStokesFE<3>            * discretiser;
    
    DoFManager  = new TStokesDoFManager<3>();
    discretiser = new TStokesFE<3>( Tabata );
    pRHS        = new TBlockVector<TFloat>;
    pX          = new TBlockVector<TFloat>;
    pexact      = new TBlockVector<TFloat>;
    presidual   = new TBlockVector<TFloat>;

    timer.start();
    
    pCNet = pFNet = NULL;
    for (i = 1; i < param.noOfNets; i++)
    {
        pCNet = pFNet;
        pFNet = pHierarchy->buildNet();
    }// for
    
    timer.stop();
    cout << "main (...) : time for refining     = " << timer << endl;
    
    cout << "main (...) : used memory           = " << TMemory::used() << endl;

    if (param.disturbNet)
        disturbNet( pFNet );
    
    pA = DoFManager->buildStiffnessMatrix( pCNet, pFNet );
    
    dim        = 0;
    noOfBlocks = pA->getNoOfRows();

    pRHS->setNoOfBlocks( noOfBlocks );
    pX->setNoOfBlocks( noOfBlocks );
    pexact->setNoOfBlocks( noOfBlocks );
    presidual->setNoOfBlocks( noOfBlocks );
    
    for ( i = 0; i < noOfBlocks - 1; i++ )
    {
        dim += (*pA)[i][i].getNoOfRows();
        (*pRHS)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*pX)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*pexact)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*presidual)[i].setSize( (*pA)[i][i].getNoOfRows() );
    }// for
    (*pRHS)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*pX)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*pexact)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*presidual)[i].setSize( (*pA)[i][0].getNoOfRows() );
    
    (*pRHS) = 0;
    

    //
    // Diskretisierung
    //
    timer.start();
    discretiser->discrDiff( pStokesEqn, pCNet, pA, pRHS);
    for ( uint j = 0; j < 3; j++)
	{
        TMatrixBlock *pC = new TMatrixBlock( (*pA)[j][j] );
        discretiser->discrConv( pStokesEqn, pCNet, pC, pRHS, j);
        (*pA)[j][j] = *pC;
        delete (pC);
	}
    discretiser->discrRhs(  pStokesEqn, pCNet, pRHS);
    timer.stop();
    cout << "main (...) : time for discretizing = " << timer << endl << endl;
    
    //cout << "rhs\n" << (*pRHS)*(*pRHS) << endl;
    //cout << "A0:\n" << (*pA)[0][0] << endl;

    if (param.draw)
        TDrawApp::drawMatrix( pA, 0.25, FALSE, "matrix" );

    cout << endl;
    
    if ((verbosity > VERBOSE_RESULT) && (param.draw))
        TDrawApp::drawNet( pFNet, "net", VRML );

    // test discretisation:
    ulong u_dim = (*pA)[0][0].getNoOfRows();
    ulong p_dim = (*pA)[3][0].getNoOfRows();
    TDynArray<TVector<3> * > arrCoord( u_dim );
    TDynArray<TVector<3> * > parrCoord( p_dim );
    discretiser->getCoord ( pCNet, 0, arrCoord, parrCoord );
    for (ulong j = 0; j < u_dim; j++)
	{
       (*pexact)[0][j] = pDirichlet[0]->eval( arrCoord[j] );
       (*pexact)[1][j] = pDirichlet[1]->eval( arrCoord[j] );
       (*pexact)[2][j] = pDirichlet[2]->eval( arrCoord[j] );
    }
    for (ulong j = 0; j < p_dim; j++)
       (*pexact)[3][j] = pPressure->eval( parrCoord[j] );

    (*presidual) = (*pA) * (*pexact) - (*pRHS);
    cout << "Residual of exact solution: " << setprecision(5) << presidual->weightedEuclidNorm() << endl;;

    //
    // solve with original matrix
    //
    TSecondNF * pPrec;
    TSolver   * pSolver = new TSchur;
    //TSolver   * pSolver = new TBicgStab;

    pPrec = new TSchur;
    
    pPrec->setMode( 0 );
    pPrec->setMaxit( 1 );
    pPrec->setVerbosity( 0 );
    ((TSchur*)pPrec)->setMaxInnerSteps( 20 );
    ((TSchur*)pPrec)->setInnerReduceBy( 1.0e-1 );

    pSolver->setMode( 1 );
    pSolver->setAccuracy( 1e-4 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 5 );
    //((TBicgStab*)pSolver)->setPrec( pPrec );
    ((TSchur*)pSolver)->setMaxInnerSteps( 10 );
    ((TSchur*)pSolver)->setInnerReduceBy( 1.0e-1 );

    (*pX) = 1;

	/*    
    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "main (...) : time for solving = " << timer << endl << endl;
	*/	

    //
    // build permutation
    //
    T3DPermBuilder * pPermBuilder = new T3DPermBuilder;
    TPermutation   * pPerm, * pInvPerm;
    
    timer.start();

    pPerm = NULL;
    
    if (param.ordering == HEURISTIC_FVS)
		// global hfvs
        //pPerm = pPermBuilder->buildFVSPerm( pFNet, & (*pA)[0][0] );
		// hfvs on flow surfaces
	    pPerm = pPermBuilder->buildFVSPerm( pFNet, pStokesEqn, & (*pA)[0][0] );

    if (param.ordering == RCM)
        pPerm = pPermBuilder->buildRCMPerm( & (*pA)[0][0] );

    if (param.ordering == CONCENTRIC_MIN)
        pPerm = pPermBuilder->buildCYCPerm( pFNet, pStokesEqn, & (*pA)[0][0] );

    if (pPerm == NULL)
        exit( 1 );
    
    timer.stop();
    cout << "main (...) : time for building permutation = " 
         << timer << endl << endl;
    
    if (pPerm == NULL)
    {
        cout << "main (...) : have NULL permutation" << endl;
        exit( 1 );
    }// if

    pInvPerm = new TPermutation( pPerm->invert() );
    
    //
    // permute matrix
    //

    pA->permute( pInvPerm );
    
    pRHS->permute( pInvPerm );
    
    if (param.draw)
        TDrawApp::drawMatrix( pA, 0.25, FALSE, "pmatrix" );
    
    //
    // solve with permuted matrix
    //
    (*pX) = 1;
    
    pA->deleteLUdecomp();
    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "main (...) : time for solving = " << timer << endl << endl;
    
    //
    // clean up
    //
    
    delete pPerm;
    delete pInvPerm;
    
    pHierarchy->free();
    delete pStokesEqn;
    delete discretiser;
    delete pRHS;
    delete pX;
    delete pA;
    delete DoFManager;
    delete pDomain;
    delete pPermBuilder;

    delete pPrec;
    delete pSolver;
  
    TMemory::statistics();
}
