//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D grids
//
// arch-tag: b70247f9-33f5-42a4-bb2d-6c8f28186d38
//

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/TFunction.hh"

#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"
#include "net/TDomain.hh"

#include "discr/TCdrEqn.hh"
#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "discr/TConvDiffHierarchy.hh"
#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/fastGaussSeidel.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/TMultigrid.hh"

#include "visual/drawapp.hh"
#include "visual/TPOVRay.hh"

class TDirichletFun : public TFunction<3>{
public:
    TDirichletFun(){};
    ~TDirichletFun(){};
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; 
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
        //  return a+b+2*c;
        return a*a+b*b+c*c;
    }
   
};
class TNeumannFun : public TFunction<3>{
public:
    TNeumannFun(){};
    ~TNeumannFun(){};
    TFloat eval(TVector <3> *av ){
        TFloat eps=1e-8;
        TVector <3> *v; v=av; 
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
        // x== 0  => -d/dx = -2x=0
        if (fabs(a) <eps)
            return 0;
        // x== 1  => d/dx = 2x=2
        if (fabs(a-1) <eps)
            return 2;
        
        if (fabs(c) <eps)
            return 0;
        if (fabs(c-1) <eps)
            return 2;
        return 0;
    }
   
};
class TZeroFun : public TFunction<3>{
public:
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 0.0;
    }
   
};
class TOneFun : public TFunction<3>{
public:
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 1.0;
    }
   
};
class TTwoFun : public TFunction<3>{
public:
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 2.0;
    }
   
};

class TMyBasisCdrEqn : public TBasisCdrEqn<3>
{
public:
    TMyBasisCdrEqn( TFloat aEpsilon = 1.0, 
                    TBlockVector<TFloat> *aConvVector = NULL)
            : TBasisCdrEqn<3>(aEpsilon,aConvVector)
    {}

    void evalDiff(TVector <3> *v, TMatrix<3>  *D)
    {
        TVector<3> * av = v;
        
        (*D)[0][0]=1;(*D)[0][1]=0; (*D)[0][2]=0;
        (*D)[1][0]=0;(*D)[1][1]=1; (*D)[1][2]=0;
        (*D)[2][0]=0;(*D)[2][1]=0; (*D)[2][2]=1;
    }

    void evalConv(TVector <3> *v, TVector<3>  *B)
    {
        TVector<3> * av = v;
        
        (*B)[0]=1;(*B)[1]=2; (*B)[2]=3;
    }


    TFloat evalReac(TVector <3> *v) const
    {
        TVector<3> * av = v;
        
        return 0.0;
    }
   
    TFloat evalRhs(TVector <3> *v ) const
    {
        TVector<3> * av = v;
        
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
     
        return -6+(2*a+4*b+6*c);
    }   
};

class TMyCdrEqn : public TCdrEqn<3>
{
public:
    TMyCdrEqn( TFloat aEpsilon = 1.0, 
               TBlockVector<TFloat> *aConvVector = NULL)
            : TCdrEqn<3>(aEpsilon,aConvVector)
    {}
};


//
// mainfunction
//
int
main( int argc, char **argv ) 
{
    uint maxNoOfNets=10;
    uint mf=2;
    TDrawApp *app = new TDrawApp;
    

    TMyBasisCdrEqn*  cdreqn = new TMyBasisCdrEqn(1.0,NULL);
 
    TDirichletFun* Dirichlet = new TDirichletFun;
    TNeumannFun* Neumann = new TNeumannFun;
    TZeroFun* Zero = new TZeroFun;
    TOneFun* One = new TOneFun;
    TTwoFun* Two = new TTwoFun;
   
 
    cdreqn->isLaplace(FALSE);
    cdreqn->Diffusion(TRUE);
    cdreqn->Convection(FALSE);
    cdreqn->setEpsilon(1);
 
  
    // read initial net and supply boundary Descriptons
    TNet<3>  *net3D = new TNet<3>( "grids/cube_2bounds.tri" );

    // TNet<3>  *net3D = new TNet<3>( "grids/cube.tri" );
    TNet<3>  *dNet;

    // in "grids/cube.tri", only one TBoundary is defined
    // therefore, one Description is enough

    TCdrBoundaryDesc<3> * boundDescN = new TCdrBoundaryDesc<3>(1);
    boundDescN->setType(0,NEU_TYPE);
    boundDescN->setFunc(0,Neumann);
    boundDescN->setDoFInfo();

    TCdrBoundaryDesc<3> * boundDescD = new TCdrBoundaryDesc<3>(1);
    boundDescD->setType(0,DIR_TYPE);
    boundDescD->setFunc(0,Dirichlet);
    boundDescD->setDoFInfo();


    TListIterator< TBoundary > iter(pDomain->getBoundaries()  );


    TBoundary *bound=iter.first();

    while ((bound = iter++) != NULL)
    {
        cout<<bound->getName()<<endl;
        if(bound->getName()=="b1")
            bound->setBoundaryDesc(boundDescD);
        if(bound->getName()=="b2")
            bound->setBoundaryDesc(boundDescN);
    } 

    // bound->setBoundaryDesc(boundDescD);
 
    // build Hierarchy
    TNetHierarchy<3> *net3DHierarchy  = new TNetHierarchy<3>();
    net3DHierarchy->add(net3D);

    for (uint i=0;i<=mf;i++)
    {
        cout<<"i: "<<i<<endl;
        net3DHierarchy->buildNet();
        cout<<" Verfeinern fertig"<<endl;
    }

    // app.drawNet( net3D, "psfiles/cnet" );
    // app.drawNet( fnet, "psfiles/fnet" );
    // app.drawNet( ffnet, "psfiles/ffnet" );
    TBlockMatrix *aA; 
    TBlockVector<TFloat> *aRhs,*aX;
    TBlockMatrix **pMatrix     = new (TBlockMatrix*)[maxNoOfNets];
    TBlockVector<TFloat> **pRHS=new (TBlockVector<TFloat>*)[maxNoOfNets];
    TBlockVector<TFloat> **pX  =new (TBlockVector<TFloat>*)[maxNoOfNets];
    TProlongation<3> **pProlongation= new (TProlongation<3>*)[maxNoOfNets];

    for (uint i=0; i<maxNoOfNets;i++)
    {
        pMatrix[i]=NULL;
        pRHS[i]=NULL;
        pX[i]=NULL;
        pProlongation[i]=NULL;
    }


    TConvDiffDoFManager<3> *DoFManager= new TConvDiffDoFManager<3>();
    TConvDiffFE<3> *discretiser= new TConvDiffFE<3>(Tabata);
  
    TConvDiffHierarchy<3>  *cdrHierarchy= new TConvDiffHierarchy<3>();
    cdrHierarchy->buildHierarchy( DoFManager,cdreqn,discretiser,
                                  net3DHierarchy,pMatrix,pRHS,pX,
                                  pProlongation);


    // app->drawNet( dnet, "psfiles/ffcube", VRML );

    // no of this net
    TListIterator< TNet<3> >  *iterNet;
    iterNet = new TListIterator< TNet<3> >( net3DHierarchy->getList() );
    // last net
    dNet = iterNet->last();
    // no of this net
    uint lastnetNumber = net3DHierarchy->getNoOfNets()-1;
    // dim of the corresponding matrix
    ulong dim = (*(pMatrix[lastnetNumber]))[0][0].getSize();
 
    aA=pMatrix[lastnetNumber];
    aRhs=pRHS[lastnetNumber];
    aX=pX[lastnetNumber];
  
    // solve
    TFastSymmetricGaussSeidel *sgs = new TFastSymmetricGaussSeidel( );
    TSymmetricGaussSeidel *ssgs = new TSymmetricGaussSeidel( );
    TBicgStab *bicgStab = new TBicgStab(ssgs );
    TCGSolver *cg = new TCGSolver(ssgs );

    bicgStab->setAccuracy( 1e-15 );
    bicgStab->setMaxit( 400 );
    bicgStab->setVerbosity( 5);
    cg->setAccuracy( 1e-15 );
    cg->setMaxit( 40 );
    cg->setVerbosity( 0);
    aX->set( 1.0 ); 
    //bicgStab->solve(aA,aRhs,aX);

    uint fNet=2;
    uint cyc=1;
    uint steps=12;
    uint psm=2;  

    TMultigrid<3> *multigrid = new TMultigrid<3>(pMatrix, pRHS, pX, 
                                                 pProlongation,pProlongation ,
                                                 fNet,  lastnetNumber, 
                                                 cyc, sgs, sgs, cg,steps);

    //sgs->setMaxit( 20 ); 
    //sgs->setVerbosity( 7 );
    //sgs->setMode(0);
    //sgs->solve(aA,aRhs,aX);


    multigrid->setPreSmoothingSteps(1);
    multigrid->setPostSmoothingSteps(1);
    multigrid->setMode( 1 );
    multigrid->setVerbosity( 7 );

    sgs->setMaxit( psm ); 
    sgs->setVerbosity( 0 );
    sgs->setMode(2);

    ssgs->setMaxit( 1 ); 
    ssgs->setVerbosity( 0 );
    ssgs->setMode(2);
    
    multigrid->solve();


     
    // check accuracy
  
    TDynArray<TVector < 3 > * > arrCoord(dim);
    discretiser->getCoord ( dNet,arrCoord);
    cout<<"size: "<<arrCoord.getSize()<<endl;
  
    TBlockVector<TFloat> *perror= new TBlockVector<TFloat>(1,dim);
    TBlockVector<TFloat> *pexact= new TBlockVector<TFloat>(1,dim);
    
    for (ulong j=0; j<dim;j++)
    {
        (*pexact)[0][j]= Dirichlet->eval ((arrCoord[j]));
        //  cout<<" X: "<<(*pX)[0][j]<<" coord: "<<*(arrCoord[j])<<endl;
    }
  
    (*perror)=(*aX)-(*pexact);
    cout<<"Fehler Diff: "<<setprecision(5)<<perror->euclidNorm()/sqrt(dim)<<endl;
   
    (*perror)=(*(pX[lastnetNumber]))-(*pexact);
    cout<<"Fehler Diff: "<<setprecision(5)<<perror->euclidNorm()/sqrt(dim)<<endl;
   

    delete discretiser;
    delete aRhs;
    delete aX;
    delete aA;
    delete DoFManager;
    delete perror;
    delete pexact;

    TMemory::statistics();
}
 
