//
//      Project         : Navier-Stokes
//      File            : main.cc
//      Author          : Sabine Gutsch/ Ronald Kriemann / Thomas Probst
//      Purpose         : test for discretization and cyclic reduction  
//                        solution (Reusken)
//                        of a convection-diffusion problem; 
//                        comparison with exact solution
//      Date            : 
//      Changes         : 
//
// arch-tag: 76514e8f-162e-48a1-be7e-b40fe58bfd4b
//

#include <stdio.h>
#include <stdlib.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"
#include "misc/matlab.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/fastGaussSeidel.hh"
#include "solver/TCyclicReduction.hh"

#include "discr/convDiffDoFManager.hh"
#include "discr/convDiffFE.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"
#include "matrix/sort.hh"
#include "matrix/TCycRedMatrixBlock.hh"

#include "visual/drawapp.hh"

#include "graph/graphtool.hh"
#include "graph/criterion.hh"
#include "graph/TCoarsener.hh"

// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y + c *e

class TMyConvDiffEqn : public TBasisEqn<2> {

public:
    TMyConvDiffEqn ( TFloat aEpsilon = 0.000001 ) 
        : TBasisEqn<2> ( aEpsilon) { };
    virtual TFloat f0 (const TVector<2> &v) {
        return ( -4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] + c(v)*( v[0] * v[0] + v[1] * v[1])) ); 
    }
    virtual TFloat dirichlet0 ( const TVector<2> &v) { 
        return ( v[0] * v[0] + v[1] * v[1] );
    }
};

int 
main( int argc, char **argv ) 
{
    uint i = 0;
    //   bool changed=FALSE;
  
    TParameter  param;
    param.steps = 12;
    param.noOfNets = 5;
    param.startLevel = 1;
    param.cycle = 1;
    param.xi =0.7;
    param.criterion = E4;
    param.epsilon = 10e-9;
    param.draw =FALSE;
    param.MaxLevel=9;
    param.MaxNoOfBlacks=20;
    param.lumping=FALSE;


    param.convDirName = "xline";
    strcpy(param.starttri, "grids/start1.tri" );
  
    param.evalCmdLine( argc, argv);
    param.endLevel = param.noOfNets-1;
    cout << "chosen parameter: " << param << endl;

  
    //
    // build and refine net
    //
  
    TTimer timer;

    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
    netHier->delReference();
  
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
  
    ulong noOfInteriorNodesNet[param.noOfNets];
   
    T2DNetHierarchy::TNetListIterator *iterNet;
    T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
    iterNet->first();
    while ((pNet = (*iterNet)++) != NULL)
    {
        noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
        cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
    }// while
   
    T2DNet  *fNet = iterNet->last();
   
    //
    // build stiffness matrix sparsity structure
    //
   
    timer.start();
   
    TBlockMatrix *pMatrix;      
    TBlockVector<TFloat> *pRHS; 
    TBlockVector<TFloat> *pX;   
   
    T2DConfDiffDoFManager *dofManager = new T2DConfDiffDoFManager();
   
    TMyConvDiffEqn *eqn = new TMyConvDiffEqn( );
  
    eqn->setEpsilon(param.epsilon );
    if (param.reactionTerm)
        eqn->setReacFunction(oneReac); 
    else 
        eqn->setReacFunction(noReac); 


    // Pointer to functions for convection direction
    if (param.convDirName == "circle")
        eqn->setConvFunctions( circle0, circle1 ); 
    if (param.convDirName == "xline")
        eqn->setConvFunctions( xline0, xline1 ); 
    if (param.convDirName == "curve")
        eqn->setConvFunctions( curve0, curve1 ); 
    if (param.convDirName == "square") 
        eqn->setConvFunctions( square0, square1 );    
    if (param.convDirName == "foursquares") 
        eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (param.convDirName == "fourcircles") 
        eqn->setConvFunctions( fourcircles0, fourcircles1 ); 
   
 

    // discretisation of convection-diffusion-reaction eqn
    T2DConvDiffFE *fe = new T2DConvDiffFE( Tabata);
     
    pMatrix = dofManager->buildStiffnessMatrix(fNet);
    ulong dim= fNet->getNoOfInteriorNodes();
   
    pRHS = new TBlockVector<TFloat>( 0.0,dim);
    pX = new TBlockVector<TFloat>( 0.0, *pRHS );
    fe->discrDiff( eqn, fNet, pMatrix, pRHS);
    fe->discrConv( eqn, fNet,  &((*pMatrix)[0][0]), pRHS); 
    if ( eqn->getReacc() !=  noReac) { // reaction term != zero
        TBlockVector<TFloat> *pM = new TBlockVector<TFloat>( 0.0, *pX);
        fe->discrMass(eqn,fNet, &((*pM)[0])); // lumped mass matrix in vector pM
        for (ulong j=0; j<dim;j++)
            (*pMatrix)[0][0][j][j] += (*pM)[0][j];
        delete pM;
    }
        
    timer.stop();
    cout << "time for discretization : " <<timer << endl << endl;
  

    // read matrix (format: Stuttgart AMG test matrices)
    /*
      ulong size[1];
      size[0]=1;
      pMatrix= new TBlockMatrix(1,size);
      param.draw=FALSE;
      matReadMatrixBlock("/tmp/AMG/unsym.mat", (*pMatrix)[0][0]);
      ulong dim= (*pMatrix)[0][0].getNoOfRows();
     
      // write Matrix as matlab file
      matWriteMatrixBlock("/tmp/matrixf.mat", (*pMatrix)[0][0], dim,"Af");
     
      // an  artificial right hand side and start vector
      pRHS = new TBlockVector<TFloat>( 100.0,dim);
      pX = new TBlockVector<TFloat>( 0.0, *pRHS );
 
      */
    


    (*pMatrix)[0][0].setOutputPrecision(6);
    //  cout << (*pMatrix)[0][0]<<endl;

    ///////////////////////////////////////////////////
    //
    // lets draw a little bit
    //
    TDrawApp *app = new TDrawApp;
   
    ///////////////////////////////////////////////////
    //
    // build graph from matrix and work with it
    //
   
    TDynArray< TVector<2> * > *arrVectors = new TDynArray< TVector<2> * >;
    fe->getCoord( fNet, *arrVectors );

   
    //drawing?
  

    timer.start();
    TCycRedMatrixBlock* Mat;
    TMemory::statistics();
    Mat = new TCycRedMatrixBlock ((*pMatrix)[0][0]);
  
    uint MSIZE=10;
    TCoarsener coarser(param.xi,param.criterion,SEMI);
    // if(param.draw)
    // Mat->setGraphName("psfiles/ppgraph");
          uint Maxlevel=Mat->decompose(coarser,1,MSIZE,param.lumping,param.MaxNoOfBlacks,param.MaxLevel);
          //uint Maxlevel=4;
    cout<<"Maxlevel: "<< Maxlevel<<endl;
    timer.stop();
    cout << "time for decompostion : " <<timer << endl << endl;
  
    //if (param.draw)
    //      app->draw( *fNet, *(Mat->getInvPerm()), "psfiles/pnet.ps" ); 
  
    
    (*(pMatrix))[0].change(0, Mat);
    // permute RHS
    (*pRHS)[0].permute(Mat->getInvPerm());

   
    /////////////////////////////////////
    //
    // solver
    //
    TVectorBlock<TFloat> *convrate= new TVectorBlock<TFloat>(0,16);
    TFastForwardGaussSeidel *fgs = new TFastForwardGaussSeidel( );
    TFastBackwardGaussSeidel *bgs = new TFastBackwardGaussSeidel( );
    TFastSymmetricGaussSeidel *sgs = new TFastSymmetricGaussSeidel( );
    TBackwardGaussSeidel *sbgs = new TBackwardGaussSeidel( );
    TSymmetricGaussSeidel *ssgs = new TSymmetricGaussSeidel( );
    TBicgStab *bicgStab = new TBicgStab( ssgs );
   
    /*
      // write matrix and rhs as matlab file

      matWriteMatrixBlock("matrixf.mat", (*pMatrix)[0][0],
      (*pMatrix)[0][0].getNoOfRows(),"Af");
     
      matWriteVectorBlock("rhs.mat",(*pRHS)[0] ,"rhs" );
      */
    
  

    TCyclicReduction *cycred = new TCyclicReduction(pMatrix,
                                                    pRHS,pX, 
                                                    param.startLevel, 
                                                    Maxlevel, 
                                                    param.cycle, sgs, bicgStab, 
                                                    param.steps);
                                                   
                                                  
    //  TBicgStab *PbicgStab = new TBicgStab(cycred );
     TBicgStab *PbicgStab = new TBicgStab(cycred );
   
       
    fgs->setMaxit( 2 );
    bgs->setMaxit( 2 );
    sgs->setMaxit( 20 );  //1
    sbgs->setMaxit( 2 );
    ssgs->setMaxit( 1 );
    cycred->setMaxit( 1 );
    fgs->setVerbosity( 0 );
    bgs->setVerbosity( 0 );
    sbgs->setVerbosity( 0 );
    sgs->setVerbosity(0 );
    ssgs->setVerbosity( 0 );
    cycred->setVerbosity(0 );
    PbicgStab->setVerbosity(2 );
    bicgStab->setVerbosity(0 );

    fgs->setMode(2);
    bgs->setMode(2);
    sbgs->setMode(2);
    sgs->setMode(0);
    ssgs->setMode(2);
    cycred->setMode(0);
    
   
    /* bicgStab->setMaxit(param.steps );
       cout<<"bicgStab with symmetric gs as preconditioner "<<endl;;
       bicgStab->setVerbosity( 127 );
       bicgStab->setMode( 0 );
       pX->set( 100.0 );
       timer.start();
       // bicgStab->solve(pMatrix,pRHS,pX);
       timer.stop();
       cout << "time for bicgStab(sgs) with symmetric gs as preconditioner " << timer  << endl<< endl;
       */
   

  	
    PbicgStab->setMaxit(param.steps );
    // inner solver
    bicgStab->setVerbosity(0);
    bicgStab->setMaxit(30);
    bicgStab->setMode( 0 );
  
    // cycred->setMaxit(1);
    cout<<"PbicgStab with cycred"<<endl;;
    PbicgStab->setVerbosity(2 );
    PbicgStab->setMode( 0 );
    pX->set( 100.0 );
    timer.start();
    PbicgStab->solve(pMatrix,pRHS,pX);
    timer.stop();
    cout << "time for PbicgStab(cycred): " << timer  << endl<< endl;
   
   

    bicgStab->setMaxit( 40);
    bicgStab->setVerbosity(0);
    bicgStab->setAccuracy( 1e-15 );
    
    cout<<"standard cycred with symmetric gs"<<endl;
    bicgStab->setVerbosity( 0 ); //0
    bicgStab->setMode( 0 ); 
  

    // cyclic reduction as solver (not as preconditioner)  

    pX->set( 100.0 ); 
    cycred->setRHS(pRHS);
    cycred->setMaxit(param.steps);
    cycred->setVerbosity( 127 );
  
    timer.start();
    cout<<"cyred start solving "<< endl; 
    // cycred->solve(); 
    cout<<"cyred solved "<< endl;
    timer.stop();
    //  (*convrate)[0]=cycred->getConvRate(); 
    cout << "time for: cycred with symmetric gs "<< timer << endl<<endl;;
  

    //
    // compare with exact solution (in function dirichlet1)
    //
    TBlockVector<TFloat> *exact = new TBlockVector<TFloat>( *pRHS );
    for ( i = 0; i < (*exact)[0].getSize(); i++) 
        (*exact)[0][i] = eqn->dirichlet0( *(*arrVectors)[i] );
    (*exact)[0].permute(Mat->getInvPerm());
  

 

    // calcultion of the exact solution (of the system)
 
    TBlockVector<TFloat> *error = new TBlockVector<TFloat>( *exact );
  
    *error -= *pX;
    cout << "(Error: " 
         << error->weightedEuclidNorm() << ")" << endl;
  
    (*pX)[0].permute(Mat->getPerm());

    TSort sort(pMatrix,pRHS,pX,fNet);
    sort.sort(lexicographical);
 
    sort.order(error);
    sort.order(exact);

    matWriteVectorBlock("x.mat",(*error)[0] ,"sol" );
  
    //
    // clean up
    //

    delete iterNet; 
    delete netHier;
	
    delete dofManager;
    delete eqn;
    delete fe;
   

    if (pX ) delete pX ;
    if (pRHS ) delete pRHS ;
    if (pMatrix ) delete pMatrix ;

    delete    Mat;
 
    delete arrVectors;
    delete app;
  
    delete fgs;
    delete bgs;
    delete sgs;
    delete sbgs;
    delete ssgs;
    delete bicgStab;

    delete cycred;
  
    delete exact;
    delete error;

 
    delete convrate;
 
    TMemory::statistics();

  

}

