//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D grids
//
// arch-tag: 0013bc24-8d41-4a4e-a7d8-8fedb7cdb9c1
//

#include "misc/memory.hh"
#include "misc/TFunction.hh"
#include "misc/matlab.hh"

#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"

#include "discr/TCdrEqn.hh"
#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "discr/TConvDiffHierarchy.hh"
#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"

#include "visual/drawapp.hh"
#include "visual/TPOVRay.hh"

class TDirichletFun : public TFunction<3>{
public:
    TDirichletFun(){};
    ~TDirichletFun(){};
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; 
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
        //  return a+b+2*c;
        return a*a+b*b+c*c;
    }
   
};
class TZeroFun : public TFunction<3>{
public:
    TZeroFun(){};
    ~TZeroFun(){};
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 0.0;
    }
   
};
class TOneFun : public TFunction<3>{
public:
    TOneFun(){};
    ~TOneFun(){};
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 1.0;
    }
   
};
class TTwoFun : public TFunction<3>{
public:
    TTwoFun(){};
    ~TTwoFun(){};
    TFloat eval(TVector <3> *av ){
        TVector <3> *v; v=av; return 2.0;
    }
   
};

class TMyBasisCdrEqn : public TBasisCdrEqn<3>{

public:

    TMyBasisCdrEqn( TFloat aEpsilon = 1.0, 
                    TBlockVector<TFloat> *aConvVector = NULL):TBasisCdrEqn<3>(aEpsilon,aConvVector)
    {};

    virtual ~TMyBasisCdrEqn () { }
   
    void evalDiff(TVector <3> *v, TMatrix<3>  *D){
        TVector<3> * av; av=v;
        (*D)[0][0]=1;(*D)[0][1]=0; (*D)[0][2]=0;
        (*D)[1][0]=0;(*D)[1][1]=1; (*D)[1][2]=0;
        (*D)[2][0]=0;(*D)[2][1]=0; (*D)[2][2]=1;
    }

    void evalConv(TVector <3> *v, TVector<3>  *B){
        TVector<3> * av; av=v;
        (*B)[0]=1;(*B)[1]=2; (*B)[2]=3;
    }


    TFloat evalReac(TVector <3> *v) const{
        TVector<3> * av; av=v;
        return 0.0;
    }
   
    TFloat evalRhs(TVector <3> *v ) const{
        TVector<3> * av; av=v;
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
     
        return -6;
    }   
};

class TMyCdrEqn : public TCdrEqn<3> {
public:

    TMyCdrEqn( TFloat aEpsilon = 1.0, 
               TBlockVector<TFloat> *aConvVector = NULL): TCdrEqn<3>(aEpsilon,aConvVector)
    {};

    virtual ~TMyCdrEqn () { }
};


//
// mainfunction
//
int
main( int argc, char **argv ) 
{  uint maxNoOfNets=10;

TDrawApp *app = new TDrawApp;
   

TMyBasisCdrEqn*  cdreqn = new TMyBasisCdrEqn(1.0,NULL);
 
TDirichletFun* Dirichlet = new TDirichletFun;
TZeroFun* Zero = new TZeroFun;
TOneFun* One = new TOneFun;
TTwoFun* Two = new TTwoFun;
   
 
/*   cdreqn->setRhsFunction(One);
     cdreqn->setRhsFunction(Dirichlet);
     cdreqn->setDiffFunctions(One,0,0);
     cdreqn->setDiffFunctions(Zero,0,1);
     cdreqn->setDiffFunctions(Zero,0,2);
     cdreqn->setDiffFunctions(Two,1,1);
     cdreqn->setDiffFunctions(Zero,1,0);
     cdreqn->setDiffFunctions(Zero,1,2);

     cdreqn->setDiffFunctions(One,2,2);
     cdreqn->setDiffFunctions(Zero,2,1);
     cdreqn->setDiffFunctions(Zero,2,0);

     */

cdreqn->isLaplace(FALSE);
cdreqn->Diffusion(TRUE);
cdreqn->Convection(FALSE);
cdreqn->setEpsilon(1);
 
  
// read initial net and supply boundary Descriptons
TNet<3>  *net3D = new TNet<3>( "grids/cube.tri" );
TNet<3>  *dNet;

// in "grids/cube.tri", only one TBoundary is defined
// therefore, one Description is enough

TCdrBoundaryDesc<3> * boundDesc = new TCdrBoundaryDesc<3>(1);
boundDesc->setType(0,DIR_TYPE);
boundDesc->setFunc(0,Dirichlet);
boundDesc->setDoFInfo();

TNet<3>::TCellListIterator iter( net3D->getList() );
bool ready=FALSE;
   
// find node with boundary pointer
TCell<3> *cell = iter.first();
TBoundary *bound;
while ((cell != NULL)&&(ready==FALSE)){
    for (uint i=0;i<4;i++){
        bound =cell->getNode(i)->getBoundary();
        if (bound!=NULL){
            ready=TRUE; // exit loop
        }
    }
    cell = iter.next();
}
bound->setBoundaryDesc(boundDesc);
 
// build Hierarchy
TNetHierarchy<3> *net3DHierarchy  = new TNetHierarchy<3>();
net3DHierarchy->add(net3D);

uint mf=2;
for (uint i=0;i<=mf;i++) {
    cout<<"i: "<<i<<endl;
    net3DHierarchy->buildNet();
    cout<<" Verfeinern fertig"<<endl;
}

// app.drawNet( net3D, "psfiles/cnet" );
//app.drawNet( fnet, "psfiles/fnet" );
//app.drawNet( ffnet, "psfiles/ffnet" );
 TBlockMatrix *aA; 
TBlockVector<TFloat> *aRhs,*aX;
  TBlockMatrix **pMatrix     = new (TBlockMatrix*)[maxNoOfNets];
TBlockVector<TFloat> **pRHS=new (TBlockVector<TFloat>*)[maxNoOfNets];
TBlockVector<TFloat> **pX  =new (TBlockVector<TFloat>*)[maxNoOfNets];
TProlongation<3> **pProlongation= new (TProlongation<3>*)[maxNoOfNets];

for (uint i=0; i<maxNoOfNets;i++){
    pMatrix[i]=NULL;
    pRHS[i]=NULL;
    pX[i]=NULL;
    pProlongation[i]=NULL;
}


TConvDiffDoFManager<3> *DoFManager= new TConvDiffDoFManager<3>();
TConvDiffFE<3> *discretiser= new TConvDiffFE<3>(Tabata);
  
TConvDiffHierarchy<3>  *cdrHierarchy= new TConvDiffHierarchy<3>();
//cdrHierarchy->buildHierarchy( DoFManager,cdreqn,discretiser,
//                              net3DHierarchy,pMatrix,pRHS,pX,
//                              pProlongation);
// app->drawNet( dnet, "psfiles/ffcube", VRML );

// test discretisation:
T3DNetHierarchy::TNetListIterator *iterNet;
iterNet = new T3DNetHierarchy::TNetListIterator( net3DHierarchy->getList() );

TNet<3> *currentNet = iterNet->first();
TNet<3> *prevNet    = iterNet->first();
uint noOfNets = net3DHierarchy->getNoOfNets();
ulong noOfDoFs[noOfNets];
ulong blockSize[1];
TBlockProlongation<3> **blockProlong = new (TBlockProlongation<3>*)[noOfNets-1];

for (uint i = 0; i < noOfNets; i++) {
    
    cout<<" i: "<<i<<endl;
    
    DoFManager->InitDoFDesc(currentNet); 
    pMatrix[i] = DoFManager->buildStiffnessMatrix(currentNet);
    noOfDoFs[i]=(*(pMatrix[i]))[0][0].getSize();
    cout<<" Dofs: "<< noOfDoFs[i]<<endl;
    if (noOfDoFs[i] >0){
        blockSize[0] = noOfDoFs[i];
        pX[i] = new TBlockVector<TFloat>( 1, blockSize );
        pRHS[i] = new TBlockVector<TFloat>( 1, blockSize );
        discretiser->discretize( cdreqn, currentNet, pMatrix[i], pRHS[i]);
        
        ulong dim=noOfDoFs[i];

        TDynArray<TVector < 3 > * > arrCoord(dim);
        discretiser->getCoord ( currentNet,arrCoord);
        ulong j;
        for (j=0;j<dim;j++)
            cout<<"("<<j<<") :"<<*(arrCoord[j])<<endl;
 }
    else{
        delete pMatrix[i];
        pMatrix[i] =NULL;
        pRHS[i]=NULL;
        pX[i]=NULL;
    }// else
 
      if ( i > 0 ) {
         
         
        blockProlong[i-1] = new TBlockProlongation<3>( prevNet,currentNet,
                                                       noOfDoFs[i-1],  
                                                       noOfDoFs[i],0 );
        blockProlong[i-1]->delReference();
        pProlongation[i-1] = new TProlongation<3>( blockProlong[i-1] );
        //  prevNet->delReference();
        //delete prevNet;
        prevNet = currentNet;
        cout<<"  blockProlong[i-1]: "<<endl<<(*blockProlong[i-1])<<endl;
 

        }// if

    
   currentNet = iterNet->next();
  

  }// for

      
// check matrix

/*
  matWriteMatrixBlock("/tmp/tpr/matrixf.mat", (*pMatrix[noOfNets-1])[0][0],
  (*pMatrix[noOfNets-1])[0][0].getNoOfRows(),"Af");
  matWriteMatrixBlock("/tmp/tpr/matrixc.mat", (*pMatrix[noOfNets-2])[0][0],
  (*pMatrix[noOfNets-2])[0][0].getNoOfRows(),"Ac");
  
  matWriteMatrixBlock( "/tmp/tpr/prol.mat",*(pProlongation[noOfNets-2]->pBlockProlong[0]), 
  pProlongation[noOfNets-2]->pBlockProlong[0]->getcDim(),"Prol" );
  */    
      

// last net
dNet=iterNet->last();
// no of this net
int lastnetNumber = net3DHierarchy->getNoOfNets()-1;
// dim of the corresponding matrix
ulong dim = (*(pMatrix[lastnetNumber]))[0][0].getSize();

  
 TDynArray<TVector < 3 > * > arrCoord(dim);
discretiser->getCoord ( dNet,arrCoord);
cout<<"size: "<<arrCoord.getSize()<<endl;
  
TBlockVector<TFloat> *perror= new TBlockVector<TFloat>(1,dim);
TBlockVector<TFloat> *pexact= new TBlockVector<TFloat>(1,dim);
   
aA=pMatrix[lastnetNumber];
aRhs=pRHS[lastnetNumber];
aX=pX[lastnetNumber];
  
// solve
TSymmetricGaussSeidel *ssgs = new TSymmetricGaussSeidel( );
TBicgStab *bicgStab = new TBicgStab(ssgs );
bicgStab->setAccuracy( 1e-15 );
bicgStab->setMaxit( 1000 );
bicgStab->setVerbosity( 4 );
aX->set( 1.0 ); 
bicgStab->solve(aA,aRhs,aX);
    
    
for (ulong j=0; j<dim;j++){
    (*pexact)[0][j]= Dirichlet->eval ((arrCoord[j]));
    //  cout<<" X: "<<(*pX)[0][j]<<" coord: "<<*(arrCoord[j])<<endl;
}
  
(*perror)=(*aX)-(*pexact);
  


delete discretiser;
delete aRhs;
delete aX;
delete aA;
delete DoFManager;
delete perror;
delete pexact;
 
   


TMemory::statistics();
}
 
