//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D grids
//
// arch-tag: 18ec1c8e-a950-4c68-9659-337e4f57df88
//

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/TFunction.hh"

#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"
#include "net/TDomain.hh"

#include "discr/TCdrEqn.hh"
#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "discr/TConvDiffHierarchy.hh"
#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"

#include "visual/drawapp.hh"
#include "visual/TPOVRay.hh"

class TDirichletFun : public TFunction<3>
{
public:
    TDirichletFun(){};
    ~TDirichletFun(){};

    TFloat eval(TVector <3> *av )
    {
        TVector <3> *v; v=av; 
        TFloat a=(*v)[0];
        TFloat b=(*v)[1];
        TFloat c=(*v)[2];
        //  return a+b+2*c;
        return a*a+b*b+c*c;
    }
};

class TMyBasisCdrEqn : public TBasisCdrEqn<3>
{
public:
    TMyBasisCdrEqn ( TFloat aEpsilon = 1.0, 
                     TBlockVector<TFloat> *aConvVector = NULL)
            : TBasisCdrEqn<3>( aEpsilon, aConvVector )
    {}

    void evalDiff ( TVector <3> *v, TMatrix<3> *D ) const
    {
        TVector<3> * av = v;
        
        (*D)[0][0]=1; (*D)[0][1]=0; (*D)[0][2]=0;
        (*D)[1][0]=0; (*D)[1][1]=1; (*D)[1][2]=0;
        (*D)[2][0]=0; (*D)[2][1]=0; (*D)[2][2]=1;
    }

    void evalConv ( TVector <3> *v, TVector<3> *B ) const
    {
        TVector<3> * av = v;
        
        (*B)[0]=1; (*B)[1]=2; (*B)[2]=3;
    }


    TFloat evalReac ( TVector <3> *v ) const
    {
        TVector<3> * av = v;

        return 0.0;
    }
   
    TFloat evalRhs ( TVector <3> *v ) const
    {
        TVector<3> * av = v;
        TFloat       a  = (*v)[0];
        TFloat       b  = (*v)[1];
        TFloat       c  = (*v)[2];
        
        return -6;
    }
};

class TMyCdrEqn : public TCdrEqn<3>
{
public:
    TMyCdrEqn ( TFloat aEpsilon = 1.0, 
                TBlockVector<TFloat> *aConvVector = NULL )
            : TCdrEqn<3>(aEpsilon,aConvVector)
    {}
};


//
// mainfunction
//
int
main( int argc, char **argv ) 
{
    //
    // parse command line
    //
    param.convDirName = "circle";
    param.setStartTri( "grids/cube.tri" );
    param.noOfNets = 3;
    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl;

    //
    // init net and discr functions
    //
    TNetHierarchy<3>  * pHierarchy = new TNetHierarchy<3>( param.startTri );
    TNet<3>           * fnet[10];

    timer.stop();
    cout << "main (...) : time for refining = " << timer << endl;

    //
    // init discr functions
    //
    TBasisCdrEqn * cdreqn    = new TMyBasisCdrEqn( 1.0, NULL );
    TFunction<3> * Dirichlet = new TDirichletFun;
   
    cdreqn->isLaplace(  FALSE );
    cdreqn->Diffusion(  TRUE  );
    cdreqn->Convection( TRUE  );
    cdreqn->setEpsilon( 1     );
  
    // supply boundary Descriptons

    // in "grids/cube.tri", only one TBoundary is defined
    // therefore, one Description is enough

    TCdrBoundaryDesc<3> * boundDescD = new TCdrBoundaryDesc<3>(1);
    boundDescD->setType( 0, DIR_TYPE );
    boundDescD->setFunc( 0, Dirichlet );
    boundDescD->setDoFInfo();

    TListIterator< TBoundary >   iterBound( pDomain->getBoundaries() );
    TBoundary                  * pBoundary;
    
    iter.first();
    while ((pBoundary = iter++) != NULL)
    {
        cout << "main (...) : boundary = " << pBoundary->getName() << endl;
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( boundDescD );
    }
    
    fnet[0] = pHierarchy->getList().getLast(); 
  
    int mf = 1;
    for (uint i = 0; i <= mf; i++)
    {
        TConvDiffDoFManager<3>  * DoFManager;
        TBlockMatrix            * A, *CC;
        ulong                     dim;
        ulong                     size[1];
        TBlockVector<TFloat>    * rhs;
        TBlockVector<TFloat>    * pX;
        TBlockVector<TFloat>    * perror;
        TBlockVector<TFloat>    * pexact;
        TConvDiffFE<3>          * discretiser;
        TMatrixBlock              C;
        TDynArray< TVector<3>* >  arrCoord;
        TSymmetricGaussSeidel   * ssgs;
        TBicgStab               * bicgStab;

        
        cout << "main (...) : mf = " << i << endl;

        fnet[i+1] = pHierarchy->buildNet();
        dnet      = fnet[i+1];
        cout<<"main (...) : net refined"<<endl;

        DoFManager = new TConvDiffDoFManager<3>();
        DoFManager->InitDoFDesc(dnet);
        
        cout << "main (...) : DoFs on this Level = " << DoFManager->getNoOfDoFs() << endl;

        A   = DoFManager->buildStiffnessMatrix(dnet,dnet);
        dim = (*A)[0][0].getNoOfRows();
        cout << "main (...) : dimension of A = " << dim << endl;
  
        size[0] = dim;
        
        rhs    = new TBlockVector<TFloat>(1,size);
        pX     = new TBlockVector<TFloat>(1,size);
        perror = new TBlockVector<TFloat>(1,size);
        pexact = new TBlockVector<TFloat>(1,size);

        // Diskretisierung
  
        discretiser = new TConvDiffFE<3>( Tabata );
   
        C = (*A)[0][0];
        discretiser->discrDiff( cdreqn, dnet, A,  rhs);
        discretiser->discrConv( cdreqn, dnet, &C, rhs);
        discretiser->discrRhs(  cdreqn, dnet, rhs);
    
        cout << "main (...) : A = " << (*A)[0][0] << endl;
        cout << "main (...) : C = " << C << endl;
    
        // test discretisation:
        
        arrCoord.setSize(dim);
    
        discretiser->getCoord ( dnet, arrCoord );
        cout << "main (...) : size = " << arrCoord.getSize() << endl;
    
        CC = new TBlockMatrix( C );
    
        // solve
        ssgs     = new TSymmetricGaussSeidel( );
        bicgStab = new TBicgStab( ssgs );
        
        bicgStab->setAccuracy( 1e-15 );
        bicgStab->setMaxit( 1000 );
        bicgStab->setVerbosity( 4 );
        pX->set( 1.0 ); 
        bicgStab->solve(A,rhs,pX);
    
        for (ulong j=0; j<dim;j++)
        {
            (*pexact)[0][j] = Dirichlet->eval( arrCoord[j] );
            //  cout<<" X: "<<(*pX)[0][j]<<" coord: "<<*(arrCoord[j])<<endl;
        }// for
    
        (*perror) = (*pX) - (*pexact);
        cout << "main (...) : diff error = "
             << setprecision(5) << perror->euclidNorm()/sqrt(dim) << endl;
    
        delete discretiser;
        delete rhs;
        delete pX;
        delete A;
        delete CC;
        delete DoFManager;
        delete perror;
        delete pexact;
        delete ssgs;
        delete bicgStab;
    }

    netHier->free();
  
    TMemory::statistics();
}
 
