//
//  Projekt   : libnet
//  file      : confDiffFE.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : class for piecewise linear finite elements for
//              the convection-diffusion equation
//
// arch-tag: 6610d139-94d3-4f6d-8e95-7b3accb1898c
//

#include "TConvDiffFE.hh"
#include "math.hh"
#include "cell.hh"
#include "TMatrix.hh"
#include "T3DFrontier.hh"
#include "T3DTetrahedron.hh"
#include "var.hh"

// constructors	
template <int N>
TConvDiffFE<N>::TConvDiffFE (TDiscrType aDiscrType,uint averb )
    : TBasisFE<N>( aDiscrType,averb) 
{
    TBasisFE<N>::setSimple(); 
    for (uint i=0;i<N+1;i++)
        for (uint j=0;j<N;j++)
            gradT[i][j]=0.0;
    noOfInnerDoFs=0;
}


template <int N>
TConvDiffFE<N>::~TConvDiffFE () {}

// get coordinates of vertices of cell
void 
TConvDiffFE<2>::getCoord ( TCell<2> *cell ) 
{
    for ( uint i = 0; i < 3; i++) {
        nodes[i] = cell->getNode(i);      
        coord[i] = *(nodes[i]);      
    }
}
// get coordinates of vertices of cell
void 
TConvDiffFE<3>::getCoord ( TCell<3> *cell ) 
{
    T3DTetrahedron* acell = ( T3DTetrahedron*)cell;
    for ( uint i = 0; i < 4; i++) {
        nodes[i]=acell->getNode(i);
        coord[i] = *(nodes[i]);  
    }
}

// write coordinates of all (interior) vertices in net
// into (global) vector coord
template <int N>
void 
TConvDiffFE<N>::getCoord ( TNet<N> *net, TDynArray<TVector <N> *  >& arrCoord, uint block ) {
    long dof;
    
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();

    arrCoord.setSize( noOfInnerDoFs );
    
    TListIterator< TCell<N> >  iter(net->getList());
    for (TCell<N> *cell = iter.first(); cell != NULL; cell = iter.next()) 
        for ( uint i = 0; i < N+1; i++) {
            if (cell->getNode(i)->getBoundary()){
                dof=cell->getNode(i)->getDoFDesc()->getName(0)  ;
                if ( dof ==NOT_DEFINED){
                    cerr<<"TConvDiffFE<N>::getCoord: dof status not defined "<<endl;
                    exit(1);
                }
                if (dof!=NO_DOF) {
                    dof+= noOfInnerDoFs;
                    arrCoord[dof]=cell->getNode(i);
                }
            }   
            else{
                dof = cell->getNode(i)->getName();
                arrCoord[dof]=cell->getNode(i);
            }
        }
}
 
template <int N>
bool 
TConvDiffFE<N>::isDof (uint i)
{
    if (dof[i] >=0)
        return TRUE;
    if (dof[i]== NO_DOF)
        return FALSE;
    if (dof[i]==NOT_DEFINED)
        cerr<<"TConvDiffFE<N>::isDof: dof status not defined "<<endl;
    return FALSE;
}

template <int N>
bool 
TConvDiffFE<N>::isNeu (uint i)
{
    int type=BoundaryDesc[i]->getType(0);
    if (type== NO_TYPE)
        cerr<<"TConvDiffFE<N>::isNeu: dof type not defined"<<endl;
   
    if (type==NEU_TYPE)
        return TRUE;
    else return FALSE;
}



template <int N>
bool 
TConvDiffFE<N>::isDir (uint i)
{
    int type=BoundaryDesc[i]->getType(0);
    if (type== NO_TYPE)
        cerr<<"TConvDiffFE<N>::isDir: dof type not defined"<<endl;
    if (type==DIR_TYPE)
        return TRUE;
    else return FALSE;
}






template <int N>
void 
TConvDiffFE<N>::discretize ( TBasisEqn< N > *aeqn, TNet< N >  *net,  
                             TBlockMatrix *A,TBlockVector<TFloat> *rhs ){


    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
    TListIterator<TCell<N> >  iter(net->getList());
    TCell<N> *cell;
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();
   
    if (verbosity & PRINT_DISCRETISATION){
        if (eqn->Diffusion())
            cout<<"discretising Diffusion" <<endl;
        cout<<"discretising Rhs" <<endl;
    }

    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        if (eqn->Diffusion())
            discrDiff  ( eqn, A, rhs );
        discrRhs(eqn, rhs,cell);
    }
    //  cout<<(*rhs)[0]<<endl;  
    // discretisation of convection  is not elementwise possible
    // therfore: call global routine
    if (eqn->Convection())
        discrConv ( aeqn, net, &((*A)[0][0]), rhs );
    // use global routine to get global stiffness matrix (instead of mass vector)
    if (eqn->Reaction())
        discrMass ( aeqn, net, A, rhs);
    
     
} 

  
        
// calculate global stiffness matrix and write it into global matrix
template <int N>
void 
TConvDiffFE<N>::discrDiff ( TBasisEqn<N> *aeqn, TNet<N> *net,  
                            TBlockMatrix *A, TBlockVector<TFloat> *rhs ) {
    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
    TListIterator< TCell<N> >  iter(net->getList());
    TCell<N> *cell;
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();


    if (eqn->Diffusion()==FALSE){
        cerr<<" TConvDiffFE<N>::discrDiff  started, but eqn says: no Diffusion " <<endl;
        return;
    }
    
    if (verbosity & PRINT_DISCRETISATION)
        cout<<"discretising Diffusion" <<endl;
    
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrDiff  ( eqn, A, rhs );
    } 
 
}
// calculate element stiffness matrix
template <int N>
void 
TConvDiffFE<N>::discrDiff ( TBasisCdrEqn<N> *eqn, TBlockMatrix *A, 
                            TBlockVector<TFloat> *rhs ) 
{
    TFloat h = 0.0;
    uint i, j;
    for ( i = 0; i < N+1; i++) {
        for ( j = 0; j < N+1; j++) {
            h=0.0;
            if ( isDof(i) ) {
                // i is interior vertex
                if (eqn->isLaplace())
                    h = gradT[i] * gradT[j] * vol * eqn->getEpsilon();
                else {
                    TMatrix<N> D;
                    eqn->evalDiff(&(coord[i]), &D);
                    h = gradT[i]*(D*gradT[j])* vol;
                }
                if (isDof(j) )  
                    // j is interior vertex ---> entry in matrix 
                    // entry in first diagonal block (Au)
                    (*A)[0][0][dof[i]][dof[j]] += h;
                else{ 
                    if (isDir(j)){
                        // j is BOUNDARY vertex ---> entry in rhs
                        (*rhs)[0][dof[i]] -=h* BoundaryDesc[j]->getFunc(0)->eval(&(coord[j]));
                    } // isDir
                } // isDof(j)
            } // isDof(i)
        } //j
    } //i
    
} // discDiff

// calculate element stiffness matrix of non-symmentric part
template <int N>
void 
TConvDiffFE<N>::discrConv ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                            TMatrixBlock *C, TBlockVector<TFloat> *rhs,
                            uint block) 
{ 
    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
   
    if (eqn->Convection()==FALSE){
        cerr<<" TConvDiffFE<N>::discrConv  started, but eqn says: no Convection " <<endl; 
        return;
    }
    // start discretisation
    if (verbosity & PRINT_DISCRETISATION)
        cout<<"discretising Convection" <<endl;
    
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();
    
    if ( TBasisFE<N>::getDiscrType()  == Null ) {
        cerr << "TConvDiff::discrConv: no discretisation type DiscrType defined\n"; 
        return;}    
    
    TDynArray<TVector <N> *  > coordVec( (*rhs)[0].getSize() );
    getCoord( net, coordVec );
    eqn->initConvVector( coordVec );
    
    TListIterator<TCell<N> >  iter(net->getList());
    TVectorBlock<TFloat> suppSize( 0.0, (*rhs)[0].getSize());
    
    TCell<N> *cell;
    if ( TBasisFE<N>::getDiscrType() == Tabata ) 
        for (cell = iter.first(); cell != NULL; cell = iter.next()) {
            initDofs( cell ); 
            updateSupportSize( suppSize );
        }
    
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDofs( cell ); 
        getCoord(cell);
        if ( TBasisFE<N>::getDiscrType() == Tabata )  
            discrConv ( eqn, C, rhs, suppSize );
        else
            discrConv ( eqn, C, rhs );
    }
    
    eqn->deleteConvVector();              
    
}


// calculate global (lumped) mass matrix and write it into global vector M
template <int N>
void 
TConvDiffFE<N>::discrMass ( TBasisEqn<N> *aeqn, TNet<N> *net,  
                            TVectorBlock<TFloat> *M, TBlockVector<TFloat> *rhs) {
        
    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
    if (eqn->Reaction()==FALSE){
        cerr<<" TConvDiffFE<N>::discrMass  started, but eqn says: no Reaction " <<endl; 
        return;
    }

    if (verbosity & PRINT_DISCRETISATION)
        cout<<"discretising Mass with single matrix "<<endl;
    
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();
    
    TListIterator<TCell<N> >  iter(net->getList());
    TCell<N> *cell;
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrMass ( eqn, M, rhs);
    }
} // discrMass

// calculate global (lumped) mass matrix and add it to stiffness matrix
template <int N>
void 
TConvDiffFE<N>::discrMass ( TBasisEqn<N> *aeqn, TNet<N> *net,  
                            TBlockMatrix *A, TBlockVector<TFloat> *rhs) {
    ulong dim=(*A)[0][0].getSize();
    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
    if (verbosity & PRINT_DISCRETISATION)
        cout<<"discretising Mass and update stiffness matrix " <<endl;
  
   // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();
    
    TVectorBlock<TFloat> M( 0.0,dim);
    TListIterator< TCell<N> >  iter(net->getList());
    
    TCell<N> *cell;
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrMass(eqn,&M,rhs);
        //discrMass(eqn,&((*A)[0][0]),rhs);
    }
    //add mass vector do main diagonal
    for (ulong j = 0; j <dim ; j++)
        (*A)[0][0][j][j] += M[j];
} // discrMass

// calculate element (lumped) mass matrix
template <int N>
void 
TConvDiffFE<N>::discrMass ( TBasisCdrEqn<N> *eqn, TVectorBlock<TFloat> *M, TBlockVector<TFloat> *rhs) {
    // simple four-node quadrature rule with lumping
    // the reaction function is assumed to be constant 
    // on the element, with value c=(1/(N+1)) sum_{i=0}^{N} reac(x[i])
    TFloat c = 0.0;
    uint i;
    for (i=0;i<N+1;i++) c+=eqn->evalReac(&(coord[i]));
    c/=(N+1);
   

    // quadrature nodes : A.H. Strout , 
    // Approximate calculation of multiple integrals, 
    // Prentice Hall 1971, p307
  
    TFloat r=(N+2-sqrt(N+2))/((N+1)*(N+2));
    TFloat s=(N+2+N*sqrt(N+2))/((N+1)*(N+2));
  
       // value on diag of element mass matrix
    TFloat vd= vol/(N+1)*c* (N*r*r+s*s);
       
    // value out of  diag of element mass matrix
    TFloat vod = vol/(N+1)*c*((N-1)*r*r+2*r*s);

     
    for (i=0;i<N+1;i++){
        if (isDof(i))
            for (uint j=0;j<N+1;j++){
                if (j==i)
                    (*M)[dof[i]]+=vd;
                else{
                    if (isDof(j)) //lumping
                        (*M)[dof[i]]+=vod;
                    else 
                        if (isDir(j)){
                            // j is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[i]] -=vod* BoundaryDesc[j]->getFunc(0)->eval(&(coord[j]));
                        } 
                }  
            }
    }
} // discrMass

// calculate element (lumped) mass matrix
template <int N>
void 
TConvDiffFE<N>::discrMass ( TBasisCdrEqn<N> *eqn, TMatrixBlock *M, 
                            TBlockVector<TFloat> *rhs) {
    // simple four-node quadrature rule with lumping
    // the reaction function is assumed to be constant 
    // on the element, with value c=(1/(N+1)) sum_{i=0}^{N} reac(x[i])
    TFloat c = 0.0;
    uint i;
    for (i=0;i<N+1;i++) c+=eqn->evalReac(&(coord[i]));
    c/=(N+1);
   
    // quadrature nodes : A.H. Strout , 
    // Approximate calculation of multiple integrals, 
    // Prentice Hall 1971, p307
    TFloat r=(N+2-sqrt(N+2))/((N+1)*(N+2));
    TFloat s=(N+2+N*sqrt(N+2))/((N+1)*(N+2));
   
     // value on diag of element mass matrix
    TFloat vd= vol/(N+1)*c* (N*r*r+s*s);
 
    // value out of  diag of element mass matrix
    TFloat vod = vol/(N+1)*c*((N-1)*r*r+2*r*s);

    for (i=0;i<N+1;i++){
        if (isDof(i)){
            for (uint j=0;j<N+1;j++)
                if ( i == j ){
                    (*M)[dof[i]][dof[i]]+= vd;
                }
                else{
                    if (isDof(j)){
                        //      (*M)[dof[i]][dof[j]]+=vol/(N+1)*c*((N-1)*r*r+2*r*s);
                        // lumping: 
                        (*M)[dof[i]][dof[i]]+=vod;
                    }
                    else
                        if (isDir(j)){
                            // j is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[i]] -=vod* BoundaryDesc[j]->getFunc(0)->eval(&(coord[j]));
                        }    
                }
        }
    }
     
 

    //  (*M)[dof[i]]=  vol/(N+1)*c* ((N*r*r+s*s) +  ((N-1)*r*r+2*r*s));
           
} // discrMass

// calculate global stiffness matrix and write it into global matrix
template <int N>
void 
TConvDiffFE<N>::discrRhs ( TBasisEqn<N> *aeqn, TNet<N> *net, TBlockVector<TFloat> *rhs ) {

    TBasisCdrEqn<N> *eqn=  (TBasisCdrEqn<N>*)aeqn;
    TListIterator< TCell<N> >  iter(net->getList());
    TCell<N> *cell;
    // init inner Dofs
    noOfInnerDoFs=net->getNoOfInteriorNodes();
    
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrRhs  ( eqn, rhs,cell );
    }
} 

void 
TConvDiffFE<2>::discrRhs  ( TBasisCdrEqn< 2 > *eqn,TBlockVector<TFloat> *rhs, TCell<2> *cell){

    // avoiding warnings...
    TBasisCdrEqn<2>* aeqn; aeqn=eqn;
    TBlockVector<TFloat> *arhs;arhs=rhs;
    TCell<2>* acell; acell=cell;
   
    
    // quadature rule: 
    // \int_{cell} f * b_i dx \approx size_of_cell \sum(x_j \in cell) 1/(N+1)(f*b_i)(x_j)
    //                            =   detJ/ fac(N) 1/(N+1) f(x_i) 
    cerr<<"TConvDiffFE<N>::discrRhs: Sorry,not implemented"<<endl; 
}

void 
TConvDiffFE<3>::discrRhs  ( TBasisCdrEqn< 3 > *eqn,TBlockVector<TFloat> *rhs, TCell<3> *cell ){
    // quadature rule: \int_{cell} f * b_i dx \approx size_of_cell \sum(x_j \in cell) 1/(N+1)   (f * b_i) (x_j)
    //                                           =   detJ/ fac(N) 1/(N+1) f(x_i) 
    
    T3DFrontier * Face;
    TVector<3>  x,y,z;
    TBoundary *bound;

    TVertex<3> *p0, *p1,*p2,*p3;
    uint j=0;
    
    TFloat h=0;

    //rhs
    for (uint  i = 0; i < 4; i++) {
        if (isDof(i)) 
            (*rhs)[0][dof[i]] += eqn->evalRhs(&(coord[i])) * vol/(4);
    }

    // Neumann Face discretisation
    for (uint  i = 0; i < 4; i++) {
        //check which face is Neumann Face
        Face=(T3DFrontier *)cell->getFrontier (i);
        bound=Face->getBoundary();
        if(bound!=NULL)
            if (bound->getBoundaryDesc()->getType(0) == NEU_TYPE){
                // Face is Neumann
                p0=Face->getNode(0);         
                p1=Face->getNode(4);         
                p2=Face->getNode(2);      
                p3=cell->getOppositeNode(i);
                
                /*cout<<" p0: "<<*p0;
                  cout<<" p1: "<<*p1;
                  cout<<" p2: "<<*p2;
                  cout<<" p3: "<<*p3<<endl;*/
                
                x=(*p1)-(*p0);
                y=(*p2)-(*p0);
                // triangle size
                TFloat size = 0.5*sqrt((x*x)*(y*y)-(x*y)*(x*y)); //= 1/2 ||x|| ||y|| \sqrt(1-cos(phi)^2)
                // quadrature rule:  
                // \int_{face} b_i*phi  dx  \approx size_of_face * (1/3) (b_i f(m_i))
                // where m_i are the face side's midpoints
                TVector<3> m0,m1,m2;
                m0=((*p1)+(*p2))*0.5; 
                m1=((*p0)+(*p2))*0.5; 
                m2=((*p0)+(*p1))*0.5;
            
                TFloat n0,n1,n2;
            
                // assume: phi_neumann is the same on each node of the face
                //   TPDEBoundaryDesc<3>* aBonDesc = (TPDEBoundaryDesc<3> *) bound->getBoundaryDesc();
               
                //two cases:
                // i)  derivatives  ux, uy, uz of the solution  are known
                //     In ths case, aBonDesc->getFunc(0) is of type TFunction<6>
                //     and expects argument x to evaluate  and  the face's normal direction
                
                // ii) as usual: phi_neu  is given

                if (bound->getBoundaryDesc()->getN() == 6){
                    TPDEBoundaryDesc<6>* aBonDesc = (TPDEBoundaryDesc<6> *) bound->getBoundaryDesc();
                    TFunction<6>* phi_neumann=aBonDesc->getFunc(0);
                    // normal direction
                    TVector<3> n = x.cross(y);
                    n *= (1.0/n.euclidNorm());
                    
                    // check if outer normal
                    z=(*p3)-(*p0);
                    if (n*z>0) // n points inside the cell
                        n *=(-1.0);
                    
                    // values at midpoints
                    TVector<6> m0,m1,m2;
                    // vectors m_j consisting of x (first 3 components)
                    //                       and n ( last 3 components)
                    for (uint l=0; l<3;l++){
                        m0[l]=((*p1)[l]+(*p2)[l])*0.5; 
                        m1[l]=((*p0)[l]+(*p2)[l])*0.5; 
                        m2[l]=((*p0)[l]+(*p1)[l])*0.5;
                        m0[l+3]=n[l];
                        m1[l+3]=n[l];
                        m2[l+3]=n[l];
                    }
                    
                    n0= phi_neumann->eval(&m0);
                    n1= phi_neumann->eval(&m1);
                    n2= phi_neumann->eval(&m2);
                  
                }
                else{
                    TPDEBoundaryDesc<3>* aBonDesc = (TPDEBoundaryDesc<3> *) bound->getBoundaryDesc();
                    TFunction<3>* phi_neumann=aBonDesc->getFunc(0);
                    // values at midpoints
                    TVector<3> m0,m1,m2;
                    m0=((*p1)+(*p2))*0.5; 
                    m1=((*p0)+(*p2))*0.5; 
                    m2=((*p0)+(*p1))*0.5;
                    n0= phi_neumann->eval(&m0);
                    n1= phi_neumann->eval(&m1);
                    n2= phi_neumann->eval(&m2);
                   
                } // else
                

                
                // integrals
                // p0:  
                h = size/3.0 *  0.5 * (n1+n2);
                
              
                // search number of p0
                j=0;
                
                while((nodes[j]!=p0) && (j<3))
                    j++;
                
                if(nodes[j]!=p0) {
                    cerr<<"TConvDiffFE<3>::discrRhs : Problem to find nodes "<<endl;
                    cout<<" p0: "<<p0<<endl;
                    cout<<" node[0]: "<<nodes[0]<<endl;
                    cout<<" node[1]: "<<nodes[1]<<endl;
                    cout<<" node[2]: "<<nodes[2]<<endl;
                    cout<<" node[3]: "<<nodes[3]<<endl;
                    
                    exit(-1);
                }
                
                if (isDof(j)) 
                    (*rhs)[0][dof[j]] +=h;
                

                // p1:  
                h = size/3.0 *  0.5 * (n0+n2);
                j=0;
                while((nodes[j]!=p1) && (j<3))
                    j++;
                if(nodes[j]!=p1) 
                    cerr<<"TConvDiffFE<3>::discrRhs : Problem to find nodes "<<endl;
                if (isDof(j)) 
                    (*rhs)[0][dof[j]] +=h;
                
                // p2:  
                h = size/3.0 *  0.5 * (n0+n1);
                j=0;
                while((nodes[j]!=p2) && (j<3))
                    j++;
                if(nodes[j]!=p2) 
                    cerr<<"TConvDiffFE<3>::discrRhs : Problem to find nodes "<<endl;

                if (isDof(j)) 
                    (*rhs)[0][dof[j]] +=h;
               
              
            } // Neumann boundary
    } // for i 
}// discrRhs 

// initialize grad, det and dof for a particular cell

void 
TConvDiffFE<2>::initDiscr ( TCell<2> *cell ) {
    // initialize dof
    initDofs(cell);
 
   // calculate determinants and their inverses
    TFloat invdetJ;
    detJ = 2 * vol;
    invdetJ = 1 / detJ;
   
    getCoord( cell );
   
    gradT[0][X] = invdetJ * (coord[1][Y]-coord[2][Y]);
    gradT[0][Y] = invdetJ * (coord[2][X]-coord[1][X]);
    gradT[1][X] = invdetJ * (coord[2][Y]-coord[0][Y]);
    gradT[1][Y] = invdetJ * (coord[0][X]-coord[2][X]);
    gradT[2][X] = invdetJ * (coord[0][Y]-coord[1][Y]);
    gradT[2][Y] = invdetJ * (coord[1][X]-coord[0][X]);
}

void 
TConvDiffFE<3>::initDiscr ( TCell<3> *cell ) {
    // initialize dofs
    initDofs(cell);

    // calculate determinants and their inverses
   
    // fac(N)= N!
    vol = cell->getSize();
    getCoord( cell );

    TMatrix<3>* C =new TMatrix<3>();
    TMatrix<3>* C_inv =new TMatrix<3>();
    // C from (3.4.68)
    // colums of C: coord(X_i)-coord(x_0)
    // colums of C^{-t} = rows of C^{-1}: gradients
     
    for (int i=0;i<3;i++)
        for (int j=0;j<3;j++)
            (*C)[j][i]=coord[i+1][j]-coord[0][j];
   
    (*C_inv)=C->inverted();
   
    // colums of C^{-t} = rows of C^{-1}: gradients
    for (int i=0;i<3;i++)
        gradT[i+1]=(*C_inv)[i];
    gradT[0]=(-1.0)*gradT[1]-gradT[2]-gradT[3];
    
    delete C;
    delete C_inv;

}

template <int N>
void 
TConvDiffFE<N>::initDiscr ( TCell<N> *cell ) {
   
    uint i;
    // initialize dof
    for ( i = 0; i < N+1; i++)
        dof[i] = cell->getNode(i)->getName();
   
    // calculate determinants and their inverses
   
    TFloat invdetJ;
   
    // fac(N)= N!
    detJ = fac(N) * cell->getSize( );
    invdetJ = 1 / detJ;
   
    getCoord( cell );
   
}

template <int N>
void 
TConvDiffFE<N>::initDofs ( TCell<N> *cell ) {
    // initialize dofs and cell volume
  
    for ( uint i = 0; i < N+1; i++){
        TBoundary *boundary=cell->getNode(i)->getBoundary();
        if (boundary != NULL){ // boundary node
            BoundaryDesc[i]=(TPDEBoundaryDesc<N>*)boundary->getBoundaryDesc();
            DoFDesc[i]=cell->getNode(i)->getDoFDesc();
            dof[i] = DoFDesc[i]->getName(0);
            
            if( dof[i] >-1) // DoF is defined and really a DoF
                dof[i] +=noOfInnerDoFs;
            
        }
        else{
            BoundaryDesc[i]=NULL;
            DoFDesc[i]=NULL;
            dof[i]=cell->getNode(i)->getName();
        }
        //     cout<<"dof: "<<dof[i]<<" "<<*(cell->getNode(i))<<endl;
    }
    vol = cell->getSize();
}

template <int N>
void 
TConvDiffFE<N>::discrConv ( TBasisCdrEqn<N> *eqn, TMatrixBlock *C, 
                            TBlockVector<TFloat> *rhs ) {
    // avoiding warnings...
    TBasisCdrEqn<N>* aeqn; aeqn=eqn;
    TMatrixBlock *aC;aC=C;
    TBlockVector<TFloat> *arhs;arhs=rhs;
    cerr<<"TConvDiffFE<N>::discrConv(not Tabata): sorry, not implemented"<<endl;

    /*
      TFloat hk = 0.0, hl = 0.0, hm = 0.0;
      TFloat alpha,beta[3][3], lambda[3][3];
      TVector<2> b, r;
     
      uint k, l, m;
      for ( k = 0; k < 3; k++ ) { 
      l = (k+1) % 3;    
      m = (k+2) % 3;
      r[0] = (1.0/6.0) * ( coord[k][1] + coord[l][1] - 2 * coord[m][1]);
      r[1] = (1.0/6.0) * ( 2 * coord[m][0] - coord[k][0] -  coord[l][0]);
      alpha = SGN(( r[0] * (coord[k][0] - coord[l][0])) +
      ( r[1] * (coord[k][1] - coord[l][1])));
  

  
      b[0] = (1.0 / 12.0) * ( 
      5 * (*(eqn->getConvb0()))(coord[k])
      + 5 * (*(eqn->getConvb0()))(coord[l])
      + 2 * (*(eqn->getConvb0()))(coord[m])) ;
      b[1] = (1.0 / 12.0) * ( 
      5 * (*(eqn->getConvb1()))(coord[k])
      + 5 * (*(eqn->getConvb1()))(coord[l])
      + 2 * (*(eqn->getConvb1()))(coord[m])) ;
  
      beta[l][k] = alpha*(b[0]*r[0]+b[1]*r[1]);                    
      beta[k][l] =-alpha*(b[0]*r[0]+b[1]*r[1]);                    
      lambda[k][l] = (*PHI)(beta[k][l] / eqn->getEpsilon());
      lambda[l][k] = 1 - lambda[k][l];

  
      for ( k = 0; k < 3; k++) { 
      l = (k+1) % 3;    
      m = (k+2) % 3;
   
      hl =  beta[k][l] * (1 - lambda[k][l]); 
      hm =  beta[k][m] * (1 - lambda[k][m]);
      hk = - hl - hm;
    
      if ( dof[k] < BOUNDARY_NAME ) {
      // k is interior vertex
      (*C)[dof[k]][dof[k]] += hk;
      
      if ( dof[l] < BOUNDARY_NAME )
      (*C)[dof[k]][dof[l]] += hl;
      else    
      (*rhs)[0][dof[k]] -= hl * eqn->dirichlet0(coord[l]);
      
      
      if ( dof[m] < BOUNDARY_NAME )
      (*C)[dof[k]][dof[m]] += hm;
      else
      (*rhs)[0][dof[k]] -= hm * eqn->dirichlet0(coord[m]);        
      }
      }
      */
}  //  discrConv 


void 
TConvDiffFE<2>::discrConv ( TBasisCdrEqn<2> *eqn, TMatrixBlock *C, 
                            TBlockVector<TFloat> *rhs, 
                            TVectorBlock<TFloat> &suppsize, 
                            uint block ) {
   
    TFloat hk = 0.0, hl = 0.0, hm = 0.0;
    TFloat det, my, beta, lambda, vl, vm, F, s;
    TVector<2> b, r, e[3];
    const TFloat eps = 1e-8;
     
    e[0] = coord[2] - coord[1]; // edges  1 ->N
    e[1] = coord[0] - coord[2]; //        N ->0 
    e[2] = coord[1] - coord[0]; //        0 ->1
     
    for ( uint k = 0; k < 2+1; k++ ) { 
        if ( isDof(k) ) {  // k is interior vertex
            uint l = (k+1) % 2+1;    
            uint m = (k+2) % 2+1;
            hk = hl = hm = 0;
            b[0] = (*(eqn->getConvVector()))[0][dof[k]]; 
            b[1] = (*(eqn->getConvVector()))[1][dof[k]];
         
            beta = b.euclidNorm();
         
            // vl > 0 if -b is left from e[m]
            vl = b[0] * (e[m][1]) - (e[m][0]) * b[1]; 
            // e[l] has "wrong" direction!
            vm = -(b[0] * (e[l][1]) - (e[l][0]) * b[1]);
            if ((fabs(vl)<eps) && (e[m]*b<0) ){       
                // convection || e[m]
                F = 0.5 * beta / e[m].euclidNorm();  
                // 0.5: each of the two cells with this edge (e[m]) 
                // contributes to the matrix entry
                hk = F; hm = 0, hl = -F;
            }
         
            if ( (fabs(vm) < eps) && (e[l] * b > 0) ){        
                // convection || e[l]
                F = 0.5 * beta / e[l].euclidNorm();
                hk = F; hl = 0, hm = -F;
            }
         
            if ( (vl > eps) && (vm < -eps)  ) { 
                // convection (-1* b) between  e[m] and e[l]
                // solve lamda b + my e[k] = - e[m] by Cramer's rule
                det = b[0] * e[k][1] - e[k][0] * b[1];
                lambda = - (e[m][0] * e[k][1] - e[k][0] * e[m][1]) / det;
                my = -(b[0] * e[m][1] - e[m][0] * b[1]) / det;
                if (lambda > eps) {
                    F = 1 / fabs(lambda);
                    if ((my < 0) || (my > 1)) cout << "my falsch: "<<my<<endl;
                    hk = F; hl = (my-1) * F; hm = - my * F; }
            
            }
         
            s = suppsize[dof[k]] / 2+1.0;
            hk *= s; hl *= s; hm *= s;
         
            // k is interior vertex
            (*C)[dof[k]][dof[k]] += hk;
         
            if (isDof(l))
                (*C)[dof[k]][dof[l]] += hl;
            else
                if (isDir(l)){
                    // l is BOUNDARY vertex ---> entry in rhs
                    (*rhs)[0][dof[k]] -=hl* BoundaryDesc[l]->getFunc(0)->eval(&(coord[l]));
                } 
     

            if ( isDof(m) )
                (*C)[dof[k]][dof[m]] += hm;
            else
                if (isDir(m)){
                    // j is BOUNDARY vertex ---> entry in rhs
                    (*rhs)[0][dof[k]] -=hm* BoundaryDesc[m]->getFunc(0)->eval(&(coord[m]));
                }   
        }
      
    }
     
} // discrConv 


void 
TConvDiffFE<3>::discrConv ( TBasisCdrEqn<3> *eqn, TMatrixBlock *C, 
                            TBlockVector<TFloat> *rhs, 
                            TVectorBlock<TFloat> &suppsize, 
                            uint block ) {
   
   
    TFloat sigma,tau,my,e,h;
    TVector<3> a,b,c,d,n,z,pki ;
    uint k1,k2,k3;
    const TFloat eps = 1e-8;
    for ( uint k = 0; k < 4; k++ ) {
        if ((isDof(k)) && (suppsize[dof[k]]>eps) ){
            k1=(k+1)%4;
            k2=(k+2)%4;
            k3=(k+3)%4;
            a=coord[k2]-coord[k1];
            b=coord[k3]-coord[k1];
            c=coord[k1]-coord[k];
            n=a.cross(b);
            z[0]=(*(eqn->getConvVector()))[0][dof[k]];
            z[1]=(*(eqn->getConvVector()))[1][dof[k]];
            z[2]=(*(eqn->getConvVector()))[2][dof[k]];
            e=-1.*(z*n);
            if ((fabs(e)>eps) && (((c*n)/e >0) )) // otherwise, z and n are parallel
            { 
                d=(z.cross(c))*(-1);
                sigma=-(b*d)/e;
                tau=(a*d)/e;
                if ((sigma>=0) &&( tau>=0) &&(sigma+tau <=1)){
                    pki=c+sigma*a+tau*b;
                    my=-(pki*z)/(pki*pki);
            
                    // my -> my* supp(b_k)/(N+1)
                    my=my*suppsize[dof[k]]/4;
               
                    //  cout<<"my: "<<setprecision(8)<<my<<" dofs: k: "<<dof[k]<<" k1: "<<dof[k1]<<" k2: "<<dof[k2]<<" k3: "<<dof[k3]<<endl;
                    //cout<<"my: "<<setprecision(8)<<my<<" sigma: "<<sigma<<" tau: "<<tau<<endl;
               
                    //k
                    (*C)[dof[k]][dof[k]]+=my;
                    // k1
                    h= -my*(1-sigma-tau);
                    if (isDof(k1))
                        (*C)[dof[k]][dof[k1]]+=h;
                    else
                        if (isDir(k1)){
                            // k1 is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[k]] -= h*BoundaryDesc[k1]->getFunc(0)->eval(&(coord[k1]));
                        } 

                    // k2
                    h= -my*sigma; 
                    if (isDof(k2))
                        (*C)[dof[k]][dof[k2]]+=h;
                    else
                        if (isDir(k2)){
                            // k1 is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[k]] -= h*BoundaryDesc[k2]->getFunc(0)->eval(&(coord[k2]));
                        } 
               
                    // k3

                    h= -my*tau; 
                    if (isDof(k3))
                        (*C)[dof[k]][dof[k3]]+=h;
                    else
                        if (isDir(k3)){
                            // k1 is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[k]] -= h*BoundaryDesc[k3]->getFunc(0)->eval(&(coord[k3]));
                        } 
                    // z almost parallel to an edge P_k P_ki
                    // therefore, set suppsize[dof[k]] :=0 to avoid another discretisation 
                    // from an neigboughring tetrahedron
             

                    if ((sigma<eps) ||(tau<eps)||(1-sigma-tau<eps))
                        suppsize[dof[k]]=0;
               
                }
                else{
                    /*   cerr<<" TConvDiffFE<3>::discrConv: sigma:"<<sigma<<" tau: "<<tau<<endl;
                         cerr<<" z: "<<z<<" a: "<<a<<" b: "<<b<< " n:"<<n<<endl;
                         cerr<<" k:"<<coord[k]<<" k1:"<<coord[k1]<<" k2:"<<coord[k2]<<" k3:"<<coord[k3]<<endl
                         */
                }
            }
        }// isDof(k)
    }//for k
} //TConvDiffFE<3>::discrConv


template <int N>
void 
TConvDiffFE<N>::updateSupportSize (TVectorBlock<TFloat> &suppSize ) {
    for ( uint k = 0; k < N+1; k++ ) {
        if (isDof(k) ) 
            // size of element :  detJ = fac(N) * cell.Size;
            suppSize[dof[k]] += vol;
      
    }
}

template class TConvDiffFE<2>;
template class TConvDiffFE<3>;
