//
//  Project   : libnet
//  File      : TStokesFE.cc
//  Author    : Thomas Probst / Sabine Gutsch
//  Purpose   : class for linear conforming 2D finite elements
//  Date      : 04.08.1998
//  Changes   : 
//  Questions : do we have to initialize floats to zero?
// arch-tag: 35498172-5e1d-47ba-ad50-fb1754734e13
//

#include "TStokesFE.hh"
#include "math.hh"
#include "cell.hh"
#include "TMatrix.hh"
#include "T3DFrontier.hh"
#include "T3DTetrahedron.hh"

// constructors	
template <int N>
TStokesFE<N>::TStokesFE ( TDiscrType aDiscrType) 
	: TBasisFE<N>( aDiscrType ) 
{
	TBasisFE<N>::setSimple(); 
    cInterior = 0;
    fInterior = 0;
}
	
// destructor
template <int N>
TStokesFE<N>::~TStokesFE () {}

// get coordinates of vertices of children cells
void 
TStokesFE<2>::getCoord ( TCell<2> *cell ) 
{
	uint i, k;
	for ( i = 0; i < 3; i++) pCoord[i] = (*cell->getNode(i));

	noOfChildren = cell->getNoOfChildren();
	for ( i = 0; i < noOfChildren; i++) 
		for ( k = 0; k < 3; k++) 
			coord[i][k] = (*cell->getChild(i)->getNode(k));
}

// get coordinates of vertices of children cells
void 
TStokesFE<3>::getCoord ( TCell<3> *cell ) 
{
    T3DTetrahedron* acell = (T3DTetrahedron*)cell;
	uint i, k;
    for ( i = 0; i < 4; i++) 
        pCoord[i] = *(acell->getNode(i));  

	noOfChildren = acell->getNoOfChildren();
	for ( i = 0; i < noOfChildren; i++) 
		for ( k = 0; k < 4; k++) 
			coord[i][k] = (*acell->getChild(i)->getNode(k));
}

// write coordinates of all (unknown) u(k)-vertices in net
// into (global) vector coord
template <int N>
void 
TStokesFE<N>::getCoord ( TNet<N> *net, TDynArray<TVector<N>*>& arrCoord,
                         uint block )
{
    uint i,j;
    long dof;
    TListIterator< TCell<N> >  iter(net->getList());
    for (TCell<N> *cell = iter.first(); cell != NULL; cell = iter.next()) 
	{
        noOfChildren = cell->getNoOfChildren();
	    for ( i = 0; i < noOfChildren; i++) 
            for ( j = 0; j < N+1; j++ )
			{
               if (cell->getChild(i)->getNode(j)->getBoundary())
			   {
                    dof = cell->getChild(i)->getNode(j)->getDoFDesc()->getName(block);
                    if (dof > -1) dof += net->getNoOfInteriorNodes();
			   }
               else
                   dof = cell->getChild(i)->getNode(j)->getName();
         
                if ( dof == NOT_DEFINED)
			    {
                    cerr << "TStokesFE<N>::getCoord: dof status not defined\n";
                    exit(1);
			    }
                if (dof != NO_DOF ) 
                    arrCoord[dof] = cell->getChild(i)->getNode(j);
			}
	}
}



// write coordinates of all (unknown) u(k)-vertices in net
// into (global) vector coord
template <int N>
void 
TStokesFE<N>::getCoord ( TNet<N> *net, uint k,
                         TDynArray<TVector<N>*>& arrCoordU,
                         TDynArray<TVector<N>*>& arrCoordP )
{
    uint i, j;
    long dof_u, dof_p;
    TListIterator< TCell<N> >  iter(net->getList());
    for (TCell<N> *cell = iter.first(); cell != NULL; cell = iter.next()) 
	{
	   // first coarse cell (p - dofs)
        for ( i = 0; i < N+1; i++) 
		{
            if (cell->getNode(i)->getBoundary())
			{
                dof_p = cell->getNode(i)->getDoFDesc()->getName(N);
                if (dof_p > -1) dof_p += cInterior;
			}
            else
                dof_p = cell->getNode(i)->getName();
     
            if ( dof_p == NOT_DEFINED)
			{
                cerr << "TStokesFE<N>::getCoord: dof status not defined\n";
                exit(1);
			}
            if (dof_p != NO_DOF ) 
                arrCoordP[dof_p] = cell->getNode(i);
		}
		// now children cells (u - dofs)
        noOfChildren = cell->getNoOfChildren();
	    for ( i = 0; i < noOfChildren; i++) 
            for ( j = 0; j < N+1; j++ )
			{
               if (cell->getChild(i)->getNode(j)->getBoundary())
			   {
                   dof_u = cell->getChild(i)->getNode(j)->getDoFDesc()->getName(k);
                   if (dof_u > -1) dof_u += fInterior;
			   }
               else
                   dof_u = cell->getChild(i)->getNode(j)->getName();
         
               if ( dof_u == NOT_DEFINED)
			   {
                    cerr << "TStokesFE<N>::getCoord: dof status not defined\n";
                    exit(1);
			   }
               if (dof_u != NO_DOF ) 
                   arrCoordU[dof_u] = cell->getChild(i)->getNode(j);
			}
	}
}

template <int N>
bool 
TStokesFE<N>::isDof (uint i, uint j, uint k)
{
   if (dof_u[i][j][k] >= 0)
      return (TRUE);
   if (dof_u[i][j][k] == NO_DOF)
      return (FALSE);
   if (dof_u[i][j][k] == NOT_DEFINED)
      cerr << "TStokesFE<N>::isDof: dof status not defined " << endl;
   return (FALSE);
}
//
//  init dofs
//
template <int N>
void 
TStokesFE<N>::initDofs ( TCell<N> *cell ) 
{
    uint i, j, k;
    // initialize dofs and cell volume
	// first coarse cell
    for ( i = 0; i < N+1; i++) 
	{
        if (cell->getNode(i)->getBoundary())
		{
            dof_p[i] = cell->getNode(i)->getDoFDesc()->getName(N);
            if (dof_p[i] > -1) dof_p[i] += cInterior;
	        //cout << "initDofs: " << dof_p[i] << endl;
		}
        else
            dof_p[i] = cell->getNode(i)->getName();
        if ( dof_p[i] == NOT_DEFINED)
		{
            cerr << "TStokesFE<N>::getCoord: dof status not defined\n";
            exit(1);
		}
	}
    vol_p = cell->getSize();
	// now children cells
    noOfChildren = cell->getNoOfChildren();
	for ( i = 0; i < noOfChildren; i++) 
	{
        for ( j = 0; j < N+1; j++ )
		{
            if (cell->getChild(i)->getNode(j)->getBoundary())
			{
                boundaryDesc_u[i][j] = (TPDEBoundaryDesc<N>*)cell->getChild(i)->getNode(j)->getBoundary()->getBoundaryDesc();
                for ( k = 0; k < N; k++ )
				{
                    dof_u[i][j][k] = cell->getChild(i)->getNode(j)->getDoFDesc()->getName(k);
				// dof is defined and really a dof
                    if (dof_u[i][j][k] > -1) dof_u[i][j][k] += fInterior;
				}
			}
            else
			{
                for ( k = 0; k < N; k++ )
                    dof_u[i][j][k] = cell->getChild(i)->getNode(j)->getName();    
                boundaryDesc_u[i][j] = NULL;
			}

            for ( k = 0; k < N; k++ )
                if ( dof_u[i][j][k] == NOT_DEFINED)
			    {
                    cerr << "TStokesFE<N>::getCoord: dof status not defined\n";
                    exit(1);
				}
		}
		/*
        cout << "initDof: ";
        for ( j = 0; j < N+1; j++ )
			cout << dof_u[i][j][0] << " ";
		cout << endl;         
		*/
        vol_u[i] = cell->getChild(i)->getSize();
	}
}
//
//  discrConv: global
//
template <int N>
void 
TStokesFE<N>::discrConv ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                          TMatrixBlock *C, TBlockVector<TFloat> *rhs,
                          uint block ) 
{  
    bool setConv = FALSE;
    TStokesBasisEqn<N> *eqn = (TStokesBasisEqn<N>*)aeqn;
	typedef TCell<N> TNetCell;
	TListIterator<TNetCell>  iter(net->getList());
    TCell<N> *cell;

    if ( TBasisFE<N>::getDiscrType()  == Null ) 
    {
        cerr << "TConvDiff::discrConv: no discretisation type DiscrType defined\n"; 
        return;
    }  

    TDynArray<TVector<N>*> coordVec( (*rhs)[block].getSize() );

    TVectorBlock<TFloat> *suppSize = new TVectorBlock<TFloat>( 0.0, (*rhs)[block].getSize());

    getCoord( net, coordVec, block );
    
    if (eqn->getConvVector() == NULL)
	{
        eqn->initConvVector( (*rhs)[block].getSize(), coordVec );
        setConv = TRUE;
	}

	for (cell = iter.first(); cell != NULL; cell = iter.next()) 
	{
		initDofs ( cell ); 
		updateSuppSize( *suppSize, block );
	}
	
	for (cell = iter.first(); cell != NULL; cell = iter.next()) 
	{
		initDofs ( cell ); 
        getCoord( cell );
        if ( TBasisFE<N>::getDiscrType() == Tabata ) 
           discrConv ( eqn, C, rhs, *suppSize, block );
        else
           discrConv ( eqn, C, rhs );
	}
	// delete convVector if it was set here
    if (setConv == TRUE) eqn->deleteConvVector();
    delete (suppSize);
	
}       

// calculate element stiffness matrix
template <int N>
void 
TStokesFE<N>::discrDiff ( TStokesBasisEqn<N> *eqn, TBlockMatrix *A, 
                          TBlockVector<TFloat> *rhs ) 
{
	TFloat h = 0.0;
    uint i, j, k, m;
	for ( i = 0; i < noOfChildren; i++) {
		for ( j = 0; j < N+1; j++) 
		    for ( k = 0; k < N+1; k++) 
                for ( m = 0; m < N; m++ )
                    if ( isDof(i, j, m) ) 
					{
                        h = gradT[i][j] * gradT[i][k] * vol_u[i] * eqn->getEpsilon();
	                    if ( isDof(i, k, m) )
			                (*A)[m][m][dof_u[i][j][m]][dof_u[i][k][m]] += h;
                        else 
                            (*rhs)[m][dof_u[i][j][m]] -= h * boundaryDesc_u[i][k]->getFunc(m)->eval(&(coord[i][k]));
					}
	}
}

template <int N>
void 
TStokesFE<N>::discrUP ( TStokesBasisEqn<N> *eqn, TBlockMatrix *A,
                        TBlockVector<TFloat> *rhs) 
{
   TFloat h = 0.0;
   uint i, j, k, m;

   for ( k = 0; k < noOfChildren; k++) 
   {
		// for noOfChildren triangles
		for ( j = 0; j < N+1; j++) 
			// for course grid nodes
			for ( i = 0; i < N+1; i++) 
			{
				// for fine grid nodes
                for (m = 0; m < N; m++) 
				{
            	    //h = - N * vol_u[k] * gradT[k][i][m] * c[k][j] / ((N+1)*fac(N));
            	    h = - vol_u[k] * gradT[k][i][m] * c[k][j] / (N+1);

            	    if ( isDof(k,i,m) ) 
				    { 	// i is interior node 
               	        (*A)[m][N][dof_u[k][i][m]][dof_p[j]] += h;  
                        (*A)[N][m][dof_p[j]][dof_u[k][i][m]] += h;  
				    }
                    else 
                        (*rhs)[N][dof_p[j]] -= h * boundaryDesc_u[k][i]->getFunc(m)->eval(&(coord[k][i]));
				}
			}
   }
}

// calculate global (lumped) mass matrix and write it into global vector M
template <int N>
void 
TStokesFE<N>::discrMass ( TBasisEqn<N> *aeqn, TNet<N> *net,  
                         TVectorBlock<TFloat> *M, 
                         TBlockVector<TFloat> *rhs) {
        
    TStokesBasisEqn<N> *eqn=  (TStokesBasisEqn<N>*)aeqn;
        typedef TCell<N> TNetCell;
		TListIterator<TNetCell>  iter(net->getList());
        TCell<N> *cell;
	    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
              initDiscr ( cell );
              discrMass ( eqn, M, rhs);
            }
} // discrMass

// calculate element (lumped) mass matrix
template <int N>
void 
TStokesFE<N>::discrMass ( TStokesBasisEqn<N> *eqn, TVectorBlock<TFloat> *M, TBlockVector<TFloat> *rhs) {

    // simple four-node quadrature rule with lumping
    // the reaction function is assumed to be constant 
    // on the element, with value c=(1/(N+1)) sum_{i=0}^{N} reac(x[i])

	// ADJUST for STOKES!!!
	/*
    TFloat c = 0.0;
    uint i;
    for (i=0;i<N+1;i++) c+=eqn->evalReac(&(coord[i]));
    c/=(N+1);

    // quadrature nodes : A.H. Strout , 
    // Approximate calculation of multiple integrals, 
    // Prentice Hall 1971, p307
  
    TFloat r=(N+2-sqrt(N+2))/((N+1)*(N+2));
    TFloat s=(N+2+N*sqrt(N+2))/((N+1)*(N+2));
  
       // value on diag of element mass matrix
    TFloat vd= vol/(N+1)*c* (N*r*r+s*s);
       
    // value out of  diag of element mass matrix
    TFloat vod = vol/(N+1)*c*((N-1)*r*r+2*r*s);

     
    for (i=0;i<N+1;i++){
        if (isDof(i))
            for (uint j=0;j<N+1;j++){
                if (j==i)
                    (*M)[dof[i]]+=vd;
                else{
                    if (isDof(j)) //lumping
                        (*M)[dof[i]]+=vod;
                    else 
                        if (isDir(j)){
                            // j is BOUNDARY vertex ---> entry in rhs
                            (*rhs)[0][dof[i]] -=vod* BoundaryDesc[j]->getFunc(0)->eval(&(coord[j]));
                        } 
                }  
            }
    }
	*/

} // discrMass
//
// discr rhs (global)
//
template <int N>
void 
TStokesFE<N>::discrRhs ( TBasisEqn<N> *aeqn, TNet<N> *net, TBlockVector<TFloat> *rhs )
{
   TStokesBasisEqn<N> *eqn =  (TStokesBasisEqn<N>*)aeqn;
   TListIterator< TCell<N> >  iter(net->getList());
   TCell<N> *cell;
   for (cell = iter.first(); cell != NULL; cell = iter.next()) 
   {
      initDiscr ( cell );
      discrRhs  ( eqn, rhs );
   }
}

template <int N>
void 
TStokesFE<N>::discrRhs( TStokesBasisEqn<N> *eqn, TBlockVector<TFloat> *rhs )
{
    uint i, j, k;
    for ( i = 0; i < noOfChildren; i++ )
        for ( j = 0; j < N+1; j++) 
            for ( k = 0; k < N; k++) 
                if ( isDof(i,j,k) ) 
                   (*rhs)[k][dof_u[i][j][k]] += eqn->evalRhs(&(coord[i][j]), k) * vol_u[i] / 4.0;
}

//
// initialize grad, det, c, dof_u and dof_p for a particular cell
//
void 
TStokesFE<2>::initDiscr ( TCell<2> *cell ) 
{
    uint i, j, k;
	
	// initialize dof_u, dof_p, number of children, vol_u
    initDofs(cell);
	// initialize coordinates
	getCoord( cell );

	// calculate determinants and their inverses
	// (triangles ordered as in refinement,
	// gradients ordered as vertices in (children) triangles)

	TFloat invdetJ[4];

	for ( i = 0; i < noOfChildren; i++) 
	{
		detJ[i] = 2 * vol_u[i];
		invdetJ[i] = 1 / detJ[i];
      
		gradT[i][0][X] = invdetJ[i] * (coord[i][1][Y]-coord[i][2][Y]);
		gradT[i][0][Y] = invdetJ[i] * (coord[i][2][X]-coord[i][1][X]);
		gradT[i][1][X] = invdetJ[i] * (coord[i][2][Y]-coord[i][0][Y]);
		gradT[i][1][Y] = invdetJ[i] * (coord[i][0][X]-coord[i][2][X]);
		gradT[i][2][X] = invdetJ[i] * (coord[i][0][Y]-coord[i][1][Y]);
		gradT[i][2][Y] = invdetJ[i] * (coord[i][1][X]-coord[i][0][X]);
	}

	// initialize c
	for ( i = 0; i < noOfChildren; i++) 
	    for ( j = 0; j < 3; j++) 
            c[i][j] = 0.0; 

	// set c
	for ( i = 0; i < noOfChildren; i++) 
		for( j = 0; j < 3; j++) {
			for( k = 0; k < 3; k++) 
				if ( cell->getChild(i)->getNode(j) == 
					 cell->getFrontier(k)->getNode(1) )
					// j is k'th midNode 
				{
					c[i][k]       += 0.5; 
					c[i][(k+1)%3] += 0.5; 
				}
			for( k = 0; k < 3; k++) 
				if ( cell->getChild(i)->getNode(j) == 
					 cell->getNode(k) )
					// j is k'th coarse node
					c[i][k] += 1.0; 
		}
}


//
// initialize grad, det, c, dof_u and dof_p for a particular cell
//
void 
TStokesFE<3>::initDiscr ( TCell<3> *cell ) 
{
    uint i, j, k, m;
    // initialize dofs
    initDofs( cell );
    getCoord( cell );

   // calculate determinants and their inverses   

   TMatrix<3>* C = new TMatrix<3>();
   TMatrix<3>* C_inv = new TMatrix<3>();
   // C from (3.4.68)
   // colums of C: coord(X_i)-coord(x_0)
   // colums of C^{-t} = rows of C^{-1}: gradients
     
	for ( k = 0; k < noOfChildren; k++) 
	{
        for (i = 0; i < 3; i++)
            for (j = 0; j < 3; j++)
                (*C)[j][i] = coord[k][i+1][j] - coord[k][0][j];
   
        (*C_inv) = C->inverted();
   
       // colums of C^{-t} = rows of C^{-1}: gradients
       for (i = 0; i < 3; i++)
           gradT[k][i+1] = (*C_inv)[i];
       gradT[k][0] = (-1.0) * gradT[k][1] - gradT[k][2] - gradT[k][3];
    }

	// initialize c
	for ( i = 0; i < noOfChildren; i++) 
	    for ( j = 0; j < 4; j++) 
            c[i][j] = 0.0; 

	// set c
    TVertex<3>* gv[3];
    TVertex<3> *node;
    for ( i = 0; i < 4; i++) 
	{ // 4 coarse grid nodes
        node = cell->getNode(i);
        bool found = FALSE; 
        j = 0; 
        while ( found == FALSE )
		{
            m = 0;
		    for ( k = 0; k < 4; k++) 
			{
                if (cell->getChild(j)->getNode(k) == node)
                    found = TRUE;
                else
                    gv[m++] = cell->getChild(j)->getNode(k);
			}
            j++;
		}

	    for ( j = 0; j < noOfChildren; j++)
		    for( k = 0; k < 4; k++)
			{
                if (cell->getChild(j)->getNode(k) == node)
                    c[j][i] += 1.0;
                else
                    for ( m = 0; m < 3; m++ )
                        if (cell->getChild(j)->getNode(k) == gv[m])
                            c[j][i] += 0.5;
			}
	}

}


// calculate global stiffnes matrix
template <int N>
void 
TStokesFE<N>::discrDiff ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                         TBlockMatrix *A, 
						 TBlockVector<TFloat> *rhs ) 
{ 
    TStokesBasisEqn<N> *eqn = (TStokesBasisEqn<N>*)aeqn;
	cInterior     = net->getNoOfInteriorNodes();
	
	typedef TCell<N> TNetCell;
	TCell<N> *cell;
	TListIterator<TNetCell>  iter(net->getList());
	for (cell = iter.first(); cell != NULL; cell = iter.next()) 
	{
		initDiscr ( cell );
		discrDiff ( eqn, A, rhs );
		discrUP ( eqn, A, rhs );
        discrRhs( eqn, rhs );
	}
}        

// upwinding 
template <int N>
void 
TStokesFE<N>::discrConv ( TStokesBasisEqn<N> *eqn, TMatrixBlock *C, 
                          TBlockVector<TFloat> *rhs ) 
{
	TFloat hk = 0.0, hl = 0.0, hm = 0.0;
	TFloat alpha, beta[3][3], lambda[3][3];
	TVector<N> b, r;
    uint j, k, l, m;

	for ( j = 0; j < noOfChildren; j++ ) {
		for ( k = 0; k < 3; k++ ) { 
			l = (k+1)%3;    
			m = (k+2)% 3;
			r[0] = (1.0/6.0)*( coord[j][k][1] + coord[j][l][1] -2*coord[j][m][1]);
			r[1] = (1.0/6.0)*(2*coord[j][m][0] - coord[j][k][0] -  coord[j][l][0]);
			alpha = SGN((r[0]*(coord[j][k][0]- coord[j][l][0])) +
					  (r[1]*(coord[j][k][1]- coord[j][l][1])));
			b[0] = (1.0/12.0)*(5*(*eqn->getConvVector())[0][dof_u[j][k][j]]+ 5*(*eqn->getConvVector())[0][dof_u[j][l][j]]+2*(*eqn->getConvVector())[0][dof_u[j][m][j]]);
			b[1] = (1.0/12.0)*(5*(*eqn->getConvVector())[1][dof_u[j][k][j]]+ 5*(*eqn->getConvVector())[1][dof_u[j][l][j]]+2*(*eqn->getConvVector())[1][dof_u[j][m][j]]);
                
			beta[l][k] = alpha*(b[0]*r[0]+b[1]*r[1]);			 
			beta[k][l] = -alpha*(b[0]*r[0]+b[1]*r[1]);			 
			lambda[k][l] = (*TBasisFE<N>::getPHI())(beta[k][l]/eqn->getEpsilon());
			lambda[l][k] = 1-lambda[k][l];
		}
            
            
		for ( k = 0; k < 3; k++) { 
			l = (k+1)%3;    
            m =(k+2)% 3;
                
			hl =  beta[k][l]*(1-lambda[k][l]); 
			hm =  beta[k][m]*(1-lambda[k][m]);
			hk =  - hl - hm;
                
			if ( dof_u[j][k][j] < 0 ) {
				// k is interior vertex
				(*C)[dof_u[j][k][j]][dof_u[j][k][j]] += hk;
                    
				if ( dof_u[j][l][j] < 0 )
					(*C)[dof_u[j][k][j]][dof_u[j][l][j]] += hl;
				else {
					(*rhs)[0][dof_u[j][k][j]] -= hl* boundaryDesc_u[k][j]->getFunc(0)->eval(&(coord[k][j]));
					(*rhs)[1][dof_u[j][k][j]] -= hl* boundaryDesc_u[k][j]->getFunc(1)->eval(&(coord[k][j]));
				}
                    
				if ( dof_u[j][m][j] < 0 )
					(*C)[dof_u[j][k][j]][dof_u[j][m][j]] += hm;
				else {
					(*rhs)[0][dof_u[j][k][j]] -= hm * boundaryDesc_u[k][j]->getFunc(0)->eval(&(coord[k][j])); 
					(*rhs)[1][dof_u[j][k][j]] -= hm * boundaryDesc_u[k][j]->getFunc(1)->eval(&(coord[k][j]));
				}
			}
                
		}
	}
} /* discrConv */

//
// discrConv: Tabata (2D)
//
void 
TStokesFE<2>::discrConv ( TStokesBasisEqn<2> *eqn, TMatrixBlock *C, 
                          TBlockVector<TFloat> *rhs, 
						  TVectorBlock<TFloat> &suppSize, uint block ) 
{ // Tabata

	TFloat hk = 0.0, hl = 0.0, hm = 0.0;
	TFloat det,my,beta, lambda,vl,vm,F,s;
	TVector<2> b, r, e[3];
	const TFloat eps = 1e-8;
    uint i, j, l, m;
   
    for ( i = 0; i < noOfChildren; i++ ) 
	{
        e[0] = coord[i][2]-coord[i][1]; // edges  1 ->2
        e[1] = coord[i][0]-coord[i][2]; //        2 ->0 
        e[2] = coord[i][1]-coord[i][0]; //        0 ->1
      
		for ( j = 0; j < 3; j++ )  
			{
                if ( isDof(i,j,block) ) 
				{	// block is interior vertex
                    l = (j+1)%3;    
                    m = (j+2)%3;
                    hk = hl = hm = 0;
                    b[0] = (*eqn->getConvVector())[0][dof_u[i][j][block]];
                    b[1] = (*eqn->getConvVector())[1][dof_u[i][j][block]];
            
                    beta = b.euclidNorm();
            
                    // vl > 0 if -b is left from e[m]
                    vl = b[0]*(e[m][1])-(e[m][0])*b[1];             
                    // e[l] has "wrong" direction!
                    vm = -(b[0]*(e[l][1])-(e[l][0])*b[1]);

                    if ( (fabs(vl) < eps) && ( e[m] * b < 0) ) 
					{        
                        // convection || e[m]
                        F = 0.5 * beta / e[m].euclidNorm();  
                        // 0.5: each of the two cell with this edge 
						// (e[m]) contributes to the matrix entry
                        hk = F; hm = 0, hl = -F;
					}
            
                    if ( (fabs(vm) < eps ) && ( e[l] * b > 0) ) 
					{    
                        // convection || e[l]
                        F = 0.5 * beta / e[l].euclidNorm();
                        hk = F; hl = 0, hm = -F;
					}
            
                    if ( ( vl > eps ) && ( vm < -eps )  ) 
					{
                        // convection (-1* b) between  e[m] and e[l]
                        // solve lamda b + my e[j]=-e[m] by Cramer's rule
                        det = b[0] * e[j][1] - e[j][0] * b[1];
                        lambda = -(e[m][0]*e[j][1]-e[j][0]*e[m][1])/det;
                        my = -(b[0]*e[m][1]-e[m][0]*b[1])/det;
                        if ( lambda > eps ) 
						{
                            F = 1/fabs(lambda);
                            if (( my < 0) || ( my > 1 )) 
                            cout<< "my falsch: "<<my<<endl;
                            hk = F;
                            hl = (my-1) * F; 
                            hm = - my * F;
						}
					}
            
                    s = suppSize[ dof_u[i][j][block] ] / 3.0;
                    hk *= s; 
                    hl *= s;
                    hm *= s;
            
         	        // j is interior vertex
				    (*C)[dof_u[i][j][block]][dof_u[i][j][block]] += hk;
            
				    if ( isDof(i,l,block) )
					    (*C)[dof_u[i][j][block]][dof_u[i][l][block]] += hl;
				    else 
					    (*rhs)[block][dof_u[i][j][block]] -= hl * boundaryDesc_u[i][l]->getFunc(block)->eval(&(coord[i][l]));
				    if ( isDof(i,m,block) )
					    (*C)[dof_u[i][j][block]][dof_u[i][m][block]] += hm;
				    else 
					    (*rhs)[block][dof_u[i][j][block]] -= hl * boundaryDesc_u[i][m]->getFunc(block)->eval(&(coord[i][m]));
				}
			}
	}
} /* discrConv Tabata (2D)*/

//
// discrConv<3> (Tabata)
//
void 
TStokesFE<3>::discrConv ( TStokesBasisEqn<3> *eqn, TMatrixBlock *C, 
                          TBlockVector<TFloat> *rhs, 
                          TVectorBlock<TFloat> &suppsize, uint block ) 
{  
	// so far only for Dirichlet (or same bc for all components,
    // otherwise one needs C1, C2, C3, suppsize1, suppsize2, suppsize3
    TFloat sigma, tau, my, e, h;
    TVector<3> a, b, c, d, n, z, pki;
    uint k1, k2, k3;
    uint i, j;
    const TFloat eps = 1e-8;

    for ( i = 0; i < noOfChildren; i++ ) 
        for ( j = 0; j < 4; j++ ) // 4 vertices in tetrahedron
		{
			if ((isDof(i,j,block)) && (suppsize[dof_u[i][j][block]] > eps) )
			{
				k1 = (j+1) % 4;
				k2 = (j+2) % 4;
				k3 = (j+3) % 4;
				a = coord[i][k2] - coord[i][k1];
				b = coord[i][k3] - coord[i][k1];
				c = coord[i][k1] - coord[i][j];
				n = a.cross(b);
				z[0] = (*(eqn->getConvVector()))[0][dof_u[i][j][block]];
				z[1] = (*(eqn->getConvVector()))[1][dof_u[i][j][block]];
				z[2] = (*(eqn->getConvVector()))[2][dof_u[i][j][block]];

				e = -1.*(z*n);
				if ((fabs(e)>eps) && (((c*n)/e >0) )) 
					// otherwise, z and n are parallel
				{
					d = (z.cross(c))*(-1);

					sigma = -(b*d)/e;
					tau = (a*d)/e;
					if ((sigma>=0) &&( tau>=0) &&(sigma+tau <=1))
					{
						pki =c+sigma*a+tau*b;
						my =-(pki*z)/(pki*pki);
						// my -> my* supp(b_k)/(N+1)
						my = my*suppsize[dof_u[i][j][block]]/4;
						//j
						(*C)[dof_u[i][j][block]][dof_u[i][j][block]] += my;
						// k1
						h= -my*(1-sigma-tau);
						if (isDof(i,k1,block))
                        (*C)[dof_u[i][j][block]][dof_u[i][k1][block]] += h;
					else
						//if (isDir(i,k1,block))
					    // k1 is BOUNDARY vertex ---> entry in rhs
                        (*rhs)[block][dof_u[i][j][block]] -= h*boundaryDesc_u[i][k1]->getFunc(block)->eval(&(coord[i][k1]));

					// k2
                    h= -my*sigma; 
                    if (isDof(i,k2,block))
                        (*C)[dof_u[i][j][block]][dof_u[i][k2][block]] += h;
                    else
                        //if (isDir(k2))
					    // k1 is BOUNDARY vertex ---> entry in rhs
                        (*rhs)[block][dof_u[i][j][block]] -= h*boundaryDesc_u[i][k2]->getFunc(block)->eval(&(coord[i][k2]));
	          
					// k3
                    h = -my*tau; 
                    if (isDof(i,k3,block))
                        (*C)[dof_u[i][j][block]][dof_u[i][k3][block]] += h;
                    else
                        //if (isDir(i,k3,block))
						// k1 is BOUNDARY vertex ---> entry in rhs
                        (*rhs)[block][dof_u[i][j][block]] -= h*boundaryDesc_u[i][k3]->getFunc(block)->eval(&(coord[i][k3]));
						
					// z almost parallel to an edge P_k P_ki
					// therefore, set suppsize[dof_u[j]] :=0 to avoid 
					// another discretisation from an neigbouring 
					//tetrahedron
             
                   if ((sigma < eps) || (tau < eps) || ((1-sigma-tau) < eps))
                       suppsize[dof_u[i][j][block]] = 0;
               
				}
			}
		}// isDof(j)
	}//for j
} //TStokesFE<3>::discrConv



template <int N>
void 
TStokesFE<N>::updateSuppSize (TVectorBlock<TFloat> &suppSize, uint k ) 
{
    uint i, j; 
    for ( i = 0; i < noOfChildren; i++ ) 
        for ( j = 0; j < N+1; j++ ) 
            if (isDof( i, j, k )) 
                suppSize[dof_u[i][j][k]] += vol_u[i];
}

template <int N>
void 
TStokesFE<N>::setInterior ( uint cInt, uint fInt )
{
    cInterior = cInt;
    fInterior = fInt;
}


template class TStokesFE<2>;
template class TStokesFE<3>;
