//
//  Projekt   : libnet
//  file      : confDiffFE.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : class for piecewise linear finite elements for
//              the convection-diffusion equation
//
// arch-tag: 3c9278f7-5a81-4504-81ce-c3541dfc7f07
//

#include "convDiffFE.hh"

// constructors	
T2DConvDiffFE::T2DConvDiffFE ( TDiscrType aDiscrType) 
        : T2DBasisFE( aDiscrType ) 
{
    PHI = simple; 
}
 
// destructor
T2DConvDiffFE::~T2DConvDiffFE () {}

// get coordinates of vertices of cell
void 
T2DConvDiffFE::getCoord ( TCell<2> *cell ) 
{
    for ( uint i = 0; i < 3; i++)
    {
        coord[i] = (*cell->getNode(i));
        dof_u[i] = cell->getNode(i)->getName();
    }// for
}

// write coordinates of all (interior) vertices in net
// into (global) vector coord
void 
T2DConvDiffFE::getCoord ( TNet<2> *net, TDynArray< TVector<2> * >& arrCoord )
{
    uint   i;
    ulong  dof;

    TListIterator< TCell<2> >  iter( net->getList() );
    TCell<2>                 * cell;

    arrCoord.setSize( net->getNoOfInteriorNodes() );
    arrCoord.setGrowBy( net->getNoOfInteriorNodes()/100 );
    
    iter.first();
    while ((cell = iter++) != NULL)
    {
        for ( i = 0; i < 3; i++ )
        {
            dof = cell->getNode( i )->getName();
            
            if ( dof < BOUNDARY_NAME )
            {
                if (arrCoord.getSize() < dof+1)
                    arrCoord.setSize( dof + 1 );
                
                arrCoord[dof] = cell->getNode(i);
            }// if
        }// for
    }// while
}

// calculate global stiffness matrix and write it into global matrix
void 
T2DConvDiffFE::discrDiff ( T2DBasisEqn *eqn, TNet<2> *net,  
                           TBlockMatrix *A, TBlockVector<TFloat> *rhs ) {
        
    BOUNDARY_NAME = ULONG_MAX - net->getNoOfBoundaryNodes();

    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
    TCell<2> *cell;
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrDiff  ( eqn, A, rhs );
    }
} 

// calculate element stiffness matrix
void 
T2DConvDiffFE::discrDiff ( T2DBasisEqn *eqn, TBlockMatrix *A, 
                           TBlockVector<TFloat> *rhs ) 
{
    TFloat h = 0.0;
    uint i, j;
        
    /*
      cout<<" Zelle: ";
      for ( i = 0; i < 3; i++) 
      cout <<" c["<<i<<"]: "<<coord[i];
      cout<<endl;
      cout<<" Dof: ";
      for ( i = 0; i < 3; i++) 
      if ( dof_u[i] < BOUNDARY_NAME ) 
      cout <<" d["<<i<<"]: "<<dof_u[i];
      else
      cout <<" d["<<i<<"]: "<<"B";
      cout<<endl;

      */

    for ( i = 0; i < 3; i++) {
        if ( dof_u[i] < BOUNDARY_NAME ) 
            // update rhs
            (*rhs)[0][dof_u[i]] += eqn->f0(coord[i]) * detJ / 6.0;
            
        for ( j = 0; j < 3; j++) {
            if ( dof_u[i] < BOUNDARY_NAME ) {
                // i is interior vertex
                h = gradT[i] * gradT[j] * detJ * 0.5 * eqn->getEpsilon();
                if ( dof_u[j] < BOUNDARY_NAME )  
                    // j is interior vertex ---> entry in matrix 
                    // entry in first diagonal block (Au)
                    (*A)[0][0][dof_u[i]][dof_u[j]] += h;
                else 
                    // j is BOUNDARY vertex ---> entry in rhs
                    (*rhs)[0][dof_u[i]] -=h*eqn->dirichlet0(coord[j]);
            }
        }
    }
}


// calculate element stiffness matrix of non-symmentric part
void 
T2DConvDiffFE::discrConv ( T2DBasisEqn *eqn, TNet<2> *net, 
                           TMatrixBlock *N, TBlockVector<TFloat> *rhs) 
{ 
    if ( DiscrType  == Null ) {
        cerr << "TConvDiff::discrConv: no discretisation type DiscrType defined\n"; 
        return;}	

    BOUNDARY_NAME = ULONG_MAX - net->getNoOfBoundaryNodes();
    //T2DVector **coordVec = new (T2DVector*)[(*rhs)[0].getSize()];
		
    TDynArray< TVector<2> * > coordVec( (*rhs)[0].getSize() );
    getCoord( net, coordVec );
		
    eqn->initConvVector( net->getNoOfInteriorNodes(), coordVec );
	
    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
		
    TVectorBlock<TFloat> suppSize( 0.0, (*rhs)[0].getSize());
		
    TCell<2> *cell;
    if ( DiscrType == Tabata ) 
        for (cell = iter.first(); cell != NULL; cell = iter.next()) {
            initDiscr( cell ); 
            getSuppSize( suppSize );
        }
			
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        getCoord( cell );
        if ( DiscrType == Tabata ) 
            discrConv ( eqn, N, rhs, suppSize );
        else
            discrConv ( eqn, N, rhs );
    }

    eqn->deleteConvVector();		
}


// calculate global (lumped) mass matrix and write it into global vector M
void 
T2DConvDiffFE::discrMass ( T2DBasisEqn *eqn, TNet<2> *net,  
                           TVectorBlock<TFloat> *M) {
        
    BOUNDARY_NAME = ULONG_MAX - net->getNoOfBoundaryNodes();

    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
    TCell<2> *cell;
    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
        initDiscr ( cell );
        discrMass ( eqn, M);
    }
} // discrMass

// calculate element (lumped) mass matrix
void 
T2DConvDiffFE::discrMass ( T2DBasisEqn *eqn, TVectorBlock<TFloat> *M) {
    TFloat h = 0.0;
    uint i, j;
    for ( i = 0; i < 3; i++) {
        for ( j = 0; j < 3; j++) {
            if ( dof_u[i] < BOUNDARY_NAME ) {
                // i is interior vertex
                h = (1.0) / (24.0) * detJ * eqn->getReacc()(coord[i]);
                if (j == i) h *= 2;
                (*M)[dof_u[i]] += h;
            }
        }
    }
  
} // discrMass

// initialize grad, det and dof_u for a particular cell
void 
T2DConvDiffFE::initDiscr ( TCell<2> *cell ) {

    uint i;
    // initialize dof_u
    for ( i = 0; i < 3; i++)
        dof_u[i] = cell->getNode(i)->getName();

    // calculate determinants and their inverses

    TFloat invdetJ;

    detJ = 2 * cell->getSize( );
    invdetJ = 1 / detJ;
      
    getCoord( cell );

    gradT[0][X] = invdetJ * (coord[1][Y]-coord[2][Y]);
    gradT[0][Y] = invdetJ * (coord[2][X]-coord[1][X]);
    gradT[1][X] = invdetJ * (coord[2][Y]-coord[0][Y]);
    gradT[1][Y] = invdetJ * (coord[0][X]-coord[2][X]);
    gradT[2][X] = invdetJ * (coord[0][Y]-coord[1][Y]);
    gradT[2][Y] = invdetJ * (coord[1][X]-coord[0][X]);
}

void 
T2DConvDiffFE::discrConv ( T2DBasisEqn *eqn, TMatrixBlock *N, 
                           TBlockVector<TFloat> *rhs ) {
    TFloat hk = 0.0, hl = 0.0, hm = 0.0;
    TFloat alpha,beta[3][3], lambda[3][3];
    TVector<2> b, r;
  
    uint k, l, m;
    for ( k = 0; k < 3; k++ ) { 
        l = (k+1) % 3;    
        m = (k+2) % 3;
        r[0] = (1.0/6.0) * ( coord[k][1] + coord[l][1] - 2 * coord[m][1]);
        r[1] = (1.0/6.0) * ( 2 * coord[m][0] - coord[k][0] -  coord[l][0]);
        alpha = SGN(( r[0] * (coord[k][0] - coord[l][0])) +
                    ( r[1] * (coord[k][1] - coord[l][1])));


 
        b[0] = (1.0 / 12.0) * ( 
            5 * (*(eqn->getConvb0()))(coord[k])
            + 5 * (*(eqn->getConvb0()))(coord[l])
            + 2 * (*(eqn->getConvb0()))(coord[m])) ;
        b[1] = (1.0 / 12.0) * ( 
            5 * (*(eqn->getConvb1()))(coord[k])
            + 5 * (*(eqn->getConvb1()))(coord[l])
            + 2 * (*(eqn->getConvb1()))(coord[m])) ;
  

        /*  b[0] = (1.0 / 12.0) * ( 5 * (*(eqn->getConvVector()))[0][dof_u[k]]
            + 5 * (*(eqn->getConvVector()))[0][dof_u[l]]
            + 2 * (*(eqn->getConvVector()))[0][dof_u[m]]);
            b[1] = (1.0 / 12.0) * ( 5 * (*(eqn->getConvVector()))[1][dof_u[k]]
            + 5 * (*(eqn->getConvVector()))[1][dof_u[l]]
            + 2 * (*(eqn->getConvVector()))[1][dof_u[m]]);
            */
                  
        beta[l][k] = alpha*(b[0]*r[0]+b[1]*r[1]);			 
        beta[k][l] =-alpha*(b[0]*r[0]+b[1]*r[1]);			 
        lambda[k][l] = (*PHI)(beta[k][l] / eqn->getEpsilon());
        lambda[l][k] = 1 - lambda[k][l];
    }
  
    for ( k = 0; k < 3; k++) { 
        l = (k+1) % 3;    
        m = (k+2) % 3;
    
        hl =  beta[k][l] * (1 - lambda[k][l]); 
        hm =  beta[k][m] * (1 - lambda[k][m]);
        hk = - hl - hm;
    
        if ( dof_u[k] < BOUNDARY_NAME ) {
            // k is interior vertex
            (*N)[dof_u[k]][dof_u[k]] += hk;
      
            if ( dof_u[l] < BOUNDARY_NAME )
                (*N)[dof_u[k]][dof_u[l]] += hl;
            else    
                (*rhs)[0][dof_u[k]] -= hl * eqn->dirichlet0(coord[l]);
      
      
            if ( dof_u[m] < BOUNDARY_NAME )
                (*N)[dof_u[k]][dof_u[m]] += hm;
            else
                (*rhs)[0][dof_u[k]] -= hm * eqn->dirichlet0(coord[m]);        
        }
    }
} /* discrConv */

void 
T2DConvDiffFE::discrConv ( T2DBasisEqn *eqn, TMatrixBlock *N, 
                           TBlockVector<TFloat> *rhs, 
                           TVectorBlock<TFloat> &suppsize ) {
    TFloat hk = 0.0, hl = 0.0, hm = 0.0;
    TFloat det, my, beta, lambda, vl, vm, F, s;
    TVector<2> b, r, e[3];
    const TFloat eps = 1e-8;
   
    e[0] = coord[2] - coord[1]; // edges  1 ->2
    e[1] = coord[0] - coord[2]; //        2 ->0 
    e[2] = coord[1] - coord[0]; //        0 ->1
   
    for ( uint k = 0; k < 3; k++ ) { 
        if ( dof_u[k] < BOUNDARY_NAME ) {  // k is interior vertex
            uint l = (k+1) % 3;    
            uint m = (k+2) % 3;
            hk = hl = hm = 0;
            b[0] = (*(eqn->getConvVector()))[0][dof_u[k]]; 
            b[1] = (*(eqn->getConvVector()))[1][dof_u[k]];
         
            beta = b.euclidNorm();
         
            // vl > 0 if -b is left from e[m]
            vl = b[0] * (e[m][1]) - (e[m][0]) * b[1]; 
            // e[l] has "wrong" direction!
            vm = -(b[0] * (e[l][1]) - (e[l][0]) * b[1]);
            if ((fabs(vl)<eps) && (e[m]*b<0) ){       
                // convection || e[m]
                F = 0.5 * beta / e[m].euclidNorm();  
                // 0.5: each of the two cells with this edge (e[m]) 
                // contributes to the matrix entry
                hk = F; hm = 0, hl = -F;
            }
         
            if ( (fabs(vm) < eps) && (e[l] * b > 0) ){        
                // convection || e[l]
                F = 0.5 * beta / e[l].euclidNorm();
                hk = F; hl = 0, hm = -F;
            }
         
            if ( (vl > eps) && (vm < -eps)  ) { 
                // convection (-1* b) between  e[m] and e[l]
                // solve lamda b + my e[k] = - e[m] by Cramer's rule
                det = b[0] * e[k][1] - e[k][0] * b[1];
                lambda = - (e[m][0] * e[k][1] - e[k][0] * e[m][1]) / det;
                my = -(b[0] * e[m][1] - e[m][0] * b[1]) / det;
                if (lambda > eps) {
                    F = 1 / fabs(lambda);
                    if ((my < 0) || (my > 1)) cout << "my falsch: "<<my<<endl;
                    hk = F; hl = (my-1) * F; hm = - my * F; }
            
            }
         
            s = suppsize[dof_u[k]] / 3.0;
            hk *= s; hl *= s; hm *= s;
         
            // k is interior vertex
            (*N)[dof_u[k]][dof_u[k]] += hk;
         
            if ( dof_u[l] < BOUNDARY_NAME )
                (*N)[dof_u[k]][dof_u[l]] += hl;
            else
                (*rhs)[0][dof_u[k]] -= hl * eqn->dirichlet0(coord[l]);
         
            if ( dof_u[m] < BOUNDARY_NAME )
                (*N)[dof_u[k]][dof_u[m]] += hm;
            else
                (*rhs)[0][dof_u[k]] -= hm * eqn->dirichlet0(coord[m]);
         
        }
      
    }
} /* discrConv */



void 
T2DConvDiffFE::getSuppSize (TVectorBlock<TFloat> &suppSize ) {
    for ( uint k = 0; k < 3; k++ ) {
        if ( dof_u[k] < BOUNDARY_NAME ) 
            suppSize[dof_u[k]] += 0.5 * detJ;
    }
}

