//
//  Project   : libnet
//  File      : convDirections.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : class for convection directions
//
// arch-tag: 24c388c2-f01b-4c13-ab13-7a970fa63ba1
//

#include "vector.hh"
#include "convDirections.hh"

//
// xline (x - direction)
//
TFloat xline  (TVector<3>) { return (1.0); }
TFloat noConv  (TVector<3>) { return (0.0); }

//
// no convection
//
TFloat noConv0  (TVector<2>) { return (0.0); }
TFloat noConv1  (TVector<2>) { return (0.0); }

//
// xline (x - direction)
//
TFloat xline0  (TVector<2>) { return (1.0); }
TFloat xline1  (TVector<2>) { return (0.0); }

//
// outboundXline
//
TFloat outboundXline0 (TVector<2> x) { return x[0] + 1.0;}
TFloat outboundXline1  (TVector<2> x) { return x[1] - 0.5;}

//
// circle centered at (0.5, 0.5)
//
TFloat circle0 ( TVector<2> v ) {
    TFloat x, y;
    //TFloat z;
    x = v[X]; y = v[Y];
    //z = sqrt( (x-0.5)*(x-0.5) + (y-0.5)*(y-0.5));
    //if ( fabs(z) < 0.000000001 ) return (0.0);
    //else return( (0.5 - y) / z );
    return( (0.5 - y) );
}
TFloat circle1 ( TVector<2> v ) {
    TFloat x, y;
    //TFloat z;
    x = v[X]; y = v[Y];
    //z = sqrt( (x-0.5)*(x-0.5) + (y-0.5)*(y-0.5));
    //if ( fabs(z) < 0.000000001 ) return (0.0);
    //else return( (x - 0.5) / z );
    return( (x - 0.5) );
}

// four circles like hotplates

TFloat fourcircles0 ( TVector<2> v )
{
    if (in(v,0,0.5,0,0.5))
        return(circle0(2.0*v));
    
    if (in(v,0.5,1,0,0.5))
    {
        v[X]-=0.5;
        return(-circle0(2.0*v));
    }
    
    if (in(v,0,0.5,0.5,1))
    {
        v[Y]-=0.5;
        return(-circle0(2.0*v));
    }
    
    if (in(v,0.5,1,0.5,1))
    {
        v[X]-=0.5;
        v[Y]-=0.5;
        return(circle0(2.0*v));
    }

    cerr << "TFloat fourcircles0: argument v" << v << " out or range: (0,1)^2" << endl;
    return 0.0;
}

TFloat fourcircles1 ( TVector<2> v ) {
    if (in(v,0,0.5,0,0.5)) return(circle1(2.0*v));
    if (in(v,0.5,1,0,0.5)){v[X]-=0.5; return(-circle1(2.0*v));}
    if (in(v,0,0.5,0.5,1)){v[Y]-=0.5; return(-circle1(2.0*v));}
    if (in(v,0.5,1,0.5,1)){v[X]-=0.5;v[Y]-=0.5; return(circle1(2.0*v));}

    cerr<<"TFloat fourcircles1: argument v"<<v<<" out or range: (0,1)^2"<<endl;
    return 0.0;

}


//
// curve (quarter of circle centered at (1.0, 1.0))
//
TFloat curve0 ( TVector<2> v )
{
    return( (1.0 - v[Y]) );
}
TFloat curve1 ( TVector<2> v )
{
    return( v[X] );
}


// square (circle centered at (0.5,0.5), but streched to the corners)
// cyclic convection only in a sector around the square's diagonals, otherwise convection 
// parallel to square's boundary. The sector has the angle "a"

TFloat square0 ( TVector<2> v ) {

    TFloat y = v[Y],x = v[X]; 
    TFloat Sx=0.5,Sy=0.5;
    TFloat w,a,b,Dx,Dy;
    TFloat P4 = PI/4.0;

    Dx=x-Sx;Dy=y-Sy;
    w=atan2(Dy,Dx);
    if (w<0) w+=2*PI;
  
    a=PI/2;a/=10.0;  
    b=P4-(a/2.0); 

    if (( 0<=w) && (w < b)) return 0;
    if (( b<=w) && (w < a+b)) return -sin(w);
    if (( a+b<=w) && (w < 3*b+a)) return -1;
    if (( 3*b+a<=w) && (w < 2*a+3*b)) return -sin(w);
    if (( 3*b+2*a<=w) && (w < 2*a+5*b)) return 0;
    if (( 5*b+2*a<=w) && (w < 3*a+5*b)) return -sin(w);
    if (( 3*a+5*b<=w) && (w < 7*b+3*a)) return 1;
    if (( 7*b+3*a<=w) && (w < 4*a+7*b)) return -sin(w);
    return 0;
  

}

TFloat square1 ( TVector<2> v ) {

    TFloat y = v[Y],x = v[X]; 
    TFloat Sx=0.5,Sy=0.5;
    TFloat w,a,b,Dx,Dy;
    TFloat P4 = PI/4.0;

    Dx=x-Sx;Dy=y-Sy;
    w=atan2(Dy,Dx);
    if (w<0) w+=2*PI;
  
    a=PI/2;a/=10.0;
    b=P4-(a/2.0); 

    if (( 0<=w) && (w < b)) return 1;
    if (( b<=w) && (w < a+b)) return cos(w);
    if (( a+b<=w) && (w < 3*b+a)) return 0;
    if (( 3*b+a<=w) && (w < 2*a+3*b)) return cos(w);
    if (( 3*b+2*a<=w) && (w < 2*a+5*b)) return -1;
    if (( 5*b+2*a<=w) && (w < 3*a+5*b)) return cos(w);
    if (( 3*a+5*b<=w) && (w < 7*b+3*a)) return 0;
    if (( 7*b+3*a<=w) && (w < 4*a+7*b)) return cos(w);
    return 1;
}

//
// triangle (quarter of circle centered at (1.0, 1.0))
//
TFloat triangle0 ( TVector<2> v ) {
    TFloat y = v[Y],x = v[X]; 
    TFloat Sx=2.0/3,Sy=1.0/3;
    TFloat w,wq=0.0,a,b,r,Dx,Dy;
    Dx=x-Sx;Dy=y-Sy;
    r=sqrt(Dx*Dx+Dy*Dy);
    w=atan2(Dy,Dx);
    if (w<0) w+=2*PI;
  
    a=atan2(2.0,1.0);b=2*PI-PI/2-2*a;

    TFloat P4 = PI/4.0;

    if ((w>=0) && ( w <= a)) wq=P4*w/a; 
    if ((w > a) && ( w <a+b)) wq= (w-a)*PI/b+P4; 
    if ((w > a+b) && ( w <2*a+b)) wq= (w-a-b)*P4/a+5*P4; 
    if (w >=2*a+b) wq=w;

    return (-sin(wq));


}
TFloat triangle1 ( TVector<2> v ) {
    TFloat y = v[Y],x = v[X]; 
    TFloat Sx=2.0/3,Sy=1.0/3;
    TFloat w,wq=0.0,a,b,r,Dx,Dy;
    Dx=x-Sx;Dy=y-Sy;
    r=sqrt(Dx*Dx+Dy*Dy);
    w=atan2(Dy,Dx);
    if (w<0) w+=2*PI;
  
    a=atan2(2.0,1.0);b=2*PI-PI/2.0-2*a;

    TFloat P4 = PI/4.0;

    if ((w>=0) && ( w <= a)) wq=P4*w/a; 
    if ((w > a) && ( w <a+b)) wq= (w-a)*PI/b+P4; 
    if ((w > a+b) && ( w <2*a+b)) wq= (w-a-b)*P4/a+5*P4; 
    if (w >=2*a+b) wq=w;

    return (cos(wq));

}


// four squares like hotplates

TFloat foursquares0 ( TVector<2> v ){

    if (in(v,0,0.5,0,0.5)) return(square0(2.0*v));
    if (in(v,0.5,1,0,0.5)){v[X]-=0.5; return(-square0(2.0*v));}
    if (in(v,0,0.5,0.5,1)){v[Y]-=0.5; return(-square0(2.0*v));}
    if (in(v,0.5,1,0.5,1)){v[X]-=0.5;v[Y]-=0.5; return(square0(2.0*v));}

    cerr<<"TFloat foursquares0: argument v"<<v<<" out or range: (0,1)^2"<<endl;
    return 0.0;

}

TFloat foursquares1 ( TVector<2> v ){

    if (in(v,0,0.5,0,0.5)) return(square1(2.0*v));
    if (in(v,0.5,1,0,0.5)){v[X]-=0.5; return(-square1(2.0*v));}
    if (in(v,0,0.5,0.5,1)){v[Y]-=0.5; return(-square1(2.0*v));}
    if (in(v,0.5,1,0.5,1)){v[X]-=0.5;v[Y]-=0.5; return(square1(2.0*v));}

    cerr<<"TFloat foursquares1: argument v"<<v<<" out or range: (0,1)^2"<<endl;
    return 0.0;

}

//
// no reaction
//
TFloat noReac  (TVector<2>) { return (0.0); }
//
// one reaction
//
TFloat oneReac  (TVector<2>) {return (1.0);}


// help functions

// v inside the square [a1,a2) x [b1,b2)  ?

bool in(TVector<2> v, TFloat a1,TFloat a2,TFloat b1,TFloat b2){
    if ((v[X]<a1) || (v[X] >=a2)) return FALSE;
    if ((v[Y]<b1) || (v[Y] >=b2)) return FALSE;
    return TRUE;
}

