//
//  Project   : libnet
//  File      : stokesFE.cc
//  Author    : Thomas Probst / Sabine Gutsch
//  Purpose   : class for linear conforming 2D finite elements
//
// arch-tag: 881088d9-7c23-4cdf-9e8d-983695f50ca5
//

#include "stokesFE.hh"

// constructors	
T2DStokesFE::T2DStokesFE ( TDiscrType aDiscrType) 
        : T2DBasisFE( aDiscrType ) 
{
	PHI = simple; 
}
	
// destructor
T2DStokesFE::~T2DStokesFE () {}

// get coordinates of vertices of children cells
void 
T2DStokesFE::getCoord ( TCell<2> *cell ) 
{
	uint i;
	for ( i = 0; i < 3; i++) pCoord[i] = (*cell->getNode(i));

	noOfChildren = cell->getNoOfChildren();
	for ( i = 0; i < noOfChildren; i++) 
		for ( uint k = 0; k < 3; k++) {
			coord[i][k] = (*cell->getChild(i)->getNode(k));
			dof_u[i][k] = cell->getChild(i)->getNode(k)->getName();
		}
}

void 
T2DStokesFE::getCoord ( TNet<2> *net, TDynArray< TVector<2> * >& arrCoordU )
{
    uint i, k;
    ulong dof;
    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
	for (TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next())
	{
		noOfChildren = cell->getNoOfChildren();
		for ( i = 0; i < noOfChildren; i++) 
			for ( k = 0; k < 3; k++) 
			{
				dof = cell->getChild(i)->getNode(k)->getName();
                if ( dof < BOUNDARY_NAME ) 
                {
                    if (arrCoordU.getSize() < dof+1)
                        arrCoordU.setSize( dof + 1 );
                    arrCoordU[dof] = cell->getChild(i)->getNode(k);
                }// if
			}
	}
}

void 
T2DStokesFE::getCoord ( TNet<2> *net, TDynArray< TVector<2> * >& arrCoordU,
						TDynArray< TVector<2> * >& arrCoordP ) 
{
    uint i, k;
    ulong dof,dof_p;
    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
	for (TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next())
	{
		noOfChildren = cell->getNoOfChildren();
	    for ( i = 0; i < noOfChildren; i++) 
		    for ( k = 0; k < 3; k++) 
			{
			    dof = cell->getChild(i)->getNode(k)->getName();
                if ( dof < BOUNDARY_NAME ) 
                {
                    if (arrCoordU.getSize() < dof+1)
                        arrCoordU.setSize( dof + 1 );
                    arrCoordU[dof] = cell->getChild(i)->getNode(k);
                }// if
            }// for
        for ( i = 0; i < 3; i++) 
		{
            dof_p = cell->getNode(i)->getName();
            if ( dof_p > BOUNDARY_NAME ) 
                dof_p = ULONG_MAX - dof_p + cInterior;
            if (arrCoordP.getSize() < dof_p+1)
                arrCoordP.setSize( dof_p + 1 );
            arrCoordP[dof_p] = cell->getNode(i);
        }

	}
}

void 
T2DStokesFE::getVertices ( TNet<2> *net, T2DVertex **VerU ){
    uint i, k;
    ulong dof;
    typedef TCell<2> TNetCell;
    TListIterator<TNetCell>  iter(net->getList());
	for (TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next()){
		noOfChildren = cell->getNoOfChildren();
	    for ( i = 0; i < noOfChildren; i++) 
		    for ( k = 0; k < 3; k++) {
			    dof = cell->getChild(i)->getNode(k)->getName();
                if ( dof < BOUNDARY_NAME ) 
                    VerU[dof] = cell->getChild(i)->getNode(k);
			}
      
	}
}



void 
T2DStokesFE::discrConv ( T2DBasisEqn *eqn, TNet<2> *net, 
                         TMatrixBlock *N, TBlockVector<TFloat> *rhs ) 
{  
	// BOUNDARY_NAME (should be ... - fineNet->getNoOfBoundaryNodes();
	BOUNDARY_NAME = ULONG_MAX - 2 * net->getNoOfBoundaryNodes();

    TVectorBlock<TFloat> suppSize( 0.0, (*rhs)[0].getSize());
	typedef TCell<2> TNetCell;

	TListIterator<TNetCell>  iter(net->getList());
    TCell<2> *cell;
	for (cell = iter.first(); cell != NULL; cell = iter.next()) {
		initDiscr ( cell ); 
        getSuppSize( suppSize );
    }

	for (cell = iter.first(); cell != NULL; cell = iter.next()) {
            getCoord( cell );
            if ( DiscrType == Tabata ) 
               discrConv ( eqn, N, rhs, suppSize );
            else
               discrConv ( eqn, N, rhs );
		}

}       

// calculate element stiffness matrix
void 
T2DStokesFE::discrDiff ( T2DBasisEqn *eqn, TBlockMatrix *A, 
                         TBlockVector<TFloat> *rhs ) 
{
	TFloat h = 0.0;
   uint i, j, k;
	for ( k = 0; k < noOfChildren; k++) 
		for ( i = 0; i < 3; i++) {
         // update rhs
         if ( dof_u[k][i] < BOUNDARY_NAME ) { // i is interior vertex
            (*rhs)[0][dof_u[k][i]] += eqn->f0(coord[k][i]) * detJ[k] / 6;
            (*rhs)[1][dof_u[k][i]] += eqn->f1(coord[k][i]) * detJ[k] / 6;
         }
			for (j = 0; j < 3; j++) 
				if ( dof_u[k][i] < BOUNDARY_NAME ) {
					h = eqn->getEpsilon() * gradT[k][i] * gradT[k][j] 
                        * 0.5 * detJ[k];
					if ( dof_u[k][j] < BOUNDARY_NAME ) {
						// j is interior vertex ---> entry in matrix 
						// entry in first diagonal block (Au)
						(*A)[0][0][dof_u[k][i]][dof_u[k][j]] += h;
						// entry in second diagonal block (Av)
						(*A)[1][1][dof_u[k][i]][dof_u[k][j]] += h; 
					}
					else {
						// j is BOUNDARY vertex ---> entry in rhs
						(*rhs)[0][dof_u[k][i]] -= h * eqn->dirichlet0(coord[k][j]);
						(*rhs)[1][dof_u[k][i]] -= h * eqn->dirichlet1(coord[k][j]);
					}
				}
      }
}

void 
T2DStokesFE::discrUP ( T2DBasisEqn *eqn, TBlockMatrix *A,
                       TBlockVector<TFloat> *rhs) 
{
   TFloat hu = 0.0, hv = 0.0;
   //TFloat coarseTriangleSize = 0.0;
   uint i, j, k;

   /*
   for ( k = 0; k < noOfChildren; k++) {
      coarseTriangleSize += detJ[k];
      for ( j = 0; j < 3; j++) 
         for ( i = 0; i < 3; i++) 
            if ( dof_u[k][i] < BOUNDARY_NAME ) { 
               hu = - detJ[k] * gradT[k][i][X] * c[j][k] / 6;
               hv = - detJ[k] * gradT[k][i][Y] * c[j][k] / 6;
               
               (*A)[0][2][dof_u[k][i]][dof_p[j]] += hu;  
               (*A)[1][2][dof_u[k][i]][dof_p[j]] += hv; 
            }
   }
   */   

   // entries caused by Lagrange factor to make p unique
   /*
   coarseTriangleSize /= 6;
   ulong pIndex  = (*A)[2][2].getNoOfRows() - 1;
   for ( j = 0; j < 3; j++) { 
      (*A)[2][2][dof_p[j]][pIndex] += coarseTriangleSize; 
      (*A)[2][2][pIndex][dof_p[j]] += coarseTriangleSize; 
   }
   */

   for ( k = 0; k < noOfChildren; k++) {
		// for noOfChildren triangles
		for ( j = 0; j < 3; j++) 
			// for 3 course grid nodes
			for ( i = 0; i < 3; i++) {
				// for 3 fine grid nodes
				// i is interior node
            hu = - detJ[k] * gradT[k][i][X] * c[k][j] / 6;
            hv = - detJ[k] * gradT[k][i][Y] * c[k][j] / 6;

            if ( dof_u[k][i] < BOUNDARY_NAME ) { 

               (*A)[0][2][dof_u[k][i]][dof_p[j]] += hu;  
               (*A)[1][2][dof_u[k][i]][dof_p[j]] += hv; 

               (*A)[2][0][dof_p[j]][dof_u[k][i]] += hu;  
               (*A)[2][1][dof_p[j]][dof_u[k][i]] += hv; 
            }
            else {
               (*rhs)[2][dof_p[j]] -= hu * eqn->dirichlet0(coord[k][i]);
               (*rhs)[2][dof_p[j]] -= hv * eqn->dirichlet1(coord[k][i]);
            }
         }
   }
}

// calculate global (lumped) mass matrix and write it into global vector M
void 
T2DStokesFE::discrMass ( T2DBasisEqn *eqn, TNet<2> *net,  
                         TVectorBlock<TFloat> *M) {
        
        BOUNDARY_NAME = ULONG_MAX - 2 * net->getNoOfBoundaryNodes();

        typedef TCell<2> TNetCell;
		TListIterator<TNetCell>  iter(net->getList());
        TCell<2> *cell;
	    for (cell = iter.first(); cell != NULL; cell = iter.next()) {
              initDiscr ( cell );
              discrMass ( eqn, M);
            }
} // discrMass

// calculate element (lumped) mass matrix
void 
T2DStokesFE::discrMass ( T2DBasisEqn *eqn, TVectorBlock<TFloat> *M) {
  TFloat h = 0.0;
  uint i, j, k;
  for ( k = 0; k < noOfChildren; k++ )
      for ( i = 0; i < 3; i++) 
          for ( j = 0; j < 3; j++) 
              if ( dof_u[k][i] < BOUNDARY_NAME ) {
                  // i is interior vertex
                  h = (1.0) / (24.0) * detJ[k] * eqn->getReacc()(coord[k][i]) ;
                  if (j == 1) h *= 2;
                  (*M)[dof_u[k][i]] += h;
			  }
} // discrMass


// initialize grad, det, c, dof_u and dof_p for a particular cell
void 
T2DStokesFE::initDiscr ( TCell<2> *cell ) 
{
	// number of children
	noOfChildren = cell->getNoOfChildren();
   
	// initialize dof_u, dof_p
	for (uint i = 0; i < 3; i++) {
		dof_p[i] = cell->getNode(i)->getName();
		if ( dof_p[i] > BOUNDARY_NAME ) 
			dof_p[i] = ULONG_MAX - dof_p[i] + cInterior;
      
		// for (uint j = 0; j < noOfChildren; j++) {
		//    dof_u[j][i] = cell->getChild(j)->getNode(i)->getName();
		// }
	}
	// calculate determinants and their inverses
	// (triangles ordered as in refinement,
	// gradients ordered as vertices in (children) triangles)

	TFloat invdetJ[4];

    uint i, j, k; 
	for ( i = 0; i < noOfChildren; i++) {
		detJ[i] = 2 * cell->getChild(i)->getSize( );
		invdetJ[i] = 1 / detJ[i];
      
		getCoord( cell );

		gradT[i][0][X] = invdetJ[i] * (coord[i][1][Y]-coord[i][2][Y]);
		gradT[i][0][Y] = invdetJ[i] * (coord[i][2][X]-coord[i][1][X]);
		gradT[i][1][X] = invdetJ[i] * (coord[i][2][Y]-coord[i][0][Y]);
		gradT[i][1][Y] = invdetJ[i] * (coord[i][0][X]-coord[i][2][X]);
		gradT[i][2][X] = invdetJ[i] * (coord[i][0][Y]-coord[i][1][Y]);
		gradT[i][2][Y] = invdetJ[i] * (coord[i][1][X]-coord[i][0][X]);
	}

	// initialize c
	for ( i = 0; i < noOfChildren; i++) 
	    for ( j = 0; j < 3; j++) 
            c[i][j] = 0.0; 

	// set c
	for ( i = 0; i < noOfChildren; i++) 
		for( j = 0; j < 3; j++) {
			for( k = 0; k < 3; k++) 
				if ( cell->getChild(i)->getNode(j) == 
					 cell->getFrontier(k)->getNode(1) )
					// j is k'th midNode 
				{
					c[i][k]       += 0.5; 
					c[i][(k+1)%3] += 0.5; 
				}
			for( k = 0; k < 3; k++) 
				if ( cell->getChild(i)->getNode(j) == 
					 cell->getNode(k) )
					// j is k'th coarse node
					c[i][k] += 1.0; 
		}
}

// calculate global stiffnes matrix
void 
T2DStokesFE::discrDiff ( T2DBasisEqn *eqn, TNet<2> *net, TBlockMatrix *A, 
						 TBlockVector<TFloat> *rhs ) 
{ 
	// BOUNDARY_NAME (should be ... - fineNet->getNoOfBoundaryNodes();
	BOUNDARY_NAME = ULONG_MAX - 2 * net->getNoOfBoundaryNodes();
	cInterior     = net->getNoOfInteriorNodes();
	
	typedef TCell<2> TNetCell;
	TCell<2> *cell;
	TListIterator<TNetCell>  iter(net->getList());
	for (cell = iter.first(); cell != NULL; cell = iter.next()) {
		initDiscr ( cell );
		discrDiff ( eqn, A, rhs );
		discrUP ( eqn, A, rhs ); //tpr
	}
}        

// upwinding 
void 
T2DStokesFE::discrConv ( T2DBasisEqn *eqn, TMatrixBlock *N, 
                         TBlockVector<TFloat> *rhs ) 
{
	TFloat hk = 0.0, hl = 0.0, hm = 0.0;
	TFloat alpha, beta[3][3], lambda[3][3];
	TVector<2> b, r;
    uint j, k, l, m;

	for ( j = 0; j < noOfChildren; j++ ) {
		for ( k = 0; k < 3; k++ ) { 
			l = (k+1)%3;    
			m = (k+2)% 3;
			r[0] = (1.0/6.0)*( coord[j][k][1] + coord[j][l][1] -2*coord[j][m][1]);
			r[1] = (1.0/6.0)*(2*coord[j][m][0] - coord[j][k][0] -  coord[j][l][0]);
			alpha = SGN((r[0]*(coord[j][k][0]- coord[j][l][0])) +
					  (r[1]*(coord[j][k][1]- coord[j][l][1])));
			b[0] = (1.0/12.0)*(5*(*eqn->getConvVector())[0][dof_u[j][k]]+ 5*(*eqn->getConvVector())[0][dof_u[j][l]]+2*(*eqn->getConvVector())[0][dof_u[j][m]]);
			b[1] = (1.0/12.0)*(5*(*eqn->getConvVector())[1][dof_u[j][k]]+ 5*(*eqn->getConvVector())[1][dof_u[j][l]]+2*(*eqn->getConvVector())[1][dof_u[j][m]]);
                
			beta[l][k] = alpha*(b[0]*r[0]+b[1]*r[1]);			 
			beta[k][l] = -alpha*(b[0]*r[0]+b[1]*r[1]);			 
			lambda[k][l] = (*PHI)(beta[k][l]/eqn->getEpsilon());
			lambda[l][k] = 1-lambda[k][l];
		}
            
            
		for ( k = 0; k < 3; k++) { 
			l = (k+1)%3;    
            m =(k+2)% 3;
                
			hl =  beta[k][l]*(1-lambda[k][l]); 
			hm =  beta[k][m]*(1-lambda[k][m]);
			hk =  - hl - hm;
                
			if ( dof_u[j][k] < BOUNDARY_NAME ) {
				// k is interior vertex
				(*N)[dof_u[j][k]][dof_u[j][k]] += hk;
                    
				if ( dof_u[j][l] < BOUNDARY_NAME )
					(*N)[dof_u[j][k]][dof_u[j][l]] += hl;
				else {
					(*rhs)[0][dof_u[j][k]] -= hl* eqn->dirichlet0(coord[j][l]);
					(*rhs)[1][dof_u[j][k]] -= hl* eqn->dirichlet1(coord[j][l]);
				}
                    
				if ( dof_u[j][m] < BOUNDARY_NAME )
					(*N)[dof_u[j][k]][dof_u[j][m]] += hm;
				else {
					(*rhs)[0][dof_u[j][k]] -= hm* eqn->dirichlet0(coord[j][m]);
					(*rhs)[1][dof_u[j][k]] -= hm* eqn->dirichlet1(coord[j][m]);
				}
			}
                
		}
	}
} /* discrConv */

void 
T2DStokesFE::discrConv ( T2DBasisEqn *eqn, TMatrixBlock *N, 
                         TBlockVector<TFloat> *rhs, 
						 TVectorBlock<TFloat> &suppSize ) 
{ // Tabata
	TFloat hk = 0.0, hl = 0.0, hm = 0.0;
	TFloat det,my,beta, lambda,vl,vm,F,s;
	TVector<2> b, r, e[3];
	const TFloat eps = 1e-8;
    uint j, k, l, m;
   
   for ( j = 0; j < noOfChildren; j++ ) {
      e[0] = coord[j][2]-coord[j][1]; // edges  1 ->2
      e[1] = coord[j][0]-coord[j][2]; //        2 ->0 
      e[2] = coord[j][1]-coord[j][0]; //        0 ->1
      
		for ( k = 0; k < 3; k++ ) { 
         if ( dof_u[j][k] < BOUNDARY_NAME ) {	// k is interior vertex
            l = (k+1)%3;    
            m = (k+2)% 3;
            hk = hl = hm = 0;
            b[0]=(*eqn->getConvVector())[0][dof_u[j][k]];
            b[1]=(*eqn->getConvVector())[1][dof_u[j][k]];
            
            beta = b.euclidNorm();
            
            // vl > 0 if -b is left from e[m]
            vl = b[0]*(e[m][1])-(e[m][0])*b[1];             
            // e[l] has "wrong" direction!
            vm = -(b[0]*(e[l][1])-(e[l][0])*b[1]);
            // cout<<" vl: "<<vl<<" vm: "<<vm<<endl;
            if ( (fabs(vl) < eps) && ( e[m] * b < 0) ) {         
               // convection || e[m]
               F = 0.5 * beta / e[m].euclidNorm();  
               // 0.5: each of the two cell with this edge (e[m]) 
               // contributes to the matrix entry
               hk = F; hm = 0, hl = -F;
               //cout<<" F: "<<F<<endl;
            }
            
            if ( (fabs(vm) < eps ) && ( e[l] * b > 0) ) {        
               // convection || e[l]
               F = 0.5*beta/e[l].euclidNorm();
               hk = F; hl = 0, hm = -F;
            }
            
            if ( ( vl > eps ) && ( vm < -eps )  ) { 
               // convection (-1* b) between  e[m] and e[l]
               // solve lamda b + my e[k] = - e[m] by Cramer's rule
               det = b[0]*e[k][1]-e[k][0]*b[1];
               lambda = -(e[m][0]*e[k][1]-e[k][0]*e[m][1])/det;
               my = -(b[0]*e[m][1]-e[m][0]*b[1])/det;
               if ( lambda > eps ) {
                  F = 1/fabs(lambda);
                  if (( my < 0) || ( my > 1 )) 
                     cout<< "my falsch: "<<my<<endl;
                  hk = F;
                  hl = (my-1) * F; 
                  hm = - my * F;
               }
            }
            
            s = suppSize[ dof_u[j][k] ] / 3.0;
            hk *= s; 
            hl *= s;
            hm *= s;
            
         	// k is interior vertex
				(*N)[dof_u[j][k]][dof_u[j][k]] += hk;
            
				if ( dof_u[j][l] < BOUNDARY_NAME )
					(*N)[dof_u[j][k]][dof_u[j][l]] += hl;
				else {
					(*rhs)[0][dof_u[j][k]] -= hl * eqn->dirichlet0(coord[j][l]);
					(*rhs)[1][dof_u[j][k]] -= hl * eqn->dirichlet1(coord[j][l]);
				}
				if ( dof_u[j][m] < BOUNDARY_NAME )
					(*N)[dof_u[j][k]][dof_u[j][m]] += hm;
				else {
					(*rhs)[0][dof_u[j][k]] -= hm * eqn->dirichlet0(coord[j][m]);
					(*rhs)[1][dof_u[j][k]] -= hm * eqn->dirichlet1(coord[j][m]);       
				}
         }
         
         /*      cout.setf(ios::fixed);
                 cout << dof_u[i][k] << " " << dof_u[i][l] << " " << dof_u[i][m] << "   ";
                 cout << k << " " << l << " " << m << "   ";
                 cout << coord[i][k] << " " << coord[i][l] << " " <<coord[i][m] << "   ";
                 cout << setprecision(5) <<hk << " " << hl <<  " " << hm<<"\n";*/
      }
   }
} /* discrConv */

void 
T2DStokesFE::getSuppSize (TVectorBlock<TFloat> &suppSize ) 
{
   uint j, k; 
	for ( j = 0; j < noOfChildren; j++ ) 
		for ( k = 0; k < 3; k++ )  
         if ( dof_u[j][k] < BOUNDARY_NAME ) 
            suppSize[dof_u[j][k]] += 0.5 * detJ[j]; 
}
