//
//  Project   : libnet
//  File      : stokesHierarchy.cc
//  Author    : Thomas Probst / Sabine Gutsch
//  Purpose   : class to build a matrix and prolongation hierarchy for
//              the Stokes equations
//
// arch-tag: be41c859-8c77-4109-82be-4d9cf0d029da
//

#include "stokesHierarchy.hh"

T2DStokesHierarchy::T2DStokesHierarchy ()
{ 
	pN       = NULL; 
    noOfNets = 0; 
}

T2DStokesHierarchy::~T2DStokesHierarchy ()
{
	if (pN){
		for (uint i = 0; i < noOfNets-1; i++ ) {
			if (pN[i]) delete pN[i];
			if (pR[i]) delete pR[i];
		}
	}
}

void 
T2DStokesHierarchy::buildHierarchy ( TBasisDoFManager<2> *dofManager,
                                     T2DBasisEqn *eqn,
                                     T2DBasisFE *fe,
                                     TNetHierarchy<2> *netHier, 
                                     TBlockMatrix **pMatrix, 
                                     TBlockVector<TFloat> **pRHS, 
                                     TBlockVector<TFloat> **pX )
{   
    TDiscrType discrType = fe->getDiscrType();
    noOfNets = netHier->getNoOfNets();

    uint noOfInteriorNodesNet[noOfNets-1];
    uint noOfBoundaryNodesNet[noOfNets-1];

    uint blockSize[3];
    TDynArray< TVector<2> * > coordVec;
    
	T2DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
    
    uint i;
	T2DNet *cNet  = iterNet.first();
	T2DNet *fNet  = iterNet.next();
   
    if (discrType != Null){
        pN = new TMatrixBlock * [noOfNets];
        pR = new TBlockVector<TFloat> * [noOfNets];
    }
 // allocation of a vector for
        // matrix blocks  pointers for the convective part;

    noOfInteriorNodesNet[0] = cNet->getNoOfInteriorNodes();
    noOfBoundaryNodesNet[0] = cNet->getNoOfBoundaryNodes();

    for (i = 0; i < noOfNets-1; i++) 
	{
        noOfInteriorNodesNet[i+1] = fNet->getNoOfInteriorNodes();
        noOfBoundaryNodesNet[i+1] = fNet->getNoOfBoundaryNodes();
		if ( noOfInteriorNodesNet[i] == 0 ) 
		{
			pMatrix[i] = NULL;
			pRHS[i] = NULL;
			pX[i] = NULL;
            if (discrType != Null) { pN[i] = NULL; pR[i] = NULL; }
		}// if
		else 
		{
			pMatrix[i] = dofManager->buildStiffnessMatrix(cNet, fNet);
			blockSize[0] = noOfInteriorNodesNet[i+1];
			blockSize[1] = noOfInteriorNodesNet[i+1];
            // '+1' is the Lagrange multiplier
			//blockSize[2] = noOfInteriorNodesNet[i] + noOfBoundaryNodesNet[i] + 1;    
			blockSize[2] = noOfInteriorNodesNet[i] + noOfBoundaryNodesNet[i];    
			pRHS[i] = new TBlockVector<TFloat>( 3, blockSize );
			pX[i] = new TBlockVector<TFloat>( 0.0, *pRHS[i] );
            
            // allocate place for the convective part:
            if (discrType != Null) {
               pN[i] = new TMatrixBlock( (*pMatrix[i])[0][0] );
               pR[i] = new TBlockVector<TFloat>( 0.0, *pRHS[i] );
            }
            
            fe->discrDiff( eqn, cNet, pMatrix[i], pRHS[i]);
            // in the startup phase, pX[i] has zero components
            if (discrType != Null) {  
                eqn->setConvVector( pX[i] ); 
      
				// if a convection direction is defined, set pX
                if ( eqn->getConvb0() != noConv0 ) {
                    fe->getCoord( cNet, coordVec );
                    eqn->initConvVector( blockSize[0], coordVec );
				}
                               
                fe->discrConv( eqn, cNet, pN[i], pR[i]);
                (*(pMatrix[i]))[0][0] += *(pN[i]); // update matrix
                (*(pMatrix[i]))[1][1] += *(pN[i]); // update matrix
                (*pRHS[i])[0] += (*pR[i])[0];
                (*pRHS[i])[1] += (*pR[i])[1]; // update rhs
			}
            // reaction term != zero
            if ( eqn->getReacc() !=  noReac) { 
              TBlockVector<TFloat> *pM = new TBlockVector<TFloat>( 0.0, *pX[i]);
              // lumped mass matrix in vector pM
              fe->discrMass(eqn, cNet, &((*pM)[0])); 
              uint j;
              for (j = 0; j < blockSize[0]; j++) {
                (*pMatrix[i])[0][0][j][j] += (*pM)[0][j];
                (*pMatrix[i])[1][1][j][j] += (*pM)[0][j];
			  }
              delete pM;
            }
            

		}// else
		
		cNet = fNet;
		fNet = iterNet.next();
	}// for
}


void 
T2DStokesHierarchy::updateHierarchy ( T2DBasisEqn *eqn,
                                      T2DBasisFE *fe,
                                      TNetHierarchy<2> *netHier, 
                                      TBlockMatrix **pMatrix, 
                                      TBlockVector<TFloat> **pRHS, 
                                      TBlockVector<TFloat> **pX )

{ 
    TDiscrType discrType = fe->getDiscrType();
    noOfNets = netHier->getNoOfNets();

    uint noOfInteriorNodesNet[noOfNets-1];
    uint noOfBoundaryNodesNet[noOfNets-1];
	
    T2DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
    
    uint i;
    T2DNet *cNet  = iterNet.first();
    T2DNet *fNet  = iterNet.next();
    
    noOfInteriorNodesNet[0] = cNet->getNoOfInteriorNodes();
    noOfBoundaryNodesNet[0] = cNet->getNoOfBoundaryNodes();

    for (i = 0; i < noOfNets-1; i++) 
	{
        noOfInteriorNodesNet[i+1] = fNet->getNoOfInteriorNodes();
        noOfBoundaryNodesNet[i+1] = fNet->getNoOfBoundaryNodes();
		if ( noOfInteriorNodesNet[i] == 0 ) 
		{
			
		}// if
		else 
		{
		 
            // in the startup phase, pX[i] has zero components
            if (discrType != Null) {  

                //restore old matrix (so its only the diffusive part)
                (*(pMatrix[i]))[0][0] -= *(pN[i]);
                (*(pMatrix[i]))[1][1] -= *(pN[i]);
			    // calculate new convection
                (*pR[i]) *= 0;
                (*pN[i]) *= 0;
                eqn->setConvVector( pX[i] );
                fe->discrConv(eqn, cNet, pN[i], pR[i]);

                // update matrix
                (*(pMatrix[i]))[0][0] += *(pN[i]); 
                (*(pMatrix[i]))[1][1] += *(pN[i]); 
			    // update rhs
                (*pRHS[i])[0] += (*pR[i])[0];
                (*pRHS[i])[1] += (*pR[i])[1]; 

				// adjust LU decomposition on coarse grid
                if ( pMatrix[i]->hasLUdecomp() ) 
				{
                    pMatrix[i]->deleteLUdecomp();
                    pMatrix[i]->calcStokesLUdecomp1();
				}
            }
            

		}// else
		
		cNet = fNet;
		fNet = iterNet.next();
	}// for
}
