//
//  Project     : libgraph
//  File        : T3DCellEdgeGraph.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a graph consisting of celledges
//
// arch-tag: 0d92b9a8-d027-4b2c-a620-9ce5d8c4a648
//

#include "timer.hh"
#include "var.hh"

#include "net.hh"
#include "T3DCellEdge.hh"
#include "T3DFrontier.hh"

#include "TConvEqn.hh"

#include "matrixblock.hh"

#include "T3DVertexData.hh"
#include "T3DCellEdgeNode.hh"

#include "T3DCellEdgeGraph.hh"

// flags for edges
#define TREATED       0x1
#define BOUNDARY_EDGE 0x2

// flags for faces provided face has leq 8 edges
#define FACE_USED      0x100
#define BOUNDARY_FACE  0x200
#define INTERIOR_FACE  0x400
#define FACE_STRONG    0x800

//////////////////////////////////////////////////
//
// constructors and destructors
//

//
// constructor
//
T3DCellEdgeGraph::T3DCellEdgeGraph ()
        : TGraph ()
{
    pMatrixBlock = NULL;
}

T3DCellEdgeGraph::T3DCellEdgeGraph ( TNet<3> * pNet, TConvEqn<3> * pEqn,
                                     TFloat kappa, TFloat gamma )
{
    T3DCellEdgeGraph::kappa = kappa;
    T3DCellEdgeGraph::gamma = gamma;
    
    pMatrixBlock = NULL;

    pFSNode = NULL;
    
    buildGraph( pNet, pEqn );
}

T3DCellEdgeGraph::T3DCellEdgeGraph ( TNet<3> * pNet, TConvEqn<3> * pEqn,
                                     TMatrixBlock * pMatrix,
                                     TFloat kappa, TFloat sigma, TFloat gamma )
{
    T3DCellEdgeGraph::kappa = kappa;
    T3DCellEdgeGraph::sigma = sigma;
    T3DCellEdgeGraph::gamma = gamma;
    
    pMatrixBlock = pMatrix;

    if (pMatrixBlock == NULL)
    {
        cout << "T3DCellEdgeGraph::T3DCellEdgeGraph (...) : got NULL matrix"
             << endl;
        exit( 1 );
    }// if
    
    pFSNode = NULL;
    
    buildGraph( pNet, pEqn );
}

//
// destructor
//
T3DCellEdgeGraph::~T3DCellEdgeGraph ()
{
}

//////////////////////////////////////////////////
//
// virtual constructors
//

//
// create new graph
//
T3DCellEdgeGraph *
T3DCellEdgeGraph::clone ()
{
    T3DCellEdgeGraph *pGraph = new T3DCellEdgeGraph;

    *pGraph = *this;

    return pGraph;
}

//////////////////////////////////////////////////
//
// misc. stuff
//

//
// access node in metagraph
//
TFSNode *
T3DCellEdgeGraph::fsNode () const
{
    return pFSNode;
}

void
T3DCellEdgeGraph::setFSNode ( TFSNode * pNode )
{
    pFSNode = pNode;
}
    

//////////////////////////////////////////////////
//
// encapsulate creation of graph
//

void
T3DCellEdgeGraph::buildGraph ( TNet<3> * pNet, TConvEqn<3> * pEqn )
{
    TTimer                       timer;
    
    uint                         i,j;
    TListIterator< TCell<3> >    iterCells;
    TCell<3>                    *pCell;
    T3DFrontier                 *pFace;
    T3DCellEdge                 *pCellEdge;
    uint                         noOfFaces, noOfEdges;
    uint                        nodeName;
    T3DCellEdgeNode             *pNode, *pAdjNode;
    TVertex<3>                  *v0, *v1;
    TVector<3>                   b;
    TEdge                       *pEdge;
    TDynArray< T3DVertexData * > arrVertexData;
    uint                         size;
    T3DVertexData               *pVData0, *pVData1;
    bool                         bStrong, bBoundaryFace;
    
    if ((pNet == NULL) || (pEqn == NULL))
    {
        cout << "T3DCellEdgeGraph::buildNet (...) : gave NULL arg" << endl;
        return;
    }// if

    //
    // we create a vertexdata for every net-node to store
    // add. data
    // first init. array holding the vertexdata objects
    // (index == node.name)
    //
    size = pNet->getNoOfInteriorNodes();
    arrVertexData.setSize( size );
    
    for (i = 0; i < size; i++)
        arrVertexData[i] = NULL;

    //
    // reset attributes in net
    //
    iterCells.setList( pNet->getList() );
    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFaces = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFaces; i++)
        {
            pFace = (T3DFrontier*) pCell->getFrontier( i );

			pFace->clearFlag( FACE_USED | BOUNDARY_FACE | INTERIOR_FACE );
            noOfEdges = pFace->getNoOfEdges();

            for (j = 0; j < noOfEdges; j++)
			{
                pCellEdge = pFace->getEdge( j );
                pCellEdge->setData( NULL );
                pCellEdge->clearFlag( TREATED | BOUNDARY_EDGE );
			}// for
		}// for
	}// while
    
    //
    // go over all cells and every face in the cell and
    // add every edge as a node to the graph.
    // also add every face as an adjacent face to the
    // edge.
    //

    timer.start();
    
    nodeName = 0;

    iterCells.setList( pNet->getList() );
    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFaces = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFaces; i++)
        {
            pFace = (T3DFrontier*) pCell->getFrontier( i );

            if ((pFace == NULL) || pFace->hasFlag( FACE_USED ))
                continue;

            pFace->setFlag( FACE_USED );
            bBoundaryFace = isBoundaryFace( pFace );
            
            noOfEdges = pFace->getNoOfEdges();

            for (j = 0; j < noOfEdges; j++)
            {
                if ((pCellEdge = pFace->getEdge( j )) == NULL)
                    continue;
                
                pNode = NULL;
                
                // check if we treated this edge before
                if ( ! pCellEdge->hasFlag( TREATED ) )
                {
                    pCellEdge->setFlag( TREATED );
                    
                    // build conv-dir at edge-midpoint
                    v0 = pCellEdge->getNode( 0 );
                    v1 = pCellEdge->getNode( 2 );

                    if ((v0 == NULL) || (v1 == NULL))
                    {
                        cerr << "T3DCellEdgeGraph::buildGraph (...) : have NULL node"
                             << endl;
                        exit( -1 );
                    }// if
                    
                    if ((v0->getBoundary() != INTERIOR) ||
                        (v1->getBoundary() != INTERIOR))
                    {
                        // no boundary edges till now
                        pCellEdge->setFlag( BOUNDARY_EDGE );
                        continue;
                    }// if

                    // check if edge is strong
                    bStrong = isStrongEdge( pCellEdge, pEqn, & b );

                    if (bStrong)
                        pFace->setFlag( FACE_STRONG );
                    
                    // catch vertexdata for net-nodes
                    if ((pVData0 = arrVertexData[v0->getName()]) == NULL)
                    {
                        pVData0 = new T3DVertexData( v0 );
                        arrVertexData[v0->getName()] = pVData0;
                    }// if
                    
                    if ((pVData1 = arrVertexData[v1->getName()]) == NULL)
                    {
                        pVData1 = new T3DVertexData( v1 );
                        arrVertexData[v1->getName()] = pVData1;
                    }// if
                    
                    // add node to graph
                    pNode = new T3DCellEdgeNode( nodeName++, pCellEdge,
                                                 pVData0, pVData1, &b );
                    pNode->setStrong( bStrong );
                    add( *pNode );
                    
                    // save created node in edge
                    pCellEdge->setData( (T3DCellEdge::data_t) pNode );
                }// if
                else
                {
                    // if edge not on boundary catch created node
                    if ( ! pCellEdge->hasFlag( BOUNDARY_EDGE ) )
                        pNode = (T3DCellEdgeNode*) pCellEdge->data();
                }// else        


                // add face if not in facelist of edge
                if ((pNode != NULL) && ( ! bBoundaryFace ))
                    pNode->addFace( pFace );
            }// for
        }// for
    }// while

    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "T3DCellEdgeGraph::buildGraph (...) : time for building nodes = "
             << timer << endl;

    //
    // the graph consists of all strong nodes (cell-edges)
    // we now created edges by evaluating the phi_f/b-functions for
    // every node
    // - outedges are defined by phi_f
    // - inedges are defined by phi_b
    //
    T3DFrontier             *pPhi_f, *pPhi_b;
    TListIterator< TNode >   iterNodes;
    T3DCellEdge             *pNodeEdge;

    timer.start();
    
    iterNodes.setList( getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        //
        // remove all weak frontiers
        //
        pNode->remWeakFaces();
    
        if (pNode->isWeak())
            continue;
        
        pNode->calcPhi( gamma );

        pNodeEdge = pNode->cellEdge();
        
        //
        // evaluate phi_f
        //
        pPhi_f = pNode->phi_f();

        if (pPhi_f != NULL)
        {
            noOfEdges = pPhi_f->getNoOfEdges();

            for (i = 0; i < noOfEdges; i++)
            {
                if ((pCellEdge = pPhi_f->getEdge( i )) == pNodeEdge)
                    continue;

                if ((pAdjNode = (T3DCellEdgeNode*) pCellEdge->data()) != NULL)
                {
                    pEdge = new TEdge( *pNode, *pAdjNode );
                    pNode->addOutEdge( *pEdge );
                }// if
            }// for
        }// if

        //
        // evaluate phi_b
        //
        pPhi_b = pNode->phi_b();

        if (pPhi_b != NULL)
        {
            noOfEdges = pPhi_b->getNoOfEdges();

            for (i = 0; i < noOfEdges; i++)
            {
                if ((pCellEdge = pPhi_b->getEdge( i )) == pNodeEdge)
                    continue;

                if ((pAdjNode = (T3DCellEdgeNode*) pCellEdge->data()) != NULL)
                {
                    pEdge = new TEdge( *pAdjNode, *pNode );
                    pNode->addInEdge( *pEdge );
                }// if
            }// for
        }// if

        // save space : remove all other faces in facelist
        if ((pPhi_f != NULL) || (pPhi_b != NULL))
            pNode->remFaceList();
    }// while

    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "T3DCellEdgeGraph::buildGraph (...) : time for building edges = "
             << timer << endl;

    //
    // to be done : clean up
    //   - remove marks of all celledges
    //
}

bool
T3DCellEdgeGraph::isStrongEdge ( T3DCellEdge * pEdge,
                                 TConvEqn<3> * pEqn,
                                 TVector<3> * pVecConvDir ) const
{
    if ((pEdge == NULL) || (pEqn == NULL) || (pVecConvDir == NULL))
    {
        cout << "T3DCellEdgeGraph::isStrongEdge (...) : gave NULL arg" << endl;
        return FALSE;
    }// if

    //
    // calc conv-dir at edgemidpoint to export data
    // to external variable
    //
    static TVector<3>  *v0, *v1;
    static TVector<3>   m;
    static TVector<3>   b;
    static TVector<3>   d;
    
    v0 = pEdge->getNode( 0 );
    v1 = pEdge->getNode( 2 );
    
    // calc dir of the edge
    d  = *v1;
    d -= *v0;
    
    // calc edge-midpoint
    m  = d;
    m *= 0.5;
    m += *v0;
    
    // calc conv-dir at midpoint
    pEqn->evalConv( &m, &b );
    
    // copy convdir to external data
    *pVecConvDir = b;

    //
    // if matrix is present, use it to check strongness
    //
//     if (pMatrixBlock != NULL)
//     {
//         uint    name0, name1;
//         TFloat  diag;

//         name0 = pEdge->getNode( 0 )->getName();
//         name1 = pEdge->getNode( 2 )->getName();

//         diag  = (*pMatrixBlock)[name0][name0];

//         if (fabs((*pMatrixBlock)[name0][name1]) > fabs(sigma * diag))
//             return TRUE;
//         else
//             return FALSE;
//     }// if
//     else
    {
        //
        // decide if edge is strong by calc the angle of the
        // conv-dir at edge-midnode with the dir of the edge
        //
    
        TFloat       norm;
        TFloat       cosAngle, angle;

        // normalize
        if ((norm = d.euclidNorm()) != 0)
            d *= 1 / norm;
        else
            return FALSE;
    
        if ((norm = b.euclidNorm()) != 0)
            b *= 1 / norm;
        else
            return FALSE;

        cosAngle = b * d;

        if (cosAngle > 1)
            cosAngle = 1;
        if (cosAngle < -1)
            cosAngle = -1;

        angle = acos( cosAngle );
    
        // apply correction to angle
//      b -= d;
//      norm = b.euclidNorm();
//      if ((norm > SQRT2) || (norm < 0))
//          angle = PI2 - angle;
    
        if ((angle >= kappa) && ((PI - angle) >= kappa))
        {
            if (norm > sigma)
                return TRUE;
            else
                return FALSE;
        }// if
        else
            return FALSE;
    }// else
}

bool
T3DCellEdgeGraph::isBoundaryFace ( T3DFrontier * pFace ) const
{
    if (pFace == NULL)
    {
        cout << "T3DCellEdgeGraph::isBoundaryFace (...) : got NULL face" << endl;
        return FALSE;
    }// if
    
    if (pFace->getBoundary() != INTERIOR)
        return TRUE;
    
    uint         i;
    uint         noOfNodes;
    TVertex<3>  *pVertex;

    noOfNodes = pFace->getNoOfNodes();
    for (i = 0; i < noOfNodes; i++)
    {
        if ((pVertex = pFace->getNode(i)) != NULL)
        {
            if (pVertex->getBoundary() != INTERIOR)
                return TRUE;
        }// if
    }// for

    return FALSE;
}

//
// check if face is used
//
bool
T3DCellEdgeGraph::isFaceUsed ( T3DFrontier * pFace ) const
{
    return pFace->hasFlag( FACE_USED );
}

//
// mark face as used
//
void
T3DCellEdgeGraph::setFaceUsed ( T3DFrontier * pFace ) const
{
    pFace->setFlag( FACE_USED );
}
