//
//  Project     : libgraph
//  File        : T3DGraph.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a graph in 3D
//
// arch-tag: af26683e-82b5-4305-abe0-f6dd14940570
//

#include "net.hh"
#include "cell.hh"
#include "T3DFrontier.hh"
#include "T3DCellEdge.hh"

#include "matrixblock.hh"

#include "T3DMarkableNode.hh"
#include "T3DCellEdgeNode.hh"
#include "T3DCellEdgeGraph.hh"
#include "criterion.hh"

#include "T3DGraph.hh"

// mark for used objects
#define USED  1

///////////////////////////////////////////
//
// constructor and destructor
//

T3DGraph::T3DGraph ()
{
}

T3DGraph::T3DGraph ( TNet<3> * pNet, TMatrixBlock * pMatrix,
                     uint criterion, TFloat xi )
{
    T3DGraph::criterion = criterion;
    T3DGraph::xi        = xi;
    T3DGraph::pMatrix   = pMatrix;

    if (pMatrix == NULL)
    {
        cout << "T3DGraph::T3DGraph (...) : got NULL matrix" << endl;
        exit( 1 );
    }// if
    
    buildGraph( pNet );
}

T3DGraph::T3DGraph ( TNet<3> * pNet,
                     T3DCellEdgeGraph * pEdgeGraph,
                     TMatrixBlock * pMatrix,
                     uint criterion, TFloat xi )
{
    T3DGraph::criterion = criterion;
    T3DGraph::xi        = xi;
    T3DGraph::pMatrix   = pMatrix;

    if (pMatrix == NULL)
    {
        cout << "T3DGraph::T3DGraph (...) : got NULL matrix" << endl;
        exit( 1 );
    }// if
    
    buildGraph( pNet, pEdgeGraph );
}

T3DGraph::~T3DGraph ()
{
}

//////////////////////////////////////////////////
//
// virtual constructors
//

//
// create new graph
//
T3DGraph *
T3DGraph::clone ()
{
    T3DGraph *pGraph = new T3DGraph;

    *pGraph = *this;

    return pGraph;
}

///////////////////////////////////////////
//
// encapsulate graph building
//

void
T3DGraph::buildGraph ( TNet<3> * pNet )
{
    if (pNet == NULL)
    {
        cout << "T3DGraph::buildGraph (...) : got NULL arg" << endl;
        return;
    }// if

    uint                       i, j;
    uint                       size, noOfEntries;
    uint                     * entries, diagIdx;
    TDynArray< TNode * >        arrNodes;
    TNode                     * pNode, * pAdjNode;
    TEdge                     * pEdge;
    TDynArray< TVector<3> * > * pArrCoord;

    size = pMatrix->getSize();

    arrNodes.setSize( size );

    //
    // catch coord for all nodes
    //

    pArrCoord = getCoord( pNet );

    if (pArrCoord == NULL)
        cout << "T3DGraph::buildGraph (...) : got NULL vector-array" << endl;
        
    if (pArrCoord->getSize() < size)
        cout << "T3DGraph::buildGraph (...) : missing some vectors" << endl;
    
    //
    // create node for each entry in the matrix
    //
    for ( i = 0; i < size; i++ )
    {
        pNode = new T3DMarkableNode( i, (*pArrCoord)[i] );

        add( * pNode );
        arrNodes[i] = pNode;
    }// for

    //
    // create edges between nodes
    //
    for ( i = 0; i < size; i++ )
    {
        noOfEntries = (*pMatrix)[i].getNoOfEntries();
        entries = (*pMatrix)[i].getIndex();
        diagIdx = entries[0];

        pNode = arrNodes[diagIdx];

        for ( j = 1; j < noOfEntries; j++ )
        {
            pAdjNode = arrNodes[ entries[j] ];
            
            if ( isStrong( pNode, pAdjNode ) )
            {
                pEdge = new TEdge( * pNode, * pAdjNode );

                pNode->addOutEdge( * pEdge );
                pAdjNode->addInEdge( * pEdge );
            }// if
        }// for
    }// for

    //
    // clean up
    //

    delete pArrCoord;
}

void
T3DGraph::buildGraph ( TNet<3> * pNet,
                       T3DCellEdgeGraph * pEdgeGraph )
{
    if (pEdgeGraph == NULL)
    {
        cout << "T3DGraph::buildGraph (...) : got NULL arg" << endl;
        return;
    }// if

    uint                     i;
    TListIterator< TNode >   iterNodes;
    T3DCellEdgeNode        * pEdgeNode;
    T3DVertexData          * pNodeData0, * pNodeData1;
    T3DMarkableNode        * pNode0 = NULL, * pNode1 = NULL;
    bool                     bNoEdge;
    uint                    name0, name1;

    TDynArray< T3DMarkableNode * >  arrNodes;
    uint                           size;

    //
    // to check if a graphnode was already created for the
    // given netnode, we store each node in an array
    //
    size = pNet->getNoOfInteriorNodes();
    arrNodes.setSize( size );
    
    for (i = 0; i < size; i++)
        arrNodes[i] = NULL;
             
    //
    // add all nodes, which do not belong to a FS in a sep. block
    //
    iterNodes.setList( pEdgeGraph->getNodeList() );
    iterNodes.first();
    while ((pEdgeNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        if ( ! pEdgeNode->isInSurface() )
        {
            bNoEdge = FALSE;
            
            //
            // extract/build nodes
            //
            pNodeData0 = pEdgeNode->nodeData( 0 );
            name0      = pNodeData0->node()->getName();
            
            if (pNodeData0->inSurface() ||
                (pNodeData0->node()->getBoundary() != INTERIOR))
            {
                bNoEdge = TRUE;
            }// if
            else
            {
                if ((pNode0 = arrNodes[name0]) == NULL)
                {
                    pNode0 = new T3DMarkableNode( name0, pNodeData0->node() );

                    add( * pNode0 );
                    arrNodes[name0] = pNode0;
                }// if
            }// else

            pNodeData1 = pEdgeNode->nodeData( 1 );
            name1      = pNodeData1->node()->getName();
            
            if (pNodeData1->inSurface() ||
                (pNodeData1->node()->getBoundary() != INTERIOR))
            {
                bNoEdge = TRUE;
            }// if
            else
            {
                if ((pNode1 = arrNodes[name1]) == NULL)
                {
                    pNode1 = new T3DMarkableNode( name1, pNodeData1->node() );

                    add( * pNode1 );
                    arrNodes[name1] = pNode1;
                }// if
            }// else

            if ( ! bNoEdge )
            {
                //
                // build edge if strong
                //
                if (isStrong( pNode0, pNode1 ))
                    buildEdge( pNode0, pNode1 );

                if (isStrong( pNode1, pNode0 ))
                    buildEdge( pNode1, pNode0 );
            }// if
        }// if
    }// while
}

//
// check if given celledge lies (partially) on a boundary
//
bool
T3DGraph::onBoundary ( T3DCellEdge * pEdge ) const
{
    if (pEdge == NULL)
    {
        cout << "T3DGraph::onBoundary (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    if ((pEdge->getBoundary() != INTERIOR) ||
        (pEdge->getNode( 0 )->getBoundary() != INTERIOR) ||
        (pEdge->getNode( 2 )->getBoundary() != INTERIOR))
        return TRUE;
    else
        return FALSE;
}

//
// look if edge from node0 to node1 is a strong edge
//
bool
T3DGraph::isStrong ( TNode * pNode0, TNode * pNode1 ) const
{
    if ((pNode0 == NULL) || (pNode1 == NULL))
    {
        cout << "T3DGraph::isStrong (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    uint   nodeName0, nodeName1;
    TFloat  diagValue, conValue0, conValue1;

    nodeName0 = pNode0->getName();
    nodeName1 = pNode1->getName();

    if (criterion == E3)
    {
        diagValue = (*pMatrix)[nodeName0][nodeName0];
        conValue0 = (*pMatrix)[nodeName0][nodeName1];
        
        if (fabs(conValue0) > fabs(xi * diagValue))
            return TRUE;
        else
            return FALSE;
    }// if

    if (criterion == E7)
    {
        conValue0 = (*pMatrix)[nodeName0][nodeName1];
        conValue1 = (*pMatrix)[nodeName1][nodeName0];
    
        if (fabs(conValue0) > fabs(xi * conValue1))
            return TRUE;
        else
            return FALSE;
    }// if

    cout << "T3DGraph::isStrong (...) : criterion not supported" << endl;
    exit( 1 );
    
    return FALSE;
}
    
//
// create an edge from node0 to node1
//
void
T3DGraph::buildEdge ( TNode * pNode0, TNode * pNode1 )
{
    TEdge  * pEdge;

    pEdge = new TEdge( * pNode0, * pNode1 );

    pNode0->addOutEdge( * pEdge );
    pNode1->addInEdge(  * pEdge );
}

//
// return array with geometry info about all nodes
//
TDynArray< TVector<3> * > *
T3DGraph::getCoord ( TNet<3> * pNet )
{
    if (pNet == NULL)
    {
        cout << "T3DGraph::getCoord (...) : got NULL arg" << endl;
        return NULL;
    }// if

    uint                        i;
    uint                        size;
    TDynArray< TVector<3> * > * pArrCoord;
    TListIterator< TCell<3> >   iterCells;
    TCell<3>                  * pCell;
    TVertex<3>                * pNode;
    uint                       nodeName;

    pArrCoord = new TDynArray< TVector<3> * >;

    size = pNet->getNoOfInteriorNodes();
    pArrCoord->setSize( size );
    
    for ( i = 0; i < size; i++ )
        (*pArrCoord)[i] = NULL;

    iterCells.setList( pNet->getList() );
    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        //
        // we assume cell to be a tetrahedron : 4 nodes
        //
        for ( i = 0; i < 4; i++ )
        {
            if ((pNode = pCell->getNode( i )) == NULL)
                continue;

            // check if node is interior node of net
            if (pNode->getBoundary() != INTERIOR)
                continue;
            
            nodeName = pNode->getName();
            
            if ((*pArrCoord)[nodeName] != NULL)
                continue;

            (*pArrCoord)[nodeName] = pNode;
        }// for
    }// while

    return pArrCoord;
}
