//
//  Project     : libgraph
//  File        : T3DSurfaceGraph.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a graph defined by flow surfaces
//
// arch-tag: 797b92c6-9553-4300-b98b-4bac8ab6a15b
//

#include "matrixblock.hh"

#include "vertex.hh"

#include "T3DCellEdgeGraph.hh"
#include "T3DCellEdgeNode.hh"
#include "T3DPlanarNode.hh"
#include "T3DVertexData.hh"
#include "T3DFaceOrient.hh"
#include "criterion.hh"

#include "T3DSurfaceGraph.hh"

/////////////////////////////////////
//
// constructor and destructor
//

T3DSurfaceGraph::T3DSurfaceGraph ( T3DCellEdgeGraph * pFlowSurface,
                                   TMatrixBlock * pMatrix,
                                   uint criterion, TFloat xi,
                                   bool bSortOFD )
{
    pMatrixBlock = pMatrix;

    if (pMatrixBlock == NULL)
    {
        cout << "T3DSurfaceGraph::T3DSurfaceGraph (...) : got NULL matrix" << endl;
        exit( 1 );
    }// if

    T3DSurfaceGraph::criterion = criterion;
    T3DSurfaceGraph::xi        = xi;

    if (bSortOFD)
        buildGraph_OFD( pFlowSurface );
    else
        buildGraph( pFlowSurface );
}

T3DSurfaceGraph::~T3DSurfaceGraph ()
{
}

/////////////////////////////////////
//
// encapsulate graph building
//

void
T3DSurfaceGraph::buildGraph ( T3DCellEdgeGraph * pFlowSurface )
{
    if (pFlowSurface == NULL)
    {
        cout << "T3DSurfaceGraph::buildGraph (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode        *pEdgeNode;
    T3DPlanarNode          *pNode0, *pNode1;

    //
    // iterate over flowsurface
    //
    iterNodes.setList( pFlowSurface->getNodeList() );
    iterNodes.first();
    while ((pEdgeNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        if ( ! pEdgeNode->isInSurface() )
        {
            cout << "T3DSurfaceGraph::buildGraph (...) : edge not in surface" << endl;
            continue;
        }// if
        
        //
        // check for every edge in flowsurface if nodes
        // are already in surfacegraph and if edge is strong
        //

        pNode0 = buildNode( pEdgeNode->nodeData( 0 ) );
        pNode1 = buildNode( pEdgeNode->nodeData( 1 ) );
        
        if (isStrong( pNode0, pNode1 ))
            buildEdge( pNode0, pNode1 );
        
        if (isStrong( pNode1, pNode0 ))
            buildEdge( pNode1, pNode0 );
    }// while
}

void
T3DSurfaceGraph::buildGraph_OFD ( T3DCellEdgeGraph * pFlowSurface )
{
    if (pFlowSurface == NULL)
    {
        cout << "T3DSurfaceGraph::buildGraph (...) : got NULL arg" << endl;
        return;
    }// if

    uint                    i;
    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode        *pEdgeNode;
    T3DPlanarNode          *pNode;

    //
    // iterate over flowsurface
    //
    iterNodes.setList( pFlowSurface->getNodeList() );
    iterNodes.first();
    while ((pEdgeNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        if ( ! pEdgeNode->isInSurface() )
        {
            cout << "T3DSurfaceGraph::buildGraph (...) : edge not in surface" << endl;
            continue;
        }// if

        //
        // build nodes and edges for each cellnode in celledge
        //
        for (i = 0; i < 2; i++)
        {
            pNode = buildNode( pEdgeNode->nodeData( i ) );

            if ( ! pNode->isSorted() )
                buildEdges_OFD( pNode, pEdgeNode );

            pNode->setSorted( TRUE );
        }// for
    }// while
}

//
// check if given edge is strong
//
const bool
T3DSurfaceGraph::isStrong ( T3DPlanarNode * pNode0,
                            T3DPlanarNode * pNode1 ) const
{
    if ((pNode0 == NULL) || (pNode1 == NULL))
    {
        cout << "T3DSurfaceGraph::isStrong (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    uint   nodeName0, nodeName1;
    TFloat  diagValue, conValue0, conValue1;

    nodeName0 = pNode0->getName();
    nodeName1 = pNode1->getName();

    if (criterion == E3)
    {
        diagValue = (*pMatrixBlock)[nodeName0][nodeName0];
        conValue0 = (*pMatrixBlock)[nodeName0][nodeName1];
        
        if (fabs(conValue0) > fabs(xi * diagValue))
            return TRUE;
        else
            return FALSE;
    }// if

    if (criterion == E7)
    {
        conValue0 = (*pMatrixBlock)[nodeName0][nodeName1];
        conValue1 = (*pMatrixBlock)[nodeName1][nodeName0];
    
        if (fabs(conValue0) > fabs(xi * conValue1))
            return TRUE;
        else
            return FALSE;
    }// if

    cout << "T3DSurfaceGraph::isStrong (...) : criterion not supported" << endl;
    exit( 1 );
    
    return FALSE;
}

//
// build a new edge from node0 to node1
//
void
T3DSurfaceGraph::buildEdge ( T3DPlanarNode * pNode0,
                             T3DPlanarNode * pNode1 ) const
{
    TWeightedMarkableEdge  * pEdge, * pNewEdge;
    TFloat                   weight;
    TListIterator< TEdge >   iterEdges;

    weight = (*pMatrixBlock)[pNode0->getName()][pNode1->getName()];

    pNewEdge = new TWeightedMarkableEdge( * pNode0, * pNode1, NO_MARK, weight );
    pNewEdge->addRef();
    pNode1->addInEdge ( * pNewEdge );

    //
    // insert edge according to weight
    //
    iterEdges.setList( pNode0->getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() < weight)
        {
            pNode0->getOutEdges().insertBefore( * pEdge, * pNewEdge );
            return;
        }// if
    }// while

    pNode0->getOutEdges().add( * pNewEdge );
}

//
// build node for given nodedata or simply return node
// if already builded
T3DPlanarNode *
T3DSurfaceGraph::buildNode ( T3DVertexData * pData )
{
    if (pData == NULL)
    {
        cout << "T3DSurfaceGraph::buildNode (...) : got NULL arg" << endl;
        return NULL;
    }// if

    if (pData->surfaceNode() == NULL)
    {
        T3DPlanarNode * pNode;

        pNode = new T3DPlanarNode( pData->node()->getName(),
                                   NO_MARK,
                                   pData->node() );

        add( * pNode );
        pData->setSurfaceNode( pNode );
    }// if

    return (T3DPlanarNode*) pData->surfaceNode();
}

//
// build edges for given node sorted by angle (OFD)
// (pNode = v and pCellEdge = e)
//
void
T3DSurfaceGraph::buildEdges_OFD ( T3DPlanarNode * pNode,
                                  T3DCellEdgeNode * pCellEdge )
{
    if ((pNode == NULL) || (pCellEdge == NULL))
    {
        cout << "T3DSurfaceGraph::buildEdges_OFD (...) : got NULL arg" << endl;
        return;
    }// if

    TList< T3DPlanarNode >  listAdjNodes;
    T3DPlanarNode         * pAdjNode = NULL, * pNode0, * pNode1;
    T3DCellEdgeNode       * pStartEdge, * pEdge, * pNewEdge;
    T3DFrontier           * pStartFace, * pFace, * pNewFace;
    bool                    bReachedEnd, bReverse;

    bReverse = bReachedEnd = FALSE;
    
    //
    // find node v != u \in e and adjacent face f with
    // f.orient = *uv*
    // (it does not matter if *uv* or *vu*, we just have
    //  to be consistent in further actions)
    //

    pEdge = pStartEdge = pCellEdge;
    
    pNode0 = buildNode( pEdge->nodeData( 0 ) );
    pNode1 = buildNode( pEdge->nodeData( 1 ) );

    if (pNode0 == pNode)
        pAdjNode = pNode1;
    else if (pNode1 == pNode)
        pAdjNode = pNode0;

    addEdge( pNode, pEdge, & listAdjNodes, bReverse );

    pFace = getFaceByOrient( pEdge, pAdjNode, pNode );
    
    if (pFace == NULL)
    {
        //
        // no face with f.orient = *uv* present, so we turn
        // the orientation and begin with other face
        // (we reached one end of the closed(!) sector)
        //

        pFace = getFaceByOrient( pEdge, pNode, pAdjNode );

        if (pFace == NULL)
        {
            cout << "T3DSurfaceGraph::buildEdges_OFD (...) : no adjacent face found" << endl;
            exit( 1 );
        }// if

        pNewEdge = otherCellEdge( pNode, pEdge, pFace );
        
        addEdge( pNode, pNewEdge, & listAdjNodes, bReverse );

        pEdge = pNewEdge;
        
        bReachedEnd = TRUE;
    }// if

    pStartFace = pFace;

    //
    // go through all adjacent faces of v and add edges
    //
    while (TRUE)
    {
        pNewFace = nextFace( pEdge, pFace );

        if (pNewFace == NULL)
        {
            // check if we reached second end of sector
            // and all is done
            if (bReachedEnd == TRUE)
                break;

            bReverse = bReachedEnd = TRUE;
            
            pNewFace = pStartFace;
            pNewEdge = otherCellEdge( pNode, pStartEdge, pNewFace );
        }// if
        else
        {
            if (pNewFace == pStartFace)
                break;

            pNewEdge = otherCellEdge( pNode, pEdge, pNewFace );
        }// else
        
        addEdge( pNode, pNewEdge, & listAdjNodes, bReverse );
        
        pEdge = pNewEdge;
        pFace = pNewFace;
    }// while

    //
    // build edges of node in graph
    //
    buildSortedEdges( pNode, & listAdjNodes );
}

//
// return adjacent face of celledge with orientation
// compatible to (pNode0,pNode1)
//
T3DFrontier *
T3DSurfaceGraph::getFaceByOrient ( T3DCellEdgeNode * pCellEdge,
                                   T3DPlanarNode * pNode0,
                                   T3DPlanarNode * pNode1 ) const
{
    if ((pCellEdge == NULL) || (pNode0 == NULL) || (pNode1 == NULL))
    {
        cout << "T3DSurfaceGraph::getFaceByOrient (...) : got NULL arg" << endl;
        return NULL;
    }// if

    uint           i;
    T3DFrontier  * pFace;

    for ( i = 0; i < 2; i++ )
    {
        pFace = pCellEdge->fsFace( i );

        if (pFace != NULL)
        {
            if (T3DFaceOrient::cmpOrient( pFace,
                                          pNode0->vector(),
                                          pNode1->vector() ) == TRUE)
                return pFace;
        }// if
    }// for

    return NULL;
}

//
// return other (than e) celledge in face with common node v
//
T3DCellEdgeNode *
T3DSurfaceGraph::otherCellEdge ( T3DPlanarNode * pNode,
                                 T3DCellEdgeNode * pCellEdge,
                                 T3DFrontier * pFace ) const
{
    if ((pNode == NULL) || (pCellEdge == NULL) || (pFace == NULL))
    {
        cout << "T3DSurfaceGraph::otherCellEdge (...) : got NULL arg" << endl;
        return NULL;
    }// if

    uint              i;
    uint              noOfEdges;
    T3DCellEdgeNode * pOtherEdge;
    T3DCellEdge     * pEdge;
    uint             nodeName;

    nodeName  = pNode->getName();
    noOfEdges = pFace->getNoOfEdges();
    
    for ( i = 0; i < noOfEdges; i++ )
    {
        pEdge = pFace->getEdge( i );
        
        if ((pOtherEdge = (T3DCellEdgeNode*) pEdge->data()) != NULL)
        {
            if (pOtherEdge == pCellEdge)
                continue;

            if (pOtherEdge->nodeData( 0 )->node()->getName() == nodeName)
                return pOtherEdge;

            if (pOtherEdge->nodeData( 1 )->node()->getName() == nodeName)
                return pOtherEdge;
        }// if
        else
            cout << "T3DSurfaceGraph::otherCellEdge (...) : ERROR" << endl;
    }// for

    cout << "T3DSurfaceGraph::otherCellEdge (...) : no other edge found" << endl;
    
    return NULL;
}

//
// return face adjacent to given edge and not the given face
//
T3DFrontier *
T3DSurfaceGraph::nextFace ( T3DCellEdgeNode * pCellEdge,
                            T3DFrontier * pFace ) const
{
    if ((pCellEdge == NULL) || (pFace == NULL))
    {
        cout << "T3DSurfaceGraph::nextFace (...) : got NULL arg" << endl;
        return NULL;
    }// if
    
    if (pFace == pCellEdge->fsFace( 0 ))
        return pCellEdge->fsFace( 1 );
    
    if (pFace == pCellEdge->fsFace( 1 ))
        return pCellEdge->fsFace( 0 );

    return NULL;
}

//
// add edge to list of incident edges of node
//
void
T3DSurfaceGraph::addEdge ( T3DPlanarNode * pNode,
                           T3DCellEdgeNode * pEdge,
                           TList< T3DPlanarNode > * pList,
                           bool bReverse )
{
    if ((pNode == NULL) || (pEdge == NULL) || (pList == NULL))
    {
        cout << "T3DSurfaceGraph::addEdge (...) : got NULL arg" << endl;
        return;
    }// if

    T3DPlanarNode  * pAdjNode;

    pAdjNode = buildNode( pEdge->nodeData( 0 ) );
    if (pAdjNode == pNode)
        pAdjNode = buildNode( pEdge->nodeData( 1 ) );

    if (pAdjNode == NULL)
    {
        cout << "T3DSurfaceGraph::addEdge (...) : have no adjacent node" << endl;
        return;
    }//if
    
    if (bReverse)
        pList->add( * pAdjNode );
    else
        pList->insert( * pAdjNode );
}

//
// build edges of node on graph w.r.t. OFD-condition
// out of given nodelist of adjacent nodes
//
void
T3DSurfaceGraph::buildSortedEdges ( T3DPlanarNode * pNode,
                                    TList< T3DPlanarNode > * pList )
{
    if ((pNode == NULL) || (pList == NULL))
    {
        cout << "T3DSurfaceGraph::buildSortedEdges (...) : got NULL arg" << endl;
        return;
    }// if

    if (pList->getSize() == 0)
        return;
    
    TListIterator< TEdge >          iterEdges;
    TList< TEdge >                  listInEdges_Old, listOutEdges_Old;
    TList< TEdge >                  listInEdges, listInEdges_Tmp;
    TList< TEdge >                  listOutEdges, listOutEdges_Tmp;
    TListIterator< T3DPlanarNode >  iterAdjNodes;
    T3DPlanarNode                 * pAdjNode;
    TEdge                         * pEdge;
    bool                            bFirstIsInEdge, bFirstIsOutEdge;
    bool                            bInEdge, bOutEdge;

    iterAdjNodes.setList( * pList );
    iterAdjNodes.first();

    bFirstIsInEdge = bFirstIsOutEdge = FALSE;

    //
    // copy all existing edges into lists and remove
    // the edgelists from node
    //
    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listInEdges_Old.add( * pEdge );
        pNode->getInEdges().remove( * pEdge );
    }// while
    
    iterEdges.setList( pNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listOutEdges_Old.add( * pEdge );
        pNode->getOutEdges().remove( * pEdge );
    }// while
    
    //
    // look if first edge in list is in- or out-edge
    // (if both is true, we look also for second edge)
    //

    iterAdjNodes.first();
    while ((pAdjNode = iterAdjNodes++) != NULL)
    {
        if ( isStrong( pAdjNode, pNode ) )
            bFirstIsInEdge = TRUE;
        
        if ( isStrong( pNode, pAdjNode ) )
            bFirstIsOutEdge = TRUE;

        if (bFirstIsInEdge || bFirstIsOutEdge)
            break;
    }// while
    
    if ( bFirstIsInEdge && bFirstIsOutEdge )
    {
        while ((pAdjNode = iterAdjNodes++) != NULL)
        {
            bInEdge  = isStrong( pAdjNode, pNode );
            bOutEdge = isStrong( pNode, pAdjNode );

            if ( bInEdge & bOutEdge )
            {
                cout << "T3DSurfaceGraph::buildSortedEdges (...) : no ofd condition at node "
                     << pNode->getName() << endl;
                
                bFirstIsInEdge  = TRUE;
                bFirstIsOutEdge = FALSE;
                break;
            }// if
            
            if ( bInEdge )
            {
                bFirstIsInEdge  = FALSE;
                bFirstIsOutEdge = TRUE;
                break;
            }// if
            
            if ( bOutEdge )
            {
                bFirstIsInEdge  = TRUE;
                bFirstIsOutEdge = FALSE;
                break;
            }// if
        }// while
    }// if

    //
    // build edges and check OFD-condition
    //

    if ( bFirstIsInEdge )
    {
        bool  bNoOutEdge, bAgainInEdge;

        bAgainInEdge = FALSE;
        bNoOutEdge   = TRUE;
        
        iterAdjNodes.first();
        while ((pAdjNode = iterAdjNodes++) != NULL)
        {
            bInEdge  = isStrong( pAdjNode, pNode );
            bOutEdge = isStrong( pNode, pAdjNode );
            
            if ( bInEdge && bNoOutEdge )
                listInEdges.add( * buildInEdge( pNode, pAdjNode, & listInEdges_Old ) );

            if ( bOutEdge && ! bAgainInEdge )
                listOutEdges.add( * buildOutEdge( pNode, pAdjNode, & listOutEdges_Old ) );

            if ( bInEdge && ! bNoOutEdge )
                listInEdges_Tmp.add( * buildInEdge( pNode, pAdjNode, & listInEdges_Old ) );

            if ( bOutEdge && bAgainInEdge )
            {
                cout << "T3DSurfaceGraph::buildSortedEdges (...) : no ofd condition at node "
                     << pNode->getName() << endl;
                listOutEdges.add( * buildOutEdge( pNode, pAdjNode, & listOutEdges_Old ) );
            }// else

            // now set flags so we do not conflict with upper if's
            if ( bOutEdge && ! bAgainInEdge )
                bNoOutEdge = FALSE;

            if ( bInEdge && ! bNoOutEdge )
                bAgainInEdge = TRUE;
        }// while
        
        iterEdges.setList( listInEdges );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addInEdge( * pEdge );
        }// while

        iterEdges.setList( listInEdges_Tmp );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addInEdge( * pEdge );
        }// while

        iterEdges.setList( listOutEdges );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addOutEdge( * pEdge );
        }// while
    }// if
    else if ( bFirstIsOutEdge )
    {
        bool  bNoInEdge, bAgainOutEdge;

        bAgainOutEdge = FALSE;
        bNoInEdge     = TRUE;
        
        iterAdjNodes.first();
        while ((pAdjNode = iterAdjNodes++) != NULL)
        {
            bOutEdge = isStrong( pNode, pAdjNode );
            bInEdge  = isStrong( pAdjNode, pNode );
            
            if ( bOutEdge && bNoInEdge )
                listOutEdges.add( * buildOutEdge( pNode, pAdjNode, & listOutEdges_Old ) );

            if ( bInEdge && ! bAgainOutEdge )
                listInEdges.add( * buildInEdge( pNode, pAdjNode, & listInEdges_Old ) );

            if ( bOutEdge && ! bNoInEdge )
                listOutEdges_Tmp.add( * buildOutEdge( pNode, pAdjNode, & listOutEdges_Old ) );

            if ( bInEdge && bAgainOutEdge )
            {
                cout << "T3DSurfaceGraph::buildSortedEdges (...) : no ofd condition at node "
                     << pNode->getName() << endl;
                listInEdges.add( * buildInEdge( pNode, pAdjNode, & listInEdges_Old ) );
            }// else

            // now set flags so we do not conflict with upper if's
            if ( bInEdge && ! bAgainOutEdge )
                bNoInEdge = FALSE;
            
            if ( bOutEdge && ! bNoInEdge )
                bAgainOutEdge = TRUE;
        }// while
        
        iterEdges.setList( listOutEdges );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addOutEdge( * pEdge );
        }// while

        iterEdges.setList( listOutEdges_Tmp );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addOutEdge( * pEdge );
        }// while

        iterEdges.setList( listInEdges );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pNode->addInEdge( * pEdge );
        }// while
    }// else if
}

//
// create edge or return edge if in given edgelist
TEdge *
T3DSurfaceGraph::buildInEdge  ( T3DPlanarNode  * pNode,
                                T3DPlanarNode  * pAdjNode,
                                TList< TEdge > * pList )
{
    if ((pNode == NULL) || (pAdjNode == NULL) || (pList == NULL))
    {
        cout << "T3DSurfaceGraph::buildInEdge (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TEdge >  iterEdges( * pList );
    TEdge                 * pEdge;

    //
    // look in given list for edge
    //
    
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        if (pEdge->getTail() == pAdjNode)
        {
            pList->remove( * pEdge );
            // we have to remove a reference because
            // we add this edge again to the node
            // (see buildSortedEdges(...))
            pEdge->delRef();
            return pEdge;
        }// if
    }// while

    //
    // edge not in list; create edge an add edge to adj. node
    // not to node itself (will be added in buildSortedEdges(...) )
    //

    pEdge = new TWeightedMarkableEdge( * pAdjNode, * pNode );
    pAdjNode->addOutEdge( * pEdge );

    return pEdge;
}

TEdge *
T3DSurfaceGraph::buildOutEdge ( T3DPlanarNode  * pNode,
                                T3DPlanarNode  * pAdjNode,
                                TList< TEdge > * pList )
{
    if ((pNode == NULL) || (pAdjNode == NULL) || (pList == NULL))
    {
        cout << "T3DSurfaceGraph::buildOutEdge (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TEdge >  iterEdges( * pList );
    TEdge                 * pEdge;

    //
    // look in given list for edge
    //
    
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        if (pEdge->getHead() == pAdjNode)
        {
            pList->remove( * pEdge );
            // we have to remove a reference because
            // we add this edge again to the node
            // (see buildSortedEdges(...))
            pEdge->delRef();
            return pEdge;
        }// if
    }// while

    //
    // edge not in list; create edge an add edge to adj. node
    // not to node itself (will be added in buildSortedEdges(...) )
    //

    pEdge = new TWeightedMarkableEdge( * pNode, * pAdjNode );
    pAdjNode->addInEdge( * pEdge );

    return pEdge;
}
