//
//  Project     : libgraph
//  File        : TPlanarNode.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a node in a planar graph
//
// arch-tag: af216a8a-88a2-462e-a378-bcf60deffd6d
//

#include "TPlanarCycle.hh"
#include "graphtool.hh"

#include "TPlanarNode.hh"

// what cycles to detect
#define MIN_CYC   0
#define MAX_CYC   1

// used flags
#define VISITED         1
#define MIN_DELETED     2
#define MAX_DELETED     4
#define IN_STACK        8

/////////////////////////////////////////////////
//
// constructor and destructor
//
    
TPlanarNode::TPlanarNode ( TMarkableNode & node,
                           TPlanarCycle * pMinCyc,
                           TPlanarCycle * pMaxCyc, 
                           TFloat w )
        : TMarkableNode(node), TFlagable<uchar>( 0 )
{
    pMinCycle = pMinCyc;
    pMaxCycle = pMaxCyc;

    _weight = w;
    
    pListBlock = NULL;
}

TPlanarNode::TPlanarNode ( uint n, int m,
                           TPlanarCycle * pMinCyc,
                           TPlanarCycle * pMaxCyc, 
                           TFloat w )
        : TMarkableNode( n, m ), TFlagable<uchar>( 0 )
{
    pMinCycle = pMinCyc;
    pMaxCycle = pMaxCyc;

    _weight = w;
    
    pListBlock = NULL;
}

TPlanarNode::~TPlanarNode ()
{
    if (pListBlock != NULL)
    {
        TListIterator<TNode>   iterNodes( * pListBlock );
        TNode                * pNode;

        while ((pNode = iterNodes++) != NULL)
        {
            pListBlock->remove( *pNode );

            if (pNode != this)
                pNode->free();
        }// while

        delete pListBlock;
    }// if
}

/////////////////////////////////////////////////
//
// access member variables
//
    
TPlanarCycle *
TPlanarNode::cycle ( uint type )
{
    switch (type)
    {
        case MIN_CYC : return pMinCycle; break;
        case MAX_CYC : return pMaxCycle; break;
        default      : return pMinCycle;
    }// switch
}

void
TPlanarNode::setCycle ( TPlanarCycle * pCycle, uint type )
{
    switch (type)
    {
        case MIN_CYC : pMinCycle = pCycle; break;
        case MAX_CYC : pMaxCycle = pCycle; break;
        default      : pMinCycle = pCycle;
    }// switch
}

const int
TPlanarNode::mark ( uint type ) const
{
    switch (type)
    {
        case MIN_CYC : return _mark;     break;
        case MAX_CYC : return _maxMark;  break;
        default      : return _mark;
    }// switch
}

void
TPlanarNode::setMark ( int m, uint type )
{
    switch (type)
    {
        case MIN_CYC : _mark    = m;  break;
        case MAX_CYC : _maxMark = m; break;
        default      : _mark    = m;
    }// switch
}

const TFloat
TPlanarNode::weight () const
{
    return _weight;
}

void
TPlanarNode::setWeight ( TFloat w )
{
    _weight = w;
}

const bool
TPlanarNode::isVisited () const
{
    return hasFlag( VISITED );
}

void
TPlanarNode::setVisited ( bool v )
{
    if (v == TRUE)
        setFlag( VISITED );
    else
        clearFlag( VISITED );
}

const bool
TPlanarNode::isDeleted ( uint type ) const
{
    switch (type)
    {
        case MIN_CYC : return hasFlag( MIN_DELETED ); break;
        case MAX_CYC : return hasFlag( MAX_DELETED ); break;
        default      : return hasFlag( MIN_DELETED );
    }// switch
}

void
TPlanarNode::setDeleted ( bool d, uint type )
{
    if (d == TRUE)
    {
        switch (type)
        {
            case MIN_CYC : setFlag( MIN_DELETED ); break;
            case MAX_CYC : setFlag( MAX_DELETED ); break;
            default      : setFlag( MIN_DELETED );
        }// switch
    }// if
    else
    {
        switch (type)
        {
            case MIN_CYC : clearFlag( MIN_DELETED ); break;
            case MAX_CYC : clearFlag( MAX_DELETED ); break;
            default      : clearFlag( MIN_DELETED );
        }// switch
    }// else
}

const bool
TPlanarNode::isInStack () const
{
    return hasFlag( IN_STACK );
}

void
TPlanarNode::setInStack ( bool b )
{
    if (b == TRUE)
        setFlag( IN_STACK );
    else
        clearFlag( IN_STACK );
}

TPlanarNode *
TPlanarNode::left ( bool bCheckSucc ) const
{
    TListIterator< TEdge > iterEdges( getOutEdges() );
    TEdge                * pEdge;
    TPlanarNode          * pAdjNode;

    iterEdges.last();
    while ((pEdge = iterEdges--) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();
        
        if ((( ! pAdjNode->isDeleted( MIN_CYC )) && (pAdjNode->mark( MIN_CYC ) == C)) ||
            (pAdjNode->isDeleted( MIN_CYC ) && (pAdjNode->cycle( MIN_CYC ) != NULL) &&
             bCheckSucc && pAdjNode->cycle( MIN_CYC )->isSucc() ))
            return pAdjNode;
    }// while

    return NULL;
}

TPlanarNode *
TPlanarNode::right ( bool bCheckCycle ) const
{
    TListIterator< TEdge > iterEdges( getOutEdges() );
    TEdge                * pEdge;
    TPlanarNode          * pAdjNode;

    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();
        
        if ((( ! pAdjNode->isDeleted( MAX_CYC )) && (pAdjNode->mark( MAX_CYC ) == C)) ||
            (pAdjNode->isDeleted( MAX_CYC ) && (pAdjNode->cycle( MAX_CYC ) != NULL) &&
             bCheckCycle && ( ! pAdjNode->cycle( MIN_CYC )->isDeleted() )))
            return pAdjNode;
    }// while

    return NULL;
}
    
/////////////////////////////////////////////////
//
// methods to manage list of weakly connected nodes
//
    
TList<TNode> *
TPlanarNode::blockList () const
{
    return pListBlock;
}

void
TPlanarNode::addLeftToBlockList  ( TNode * pNode )
{
    if (pNode != NULL)
    {
        if (pListBlock == NULL)
        {
            pListBlock = new TList<TNode>;
            pListBlock->insert( *this );
        }// if

        pNode->addReference();
        pListBlock->add( *pNode );
    }//if
}

void
TPlanarNode::addRightToBlockList ( TNode * pNode )
{
    if (pNode != NULL)
    {
        if (pListBlock == NULL)
        {
            pListBlock = new TList<TNode>;
            pListBlock->insert( *this );
        }// if

        pNode->addReference();
        pListBlock->insert( *pNode );
    }//if
}
    
/////////////////////////////////////////////////
//
// virtual constructors
//
    
TPlanarNode *
TPlanarNode::clone ()
{
    return new TPlanarNode( name, _mark,
                            pMinCycle, pMaxCycle,
                            _weight );
}

TPlanarNode *
TPlanarNode::create ( uint n )
{
    return new TPlanarNode( n );
}
