//
//  Project     : libgraph
//  File        : T3DFlowSurface.hh
//  Author      : Ronald Kriemann
//  Purpose     : class for building flow surfaces in a T3DCellEdgeGraph
//
// arch-tag: 2d20bfd9-189e-4545-b1d5-096a02bad1c1
//

#ifndef __T3DFLOWSURFACE_HH
#define __T3DFLOWSURFACE_HH

#include "memory.hh"

#include "T3DFrontier.hh"

#include "T3DCellEdgeGraph.hh"
#include "T3DCellEdgeNode.hh"

//
//
//
class T3DFlowSurface : public TMemory
{
private:
    // actually handled graph
    T3DCellEdgeGraph  *pFlowSurface;

    // lower bound for two neighbouring faces
    TFloat             alpha, beta;
    
public:
    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    T3DFlowSurface ();

    ~T3DFlowSurface ();

    /////////////////////////////////////////////////
    //
    // methods for building a list of graphs, each
    // representing a flow surface
    //

    TList< T3DCellEdgeGraph > * buildFlowSurfaces ( T3DCellEdgeGraph *,
                                                    TFloat alpha,
                                                    TFloat beta,
                                                    uint method );

protected:
    // start a new flow surface
    void startFlowSurface ( T3DCellEdgeNode * pNode,
                            uint method );
    
    // recursion/iteration for building one flow surface
    void buildFS_DFS ( T3DCellEdgeNode * pNode,
                       T3DFrontier * pPhi,
                       int func );

    void buildFS_BFS ( T3DCellEdgeNode * pStartNode );

    // expand flow surface by adj. nodes of given node
    void expandFS ( T3DCellEdgeNode * pNode,
                    T3DFrontier * pPhi,
                    int func );
    
    // calc angle between two faces
    TFloat calcFaceAngle ( T3DCellEdge * pEdge,
                           T3DFrontier * pFace1, T3DFrontier * pFace2 ) const;

    // check conditions for flow surfaces
    bool checkS1 ( T3DCellEdgeNode * pNode,
                   T3DFrontier * pFace,
                   int func ) const;

    bool checkS2 ( T3DCellEdgeNode * pNode,
                   T3DFrontier * pFace,
                   int func ) const;

    T3DFrontier * checkFS ( T3DCellEdgeNode * pNode,
                            T3DFrontier * pFace,
                            int & func ) const;

    // add a node to actual flow surface
    void addFSNode ( T3DCellEdgeNode * pNode,
                     T3DFrontier * pPhi,
                     int func );

    // check if given node is a suitable startnode
    bool isStartNode ( T3DCellEdgeNode * pNode ) const;

    // virtually (un)delete given flow surface
    void deleteFS ( T3DCellEdgeGraph * pGraph );
    void unDeleteFS ( T3DCellEdgeGraph * pGraph );

    // determine number of used (in a flowsurface) celledges
    // of given face
    uint noOfUsedEdges ( T3DFrontier * pFace );

    // delete all edges from/to node
    void delEdges ( T3DCellEdgeNode * pNode );

    // order given list of flowsurfaces
    void orderFS ( T3DCellEdgeGraph * pEdgeGraph,
                   TList< T3DCellEdgeGraph > * pListFS );
};

#endif  // __T3DFLOWSURFACE_HH
