//
//  Project     : libgraph
//  File        : T3DPermBuilder.hh
//  Author      : Ronald Kriemann
//  Purpose     : class to handle building of permutations in 3D
//
// arch-tag: 210ebc7a-161c-4432-b8e7-594875a00f5b
//

#ifndef __T3DPERMBUILDER_HH
#define __T3DPERMBUILDER_HH

#include "memory.hh"

#include "TPermutation.hh"

#include "blockinfo.hh"

// defines for multigridtype
#define MGSORT_PROJ    1
#define MGSORT_NESTED  2
#define MGSORT_INDEP   3

// forward decl.
template <int N>   class TNet;
template <int N>   class TNetHierarchy;
template <class T> class TList;
class TMatrixBlock;
class T3DCellEdgeGraph;
class T3DSurfaceGraph;

//
//
//
class T3DPermBuilder : public TBlockInfo
{
protected:
    // hold permutationblocks
    TPermutation        permF, permL, permOther;

    // last pos in permOther
    uint                lastOther;

    // hold blockstructure of last permutation built
    TDynArray< uint >  arrBlockStruct;

    // type of ordering; of multi-grid perm-building
    uint                orderType;
    uint                multiGridSort;
    
public:
    ////////////////////////////////////////////
    //
    // constructor and destructor
    //
    T3DPermBuilder ();
    T3DPermBuilder ( uint type, uint mgSort );

    virtual ~T3DPermBuilder ();

    ////////////////////////////////////////////
    //
    // access member variables
    //

    TDynArray< uint > * blockStructure ();
    
    ////////////////////////////////////////////
    //
    // basic methods for setting up permutations
    //

	// initialize var. and perm.
	virtual void init ( uint size );

	// add FVS to permutation
    virtual void addFVS ( TList< TNode > & fvs );

	// add F,L,C or other node(s) to permutation
	virtual void addFNode ( TNode & node );
	virtual void addLNode ( TNode & node );
	virtual void addCycle ( TGraph & cycle );
    virtual void addOther ( TNode & node );
    
	// assemble permutation
	virtual void buildPerm ( TPermutation * pPerm );

    ////////////////////////////////////////////
    //
    // various high level methods for perm building
    // - single net
    //

    virtual TPermutation * buildCYCPerm ( TNet<3>      * pNet,
                                          TConvEqn<3>  * pEqn,
                                          TMatrixBlock * pMatrix );

    virtual TPermutation * buildFVSPerm ( TNet<3>      * pNet,
                                          TConvEqn<3>  * pEqn,
                                          TMatrixBlock * pMatrix,
                                          uint           fvsMethod );

    virtual TPermutation * buildFVSPerm ( TMatrixBlock * pMatrix );
    virtual TPermutation * buildFVSPerm ( TNet<3>      * pNet,
                                          TMatrixBlock * pMatrix );

    virtual TPermutation * buildRCMPerm ( TMatrixBlock * pMatrix );
    
    ////////////////////////////////////////////
    //
    // various high level methods for perm building
    // - multi net
    // - matrix and perm arrays must be initialized
    //

    virtual void buildMGCYCPerm ( TNetHierarchy<3> * pNetHier,
                                  TConvEqn<3>      * pEqn,
                                  TBlockMatrix    ** pMatrix,
                                  TPermutation    ** pPermHier,
                                  uint               startLevel,
                                  uint               endLevel );

    virtual void buildMGFVSPerm ( TNetHierarchy<3> * pNetHier,
                                  TConvEqn<3>      * pEqn,
                                  TBlockMatrix    ** pMatrix,
                                  TPermutation    ** pPermHier,
                                  uint               startLevel,
                                  uint               endLevel,
                                  uint               fvsMethod );
    virtual void buildMGFVSPerm ( TBlockMatrix    ** pMatrix,
                                  TPermutation    ** pPermHier,
                                  uint               startLevel,
                                  uint               endLevel );
    
    virtual void buildMGRCMPerm ( TBlockMatrix    ** pMatrix,
                                  TPermutation    ** pPermHier,
                                  uint               startLevel,
                                  uint               endLevel );
    
protected:
    ////////////////////////////////////////////
    //
    // encapsulate various steps
    //

    // build graph consisting of cell edges
    T3DCellEdgeGraph * buildEdgeGraph ( TNet<3> * pNet,
                                        TConvEqn<3> * pEqn,
                                        TMatrixBlock * pMatrix );

    // build flow surfaces
    TList< T3DCellEdgeGraph > * buildFS ( T3DCellEdgeGraph * pEdgeGraph );

    // build surface graphs
    TList< T3DSurfaceGraph > * buildSurfaceGraphs ( TList< T3DCellEdgeGraph > * pListFS,
                                                    TMatrixBlock * pMatrix,
                                                    bool bSortOFD = FALSE );

    // care about edges not in a flowsurface
    void orderNonFSNodes ( TNet<3> * pNet,
                           T3DCellEdgeGraph * pEdgeGraph,
                           TMatrixBlock * pMatrix,
                           T3DPermBuilder * pPermBuilder = NULL );

    // put perm2 at bottom of perm1
    void concatPerm ( TPermutation * perm1,
                      TPermutation * perm2 ) const;

    // add a block to blockstructure
    void addBlock ( uint size );

    // check if given graph is cyclic
    void checkCyclic( TGraph * pGraph );
};

#endif  // __T3DPERMBUILDER_HH
