//
//  Project :  libvisual
//  File    :  T3DPrinter.hh
//  Purpose :  abstract baseclass for a file-based 3D output interface
//  Author  :  Ronald Kriemann
//
// arch-tag: 6da2be83-d6bc-47d6-97fc-cba0e184ea88
//

#ifndef __T3DPRINTER_HH
#define __T3DPRINTER_HH

#include <stdio.h>

#include "memory.hh"
#include "dynarray.hh"
#include "list.hh"
#include "TBTree.hh"

#include "vector.hh"

#include "printer.hh"
#include "styles.hh"

// thickness of a line (simulated with a thin cylinder)
#define LINE_THICKNESS   0.0015

#define FILLED TRUE

#define PIGMENT  1
#define AMBIENT  2

// different states
enum TStateType
{
    UNDEF, READY, BUILD_POLY
};

//
// data to store edges
//
class TEdgeData
{
protected:
    TVector<3>  *pNode0, *pNode1;
    
public:
    // constructor
    TEdgeData ();
    TEdgeData ( TVector<3> *v0, TVector<3> *v1 );

    // access nodes
    TVector<3> * node0 () const;
    TVector<3> * node1 () const;
    
    // some operators
    TEdgeData & operator = ( const TEdgeData & data );
    bool operator == ( const TEdgeData & data ) const;
};

//
// data for a polygon while building
//
class TPolygonData
{
public:
    // data for edge-based drawing
    TVector<3>  *pNode0, *pNode1, *pFirstNode;

    // data for filled polygons
    TList< TVector<3> >  listNodes;

    // indicate ifpolygon is filled
    bool filled;
};

//
// compare two edgedata objects
//
class TEdgeCmp
{
public:
    static int compare ( const TEdgeData & e1, const TEdgeData & e2 );
};

//
//
//
class T3DPrinter
{
protected:
    // file name and handle
    char        _fileName[256];
    FILE       *_file;

    // level of indentation
    int         _indentLevel;

    // working state (create polygon etc)
    TStateType  _state;

    // default color
    TColorType  _color;

    // default font
    TFontType   _font;

    // list with all edges, we had in the file
    // (so we do not draw a thing twice)
    TBTree< TEdgeData, TEdgeCmp >  treeEdges;

    // data for building different things
    TPolygonData  _polygonData;
    
public:
    // constructor
    T3DPrinter ( const char *fileName );

    // destructor
    virtual ~T3DPrinter ();

    // access data
    const int indentLevel () const;
    const TStateType   & state () const;
    const TColorType   & color () const;
    const TFontType    & font () const;
    const TPolygonData & polygonData () const;
    
    // begin/end printing
    virtual void begin ();
    virtual void end ();

    // indent commands
    virtual void indent ();
    virtual void indentUp ();
    virtual void indentDown ();
    virtual void indentWrite ( const char *string );

    // begin/end new (sub) section 
    virtual void beginSection ( const char *name ) = 0;
    virtual void endSection () = 0;

    //
    // misc. draw routines
    //
    virtual void beginPolygon ( bool filled = FALSE );
    virtual void addNode ( TVector<3> *v );
    virtual void endPolygon ();

    virtual void drawText ( TVector<3> *orig, const char *text ) = 0;
                    
    virtual void drawCoordSystem ( TFloat length = 1 );

    virtual void drawLine ( TVector<3> *orig, TVector<3> *dest ) = 0;
    
    virtual void drawCylinder ( TVector<3> *orig, TVector<3> *dest,
                                TFloat radius ) = 0;

    virtual void drawCone ( TVector<3> *orig, TFloat origRad,
                            TVector<3> *dest, TFloat destRad ) = 0;

    virtual void drawArrow ( TVector<3> *orig, TVector<3> *dest,
                             TFloat radius, TFloat lengthArrow = (2.0 / 3.0) );

    virtual void drawSphere ( TVector<3> * orig, TFloat radius ) = 0;
    
    //
    // misc. style settings
    //
    virtual void setRGBColor ( uint rgb, uint type = PIGMENT );
    virtual void setRGBColor ( TFloat r, TFloat g, TFloat b, uint type = PIGMENT );
    virtual void setHSVColor ( TFloat h, TFloat s, TFloat v, uint type = PIGMENT );
    virtual void setColor ( const TColorType & c );
    virtual void setFilter ( TFloat filter );

    virtual void setFont ( TFloat size, int font = -1, int style = -1,
                           TFloat width = -1, TFloat spacing = -1,
                           int justification = -1 );
    virtual void setFont ( const TFontType & f );

    // transform coord into local representation
    virtual TFloat transX ( TVector<3> * v ) const;
    virtual TFloat transY ( TVector<3> * v ) const;
    virtual TFloat transZ ( TVector<3> * v ) const;
    virtual void trans ( TVector<3> * v ) const;

    // actually draw the polygon
    virtual void drawPolygon () = 0;

    // write color/font def to file
    virtual void writeColor () = 0;
    virtual void writeFont () = 0;

private:
    // look if the given edge is in the edgelist
    bool isInEdgeList ( TEdgeData & edge );

    // correct given rgb value
    void checkColor ( TColorType & c );
};

#endif  // __T3DPRINTER_HH
