#if !defined(__TBTREE_HH)
#define __TBTREE_HH
//
//  Project   : libmisc
//  File      : TBTree.hh
//  Author    : Ronald Kriemann
//  Purpose   : template class for a binary search tree
//
// arch-tag: 9a74fcf0-6c1e-4eeb-ae1e-524e6781e9db
//

#include "defines.hh"
#include "memory.hh"

#include "TSList.hh"
#include "dynarray.hh"

#define CMP_BIGGER   1
#define CMP_EQUAL    0
#define CMP_SMALLER -1

/////////////////////////////////////////////////
//
// class for a item in a btree
//
template <class T>
class TBTreeItem : public TMemory
{
public:
    // the stored element
    T               element;

    // parent item
    TBTreeItem<T>  *pParent;

    // child items
    TBTreeItem<T>  *pLeft, *pRight;

    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    TBTreeItem ( const T & e,
                 TBTreeItem<T> * parent,
                 TBTreeItem<T> * left,
                 TBTreeItem<T> * right )
    {
        element = e;
        pParent = parent;
        pLeft   = left;
        pRight  = right;
    }
};

//
// class for a binary search tree storing elements of type T
// CMP must have a static method "int compare ( T, T )"
// with return values:
//   - CMP::compare( e1, e2 ) == CMP_BIGGER  iff e1 >  e2
//   - CMP::compare( e1, e2 ) == CMP_EQUAL   iff e1 == e2
//   - CMP::compare( e1, e2 ) == CMP_SMALLER iff e1 <  e2
//
template <class T, class CMP>
class TBTree : public TMemory
{
protected:

    /////////////////////////////////////////////////
    //
    // member variables
    //

    // root of btree
    TBTreeItem<T>  *pRoot;

    // number of stored elements
    uint            nSize;

public:
    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    TBTree ();

    virtual ~TBTree ();

    /////////////////////////////////////////////////
    //
    // access member variables
    //

    virtual const uint size () const;
    
    /////////////////////////////////////////////////
    //
    // methods for manipulating btree
    //

    // insert an element
    virtual void insert ( const T & e );

    // remove an element
    virtual void remove ( const T & e );

    // remove all elements in tree
    virtual void removeAll ();
    
    /////////////////////////////////////////////////
    //
    // search/sort related methods
    //

    // return minimal/maximal element in tree
//     virtual const T & min () const;
//     virtual const T & max () const;

    // look if given element is in tree
    virtual const bool isIn ( const T & e ) const;

    /////////////////////////////////////////////////
    //
    // methods to convert from/into various
    // data structures
    //

    virtual void fromList ( TSList< T > * pList );
    virtual TSList< T > * toList () const;

    virtual void fromArray ( TDynArray< T > * pArray );
    virtual TDynArray< T > * toArray () const;
    
protected:
    /////////////////////////////////////////////////
    //
    // misc. help-methods
    //

    // recursivly remove all items below root
    virtual void removeAll ( TBTreeItem<T> * pItem );

    // encapsulate creation of new tree items
    TBTreeItem<T> * createItem ( const T & element,
                                 TBTreeItem<T> * parent,
                                 TBTreeItem<T> * left,
                                 TBTreeItem<T> * right ) const;

    // return pointer to minimal/maximal items in subtree
    virtual TBTreeItem<T> * minItem ( TBTreeItem<T> * pRootItem ) const;
    virtual TBTreeItem<T> * maxItem ( TBTreeItem<T> * pRootItem ) const;

    // add recursivly elements to list or array
    void addToList  ( TBTreeItem<T> * pRootItem, TSList< T > * pList ) const;
    void addToArray ( TBTreeItem<T> * pRootItem, TDynArray< T > * pArray,
                      uint & pos ) const;
};

/////////////////////////////////////////////////
//
// implementation
//

/////////////////////////////////////////////////
//
// constructors and destructors
//

template <class T, class CMP>
TBTree<T,CMP>::TBTree ()
{
    pRoot = NULL;
    nSize = 0;
}

template <class T, class CMP>
TBTree<T,CMP>::~TBTree ()
{
    if (pRoot != NULL)
        removeAll();
}

/////////////////////////////////////////////////
//
// access member variables
//

template <class T, class CMP>
const uint
TBTree<T,CMP>::size () const
{
    return nSize;
}
    
/////////////////////////////////////////////////
//
// methods for manipulating btree
//

//
// insert an element
//
template <class T, class CMP>
void
TBTree<T,CMP>::insert ( const T & e )
{
    TBTreeItem<T>  *pItem;

    if (pRoot == NULL)
        pRoot = createItem( e, NULL, NULL, NULL );
    else
    {
        pItem = pRoot;

        while (TRUE)
        {
            if (CMP::compare( e, pItem->element ) == CMP_BIGGER)
            {
                if (pItem->pRight == NULL)
                {
                    pItem->pRight = createItem( e, pItem, NULL, NULL );
                    break;
                }// if
                else
                    pItem = pItem->pRight;
            }// if
            else
            {
                if (pItem->pLeft == NULL)
                {
                    pItem->pLeft = createItem( e, pItem, NULL, NULL );
                    break;
                }// if
                else
                    pItem = pItem->pLeft;
            }// else
        }// while
    }// else

    nSize++;
}

//
// remove an element
//
template <class T, class CMP>
void
TBTree<T,CMP>::remove ( const T & e )
{
    TBTreeItem<T>  *pItem, *pNewNode;
    int             cmpResult;
    
    pItem = pRoot;
    
    while (TRUE)
    {
        if (pItem == NULL)
            return;

        cmpResult = CMP::compare( e, pItem->element );

        if (cmpResult == CMP_EQUAL)
        {
            if ((pItem->pLeft != NULL) && (pItem->pRight != NULL))
            {
                // look for smallest entry on right subtree
                pNewNode = minItem( pItem->pRight );

                if (pNewNode->pParent != pItem)
                {
                    pNewNode->pParent->pLeft = pNewNode->pRight;
                    pNewNode->pRight = pItem->pRight;
                }// if
                
                pNewNode->pLeft  = pItem->pLeft;
            }// if
            else
            {
                pNewNode = pItem->pLeft == NULL ? pItem->pRight : pItem->pLeft;
            }// else

            if (pNewNode != NULL)
                pNewNode->pParent = pItem->pParent;
            
            if (pItem->pParent != NULL)
            {
                if (pItem->pParent->pLeft == pItem)
                    pItem->pParent->pLeft  = pNewNode;
                else
                    pItem->pParent->pRight = pNewNode;
            }// if
            else
                pRoot = pNewNode;

            delete pItem;
            nSize = (--nSize) < 0 ? 0 : nSize;

            return;
        }// if
        else
        {
            if (cmpResult == CMP_BIGGER)
                pItem = pItem->pRight;
            else
                pItem = pItem->pLeft;
        }// else
    }// while
}

//
// remove all elements in tree
//
template <class T, class CMP>
void
TBTree<T,CMP>::removeAll ()
{
    if (pRoot != NULL)
        removeAll( pRoot );
}

//
// recursivly remove all items below root
//
template <class T, class CMP>
void
TBTree<T,CMP>::removeAll ( TBTreeItem<T> * root )
{
    if (root == NULL)
        return;
    
    if (root->pLeft != NULL)
        removeAll( root->pLeft );

    if (root->pRight != NULL)
        removeAll( root->pRight );

    delete root;
}

/////////////////////////////////////////////////
//
// search/sort related methods
//

//
// return minimal/maximal relement in tree
//
// template <class T, class CMP>
// const T &
// TBTree<T,CMP>::min () const
// {
//     TBTreeItem<T>  *pItem = minItem( pRoot );
    
//     if (pItem != NULL)
//         return pItem->element;

//     cout << "TBTree::min () : tree is empty" << endl;
// }

// template <class T, class CMP>
// const T &
// TBTree<T,CMP>::max () const
// {
//     TBTreeItem<T>  *pItem = maxItem( pRoot );
    
//     if (pItem != NULL)
//         return pItem->element;

//     cout << "TBTree::max () : tree is empty" << endl;
// }

//
// return pointer to minimal/maximal items in subtree
//
template <class T, class CMP>
TBTreeItem<T> *
TBTree<T,CMP>::minItem ( TBTreeItem<T> * pRootItem ) const
{
    TBTreeItem<T>  *pItem;
    
    if (pRootItem != NULL)
    {
        pItem = pRootItem;
        
        while (pItem->pLeft != NULL)
            pItem = pItem->pLeft;
        
        return pItem;
    }// if
    
    return NULL;
}

template <class T, class CMP>
TBTreeItem<T> *
TBTree<T,CMP>::maxItem ( TBTreeItem<T> * pRootItem ) const
{
    TBTreeItem<T>  *pItem;
    
    if (pRootItem != NULL)
    {
        pItem = pRootItem;
        
        while (pItem->pRight != NULL)
            pItem = pItem->pRight;
        
        return pItem;
    }// if
    
    return NULL;
}

//
// look if given element is in tree
//
template <class T, class CMP>
const bool
TBTree<T,CMP>::isIn ( const T & e ) const
{
    TBTreeItem<T>  *pItem;
    int             cmpResult;
    
    pItem = pRoot;
    
    while (TRUE)
    {
        if (pItem == NULL)
            return FALSE;

        cmpResult = CMP::compare( e, pItem->element );

        if (cmpResult == CMP_EQUAL)
            return TRUE;
        else
        {
            if (cmpResult == CMP_BIGGER)
                pItem = pItem->pRight;
            else
                pItem = pItem->pLeft;
        }// else
    }// while

    // we'll never reach this
    return FALSE;
}

//
// encapsulate creation of new tree items
//
template <class T, class CMP>
TBTreeItem<T> *
TBTree<T,CMP>::createItem ( const T & element,
                            TBTreeItem<T> * parent,
                            TBTreeItem<T> * left,
                            TBTreeItem<T> * right ) const
{
    return new TBTreeItem<T>( element, parent, left, right );
}

/////////////////////////////////////////////////
//
// methods to convert from/into various
// data structures
//

template <class T, class CMP>
void
TBTree<T,CMP>::fromList ( TSList< T > * pList )
{
    if (pList == NULL)
        return;
    
    TSListIterator< T > iterator;
    T                 * pElement;

    //
    // we create a copy of the list as a tree, so we
    // delete all present elements in tree
    //

    removeAll();

    //
    // now store elements from list
    //

    iterator.setList( pList );
    iterator.first();
    while ((pElement = iterator++) != NULL)
    {
        insert( * pElement );
    }// while
}

template <class T, class CMP>
TSList< T > *
TBTree<T,CMP>::toList () const
{
    TSList< T > * pList;

    pList = new TSList< T >;

    addToList( pRoot, pList );
    
    return pList;
}

template <class T, class CMP>
void
TBTree<T,CMP>::fromArray ( TDynArray< T > * pArray )
{
    if (pArray == NULL)
        return;
    
    uint  i;
    uint  size;

    size = pArray->getSize();

    //
    // we create a copy of the array as a tree, so we
    // delete all present elements in tree
    //

    removeAll();

    //
    // now store elements from array
    //
    for ( i = 0; i < size ; i++ )
    {
        insert( (*pArray)[i] );
    }// for
}

template <class T, class CMP>
TDynArray< T > *
TBTree<T,CMP>::toArray () const
{
    TDynArray< T > * pArray;
    uint             pos;

    pArray = new TDynArray< T >;

    //
    // init array with correct size
    //

    pArray->setSize( size() );

    pos = 0;
    addToArray( pRoot, pArray, pos );
    
    return pArray;
}

//
// add recursivly elements to list or array
//
template <class T, class CMP>
void
TBTree<T,CMP>::addToList  ( TBTreeItem<T> * pRootItem,
                            TSList< T > * pList ) const
{
    if (pRootItem->pLeft != NULL)
        addToList( pRootItem->pLeft, pList );

    pList->append( pRootItem->element );

    if (pRootItem->pRight != NULL)
        addToList( pRootItem->pRight, pList );
}

template <class T, class CMP>
void
TBTree<T,CMP>::addToArray ( TBTreeItem<T> * pRootItem,
                            TDynArray< T > * pArray,
                            uint & pos ) const
{
    if (pRootItem->pLeft != NULL)
        addToArray( pRootItem->pLeft, pArray, pos );

    (*pArray)[ pos++ ] = pRootItem->element;

    if (pRootItem->pRight != NULL)
        addToArray( pRootItem->pRight, pArray, pos );
}

#endif  // __TBTREE_HH
