//
//  Project   : libnet
//  File      : confDiffFE.hh
//  Author    : Sabine Gutsch / Thomas Probst 
//  Purpose   : class for piecewise linear finite elements for
//              the convection-diffusion equation
//
// arch-tag: e391297e-d051-48b6-bf3f-8611df0413ea
//

#ifndef __TCONVDIFFFE_HH
#define __TCONVDIFFFE_HH

//
// class for a piecewise linear element discretization for
// the convection-diffusion equation
//

#include "TBasisFE.hh"
#include "TBasisCdrEqn.hh"
#include "parray.hh"
#include "TPDEBoundaryDesc.hh"
#include "TDoFDesc.hh"

template <int N>
class TConvDiffFE : public TBasisFE< N > 
{
protected:

    // gradients of basis functions on cell
    TVector< N > gradT[N+1];
    // determinants of transformation for cell
    TFloat     detJ;
    // dofs for u 
    long       dof[N+1];
    // cell's nodes
    TVertex< N > * nodes[N+1]; 
    
    // dof descriptions for u 
    TPDEBoundaryDesc<N> *  BoundaryDesc[N+1];
  
    // dof descriptions for u 
    TDoFDesc *  DoFDesc[N+1];

    // coordinates of vertices of cell
    TVector< N > coord[N+1];   
   
    //  size of the element;
    TFloat vol;

    // boundary names are larger than BOUNDARY_NAME
    uint noOfInnerDoFs;  // = net.noOfInteriorNodes()

public:
    // constructors	
    TConvDiffFE ( TDiscrType aDiscrType = Tabata, uint averb =0 );
 
  // destructor
  ~TConvDiffFE ();

  // get coordinates of vertices of cell
  void getCoord ( TCell< N > *cell );

  // write coordinates of all (interior) vertices in net
  // into (global) vector coord
  void getCoord ( TNet< N >  *net, TDynArray<TVector < N > * >& arrCoord, uint block = 0 );

  // initialize grad, det, c and  dof_u for a particular cell
  void initDiscr ( TCell< N > *cell );

 // initialize  dof_u for a particular cell
  void initDofs ( TCell< N > *cell );


    // bool: whether  Node(i)  is dof or not
    bool isDof (uint i);
 
    // bool: whether  Node(i)  is on Dirichlet boundary or not 
    bool isDir (uint i);
 
    // bool: whether  Node(i)  is on Neumann boundary or not 
    bool isNeu (uint i);
 
  

    // calculate global stiffness matrix and write it into global matrix
    virtual void discretize ( TBasisEqn< N > *eqn, TNet< N >  *net,  
                              TBlockMatrix *A,  TBlockVector<TFloat> *rhs );
   
	
  // calculate global stiffness matrix and write it into global matrix
  virtual void discrDiff ( TBasisEqn< N > *aeqn, TNet< N >  *net,  
                           TBlockMatrix *A, TBlockVector<TFloat> *rhs );
   
    // calculate global stiffness matrix of non-symmentric part
    virtual void discrConv ( TBasisEqn< N > *aeqn, TNet< N > *net, 
                             TMatrixBlock *C, TBlockVector<TFloat> *rhs,
                             uint block = 0 );
  
    virtual void discrRhs ( TBasisEqn<N> *aeqn, TNet<N> *net, TBlockVector<TFloat> *rhs ); 
  
    // calculate global  (lumped) mass matrix and write it into global vector
    virtual void discrMass ( TBasisEqn< N > *aeqn, TNet< N > *net,  
                             TVectorBlock<TFloat> *M,TBlockVector<TFloat> *rhs);

 
    // calculate global (lumped) mass matrix and add it to stiffness matrix
    virtual void discrMass ( TBasisEqn<N> *aeqn, TNet<N> *net,  
                             TBlockMatrix *A,TBlockVector<TFloat> *rhs);

    // calculate element stiffness matrices
    virtual void discrDiff ( TBasisCdrEqn< N > *eqn, TBlockMatrix *A, 
                             TBlockVector<TFloat> *rhs );
    virtual void discrConv ( TBasisCdrEqn< N > *eqn, TMatrixBlock *C, 
                             TBlockVector<TFloat> *rhs );
    virtual void discrConv ( TBasisCdrEqn< N > *eqn, TMatrixBlock *C, 
                             TBlockVector<TFloat> *rhs,  
                             TVectorBlock<TFloat> &suppSize, 
                             uint block = 0 );
    virtual void discrRhs  ( TBasisCdrEqn< N > *eqn,TBlockVector<TFloat> *rhs, TCell<N> *cell );

    virtual void  discrMass ( TBasisCdrEqn< N > *eqn, TVectorBlock<TFloat> *M,TBlockVector<TFloat> *rhs);
  
    virtual void  discrMass ( TBasisCdrEqn<N> *eqn, TMatrixBlock *M,TBlockVector<TFloat> *rhs);

    // help routines for discretisation
 
    void updateSupportSize ( TVectorBlock<TFloat> &suppSize );

  
};

#endif  // __TCONVDIFFFE_HH
