//
//  Project   : libnet
//  File      : TMatrix.hh
//  Author    : Ronald Kriemann
//  Purpose   : template class for a matrix (template in dimension)
//
// arch-tag: aabb642c-404c-4254-8881-1ce19b293f5c
//

#ifndef __TMATRIX_HH
#define __TMATRIX_HH

#include <math.h>

#include "memory.hh"
#include "multiref.hh"
#include "defines.hh"

#include "vector.hh"

// forward friend decl.
template <int N> class TMatrix;
template <int N> TMatrix<N> operator *  ( const TFloat & scalar, TMatrix<N> & matrix );
template <int N> TMatrix<N> operator *  ( const TMatrix<N> & matrix1,
                                          const TMatrix<N> & matrix2 );
template <int N> ostream &  operator << ( ostream & os, const TMatrix<N> & matrix );

//
// - this class is a template class in the dimension of the space
// - for 2d and for 3d,
//   for NxN matrices
//
template <int N>
class TMatrix : public TMemory 
{
protected:
    // array for the coord.
    TVector<N> rows[N];

public:
    // constructors (for 2d and 3d)
    TMatrix ();
    TMatrix( const TMatrix<N>& matrix );
   
    // destructor
    ~TMatrix();

    // add operators
    TMatrix<N>& operator+= ( const TMatrix<N>& matrix );
//    TMatrix<N>  operator+ ( const TMatrix<N>& matrix );

    // sub-and-set operator
    TMatrix<N>& operator-= ( const TMatrix<N>& matrix );
//    TMatrix<N>  operator- ( const TMatrix<N>& matrix );

    // scalar multiplikation (from left and right)
    TMatrix<N>& operator*= ( TFloat scalar );
//    TMatrix<N>  operator* ( TFloat scalar );
#ifdef EGCS
    friend TMatrix<N> operator * <> ( const TFloat & scalar, TMatrix<N>& matrix );
#elif GCC
    friend TMatrix<N> operator *    ( TFloat scalar, TMatrix<N>& matrix );
#endif

#ifdef EGCS
    friend TMatrix<N> operator * <> ( const TMatrix<N> & matrix1, const TMatrix<N> & matrix2 );
#elif GCC
    friend TMatrix<N> operator *    ( const TMatrix<N> & matrix1, const TMatrix<N> & matrix2 );
#endif

    //matrix-vector multiplication
    TVector<N> operator * ( const TVector<N> & vec );
   
    // subscript operator
    TVector<N> & operator[] ( uint i );
    const TVector<N> & operator[] ( uint i ) const;

    // copy operator
    TMatrix<N> & operator = ( const TMatrix<N> & matrix );

    // comparing
    const bool operator == ( const TMatrix<N> & matrix ) const;
    const bool operator != ( const TMatrix<N> & matrix ) const;

    // inverse 
    TMatrix<N> inverted();
  
    // transpose
    TMatrix<N> transposed();

    // output
    friend ostream & operator << <> ( ostream & os, const TMatrix<N> & matrix );
};

// some needed types
typedef TMatrix<2> T2DMatrix;
typedef TMatrix<3> T3DMatrix;

typedef TMatrix<2>* P2DMatrix;
typedef TMatrix<3>* P3DMatrix;

///////////////////////////////////////////////////////////////
//
// implementation
//

//
// constructors
//
template <int N>
TMatrix<N>::TMatrix( const TMatrix<N> & matrix ) 
{ 
    (*this) = matrix; 
}

//
// destructor
//
template <int N>
TMatrix<N>::~TMatrix() 
{}

//
// add operator
//
// template <int N>
// TMatrix<N> 
// TMatrix<N>::operator+ ( const TMatrix<N>& matrix ) 
// {
//     TMatrix<N>  matTmp( *this );
//     return (matTmp += matrix);
// }

//
// sub operators
//
// template <int N>
// TMatrix<N> 
// TMatrix<N>::operator- ( const TMatrix<N>& matrix ) 
// {
//     TMatrix<N>  matTmp( *this );
//     return (matTmp -= matrix);
// }

//
// scalar multiplication
//
// template <int N>
// TMatrix<N>
// TMatrix<N>::operator* ( TFloat scalar ) 
// {
//     TMatrix<N>  matTmp( *this );
//     return (matTmp *= scalar);
// }

template <int N>
TMatrix<N> 
operator * ( const TFloat & scalar, TMatrix<N>& matrix )
{
    TMatrix<N>  matTmp( matrix );

    return matTmp *= scalar;
}

template <int N>
TMatrix<N> 
operator * ( const TMatrix<N> & matrix1, const TMatrix<N> & matrix2 )
{
    TMatrix<N>  matTmp;
    uint        i, j, k;
    TFloat      s;

    for ( i = 0; i < N; i++ )
    {
        for ( j = 0; j < N; j++ )
        {
            s = 0;
            
            for ( k = 0; k < N; k++ )
                s += matrix1[i][k] * matrix2[k][j];
            
            matTmp[i][j] = s;
        }// for
    }// for

    return matTmp;
}

// transpose
template <int N>
TMatrix<N> 
TMatrix<N>::transposed ()
{
    TMatrix<N>  matTmp;
    int         i, j;
    
    for ( i = 0; i < N; i++ )
    {
        for ( j = 0; j < N; j++ )
            matTmp[i][j] = (*this)[j][i];
    }// for

    return matTmp;
}

#endif  // __TMATRIX_HH
