//
//  Project     : libgraph
//  File        : blockinfo.hh
//  Author      : Ronald Kriemann
//  Purpose     : class which builds blockinfo of F, L and C blocks
//  Date        : 15.08.1997
//
// arch-tag: 00e307a3-6aec-42f8-9000-ee74837cbb99
//

#ifndef __BLOCKINFO_HH
#define __BLOCKINFO_HH

#include "memory.hh"

#include "vector.hh"
#include "blockvector.hh"
#include "blockmatrix.hh"
#include "basisFE.hh"

#include "graphtool.hh"

// forward decl
template <int N> class TBasisFE;
template <int N> class TNet;
class TGraph;

// different types of blockstructure
#define NONE           0
#define PLANAR_FVS     1
#define CONCENTRIC     2
#define CONCENTRIC_MIN 3 // contain CONCENTRIC in a bitwise sense
#define CONCENTRIC_MAX 6 // contain CONCENTRIC in a bitwise sense
#define RCM            8
#define HEURISTIC_FVS  16
#define CROSSWIND_WO   32
#define CROSSWIND_WI   33
#define ORDER_ALL      64

// different sort methods
#define SORT_OFD               0
#define SORT_EDGE_WEIGHT       1
#define SORT_SUCC_EDGE_WEIGHT  2

//
//
//
class TBlockInfo : public TMemory
{
protected:
	// vector with blockinfo of F, L and C nodes
	TVectorBlock<uint>  vecBlocks;

	// vectors with size info of of F, L and C blocks
	TVectorBlock<uint>  vecSizeF_C, vecSizeL;

	// last pos of F(C) and L nodes in the blockvector
	uint  lastF_CNode, firstLNode;

	// last pos of blocks in the size vectors
	uint  lastF_CBlock, lastLBlock;

	// no of FVS nodes
	uint  noOfFVSNodes;

	// which sort-method to use
	uint  sortMethod;

	// names of outputfiles
	char  szCycleName[256];
	char  szGraphName[256];

	// and should we draw something
	bool  bDrawCycles;
	bool  bDrawGraph;

	// should we use parallel node-blocking
	bool  bUseBlocks;

    // array holding orig. nodes
    TDynArray< TNode * >  arrNodes;
    
public:
	// constructor and destructor
	TBlockInfo ();

    virtual ~TBlockInfo ();
    
	// init the vectors and var. with approp. values
	virtual void init ( uint size );

	// set size of fvs
	virtual void setFVSSize ( uint size );

	// add a FVS to the vectors
    virtual void addFVS ( TList<TNode> &fvs );

    // "remove" FVS from graph
    void removeFVS ( TList< TNode > * pFVS, TGraph * pGraph ) const;

	// add a F,L or C node(s) to the vector
	virtual void addFNode ( TNode &node );
	virtual void addLNode ( TNode &node );
	virtual void addCycle ( TGraph &cycle );

	// access sort method
	virtual void setSortMethod ( int sort );

	// access filenames
	virtual void setCycleName ( const char *name = NULL );
	virtual void setGraphName ( const char *name = NULL );

	// set block-using
	virtual void setBlocking ( bool useBlocks );

	// simply put permutation into vector
	virtual void buildPerm ( TVectorBlock<uint> &permute );

	//
	// build permutation vector (for all situations ;-)
	//
	// for concentric cycles
	virtual void buildCYCPerm ( TVectorBlock<uint> &permute, 
                                TVectorBlock<uint> &vecSize );
	virtual void buildCYCPerm ( TVectorBlock<uint> &permute, 
                                TVectorBlock<uint> &vecSize,
                                TDynArray< TVector<2> * > &arrVectors,
                                TMatrixBlock &mat, 
                                TFloat xi,
                                uint criterion, 
                                uint ordering = CONCENTRIC_MIN);
	virtual void buildCYCPerm ( TVectorBlock<uint> &permute, 
                                TVectorBlock<uint> &vecSize,
                                TDynArray< TVector<2> * > &arrVectors,
                                TMatrixBlock &diff,
                                TMatrixBlock &conv,
                                TFloat xi,
                                uint criterion, 
                                uint ordering = CONCENTRIC_MIN);

	// for fvs
	virtual void buildFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize );
	virtual void buildFVSPerm ( TVectorBlock<uint> &permute,
                                uint &fvsSize,
                                TDynArray< TVector<2> * > &arrVectors,
                                TMatrixBlock &mat, 
                                TFloat xi,
                                uint criterion,
                                uint searchType,
                                uint fvsType );
	virtual void buildFVSPerm ( TVectorBlock<uint> &permute,
                                uint &fvsSize,
                                TDynArray< TVector<2> * > &arrVectors,
                                TMatrixBlock &diff,
                                TMatrixBlock &conv,
                                TFloat xi,
                                uint criterion,
                                uint searchType,
                                uint fvsType );

	// for planar fvs
	virtual void buildPlanarFVSPerm ( TVectorBlock<uint> &permute, 
                                      uint &fvsSize );
    virtual void buildPlanarFVSPerm ( TVectorBlock<uint> &permute,
                                      uint &fvsSize,
                                      TDynArray< TVector<2> * > &arrVectors,
                                      TMatrixBlock &mat, 
                                      TFloat xi, uint criterion,
                                      uint searchType = SEARCH_DFS );
	virtual void buildPlanarFVSPerm ( TVectorBlock<uint> &permute,
                                      uint &fvsSize,
                                      TDynArray< TVector<2> * > &arrVectors,
                                      TMatrixBlock &diff, TMatrixBlock &conv,
                                      TFloat xi, uint criterion,
                                      uint searchType = SEARCH_DFS );

	// for heuristic fvs
	virtual void buildHeuristFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize );
	virtual void buildHeuristFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                                       TDynArray< TVector<2> * > &arrVectors,
                                       TMatrixBlock &mat, 
                                       TFloat xi, uint criterion,
                                       uint searchType = SEARCH_DFS );
	virtual void buildHeuristFVSPerm ( TVectorBlock<uint> &permute,
                                       uint &fvsSize,
                                       TDynArray< TVector<2> * > &arrVectors,
                                       TMatrixBlock &diff,
                                       TMatrixBlock &conv,
                                       TFloat xi, uint criterion,
                                       uint searchType = SEARCH_DFS );

	// for rcm
	virtual void buildRCMPerm ( TVectorBlock<uint> &permute, TMatrixBlock &mat );

	// for multigrid (concentric cycles)
	virtual void buildCYCProjPerm ( TVectorBlock<uint> **arrPermVec,
                                    TVectorBlock<uint> **arrSizeVec,
                                    T2DBasisFE &fe,
                                    const TList< TNet<2> > &listNets,
                                    TBlockMatrix **arrMatrix,
                                    TFloat xi, uint criterion, 
                                    uint ordering = CONCENTRIC_MIN );
	virtual void buildCYCIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                     TVectorBlock<uint> **arrSizeVec,
                                     T2DBasisFE &fe,
                                     const TList< TNet<2> > &listNets,
                                     TBlockMatrix **arrMatrix,
                                     TFloat xi, uint criterion, 
                                     uint ordering,
                                     uint startLevel, uint endLevel );
	virtual void buildCYCNestedPerm ( TVectorBlock<uint> **arrPermVec,
                                      TVectorBlock<uint> **arrSizeVec,
                                      T2DBasisFE &fe,
                                      const TList< TNet<2> > &listNets,
                                      TBlockMatrix **arrMatrix,
                                      TFloat xi, uint criterion, 
                                      uint ordering = CONCENTRIC_MIN );

	// for multigrid (planar fvs)
	virtual void buildFVSProjPerm ( TVectorBlock<uint> **arrPermVec,
                                    TVectorBlock<uint> *arrFVSSize,
                                    T2DBasisFE &fe,
                                    const TList< TNet<2> > &listNets,
                                    TBlockMatrix **arrMatrix,
                                    TFloat xi, uint criterion,
                                    uint searchType, uint fvsType );
	virtual void buildFVSIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                     TVectorBlock<uint> *arrFVSSize,
                                     T2DBasisFE &fe,
                                     const TList< TNet<2> > &listNets,
                                     TBlockMatrix **arrMatrix,
                                     TFloat xi, uint criterion,
                                     uint searchType, uint fvsType,
                                     uint startLevel, uint endLevel );
	virtual void buildFVSNestedPerm ( TVectorBlock<uint> **arrPermVec,
                                      TVectorBlock<uint> *arrFVSSize,
                                      T2DBasisFE &fe,
                                      const TList< TNet<2> > &listNets,
                                      TBlockMatrix **arrMatrix,
                                      TFloat xi, uint criterion,
                                      uint searchType, uint fvsType );

	// for multigrid (rcm)
	virtual void buildRCMProjPerm ( TVectorBlock<uint> **arrPermVec,
                                    TBlockMatrix **arrMatrix,
                                    uint noOfNets );
	virtual void buildRCMIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                     TBlockMatrix **arrMatrix,
                                     uint startLevel, uint endLevel );
};

///////////////////////////////////////////////////////
//
// additional functions for permutation handling
//

void invertPerm ( TVectorBlock<uint> &invPerm, TVectorBlock<uint> &perm );

void reverseOrder ( TVectorBlock<uint> &perm );

bool orderBlocks ( TVectorBlock<uint> *PermVec,
				   TVectorBlock<uint> *cycle,  TBlockMatrix *Matrix );


#endif  // __BLOCKINFO_HH
