//
//  Project  : libnet
//  File     : blockmatrix.hh
//  Author   : Sabine Gutsch / Ronald Kriemann / Thomas Probst   
//  Purpose  : class for a (sparse) blockmatrix 
//
// arch-tag: e81c428f-9009-4420-8d80-675fd53caa7c
//

#ifndef __BLOCKMATRIX_HH
#define __BLOCKMATRIX_HH

#include <string.h>
#include "matrixblock.hh"

//
// class for a row of blocks
//
class TMatrixBlockRow : public TMemory 
{
protected:
	// blocks in a blockrow
	TMatrixBlock  **pColumns;
	// number of blocks
	uint		  n;

public:
	// constructors
	TMatrixBlockRow ();
	TMatrixBlockRow ( uint number );
	TMatrixBlockRow ( uint number, uint * size );
	TMatrixBlockRow ( const TMatrixBlockRow& Row );
	TMatrixBlockRow ( TMatrixBlock * pMatrix );

	// destructor
	~TMatrixBlockRow ();

	// return the number of blocks
	uint getNoOfBlocks () const;

    // set a matrixblock to given matrix
    void change ( uint i, TMatrixBlock * pMatrix );
    
	// index operator
	TMatrixBlock& operator[] ( uint i );
	const TMatrixBlock& operator[] ( uint i ) const;
};


//
// class TBlockMatrix : an array of TMatrixBlockRow
//
class TBlockMatrix : public TMemory 
{
protected:
	//  the stored blockrows
	TMatrixBlockRow  **pRows;
	// number of blockrows
	uint			 n;
    // pointer to blockmatrix that contains its LU-decomposition
    TBlockMatrix     *pLUdecomposition;

public:
	// constructors
	TBlockMatrix ();
	TBlockMatrix ( uint number );
	TBlockMatrix ( uint number, uint *size );
	TBlockMatrix ( const TBlockMatrix & bm );
	TBlockMatrix ( const TMatrixBlock & m );
	TBlockMatrix ( TMatrixBlock * pMatrix );

	// destructor
	~TBlockMatrix ();

	// get number of rows
	const uint getNoOfRows() const;

	// index operator [] such that [][][][] is defined for matrix
	TMatrixBlockRow& operator[] ( uint i );
	const TMatrixBlockRow& operator[] ( uint i ) const;

	// get tranposed j'th row of i'th matrix block row 
	// for normal blockvectors
	TBlockVector<TFloat> getTransposedRow( uint i, uint j, uint *size);
	// for sparsevectors with copy
	TSparseBlockVector getTransposedRow( uint i, uint j );
	// for sparsevectors without copy
	void getTransposedRow( TSparseBlockVector& vec, uint i, uint j );

	// matrix - vector multiplication (y = Ax)
    void matmul ( TBlockVector< TFloat > & x, const TBlockVector< TFloat > & y ) const; 
	TBlockVector<TFloat> operator* ( const TBlockVector<TFloat>& vec ) const;

	// multiplication with transpose matrix
    TBlockVector<TFloat> transposeMult ( const TBlockVector<TFloat>& vec ) const;

	// diag - vector multiplication (we assume, that all blocks on the blockdiag
	// are mxm - mat.)
	TBlockVector<TFloat> diagMult ( const TBlockVector<TFloat>& vec ) const;

	// invDiag - vector multiplication
	TBlockVector<TFloat> invDiagMult ( const TBlockVector<TFloat> &vec ) const;

	// blockDiag   - vector multiplication
	TBlockVector<TFloat> blockDiagMult ( const TBlockVector<TFloat>& vec ) const;

	// blockBlockDiag   - vector multiplication
    // performs blockiterations on the matrixblocks
	TBlockVector<TFloat> blockBlockDiagMult ( const TBlockVector<TFloat>& vec ) const;

	// invLower - vector multiplication
	TBlockVector<TFloat> invLowerMult ( const TBlockVector<TFloat>& vec ) const;

	// invUpper - vector multiplication
	TBlockVector<TFloat> invUpperMult ( const TBlockVector<TFloat>& vec ) const;

	// invBlockDiag - vector multiplication  (iterate invdiagblocks)

	// blockUpper - vector multiplication
	TBlockVector<TFloat> blockUpperMult ( const TBlockVector<TFloat> &vec ) const;

	// invBlockUpper - vector multiplication (iterate invdiagblocks)

	// blockLower - vector multiplication
	TBlockVector<TFloat> blockLowerMult ( const TBlockVector<TFloat> &vec ) const;

	// invBlockLower - vector multiplication (iterate invdiagblocks)

	// permutation of the matrix 
	// (p: row and column permutation velocity unknowns, not pressure)
	void permute ( const TVectorBlock<uint> *p );
	void permute ( const TPermutation * pPerm );

	// energynorm of a given vector
	TFloat energyNorm ( const TBlockVector<TFloat>& vec ) const;

	// row sum norm (infinity norm) of blockmatrix (with scaled rows/cols)
    TFloat rowSumNorm ( TFloat *scalRows = NULL, TFloat *scalCols = NULL ) const;

	// invBlockLowerMult based on concentric cycles for scalar problems
    TBlockVector<TFloat> invBlockLowerMult ( const TBlockVector<TFloat> &vec ) const;

	// invBlockUpperMult based on concentric cycles for scalar problems
    TBlockVector<TFloat> invBlockUpperMult ( const TBlockVector<TFloat> &vec ) const;

	// add or subtract factor times identity matrix to blockmatrix
    void addIdentity ( const TFloat factor );
    void subtractIdentity ( const TFloat factor );

	// check if decomposition exists already
    bool hasLUdecomp(); 
	// calculate (exact) LU decomposition
    void calculateLUdecomp();
	// calculate incomplete LU decomposition 
    void calculateILUdecomp();
    // delete LU decomposition
    void deleteLUdecomp();
	// calculate (exact) LU decomposition for Stokes
    void calcStokesLUdecomp1();
    void calcStokesLUdecomp2();
	// exact solver based on LU decomposition
    TBlockVector<TFloat> solveLUdecomp( const TBlockVector<TFloat> &vec ) const;
	// get LU matrix
    TBlockMatrix* getLUdecomposition() { return (pLUdecomposition); }
};

#endif  // __BLOCKMATRIX_HH
