//
//  Project  : libnet
//  File     : blockvector.hh
//  Author   : Sabine Gutsch / Ronald Kriemann / Thomas Probst
//  Purpose  : class for a blockvector
//
// arch-tag: b7c7b412-69d4-432e-a316-c1ccfa5631c9
//

#ifndef __BLOCKVECTOR_HH
#define __BLOCKVECTOR_HH

#include "defines.hh"
#include "memory.hh"
#include "dynarray.hh"

// forward declarations
class TMatrixData;
class TMatrixBlock;
class TBlockMatrix;
class TSparseBlockVector;
class TPermutation;
template <class T> class TVectorBlock;
template <class T> class TBlockVector;

// forward decl for friend methods
template <class T>
ostream & operator << ( ostream &, const TVectorBlock<T> & );

template <class T>
TBlockVector<T> operator * ( const T &, const TBlockVector<T> & );

template <class T>
ostream & operator << ( ostream &, const TBlockVector<T> & );

//
// class TVectorBlock (template in storagetype)
// - a single block in a BlockVector
// - uses TDynArray to store elements
//
template <class T>
class TVectorBlock : public TDynArray<T>
{
protected:
    // precision for cout
    uint          precision;

public:
    // constructor
    TVectorBlock ();
    TVectorBlock ( const TVectorBlock<T>& v );
    TVectorBlock ( const T & f, uint n );
    
    // destructor
    ~TVectorBlock ();

    // set vectorBlock
    void setVectorBlock ( const TVectorBlock<T>& v );
    
    // set all vector components to scalar
    void set ( const T& scalar );

    void setPrecision (const uint prec );
    uint getPrecision () const;

    // copy operator
    TVectorBlock<T>& operator= ( const T & scalar );

    // init vec with same structure, but no data
    void clone ( TVectorBlock< T > & vec ) const;

    // vectorblock addition ( we assume, that the vectorblocks
    // have same structure )
    TVectorBlock<T>& operator+= ( const TVectorBlock<T>& vec );
    TVectorBlock<T>  operator+ ( const TVectorBlock<T>& vec ) const;
    // for implementation see sparsevector.hh
    TVectorBlock<T>& operator+= ( const TMatrixData& data );
    
    // vectorblock subtraction ( we assume, that the vectorblocks
    // have same structure )
    TVectorBlock<T>& operator-= ( const TVectorBlock<T>& vec );
    TVectorBlock<T>  operator- ( const TVectorBlock<T>& vec ) const;
    // for implementation see sparsevector.hh
    TVectorBlock<T>& operator-= ( const TMatrixData& data );
    
    // scalar mult.
    TVectorBlock<T>& operator*= ( const T& scalar );

    // inner product (we assume that the vectorblocks have same structure)
    T operator* ( const TVectorBlock<T>& vec ) const;

    //
    // some blas routines
    //

    // x = x + alpha y
    const TVectorBlock< T > & daxpy ( const TVectorBlock< T > & y, const T & alpha );

    // x = x + alpha A y
    const TVectorBlock< T > & matmulplus ( const TMatrixBlock      & A,
                                           const TVectorBlock< T > & y,
                                           const T                 & alpha );
    
    // permutation
    void permute ( const TVectorBlock<uint> *p );
    void permute ( const TPermutation * pPerm );

    // output
    friend ostream& operator << <> ( ostream& os, const TVectorBlock<T>& vectorblock );
};

//
// class TBlockVector (template in storagetype)
//
template<class T>
class TBlockVector : public TDynArray< TVectorBlock<T> >
{
public:
    // constructors
    // standard
    TBlockVector ();

    // create blockstructure with one-sized block
    TBlockVector ( uint number, uint size );

    // set all n = number blocks to size size[]
    TBlockVector ( uint number, uint * size );

    // copy constructor
    TBlockVector ( const TBlockVector<T> &v );

    // constructor with vector block
    TBlockVector ( const TVectorBlock<T> &v );
   
    // take the blockstructure from v but init. the vector with f
    TBlockVector ( const T& scalar, const TBlockVector<T> &v );

    // destructor
    virtual ~TBlockVector ();

    // set all vector components to scalar
    void set ( const T& scalar );

    // get/set number of blocks
    uint getNoOfBlocks () const;
    void setNoOfBlocks ( uint no );
    
    // return size (sum of sizes of all blocks)
	virtual void setSize ( uint n ) { TDynArray< TVectorBlock<T> >::setSize(n); }
    virtual const uint getSize () const;

    // vector addition ( we assume, that the vectors have same structure )
    TBlockVector<T>& operator+= ( const TBlockVector<T>& vec );
    TBlockVector<T>& operator+= ( const TSparseBlockVector& vec );
    TBlockVector<T> operator+ ( const TBlockVector<T>& vec );

    // vector subtraction ( we assume, that the vectors have same structure )
    TBlockVector<T>& operator-= ( const TBlockVector<T>& vec );
    TBlockVector<T>& operator-= ( const TSparseBlockVector& vec );
    TBlockVector<T> operator- ( const TBlockVector<T>& vec );

    // scalar mult.
    TBlockVector<T>& operator*= ( const T& scalar );
    TBlockVector<T>  operator* ( const T& scalar );

    // inner product ( we assume, that the vectors have same structure )
    T operator* ( const TBlockVector<T>& vec ) const;

    //
    // some blas routines
    //

    // x = x + alpha y
    const TBlockVector< T > & daxpy ( const TBlockVector< T > & y, const T & alpha );

    // x = x + alpha A y
    const TBlockVector< T > & matmulplus ( const TBlockMatrix      & A,
                                           const TBlockVector< T > & y,
                                           const T                 & alpha );
    
    // copy operator
    TBlockVector<T> & operator= ( const T & scalar );

    // init vec with same structure, but no data
    void clone ( TBlockVector< T > & vec ) const;

    // euclid norm
    const TFloat euclidNorm ( TFloat *scal = NULL ) const;
    
   // weighted euclid norm
    const TFloat weightedEuclidNorm ( TFloat *scal = NULL ) const;
        
    // euclid norm of a single block
    const TFloat euclidBlockNorm ( uint i ) const;

    // weighted euclid norm of a single block
    const TFloat weightedEuclidBlockNorm ( uint i ) const;
    
    // maximum norm
    const TFloat maxNorm () const;

    // maximum norm of a single block
    const TFloat maxNorm ( uint i) const;

    // for energynorm see TBlockMatrix

    // permute first n-1 blocks
    void permute ( const TVectorBlock<uint> *p );
    void permute ( const TPermutation * pPerm );

    // output
    friend ostream & operator << <> ( ostream& os, const TBlockVector<T>& vector );
};


/////////////////////////////////////////////////////////////
//
// friend methods
//

//
// output
//
template <class T>
ostream&
operator << (ostream& os, const TVectorBlock<T>& vectorblock) 
{
    uint size = vectorblock.getSize();
    
    os.setf( ios::fixed, ios::floatfield );
    os.precision( vectorblock.getPrecision() );
    
    for (uint j = 0; j < size; j++)
        os << vectorblock[j] << endl;
    
    os.precision(6);
    return os;
}

//
// scalar mult.
//
template <class T>
TBlockVector<T>
operator * ( const T & scalar, const TBlockVector<T>& vec ) 
{
    TBlockVector<T> vecTmp( vec );

    return (vecTmp *= scalar);
}

//
// output
//
template <class T>
ostream&
operator << (ostream& os, const TBlockVector<T>& vector) 
{
    uint  i, j;
    uint  blocks, size;
    
    os.setf( ios::fixed, ios::floatfield);
    os.precision(6);

    blocks = vector.getNoOfBlocks();
    for (i = 0; i < blocks; i++)
    {
        size = vector[i].getSize();
        
        for (j = 0; j < size; j++)
            os << vector[i][j] << endl;
    }// for
    
    return os;
}

#endif  // __BLOCKVECTOR_HH
