//
//  Project   : libnet
//  File      : cell.hh
//  Author    : Ronald Kriemann / Sabine Gutsch
//  Purpose   : class for a cell (smallest part of a net (see net.hh))
//
// arch-tag: 5f6e9f53-b2a3-46b3-bb58-7f9c28c3233c
//

#ifndef __CELL_HH
#define __CELL_HH

#include "memory.hh"
#include "defines.hh"
#include "multiref.hh"
#include "dynarray.hh"

#include "vertex.hh"
#include "frontier.hh"

//
// color (state) of refinement :
//     WHITE : cell must be refined
//     GRAY  : cell may be refined
//     BLACK : cell must not be refined
//
enum RefinementColor { WHITE , GRAY , BLACK };

//
// a cell is a set of frontiers
//
template <int N>
class TCell : public TMemory, public TMultiReference 
{
protected:
    // pointer to parentcell
    // TCell<N>          *pcellParent;

    // list with the frontiers which define the cell
    TFrontier<N>     **arrFrontiers;
    uint              noOfFrontiers;

    // list of children cells
    TCell<N>         **arrChildren;
    uint              noOfChildren;

    // color(state) of refinement
    // RefinementColor    colRefinement;

    // size of cell (in 2D the area)
    TFloat             fSize;

public:
    // constructors
    TCell ( TCell<N>* parent = NULL );

    // destructor
    virtual ~TCell ();

    // init memory for used arrays
    virtual void initChildPtr ( uint n );
    virtual void initFrontierPtr ( uint n );

    // return i'th frontier
    virtual TFrontier<N>* getFrontier ( uint i ) const;

    // return number of frontiers
    virtual const uint getNoOfFrontiers () const;

    // return i'th child (only defined if cell is refined)
    virtual TCell<N>* getChild ( uint i ) const;

    // get actual number of children
    virtual const uint getNoOfChildren() const;

    // return parentcell
    // virtual TCell<N>* getParent () const;

    // set parentcell
    // virtual void setParent ( TCell<N>* cell );

    // return refinement color
    // virtual RefinementColor getRefinement () const;

    // adjust things before real refinement
    virtual void prepareRefine () = 0;

    // refine yourself
    virtual void refine ()  = 0;

    // and the opposite : coarse
    virtual void coarsen () = 0;

    // set mark of frontier for refinement
    virtual void setMark ( uint i );

    // set mark automagically
    virtual void setMark ();

    // get i'th vertex of this cell (node i is first node of frontier i)
    virtual TVertex<N>* getNode ( uint i ) const;

    // return vertex opposite of i'th frontier
    virtual TVertex<N>* getOppositeNode ( uint i ) const;

    // calculate size of cell (2D: area, 3D: volume)
    virtual TFloat calcSize ();

    // get size of cell (2D: area, 3D: volume)
    virtual TFloat getSize ();
};

//////////////////////////////////////////////////
//
// implementation
//

//
// constructors
//
template <int N>
TCell<N>::TCell ( TCell<N> * ) 
{
    // pcellParent = parent;

    arrChildren  = NULL;
    noOfChildren = 0;

    arrFrontiers  = NULL;
    noOfFrontiers = 0;

    // colRefinement = WHITE;
    fSize         = 0.0;
}

//
// destructor
// remove all objects which are used only by this cell and
// all pointers to this cell
//
template <int N>
TCell<N>::~TCell () 
{
    uint i;             
    TFrontier<N>  *pFrontier;
    TCell<N>      *pChild;

    if (arrChildren != 0)
    {
        for (i = 0; i < noOfChildren; i++)
        {
            pChild = arrChildren[i];

            if ((pChild != NULL) && (pChild != this))
            {
                // pChild->setParent( NULL );
                if (pChild->delReference()) 
                    delete pChild;
            }// if
        }// for

        delete arrChildren;
    }// if

    if (arrFrontiers != 0)
    {
        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = arrFrontiers[i];

            if (pFrontier != NULL)
            {
                if (pFrontier->delReference())
                    delete pFrontier;
                else
                    pFrontier->clearCell(this);
            }// if
        }// for
        
        delete arrFrontiers;
    }// if
}

//
// init memory for used arrays
//
template <int N>
void 
TCell<N>::initChildPtr ( uint n )
{
    uint  i;

    if (n == noOfChildren) 
        return;

    if (arrChildren != NULL) 
        delete arrChildren;

    if (n == 0)
    {
        arrChildren  = NULL;
        noOfChildren = 0;
    }// if
    else
    {
        arrChildren  = new TCell<N> * [n];
        noOfChildren = n;

        for (i = 0; i < noOfChildren; i++)
            arrChildren[i] = NULL;
    }// else
}

template <int N>
void 
TCell<N>::initFrontierPtr ( uint n )
{
    uint  i;

    if (n == noOfFrontiers) 
        return;

    if (arrFrontiers != NULL) 
        delete arrFrontiers;

    if (n == 0)
    {
        arrFrontiers  = NULL;
        noOfFrontiers = 0;
    }// if
    else
    {
        arrFrontiers  = new TFrontier<N> * [n];
        noOfFrontiers = n;

        for (i = 0; i < noOfFrontiers; i++)
            arrFrontiers[i] = NULL;
    }// else
}

//
// return i'th frontier
//
template <int N>
TFrontier<N>*  
TCell<N>::getFrontier ( uint i ) const 
{
    if (i < noOfFrontiers)
        return arrFrontiers[i];
    else
    {
        cerr << "TCell<N>::getFrontier (uint) : index out of range" << endl;
        return NULL;
    }// else
}

//
// return number of frontiers
//
template <int N>
const uint 
TCell<N>::getNoOfFrontiers () const 
{ 
    return noOfFrontiers;
}

//
// return i'th child (only defined if cell is refined)
//
template <int N>
TCell<N>* 
TCell<N>::getChild ( uint i ) const 
{ 
    if (i < noOfChildren)
        return arrChildren[i];
    else
    {
        cerr << "TCell<N>::getChild (uint) : index out of range" << endl;
        return NULL;
    }// else
}

//
// get actual number of children
//
template <int N>
const uint 
TCell<N>::getNoOfChildren () const
{ 
    return noOfChildren; 
}

//
// return parentcell
//
// template <int N>
// TCell<N>* 
// TCell<N>::getParent () const 
// { 
//     return pcellParent;
// }

//
// set parentcell
//
// template <int N>
// void 
// TCell<N>::setParent ( TCell<N> *cell )
// { 
//     pcellParent = cell; 
// }

//
// return refinement color
//
// template <int N>
// RefinementColor 
// TCell<N>::getRefinement () const 
// { 
//     return colRefinement;
// }

//
// set mark of frontier for refinement
//
template <int N>
void 
TCell<N>::setMark ( uint )
{}

//
// set mark automagically
//
template <int N>
void 
TCell<N>::setMark () 
{}

//
// get i'th vertex of this cell (node i is first node of frontier i)
//
template <int N>
TVertex<N>* 
TCell<N>::getNode ( uint i ) const
{
    return getFrontier(i)->getNode(this, 0);
}

//
// get size of cell (2D: area, 3D: volume)
//
template <int N>
TFloat 
TCell<N>::getSize ()
{ 
    return fSize;
}

// some used types
typedef TCell<2>  T2DCell;
typedef TCell<3>  T3DCell;

#endif  // __CELL_HH
