//
//  Project :  libvisual
//  File    :  drawapp.hh
//  Purpose :  class for printing the different things
//  Author  :  Ronald Kriemann
//
// arch-tag: 99189f17-c1c1-4890-8886-0b8e82d98748
//

#ifndef __DRAWAPP_HH
#define __DRAWAPP_HH

#include "defines.hh"

#include "net.hh"

#include "blockmatrix.hh"
#include "blockvector.hh"

#include "basisEqn.hh"

// which device for 3D
#define VRML    0x1
#define POVRAY  0x2

// forward decl
class T3DCellEdgeGraph;
class T3DSurfaceGraph;
class T3DGraph;
class T3DMeshGenerator;
template <int N> class TBasisEqn;
template <int N> class TConvEqn;
class T3DPrinter;
class TView;
class TNode;
class TGraph;
class TPermutation;

//
// by calling draw() you decide what to draw
//
class TDrawApp 
{
public:
    // constructor
    TDrawApp ();

    ////////////////////////////////////////////
    //
    // draw a vector field
    //
    static void draw ( TDynArray< TVector<2> * >& vertices,
                       TBlockVector<TFloat>& solution,
                       const T2DNet& net,
                       const char *filename,
                       bool scale = FALSE );

    ////////////////////////////////////////////
    //
    // draw a net
    //
    static void draw ( T2DNet& net,
                       const char *filename );
    static void draw ( T2DNet& net, TVectorBlock<uint>& permutation,
                       const char *filename );

    static void drawNet ( T3DNet *net, const char *filename,
                          int dev = VRML | POVRAY );

    static void drawMG ( T3DMeshGenerator * mg, const char *filename,
                         int dev = VRML | POVRAY );

    ////////////////////////////////////////////
    //
    // draw a matrix
    //
    static void drawMatrix ( TBlockMatrix * pMatrix, TFloat xi,
                             const char *filename );

    static void drawMatrix ( TMatrixBlock * pMatrix, TFloat xi,
                             const char *filename );

    static void drawMatrix ( TMatrixBlock * pMatrix, TFloat xi,
                             TDynArray< uint > * pBlockStruct,
                             const char *filename );

    static void drawMatrix ( TBlockMatrix * pMatrix, TFloat xi, bool smaller,
                             const char *filename );

    static void drawMatrix ( TMatrixBlock * pMatrix, TFloat xi,
                             bool smaller, const char * filename );

    ////////////////////////////////////////////
    //
    // draw a graph
    //
    static void draw ( TGraph& graph, const char *filename, bool drawDel = FALSE );
    static void draw ( TGraph& graph, TVectorBlock<uint>& permutation,
                       const char *filename, bool drawDel = FALSE );
    static void draw ( TList<TGraph>& cycles, const char *filename );
    static void draw ( TList<TGraph>& cycles, TVectorBlock<uint>& permutation, 
                       const char *filename );
    static void draw ( TList<TNode>& listCycNodes, const char *filename );

    static void drawGraph ( T3DCellEdgeGraph * pGraph,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawGraph ( T3DSurfaceGraph * pGraph,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawGraph ( T3DSurfaceGraph * pGraph, TPermutation * pPerm,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawGraph ( TList< T3DSurfaceGraph > * pList,
                            TPermutation * pPerm,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawGraph ( T3DGraph * pGraph,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawGraph ( T3DGraph * pGraph, TPermutation * pPerm,
                            const char *filename,
                            int dev = VRML | POVRAY );

    static void drawFlowSurfaces ( TList< T3DCellEdgeGraph > * pFSList,
                                   const char *filename,
                                   int dev = VRML | POVRAY );

    static void drawCycles ( TList< TGraph > * pListCycles,
                             const char *filename,
                             int dev = VRML | POVRAY );

    ////////////////////////////////////////////
    //
    // draw flow
    //
    static void drawFlow ( T2DBasisEqn * pEqn,
                           TFloat stepX, TFloat stepY,
                           const char * filename );

    static void drawFlow ( TConvEqn<3> * pEqn,
                           TFloat stepX, TFloat stepY, TFloat stepZ,
                           const char * filename, int dev = VRML | POVRAY );
    
    ////////////////////////////////////////////
    //
    // draw streaklines
    //
    // static void drawStreakLine ( T2DNet &net, TBlockVector<TFloat> &velocity,
    //                              T2DBasisEqn &eqn,
    //                              const char *filename );

    ////////////////////////////////////////////
    //
    // draw convergence rates
    //
    static void drawConv ( TDynArray<TFloat> *parrConv, const char *filename );

protected:
    ////////////////////////////////////////////
    //
    // misc. helpfull methods
    //

    // draw given view via postscript
    static void drawPS ( TView * pView, const char * filename );

    // draw given view according to device
    static void draw ( TView * pView, const char * filename, int dev );
    
    // draw given view via VRML
    static void drawVRML ( TView * pView, const char *filename );

    // draw given view via POVRay
    static void drawPOVRay ( TView * pView, const char *filename );
};

#endif // __DRAWAPP_HH
