#ifndef __LIST_HH
#define __LIST_HH
//
//  Project   : libmisc
//  File      : list.hh
//  Author    : Ronald Kriemann
//  Purpose   : Template class for a double concatenated list
//  Date      : 20.04.1997
//  Problems  : - TListIterator::getItem() can return a NULL-pointer
//                but in TList::remove(Item&) we expect a reference !
//
// arch-tag: 7a55e3c8-5536-43cd-abea-cd43ceaf12f3
//

#include "defines.hh"
#include "memory.hh"

////////////////////////////////////////////////////
//
// class for a single item in the list
//
template <class T> 
class TItem // : public TMemory
{
public:
	T         *pElement;     // pointer to data
	TItem<T>  *pPrev;        // the previous element
	TItem<T>  *pNext;        // next element in list
	
public:
	TItem ( T        *element, 
			TItem<T> *prev = NULL, 
			TItem<T> *next = NULL )
	{
		pElement = element;
		pPrev    = prev;
		pNext    = next;
	}
};

////////////////////////////////////////////////////
//
// forward declaration for the iterator
//
template<class T> class TListIterator;

////////////////////////////////////////////////////
// 
// Template class list
//
// - stores pointers to objects in the list
//
template <class T> 
class TList : public TMemory 
{
	// set the iterator as friend
	friend class TListIterator<T>;

protected:
	TItem<T>  *pFirst;       // pointer to the first element
	TItem<T>  *pLast;        // and to the last one
	TItem<T>  *pActItem;     // act element in list access
	int        nSize;        // act. size of the list

public:
	// constructors
	TList ();

	// destructor
	virtual ~TList ();

	// return the number of items stored
	const int getSize () const;

	// return first (last) pointer in list
	T* getFirst ();

	T* getLast ();

	// add a new object to the list
	virtual void add ( T &object );

	// add a new object to the begin of the list
	virtual void insert ( T &object );

	// insert item after given item
	virtual void insertAfter ( T &after, T &object );

	// insert item before given item
	virtual void insertBefore ( T &before, T &object );

	// replace one item by another
	virtual void replace ( T &oldObject, T &newObject );

	// remove item from list (access by element)
	// (NOT THE ELEMENT HIMSELF !!!)
	virtual void remove ( T &object );

	// remove the item (access by container class)
	virtual void remove ( TItem<T> &item );

	// search for an object and return the item
	virtual TItem<T>* search ( T &object );

	// clear the list (NOT THE ELEMENTS !!!)
	virtual void clear ();
};

//////////////////////////////////////////////////////////
// 
// class for an iterator for the previous list
//
template<class T>
class TListIterator : public TMemory 
{
protected:
	TItem<T>  *pCurrItem;  // current selected item
	TList<T>  *pList;      // used list for work

public:
	// constructor
	TListIterator ();
	
	// constructor
	TListIterator ( const TList<T> &list );

	// set the list, we iterate
	void setList ( const TList<T> &list );

	// return the act. item
	TItem<T>* getItem ();

	// return next object from list
	T* next ();

	// return the prev object from list
	T* prev ();

	// set counter to begin of list
	T* first ();

	// set counter to end of list
	T* last ();

	// search for an object and set current pointer to it
	bool search ( T &object );

	//
	// operators for easier accessing
	//

	// preincrement
	T* operator ++ ();

	// postincrement
	T* operator ++ (int);

	// predecrement
	T* operator -- ();

	// postdecrement
	T* operator -- (int);

	// access actual element
	T* operator () ();

    //
    // methods for O(1) modification of lists
    //

    // remove act. element in list and inc. iterator
    // (for exact specs see implementation below)
    void remCurrAndInc ();
};

//////////////////////////////////////////////////////////////
//
// TList (Implementation)
//

//
// constructors
//
template <class T>
TList<T>::TList () 
{
	pFirst = pLast = pActItem = NULL;
	nSize = 0;
}

//
// destructor
//
template <class T>
TList<T>::~TList () 
{
	if (nSize > 0) clear(); 
}

//
// return the number of items stored
//
template <class T>
const int 
TList<T>::getSize () const 
{ 
	return nSize; 
}

//
// return first (last) pointer in list
//
template <class T>
T* 
TList<T>::getFirst () 
{ 
	if (pFirst != NULL)
    {
        pActItem = pFirst;
		return pFirst->pElement;
    }// if
	else
		return NULL; 
}

template <class T>
T* 
TList<T>::getLast ()
{ 
	if (pLast != NULL)
    {
        pActItem = pLast;
		return pLast->pElement;
    }// if
	else
		return NULL;
}

//
// add a new object to the end of list
//
template <class T>
void 
TList<T>::add ( T &object ) 
{
	if (pFirst == NULL)
	{
		pLast = pFirst = new TItem<T>( &object );
	}// if
	else
	{
		pLast->pNext = new TItem<T>( &object, pLast );
		pLast = pLast->pNext;
	}// else

	nSize++;
}

//
// add a new object to the begin of the list
//
template <class T>
void 
TList<T>::insert ( T &object ) 
{
	if (pFirst == NULL)
	{
		pLast = pFirst = new TItem<T>( &object );
	}// if
	else
	{
		pFirst->pPrev = new TItem<T>( &object, NULL, pFirst );
		pFirst = pFirst->pPrev;
	}// else

	nSize++;
}

//
// insert item after given item
//
template <class T>
void 
TList<T>::insertAfter ( T &after, T &object ) 
{
	TItem<T> *pItem = search( after );

	if (pItem != NULL)
	{
		TItem<T> *pTmp = new TItem<T>( &object, pItem, pItem->pNext );

		if (pItem->pNext != NULL) 
			pItem->pNext->pPrev = pTmp;

		if (pLast == pItem) 
			pLast = pTmp;

		pItem->pNext = pTmp;

		nSize++;
	}// if
	else
		cout << "TList::insertAfter (T*) : could not insert object" << endl;
}

//
// insert item before given item
//
template <class T>
void 
TList<T>::insertBefore ( T &before, T &object ) 
{
	TItem<T> *pItem = search( before );

	if (pItem != NULL)
	{
		TItem<T> *pTmp = new TItem<T>( &object, pItem->pPrev, pItem );

		if (pItem->pPrev != NULL) 
			pItem->pPrev->pNext = pTmp;

		if (pFirst == pItem) 
			pFirst = pTmp;

		pItem->pPrev = pTmp;

		nSize++;
	}// if
	else
		cout << "TList::insertBefore (T*) : could not insert object" << endl;
}

//
// replace one item by another
//
template <class T>
void 
TList<T>::replace ( T &oldObject, T &newObject )
{
	TItem<T>  *pItem = search( oldObject );

	if (pItem != NULL)
	{
		pItem->pElement = (T*) &newObject;
	}// if
	else
		cout << "TList::replace (T&,T&) : could not replace object" << endl;
}

//
// remove item from list (access by element)
// (NOT THE ELEMENT HIMSELF !!!)
//
template <class T>
void 
TList<T>::remove ( T &object ) 
{
	TItem<T> *pItem = search( object );

	if (pItem != NULL) 
	{
		if (pItem == pFirst)
		{
			if (pItem == pLast)
				pFirst = pLast = NULL;
			else 
			{
				pFirst = pItem->pNext;
				pFirst->pPrev = NULL;
			}// else
		}// if
		else
		{
			if (pItem == pLast)
			{
				pLast = pItem->pPrev;
				pLast->pNext = NULL;
			}// if
			else
			{
				pItem->pPrev->pNext = pItem->pNext;
				pItem->pNext->pPrev = pItem->pPrev;
			}// else
		}// else

		if (pActItem == pItem) pActItem = NULL;

		delete pItem;
		nSize--;
	}// if
	else
		cout << "TList::remove(T*) : object not in list" << endl;
}

//
// remove the item (access by container class)
//
template <class T>
void 
TList<T>::remove ( TItem<T> &item )
{
	TItem<T> *pItem = &item;

	if (pItem == pFirst)
	{
		if (pItem == pLast)
			pFirst = pLast = NULL;
		else 
		{
			pFirst = pItem->pNext;
			pFirst->pPrev = NULL;
		}// else
	}// if
	else
	{
		if (pItem == pLast)
		{
			pLast = pItem->pPrev;
			pLast->pNext = NULL;
		}// if
		else
		{
			pItem->pPrev->pNext = pItem->pNext;
			pItem->pNext->pPrev = pItem->pPrev;
		}// else
	}// else

	if (pActItem == pItem) pActItem = NULL;

	delete pItem;
	nSize--;
}

//
// search for an object and return the item
//
template <class T>
TItem<T>* 
TList<T>::search ( T &object ) 
{
	TItem<T> *pItem   = pFirst;
	T        *pObject = &object;

	if ((pActItem != NULL) && (pActItem->pElement == pObject))
	{
		return pActItem;
	}// if

	while (pItem != NULL)
	{
		if (pItem->pElement == pObject)
		{
			pActItem = pItem;
			return pItem;
		}// if

		pItem = pItem->pNext;
	}// while

	return NULL;
}

//
// clear the list (NOT THE ELEMENTS !!!)
//
template <class T>
void 
TList<T>::clear ()
{
	if (nSize == 0) return;

	TItem<T> *pItem = pFirst;
	TItem<T> *pOld;

	while (pItem != NULL)
	{
		pOld = pItem->pNext;
		if (pItem != NULL) delete pItem;
		pItem = pOld;
	}// while

	pFirst = pLast = pActItem = NULL;
	nSize  = 0;
}

//////////////////////////////////////////////////////////////
//
// TListIterator (Implementation)
//

//
// constructor
//
template <class T>
TListIterator<T>::TListIterator ()
{
	pCurrItem = NULL;
	pList     = NULL;
}
	
template <class T>
TListIterator<T>::TListIterator ( const TList<T> &list )
{
	setList( list );
}

//
// set the list, we iterate
//
template <class T>
void 
TListIterator<T>::setList ( const TList<T> &list )
{
	pList = (TList<T>*) &list;

	if (pList != NULL)
		pCurrItem = pList->pFirst;
	else
		pCurrItem = NULL;
}

//
// return the act. item
//
template <class T>
TItem<T>* 
TListIterator<T>::getItem ()
{
	return pCurrItem;
}

//
// return next object from list
//
template <class T>
T* 
TListIterator<T>::next () 
{
	return ++(*this);
}

//
// return the prev object from list
//
template <class T>
T* 
TListIterator<T>::prev () 
{
	return --(*this);
}

//
// set counter to begin of list
//
template <class T>
T* 
TListIterator<T>::first () 
{
	if (pList == NULL) return NULL;

	pCurrItem = pList->pFirst;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return pCurrItem->pElement;
	}// if
	else
		return NULL;
}

//
// set counter to end of list
//
template <class T>
T* 
TListIterator<T>::last () 
{
	if (pList == NULL) return NULL;

	pCurrItem = pList->pLast;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return pCurrItem->pElement;
	}// if
	else
		return NULL;
}

//
// search for an object and set current pointer to it
//
template <class T>
bool 
TListIterator<T>::search ( T &object )
{
	if (pList != NULL)
	{
		pCurrItem = pList->search( object );

		if (pCurrItem == NULL)
			return FALSE;
		else
			return TRUE;
	}// if
	else
	{
		pCurrItem = NULL;
		return FALSE;
	}// else
}

//
// preincrement
//
template <class T>
T* 
TListIterator<T>::operator ++ ()
{
	if ((pCurrItem != NULL) && (pCurrItem->pNext != NULL))
	{
		pCurrItem = pCurrItem->pNext;
		pList->pActItem = pCurrItem;
		return pCurrItem->pElement;
	}// if

	return NULL;
}

//
// postincrement
//
template <class T>
T* 
TListIterator<T>::operator ++ (int)
{
	T* tmp;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		tmp = pCurrItem->pElement;
		pCurrItem = pCurrItem->pNext;
	}// if
	else
		tmp = NULL;

	return tmp;
}

//
// predecrement
//
template <class T>
T* 
TListIterator<T>::operator -- ()
{
	if ((pCurrItem != NULL) && (pCurrItem->pPrev != NULL))
	{
		pCurrItem = pCurrItem->pPrev;
		pList->pActItem = pCurrItem;
		return pCurrItem->pElement;
	}// if

	return NULL;
}

//
// postdecrement
//
template <class T>
T* 
TListIterator<T>::operator -- (int)
{
	T* tmp;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		tmp = pCurrItem->pElement;
		pCurrItem = pCurrItem->pPrev;
	}// if
	else
		tmp = NULL;

	return tmp;
}

//
// access actual element
//
template <class T>
T* 
TListIterator<T>::operator () ()
{
	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return pCurrItem->pElement;
	}// if
	else
		return NULL;
}

//
// methods for O(1) modification of lists
//

//
// remove act. element in list
// - if curritem == first => curritem becomes first->next
// - if curritem == last => curritem becomes NULL
// - else curritem becomes curritem->next
//
template <class T>
void
TListIterator<T>::remCurrAndInc ()
{
    if (pCurrItem != NULL)
    {
        TItem<T>  * pNewCurrItem;

        if (pCurrItem == pList->pFirst)
            pNewCurrItem = pCurrItem->pNext;
        else if (pCurrItem == pList->pLast)
            pNewCurrItem = NULL;
        else
            pNewCurrItem = pCurrItem->pNext;

        pList->remove( * pCurrItem );
        pCurrItem = pList->pActItem = pNewCurrItem;
    }// if
}

#endif   // __LIST_HH
