//
//  Project  : libnet
//  File     : matrixblock.hh
//  Author   : Sabine Gutsch / Ronald Kriemann / Thomas Probst   
//  Purpose  : class for a (sparse) matrixblock 
//
// arch-tag: f77e4fbc-c657-455e-ab20-60b61a211175
//

#ifndef __MATRIXBLOCK_HH
#define __MATRIXBLOCK_HH

#include "defines.hh"
#include "memory.hh"
#include "dynarray.hh"
#include "multiref.hh"

#include "blockvector.hh"
#include "sparsevector.hh"

// forward decl.
class TPermutation;

// Decomposition of a MatrixBlock
enum TDecompType { NONE, LU, CYC_LU, BLOCK_CYC_LU, ILU, CYC_SCHUR, SCHUR, BLOCK_DIAG };

//
// class for a block in blockmatrix
//
class TMatrixBlock : public TDynArray<TMatrixData>, public TMultiReference
{
protected:
    // precision for ostream;
    uint outputPrecision;

    // decomposition of matrix block
    TDecompType         decompType;

    // pointer to decomposition
    TMatrixBlock        *pDecomp;

    // vector containing information about block sizes in a 
    // block decomposition
    TVectorBlock<uint> *decompBlocks;
    uint               noOfDecompBlocks;

    // array of pointers to the diagonal (matrix) blocks
    TMatrixBlock        **diagBlocks;
  
public:
    // constructors
    TMatrixBlock ();
    TMatrixBlock ( uint rows );
    TMatrixBlock ( const TMatrixBlock & m );

    // destructor
    virtual ~TMatrixBlock ();
  
    // copy operator =
    TMatrixBlock& operator= ( const TMatrixBlock &m );
  
    // set a new size for the number of rows
    void setNoOfRows ( uint i );
  
    // return number of rows
    const uint getNoOfRows () const;
  
    // return DecompType
    TDecompType getDecompType ( ) const;
   
    // set DecompType
    void  setDecompType (TDecompType aDecompType);
   
    // return Pointer to Decomposition
    TMatrixBlock* getDecomposition() const;
   
    //set Pointer to Decomposition
    void setDecomposition (TMatrixBlock *apDecomp );
 
    // return Pointer to decompBlocks
    TVectorBlock<uint>* getDecompBlocksPointer() const;
 
    // return   *decompBlocks[i] 
    uint getDecompBlocksEntry(const uint i) const ;
  
    // return number of DecompBlocks()
    uint getNoOfDecompBlocks() const;

    // return Pointer to diagonal subblock
    TMatrixBlock* getDiagBlock( uint i ) const;

    // return DecompTypeName
    void printDecompTypeName ();
  
    // prints DecompTypename of Type
    void printDecompTypeName( TDecompType Type );

    // clear old decompostion
    void clearDecomposition();
  
    // calculate decoposition
    void decompose ( TDecompType Type, 
                     TVectorBlock<uint> *aDecompBlocks = NULL);
    // calculate (LU)^{-1} v
    TVectorBlock<TFloat> solveDecomp ( const TVectorBlock<TFloat> &v );
  
    // set sparsity pattern to cyclic (upper) bidiagonal matrixblock
    void setCyclicPattern ();

    // create diagonal blocks 
    void createDiagBlock ( uint blockNo, uint startIndex, 
                            uint endIndex );
   
    // create bidiagonal blocks ()
    void createBidiagBlocks ( uint blockNo, uint startIndex, 
                              uint endIndex );
  
   
    // block addition (we assume, that the other block has the same
    // index structure and same size of contents)
    TMatrixBlock& operator+= ( const TMatrixBlock& block );
    TMatrixBlock  operator+  ( const TMatrixBlock& block ) const;
  
    // block subtraction (we assume, that the other block has the same
    // index structure and same size of contents)
    TMatrixBlock& operator-= ( const TMatrixBlock& block );
    TMatrixBlock  operator-  ( const TMatrixBlock& block ) const;
  
    // scalar multiplication
    TMatrixBlock& operator*= ( TFloat scalar );
    TMatrixBlock  operator*  ( TFloat scalar ) const;
    friend TMatrixBlock operator* ( TFloat scalar, const TMatrixBlock& block );
  
    // block - vectorblock multiplication 
    TVectorBlock<TFloat> operator* ( const TVectorBlock<TFloat>& v ) const;
  
    // row sum norm (infinity norm)
    TFloat rowSumNorm ( ) const;

    // block - vectorblock multiplication  with transpose matrix
    void transposeMult (const TVectorBlock<TFloat>& v, 
                        TVectorBlock<TFloat>& w ) const;

    // multiply with diagonal of block (for full vectors)
    TVectorBlock<TFloat> diagMult ( const TVectorBlock<TFloat> &v )    const;
    // multiply with inverse of diagonal part of block (for full vectors)
    // (only meaningful for quadratic matrices)
    TVectorBlock<TFloat> invDiagMult ( const TVectorBlock<TFloat> &v )    const;
    // for sparse vectors
    TMatrixData invDiagMult ( const TMatrixData& data ) const;
    
    // multiplication with block diagonal
    TVectorBlock<TFloat> blockDiagMult ( const TVectorBlock<TFloat> &v ) const;

    // multiply with inverse of lower triangular part of block
    // (only meaningful for quadratic matrices)
    TVectorBlock<TFloat> invLowerMult ( const TVectorBlock<TFloat> &v ) const;
    TVectorBlock<TFloat> invLowerMult ( const TVectorBlock<TFloat> &v,
                                        const TFloat & diagDamp ) const;
    TVectorBlock<TFloat> invUnitLowerMult ( const TVectorBlock<TFloat> &v ) const;

    // multiply with inverse of block lower triangular part of block
    // (only meaningful for quadratic matrices)
    // assumes ordering with concentric cycles and corresponding
    // decomposition
    TVectorBlock<TFloat> invBlockLowerMult ( const TVectorBlock<TFloat> &v ) const;
    // for sparse vectors (tpr)
    TVectorBlock<TFloat> invLowerMult ( const TMatrixData& data ) const;

    // multiply with inverse of upper triangular part of block
    // (only meaningful for quadratic matrices)
    TVectorBlock<TFloat> invUpperMult ( const TVectorBlock<TFloat> &v ) const;
    TVectorBlock<TFloat> invUpperMult ( const TVectorBlock<TFloat> &v,
                                        const TFloat & diagDamp ) const;

    // multiply with inverse of block upper triangular part of block
    // (only meaningful for quadratic matrices)
    // assumes ordering with concentric cycles and corresponding
    // decomposition
    TVectorBlock<TFloat> invBlockUpperMult (const TVectorBlock<TFloat> &v) const;

    // get transpose of j'th row
    TVectorBlock<TFloat> getTransposedRow ( uint j, uint size ) const;

    // get column as vectorblock
    TVectorBlock<TFloat> getColumn ( uint j ) const;

    // permutation for matrix block
    //  p: row permutation, q: column permutation
    virtual void permute ( const TVectorBlock<uint> *p,
                           const TVectorBlock<uint> *q );
    virtual void permute ( const TPermutation * pPerm1,
                           const TPermutation * pPerm2 );

    // output
    void setOutputPrecision (const uint prec ) { outputPrecision = prec;}
    uint getOutputPrecision () const { return (outputPrecision); }
    friend ostream& operator << (ostream& os, TMatrixBlock& matrixblock);
};

#endif  // __MATRIXBLOCK_HH
