//
//  Project     : libgraph
//  File        : TPlanarFVS.hh
//  Author      : Ronald Kriemann
//  Purpose     : class building FVS for planar graphs
//
// arch-tag: 9d2a5da9-9167-42c3-b5ae-0312417a32a4
//

#ifndef __TPLANARFVS_HH
#define __TPLANARFVS_HH

#include "memory.hh"
#include "list.hh"

#include "basenode.hh"
#include "basegraph.hh"

// different kind of cycles
#define MIN_CYC  0
#define MAX_CYC  1

// forward decl.
class TPlanarNode;
class T2DPlanarNode;
class TCycleNode;
class TPlanarCycle;

//
// This class builds a Feedback Vertex Set of a given graph
// provided the graph fulfills the OFD condition
//
class TPlanarFVS : public TMemory
{
protected:
    // graph we build fvs of
    TGraph                      * pGraph;

    // stack for div. DFS runs
    TList< TPlanarNode >          stack;
    TListIterator< TPlanarNode >  iterStack;
    
public:
    ///////////////////////////////////////////
    //
    // constructor
    //
    TPlanarFVS ();

    ///////////////////////////////////////////
    //
    // builds the FVS
    //
    TList< TNode > * buildFVS ( TGraph & graph );
    TList< TNode > * buildFVS_SCC ( TGraph * pGraph );

protected:
    ///////////////////////////////////////////
    //
    // misc. help methods 
    //
    
    // build graph consisting of cycles
    void buildZNodes ( TGraph * pZ, uint type );
    void buildZEdges ( TGraph * pZ, uint type );

    // insert given cycle into Z+/Z-
    void insertCycle ( TCycleNode * pZNode, TGraph * pZ, uint type );

    // try to reach other cycles via DFS
    void reachForward ( TPlanarNode * pNode, TPlanarCycle * pCycle,
                        TGraph * pZ, uint type );
    void reachBackward ( TPlanarNode * pNode, TPlanarCycle * pCycle,
                         TGraph * pZ, uint type );
    
    // find suitable startnode for buildCycle
    TPlanarNode * findStartNode ( TCycleNode * pLastCycle,
                                  TGraph * pZ, uint type );
    
    // maps a given node to a minimal/maximal cycle
    TPlanarCycle * buildCycle ( TPlanarNode * pStartNode, uint type );
    TPlanarCycle * buildUpdatedCycle ( TPlanarNode * pStartNode,
                                       TPlanarCycle * pOldCycle,
                                       uint type );

    // delete nodes on given cycle and apply add. info for nodes
    void deleteCycle ( TPlanarCycle * pCycle, uint type );
    
    // check if different cycle is reached and build edges if needed
    bool checkForCycle ( TPlanarNode * pNode, TEdge * pEdge,
                         TPlanarCycle * pCycle, uint type );
    
    // look if given edge is on left side of given cycle
    bool isLeftEdge ( TEdge * pTestEdge, TPlanarCycle * pCycle, uint type );

    // look if graph is cyclic or not (graph is represented by Z+/Z-)
    bool cyclic ( TGraph * pZPlus, TGraph * pZMinus );

    // return node in T+/T- ( a feedback vertex )
    TPlanarNode * getTPlusNode  ( TGraph * pZPlus );
    TPlanarNode * getTMinusNode ( TGraph * pZMinus );

    // add node to fvs
    void addFV ( TPlanarNode * pNode, TList< TNode > * pListFVS,
                 TGraph * pZPlus, TGraph * pZMinus );

    // find old cycles touched by newcycle (except oldcycle)
    void findTouchedCycles ( TPlanarCycle          * pNewCycle,
                             TPlanarCycle          * pOldCycle,
                             TList< TPlanarCycle > * pListTouchedCycles,
                             uint                    type );

    // remove given list of cycles from graph and from Z
    void removeCycles ( TList< TPlanarCycle > * pListCycles,
                        TGraph                * pZ,
                        uint                    type );
    
    // update Z graph after feedback vertex was deleted
    void updateZ ( TPlanarNode * pNode, TGraph * pZ, uint type );
    
    // undelete all nodes in given cycle
    void undelete ( TPlanarCycle * pCycle, bool bDelCycle, uint type );
    
    // a local version of the numbering algorithm
    void localNumber ( uint type, bool bDelNonCyclic );
    void localNumberCycle ( TPlanarCycle * pCycle, uint type, bool bDelNonCyclic );
    void localNumberFV ( TPlanarNode * pNode, uint type );
    void localSetF ( TPlanarNode * pNode, uint type,
                     bool bDelNonCyclic, bool bCheckCycle );
    void localSetL ( TPlanarNode * pNode, uint type,
                     bool bDelNonCyclic, bool bCheckCycle );

    //
    // testversion of adding feedback vertices by succ/pred of found feedback vertex
    //

    void markPred ( TCycleNode * pNode );
    void unmarkPred ( TCycleNode * pNode );

    void detectSuccFV ( TCycleNode * pMinNode, TCycleNode * pMaxNode,
                        TList< TNode > * pListFVS );

    void rebuildZ ( TGraph * pZPlus, TGraph * pZMinus, TList< TNode > * pListFVS );
};

#endif  // __TPLANARFVS_HH
