//
//  Project   : libnet
//  File      : schur.hh
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : Schur complement solver in second normal form
//
// arch-tag: 7fb3cfa3-8359-4c21-8384-ebf62ebeb81f
//

#ifndef __SCHUR_HH
#define __SCHUR_HH

#include "secondNF.hh"

#define UNSYMMETRIC 0
#define SYMMETRIC   1
#define BACKWARD    0
#define TWOBACKWARD 2

//
// class for block Schur complement iteration 
// (Schur complement approximation for the last block)
//
//    ( I     -C^{-1} B^T )   ( C     )^{-1}
//    (            I      )   ( B  -S )
//
//    where S = B C^{-1} B^T, and C^{-1} is approximated with
//    one forward Gauss-Seidel step, S by its diagonal
//
// !!! NOT USEFUL AS PRECONDITIONER !!!
//
class TSchur : public TSecondNF {
    uint maxInnerSteps;
    uint totalInnerSteps;
    TFloat innerReduceBy;
    TFloat tol;
    uint problemType;        // 1: symmetric, 0: nonsymmetric
    // Schur complement wrt first dim... unknowns
    uint dimSchurComplement; 
    // explicit approximate Schur complement
    TMatrixBlock schurComplement; 
    uint innerGS;            // 1: symmetric, 0: backward
 
public:
	// constructor
    TSchur ( TBlockMatrix *paA = NULL, 
             TBlockVector<TFloat> *paRHS = NULL, 
             TBlockVector<TFloat> *paX = NULL, 
             uint amaxiter = DEFAULT_MAX_ITERATION,
             TFloat aaccur = DEFAULT_ACCURACY, 
             uint averb = 0,
             TFloat aDamp = 1.0 );
	// destructor
    ~TSchur() {schurComplement.delReference();}

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &r );

	// WInvers for Stokes problem and scalar problem
    TBlockVector<TFloat> WInversStokes ( const TBlockVector<TFloat> &r );
    TBlockVector<TFloat> WInversScalar ( const TBlockVector<TFloat> &r );

	// set/get maximum number of inner steps
    void setMaxInnerSteps( uint aSteps ) { maxInnerSteps = aSteps; }
    uint getMaxInnerSteps( ) { return (maxInnerSteps); }

	// set/get actual number of inner steps
    uint getTotalInnerSteps( ) { return (totalInnerSteps); }
    void initTotalInnerSteps( ) { totalInnerSteps = 0; }

	// set/get innerReduceBy
    void setInnerReduceBy( TFloat r ) { innerReduceBy = r; }
    TFloat getInnerReduceBy( ) { return (innerReduceBy); }

	// set/get problem type
    void setProblemType( uint type ) { problemType = type; }
    uint getProblemType() { return (problemType); }

	// set alpha
    //void setAlpha( TFloat aAlpha ) { alpha = aAlpha; }

	// perform inner iterations to solve for (symmetric) Schur complement
    TVectorBlock<TFloat> innerSchurSymm( TVectorBlock<TFloat> &b );

	// perform inner iterations for nonsymmetric Schur complement
    TVectorBlock<TFloat> innerSchurNonSymm( TVectorBlock<TFloat> &b );

	// set inner Gauss-Seidel method to symmetric or backward
    void setInnerGS( uint aGS ) { innerGS = aGS; }

	// perform inner iterations for Schur complement of scalar equation
    TVectorBlock<TFloat> innerSchurScalar( TVectorBlock<TFloat> &b );

    // set/get size of Schur complement (for scalar problems) 
    void setDimSchurComplement( uint size ) { dimSchurComplement = size; }
    uint getDimSchurComplement( ) { return (dimSchurComplement); }

	// calculate Schur complement approximation
    void calcSchurComplement();

}; 



#endif  // __SCHUR_HH
