//
//  Project   : libnet
//  File      : secondNF.hh
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : (classical) linear solvers in second normal form
//
// arch-tag: e20d9126-564e-4d11-8b95-0f6607eb3487
//

#ifndef __SECONDNF_HH
#define __SECONDNF_HH

#include "solver.hh"

// abbr. for solvers
#define TFGS  TForwardGaussSeidel
#define TBGS  TBackwardGaussSeidel
#define TSGS  TSymmetricGaussSeidel

class TSecondNF : public TSolver {
protected:
    TFloat               dampFactor;
    TBlockVector<TFloat> r;
   public:
	// constructor
    TSecondNF (TBlockMatrix *paA, TBlockVector<TFloat> *paRHS, 
               TBlockVector<TFloat> *paX,
               uint amaxiter = DEFAULT_MAX_ITERATION, 
               TFloat aaccur = DEFAULT_ACCURACY, 
               uint averb = 0,
               TFloat aDamp = 1.0 );
	// destructor
    virtual ~TSecondNF ();

	// calculate WInvers
    virtual TBlockVector<TFloat> WInvers (const TBlockVector<TFloat> &v ) = 0;

	// one iteration step
    int step ();

	// set and get damping factor
    void setDampFactor ( const TFloat & aDamp );
    const TFloat & getDampFactor () const;

    // update internal variables if nessecary
    virtual void update ( const TBlockVector< TFloat > & vecX );
};

//
// class for Jacobi iteration
//
class TJacobi : public TSecondNF {
public:
	// constructor
    TJacobi ( TBlockMatrix *paA = NULL, 
              TBlockVector<TFloat> *paRHS = NULL, 
               TBlockVector<TFloat> *paX = NULL,  
               uint amaxiter = DEFAULT_MAX_ITERATION, 
               TFloat aaccur = DEFAULT_ACCURACY, 
               uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
}; 

//
// class for Richardson iteration
//
class TRichardson : public TSecondNF {
public:
	// constructor
    TRichardson ( TBlockMatrix *paA = NULL, 
                  TBlockVector<TFloat> *paRHS = NULL, 
                  TBlockVector<TFloat> *paX = NULL, 
                  uint amaxiter = DEFAULT_MAX_ITERATION, 
                  TFloat aaccur = DEFAULT_ACCURACY, 
                  uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
}; 


//
// class for forward Gauss-Seidel iteration
//
class TForwardGaussSeidel : public TSecondNF {
public:
	// constructor
    TForwardGaussSeidel ( TBlockMatrix *paA = NULL, 
                          TBlockVector<TFloat> *paRHS = NULL,
                          TBlockVector<TFloat> *paX = NULL,
                          uint amaxiter = DEFAULT_MAX_ITERATION, 
                          TFloat aaccur = DEFAULT_ACCURACY, 
                          uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
}; 

//
// class for backward Gauss-Seidel iteration
//
class TBackwardGaussSeidel : public TSecondNF {
public:
	// constructor
    TBackwardGaussSeidel ( TBlockMatrix *paA = NULL, 
                           TBlockVector<TFloat> *paRHS = NULL,
                           TBlockVector<TFloat> *paX = NULL,
                           uint amaxiter = DEFAULT_MAX_ITERATION, 
                           TFloat aaccur = DEFAULT_ACCURACY, 
                           uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
}; 

//
// class for symmetric Gauss-Seidel iteration
//
class TSymmetricGaussSeidel : public TSecondNF {
public:
	TSymmetricGaussSeidel ( TBlockMatrix *paA = NULL, 
                            TBlockVector<TFloat> *paRHS = NULL,
                            TBlockVector<TFloat> *paX = NULL,
                            uint amaxiter = DEFAULT_MAX_ITERATION, 
                            TFloat aaccur = DEFAULT_ACCURACY, 
                            uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
};

//
// class for symmetric Gauss-Seidel iteration
//
class TILU : public TSecondNF {
public:
	TILU ( TBlockMatrix *paA = NULL, 
                            TBlockVector<TFloat> *paRHS = NULL,
                            TBlockVector<TFloat> *paX = NULL,
                            uint amaxiter = DEFAULT_MAX_ITERATION, 
                            TFloat aaccur = DEFAULT_ACCURACY, 
                            uint averb = 0, TFloat aDamp = 1.0 );

    // calculate WInvers
    TBlockVector<TFloat> WInvers ( const TBlockVector<TFloat> &v );
};

#endif  // __SECONDNF_HH
