//
//  Project   : libnet
//  File      : vector.hh
//  Author    : Ronald Kriemann
//  Purpose   : template class for a vector (template in dimension)
//
// arch-tag: 214caf62-726a-4af5-a3bc-ea302f76db8e
//

#ifndef __VECTOR_HH
#define __VECTOR_HH

#include <math.h>

#include "memory.hh"
#include "multiref.hh"
#include "defines.hh"

// for easier access
#define X  0
#define Y  1
#define Z  2

// forward decl for friend methods
template <int N> class TVector;
template <int N> TVector<N> operator *  ( const TFloat &, const TVector<N>& );
template <int N> ostream &  operator << ( ostream &, const TVector<N> & );

//
// - this class is a template class in the dimension of the space
// - for 2d and for 3d, it has specilized methods if needed (for speed)
// - all usual operators (+,-,* usw) are defined
//
template <int N>
class TVector : public TMemory 
{
protected:
    // array for the coord.
    TFloat  arrCoord[N];

public:
    // constructors (for 2d and 3d)
    TVector ();
    TVector ( TFloat coord[] );
    TVector ( const TVector<N> & vector );
    TVector ( const TFloat & x, const TFloat & y );
    TVector ( const TFloat & x, const TFloat & y, const TFloat & z );

    // destructor
    ~TVector();

    // add operators
    TVector<N>& operator+= ( const TVector<N>& vector );
    TVector<N>  operator+ ( const TVector<N>& vector );

    // sub-and-set operator
    TVector<N>& operator-= ( const TVector<N>& vector );
    TVector<N>  operator- ( const TVector<N>& vector );

    // scalar multiplikation (from left and right)
    TVector<N>& operator*= ( const TFloat & scalar );
    TVector<N>  operator* ( const TFloat & scalar );

    // inner produkt (euclid)
    TFloat operator* ( const TVector<N>& vector ) const;
    // euclid norm
    TFloat euclidNorm () const;
    // vector product (only for N==3): (*this) x v2
    TVector<N> cross ( const TVector<N>& v2 ); 

    // subscript operator
    TFloat& operator[] ( uint i );
    const TFloat& operator[] ( uint i ) const;

    // copy operator
    TVector<N> & operator = ( const TVector<N> & vector );
    TVector<N> & operator = ( const TFloat & scalar );

    // comparing
    const bool operator == ( const TVector<N> &vector ) const;
    const bool operator != ( const TVector<N> &vector ) const;

    // output
    friend ostream & operator << <> ( ostream & os, const TVector<N> & vector );
};

// some needed types
typedef TVector<2> T2DVector;
typedef TVector<3> T3DVector;

///////////////////////////////////////////////////////////////
//
// implementation
//

//
// constructors
//
template <int N>
TVector<N>::TVector( const TVector<N>& vector ) 
{ 
    (*this) = vector; 
}

//
// destructor
//
template <int N>
TVector<N>::~TVector() 
{}

//
// add operator
//
template <int N>
TVector<N> 
TVector<N>::operator+ ( const TVector<N>& vector )
{
    TVector<N>  vecTmp( *this );
    return vecTmp += vector;
}

//
// sub operators
//
template <int N>
TVector<N> 
TVector<N>::operator- ( const TVector<N>& vector ) 
{
    TVector<N>  vecTmp( *this );
    return vecTmp -= vector;
}

//
// scalar multiplication
//
template <int N>
TVector<N>
TVector<N>::operator* ( const TFloat & scalar ) 
{
    TVector<N>  vecTmp( *this );
    return vecTmp *= scalar;
}

template <int N>
TVector<N> 
operator* ( const TFloat & scalar, const TVector<N> & vector ) 
{
    TVector<N>  vecTmp( vector );
    return vecTmp *= scalar;
}

//
// euclid norm
//
template <int N>
inline
TFloat 
TVector<N>::euclidNorm () const 
{ 
    return (sqrt((*this) * (*this))); 
}

//
// ouput
//
template <int N>
ostream&
operator << (ostream& os, const TVector<N>& vector)
{
    int  i;
    
    os.setf( ios::fixed, ios::floatfield);
    os.precision(2);

    os << " ( ";
    for ( i = 0; i < N-1; i++)
        os << vector[i] << " , ";

    os << vector[N-1] << " ) ";

    return os;
}

#endif  // __VECTOR_HH
