//
//  Project     : libnet
//  File        : TMatrixData.cc
//  Author      : Ronald Kriemann 
//  Purpose     : class for a sparse row in a matrix
//
// arch-tag: a1c9cb12-8c13-4098-92f5-c34e13c33cbe
//

extern "C" {
#include <string.h>

#include <math.h>
}

#include "criterion.hh"

#include "cmdline.hh"

#include "TPermutation.hh"
#include "blockvector.hh"

#include "TMatrixData.hh"

bool TMatrixData::bUseCriterion = FALSE;

//
// constructors
//
TMatrixData::TMatrixData () 
{
    noOfEntries = 0; 
    arrIndex = NULL;
    arrValue = NULL; 
    writeable = FALSE;
}
TMatrixData::TMatrixData (uint s) 
{
    noOfEntries = s;
    arrIndex = new uint[noOfEntries];
    // fill with 0xFF, so all indices have max. value (4*10^9)
    memset( arrIndex, 0xFF, sizeof(uint)*noOfEntries );
    arrValue = new TFloat[noOfEntries];
    writeable = FALSE;
}
TMatrixData::TMatrixData ( const TMatrixData& data ) 
{
    noOfEntries = 0;
    arrIndex = NULL;
    arrValue = NULL;
    writeable = FALSE;
    (*this) = data;
}

// destructor
TMatrixData::~TMatrixData () 
{
    if (arrIndex != NULL) delete[] arrIndex;
    if (arrValue != NULL) delete[] arrValue;        
} 

//
// set usage of criterion in output
//
void
TMatrixData::useCriterion ( bool b )
{
    bUseCriterion = b;
}
    
//
// change writeable status
//
void
TMatrixData::writeableStatus ( const bool flag )
{
    writeable = flag;
}

//
// set the indexarray and the valuearray
//
void 
TMatrixData::set ( uint s, uint *index, TFloat *values ) 
{
    if (s != noOfEntries)
    {
        if (arrIndex != NULL) delete[] arrIndex;
        if (arrValue != NULL) delete[] arrValue;        
            
        noOfEntries = s;
        arrIndex = new uint[noOfEntries];
        arrValue = new TFloat[noOfEntries];
    }// if

    memcpy( arrIndex, index , sizeof(uint)*noOfEntries );
    memcpy( arrValue, values, sizeof(TFloat)*noOfEntries );
}

void
TMatrixData::set ( const TMatrixData &md ) 
{
    //        set( md.getNoOfEntries(), md.getIndex(), md.getValue() );
    (*this) = md;
}

//
// keep the first s entries of the index- and valuearray
//
void 
TMatrixData::set ( uint s ) 
{
    noOfEntries = s;
    
    if ( (arrIndex != NULL) && (arrValue != NULL) ) 
    {
        uint  *newIndex = new uint[noOfEntries];
        TFloat *newValue = new TFloat[noOfEntries];
        
        memcpy( newIndex, arrIndex, sizeof(uint)*noOfEntries );
        memcpy( newValue, arrValue, sizeof(TFloat)*noOfEntries );
        
        delete[] arrIndex;
        delete[] arrValue; 
        
        arrIndex = new uint[noOfEntries];
        arrValue = new TFloat[noOfEntries];
        
        memcpy( arrIndex, newIndex, sizeof(uint)*noOfEntries );
        memcpy( arrValue, newValue, sizeof(TFloat)*noOfEntries );
        
        delete[] newIndex;
        delete[] newValue;
    }// if
    else 
    {
        arrIndex = new uint[noOfEntries];
        arrValue = new TFloat[noOfEntries];
    }// else
}

// get noOfEntries
uint 
TMatrixData::getNoOfEntries () const
{
    return noOfEntries;
}

// return number of the i'th index
uint 
TMatrixData::getIndex( uint i ) const
{
    return arrIndex[i];
}

uint* 
TMatrixData::getIndex( ) const
{
    return (arrIndex);
}

// set number of the i'th index
void 
TMatrixData::setIndex( uint i, uint j, TFloat k ) 
{ 
    arrIndex[i] = j; 
    arrValue[i] = k; 
}

// return i'th value
TFloat 
TMatrixData::getValue( uint i ) const
{
    return (arrValue[i]);
}

TFloat* 
TMatrixData::getValue( ) const
{
    return (arrValue);
}

// set i'th value
void 
TMatrixData::setValue( uint i, TFloat val ) const
{
    arrValue[i] = val;
}

//
// has Index i within the first j indices
//
bool 
TMatrixData::hasIndex( uint i, uint j ) 
{
    uint k = 0;
    while( (k < j) && (k < noOfEntries)) {
        if (arrIndex[k] == i)
            return (TRUE);
        k++;
    }
    return (FALSE);
}   

//
// copy operator
//
TMatrixData& 
TMatrixData::operator= ( const TMatrixData& data )
{
    if (noOfEntries != data.noOfEntries)
    {
        if (arrIndex != NULL) delete[] arrIndex;
        if (arrValue != NULL) delete[] arrValue;

        noOfEntries = data.noOfEntries;

        arrIndex = new uint[noOfEntries];
        arrValue = new TFloat[noOfEntries];
    }// if

    memcpy( arrIndex, data.arrIndex, sizeof(uint)*noOfEntries );
    memcpy( arrValue, data.arrValue, sizeof(TFloat)*noOfEntries );

    return (*this);
}

//
// index operator 
//
TFloat& 
TMatrixData::operator[] ( uint i ) 
{
    uint *p = arrIndex;

    for (uint j = 0; j < noOfEntries; j++)
        if (*p++ == i)
            return (arrValue[j]);

    if (writeable)
    { // create new matrix entry
        //cout<< "    creating new matrix entry ["<<i<<"]"<<endl;
        if ( (arrIndex != NULL) && (arrValue != NULL) ) 
        {
            uint  *newIndex = new uint[noOfEntries+1];
            TFloat *newValue = new TFloat[noOfEntries+1];
         
            memcpy( newIndex, arrIndex, sizeof(uint)*noOfEntries );
            memcpy( newValue, arrValue, sizeof(TFloat)*noOfEntries );
         
            delete[] arrIndex;delete[] arrValue; 
         
         
            arrIndex = new uint[noOfEntries+1];
            arrValue = new TFloat[noOfEntries+1];
         
             
            memcpy( arrIndex, newIndex, sizeof(uint)*noOfEntries );
            memcpy( arrValue, newValue, sizeof(TFloat)*noOfEntries );
            arrIndex[noOfEntries]=i;
            arrValue[noOfEntries]=0.0;
         
            noOfEntries++;
            delete[] newIndex;   delete[] newValue;
        }// if
        else {
            noOfEntries=1;
            arrIndex = new uint[noOfEntries];
            arrValue = new TFloat[noOfEntries];
            arrIndex[noOfEntries-1]=i;
            arrValue[noOfEntries-1]=0.0;
        }// else
      
        return arrValue[noOfEntries-1];
    } 

    cerr << "TMatrixData::operator[] (uint) : index " << i 
         << " doesn't exit" << endl;
 
    // no index, no element
    return (arrValue[0]);
    //return (zero);
}
const TFloat& 
TMatrixData::operator[] ( uint i ) const 
{
    uint *p = arrIndex;

    for (uint j = 0; j < noOfEntries; j++)
        if (*p++ == i)
            return (arrValue[j]);
  
    cerr << "const TMatrixData::operator[] (uint) const : index " << i 
         << " doesn't exit" << endl;

    // no index, no element
    return (arrValue[0]);
    //return (zero);
}

//
// operator * with a scalar
//
TMatrixData& 
TMatrixData::operator*= ( TFloat scalar ) 
{
    for (uint i = 0; i < noOfEntries; i++)
        arrValue[i] *= scalar;

    return (*this);
}

//
// operator * (between matrix data and vector block)
//
TFloat
operator* ( const TMatrixData& matrixdata,
            const TVectorBlock<TFloat>& v) 
{
    TFloat fResult = 0.0;

    for (uint i = 0; i < matrixdata.noOfEntries; i++) 
        fResult += matrixdata.arrValue[i] * v[matrixdata.arrIndex[i]]; 

    return (fResult);
}

//
// inner product between TMatrixData as VectorBlocks
// (assumption: both matrixdata objects have same index structure)
//
TFloat 
TMatrixData::operator* ( const TMatrixData& data )
{
    TFloat res = 0.f;

    if (noOfEntries != data.noOfEntries)
        cerr << "TMatrixData::operator* (TMatrixData&) : inequal structure" << endl;

    for (uint i = 0; i < noOfEntries; i++)
        res += arrValue[i] * data.arrValue[i];

    return (res);
}

//
// operator += with another blockline (assume, that they have 
// same indexstructure, and all indices are set correct)
//
TMatrixData& 
TMatrixData::operator+= ( const TMatrixData& matrixdata ) 
{
    for (uint i = 0; i < noOfEntries; i++)
        arrValue[i] += matrixdata.arrValue[i];

    return (*this);
}

//
// operator -= with another blockline (assume, that they have 
// same indexstructure, and all indices are set correct)
//
TMatrixData& 
TMatrixData::operator-= ( const TMatrixData& matrixdata ) 
{
    for (uint i = 0; i < noOfEntries; i++)
        arrValue[i] -= matrixdata.arrValue[i];

    return (*this);
}

//
// invLowerMult : multiplication 
// (assumes diagonal entry in arrValue[0] )
//
TFloat 
TMatrixData::invLowerMult ( const TVectorBlock<TFloat> & v ) const
{
    TFloat result;
    uint   diag = arrIndex[0];

    result = v[diag];

    for (uint i = 1; i < noOfEntries; i++)
    {
        if ( arrIndex[i] < diag ) 
            result -= arrValue[i] * v[arrIndex[i]];
    }// for

    result *= 1.0 / arrValue[0];

    return result;
}

TFloat
TMatrixData::invLowerMult ( const TVectorBlock<TFloat> & v,
                            const TFloat & diagDamp ) const
{
    TFloat  result;
    uint    diag = arrIndex[0];

    result = v[diag];

    for (uint i = 1; i < noOfEntries; i++)
    {
        if ( arrIndex[i] < diag ) 
            result -= arrValue[i] * v[arrIndex[i]];
    }// for

    result *= 1.0 / ( diagDamp * arrValue[0] );

    return result;
}

//
// invLowerMult : multiplication 
// (assumes diagonal entry in arrValue[0] )
//
TFloat 
TMatrixData::invUnitLowerMult ( const TVectorBlock<TFloat> &v) const 
{
	TFloat fResult;
	uint diag = arrIndex[0];
	fResult = v[diag];

	for (uint i = 1; i < noOfEntries; i++) 
		if ( arrIndex[i] < diag ) 
			fResult -= arrValue[i] * v[arrIndex[i]]; 

	return (fResult);
}


//
// lowerMult : multiplication 
// multiply with all lower entries < upperLimit
//
TFloat 
TMatrixData::lowerMult ( const TVectorBlock<TFloat> &v,
                         uint upperLimit ) const 
{
    TFloat fResult;
    uint diag = arrIndex[0];
    fResult = 0.0;

    for (uint i = 1; i < noOfEntries; i++) 
        if ( (arrIndex[i] < diag) && (arrIndex[i] < upperLimit) ) 
            fResult += arrValue[i] * v[arrIndex[i]]; 

    return (fResult);
}

//
// invUpperMult : multiplication 
// (assumes diagonal entry in arrValue[0] )
//
TFloat 
TMatrixData::invUpperMult ( const TVectorBlock<TFloat> &v, 
                            uint upperLimit ) const 
{
    TFloat fResult;
    uint diag = arrIndex[0];
    fResult = v[diag];

    for (uint i = 1; i < noOfEntries; i++) 
        if ( (arrIndex[i] > diag) && (arrIndex[i] < upperLimit) ) 
            fResult -= arrValue[i] * v[arrIndex[i]]; 

    fResult *= (1 / arrValue[0]);

    return (fResult);
}

TFloat
TMatrixData::invUpperMult ( const TVectorBlock<TFloat> &v, 
                            const TFloat & diagDamp ) const
{
    TFloat  result;
    uint    i;
    uint    diag = arrIndex[0];

    result = v[diag];

    for ( i = 1; i < noOfEntries; i++ ) 
        if (arrIndex[i] > diag)
            result -= arrValue[i] * v[arrIndex[i]]; 

    return result *= (1 / ( diagDamp * arrValue[0] ));
}

//
// upperMult : multiplication 
// multiply with all entries > blockIndex
//
TFloat 
TMatrixData::upperMult ( const TVectorBlock<TFloat> &v, 
                         uint lowerLimit ) const 
{
    TFloat fResult;
    uint diag = arrIndex[0];
    fResult = 0;

    for (uint i = 1; i < noOfEntries; i++) 
        if ( (arrIndex[i] > diag) && (arrIndex[i] > lowerLimit) ) 
            fResult += arrValue[i] * v[arrIndex[i]]; 

    return (fResult);
}

//
// multiply with all entries between upper and lower:
// ( lowerLimit <= entry index < upperLimit )
//
TFloat 
TMatrixData::mult ( const TVectorBlock<TFloat> &v,
                    uint lowerLimit, uint upperLimit ) const
{
    TFloat fResult = 0.0;

    for (uint i = 0; i < noOfEntries; i++) 
        if ( (arrIndex[i] >= lowerLimit) && (arrIndex[i] < upperLimit) ) 
            fResult += arrValue[i] * v[arrIndex[i]]; 

    return (fResult);
}

//
// permutation for row of matrixblock
//
void 
TMatrixData::permute ( const TVectorBlock<uint> * q, const TMatrixData & md ) 
{
    uint   mdEntries = md.getNoOfEntries();
    uint   newIndex[ mdEntries ];        
    TFloat newValue[ mdEntries ];        

    for (uint i = 0; i < mdEntries; i++) 
    {
        newIndex[i] = (*q)[ md.getIndex(i) ];
        newValue[i] = md.getValue(i);
    }// for
        
    set ( mdEntries, newIndex, newValue );
}

void
TMatrixData::permute ( const TPermutation * pPerm, const TMatrixData & md )
{
    uint   i;
    uint   mdEntries = md.getNoOfEntries();
    uint   newIndex[ mdEntries ];        
    TFloat newValue[ mdEntries ];        

    for (i = 0; i < mdEntries; i++)
    {
        newIndex[i] = (*pPerm)[ md.getIndex(i) ];
        newValue[i] = md.getValue(i);
    }// for
        
    set( mdEntries, newIndex, newValue );
}

//
// output
//
ostream &
operator<< ( ostream & os, const TMatrixData & data )
{
    uint    i;
    TFloat  diagValue = data.arrValue[0];
    
//    os.setf( ios::fixed, ios::floatfield);

    for ( i = 0; i < data.noOfEntries; i++ )
    {
        if ((i > 0) && (TMatrixData::bUseCriterion))
        {
            if (param.criterion == E3)
            {
                if ( ! (fabs(data.arrValue[i]) > fabs(param.xi * diagValue)) )
                    continue;
            }// switch
        }// if

        os << "(" << data.arrIndex[i] << ") " << data.arrValue[i] << " , ";
    }// for

    return (os);
}
