//
//  Project     : libnet
//  File        : blockmatrix.cc
//  Author      : Sabine Gutsch / Ronald Kriemann / Thomas Probst
//  Purpose     : a matrix is a TDynArray with type TMatrixData
//
// arch-tag: c3c2b44e-7665-46af-97b9-6711d262fc10
//

#include <math.h>

#include "blockmatrix.hh"

///////////////////////////////////////////////////////////////
//
// TMatrixBlockRow
//

//
// constructors
//
TMatrixBlockRow::TMatrixBlockRow () 
{
	n = 0;
	pColumns = NULL;
}

//
//
//
TMatrixBlockRow::TMatrixBlockRow ( uint number ) 
{
	// first create an array of n pointers to TMatrixBlock
	n = number;
	pColumns = new TMatrixBlock * [n];
    
	// then init. every pointer
	for (uint i = 0; i < n; i++)
		pColumns[i] = new TMatrixBlock;
}

//
// init. every block with an individual size
//
TMatrixBlockRow::TMatrixBlockRow ( uint number, uint * size ) 
{
	// first create an array of n pointers to TMatrixBlock
	n = number;
	pColumns = new TMatrixBlock * [n];
    
	// then init. every pointer with a matrixblock of size[]
	for (uint i = 0; i < n; i++)
		pColumns[i] = new TMatrixBlock( size[i] );
}

//
//  copy constructor
//
TMatrixBlockRow::TMatrixBlockRow (const TMatrixBlockRow& Row ) 
{
	// first create an array of n pointers to TMatrixBlock
	n = Row.getNoOfBlocks();
	pColumns = new TMatrixBlock * [n];
    
	// then init. every pointer with a matrixblock of size[]
	for (uint i = 0; i < n; i++)
		pColumns[i] = new TMatrixBlock( *(Row.pColumns[i]) );
}

TMatrixBlockRow::TMatrixBlockRow ( TMatrixBlock * pMatrix )
{
    n = 1;
	pColumns = new TMatrixBlock * [n];
    pColumns[0] = pMatrix;
    pMatrix->addRef();
}

//
// destructor
//
TMatrixBlockRow::~TMatrixBlockRow () 
{
	if (pColumns != NULL)
	{
		for ( uint i = 0; i < n; i++ )
        {
            if (pColumns[i] != NULL)
                pColumns[i]->free();
        }// for
        
		delete[] pColumns;
	}// if
}

//
// return the number of blocks
//
uint 
TMatrixBlockRow::getNoOfBlocks () const 
{ 
	return n;
}

//
// set a matrixblock to given matrix
//
void
TMatrixBlockRow::change ( uint i, TMatrixBlock * pMatrix )
{
    if (i < n)
    {
        if (pColumns[i] != NULL)
            pColumns[i]->free();
        
        pColumns[i] = pMatrix;
        pMatrix->addRef();
    }// if
    else
        cout << "TMatrixBlockRow::change (...) : index out of range" << endl;
}
    
//
// index operator
//
TMatrixBlock& 
TMatrixBlockRow::operator[] ( uint i ) 
{ 
	if (i >= n)
		cerr << "TMatrixBlockRow::operator[] (" << i << ") : "
			 << "out of bound (n=" << n << ")" << endl;

	return *(pColumns[i]); 
}

const TMatrixBlock& 
TMatrixBlockRow::operator[] ( uint i ) const 
{ 
	if (i >= n)
		cerr << "const TMatrixBlockRow::operator[] (" << i << ") const : "
			 << "out of bound (n=" << n << ")" << endl;

	return *(pColumns[i]); 
}

//////////////////////////////////////////////////////
//
//  TBlockMatrix
//

//
// constructors
//
TBlockMatrix::TBlockMatrix () 
{
	pRows = NULL;
	n = 0;
    pLUdecomposition = NULL;
}

//
//
//
TBlockMatrix::TBlockMatrix (uint number) 
{
    pLUdecomposition = NULL;
    
	// first create an array of n pointers to TMatrixBlockRow
	n = number;
	pRows = new TMatrixBlockRow * [n];
    
	// then init. every pointer with a blockrow
	for (uint i = 0; i < n; i++)
		pRows[i] = new TMatrixBlockRow;
}

//
// init. every row with an individual size
//
TBlockMatrix::TBlockMatrix ( uint number, uint *size ) 
{
    pLUdecomposition = NULL;
    
	// first create an array of n pointers to TMatrixBlockRow
	n = number;
	pRows = new TMatrixBlockRow * [n];
    
	// then init. every pointer with a blockrow of a given size
	for (uint i = 0; i < n; i++)
		pRows[i] = new TMatrixBlockRow(n, (size + (n*i)) );
}

//
// constructor with other matrix
//
TBlockMatrix::TBlockMatrix ( const TBlockMatrix& bm ) 
{
    pLUdecomposition = NULL;
    
	n = bm.getNoOfRows(); 
	pRows = new TMatrixBlockRow * [n];
    
	for (uint i = 0; i < n; i++)
		pRows[i] = new TMatrixBlockRow(bm[i] );
}

//
// constructor with matrix block
//
TBlockMatrix::TBlockMatrix ( const TMatrixBlock &m ) 
{
    pLUdecomposition = NULL;
    
	n = 1;
	pRows = new TMatrixBlockRow * [1];
	pRows[0] = new TMatrixBlockRow( 1 );
	(*pRows)[0][0] = m;
}

TBlockMatrix::TBlockMatrix ( TMatrixBlock * pMatrix )
{
    pLUdecomposition = NULL;
	n = 1;
    
	pRows    = new TMatrixBlockRow * [1];
    pRows[0] = new TMatrixBlockRow( pMatrix );
}

//
// destructor
//
TBlockMatrix::~TBlockMatrix () 
{
	if (pRows != NULL)
	{
		for (uint i = 0; i < n; i++)
			delete pRows[i];
		delete[] pRows;
	}// if
    if (pLUdecomposition != NULL)
        delete (pLUdecomposition);

}

//
// get number of rows
//
const uint 
TBlockMatrix::getNoOfRows() const 
{ 
	return n; 
}

//
// index operator [] such that [][][][] is defined for matrix
//
TMatrixBlockRow& 
TBlockMatrix::operator[] ( uint i ) 
{ 
	if (i >= n)
		cerr << "TBlockMatrix::operator[] (" << i << ") : "
			 << "out of bound (n=" << n << ")" << endl;

	return *(pRows[i]);
}

const TMatrixBlockRow& 
TBlockMatrix::operator[] ( uint i ) const 
{ 
	if (i >= n)
		cerr << "const TBlockMatrix::operator[] (" << i << ") const : "
			 << "out of bound (n=" << n << ")" << endl;

	return *(pRows[i]); 
}

//
// matrix - vector multiplication
//
void
TBlockMatrix::matmul ( TBlockVector< TFloat > & x, const TBlockVector< TFloat > & y ) const
{
	uint       i, j, k;
    uint      size;
	TFloat     fResult;

    y.clone( x );
    
	// perform matrix * vector with each blockrow
	for (i = 0; i < n; i++)
    {
        size = y[i].getSize();
        
		// for all elements in a vectorblock
		for (j = 0; j < size; j++)
        {
			fResult = 0.0;
				
			// go over all blocks
			for (k = 0; k < n; k++)
            {
				// and use * operator in matrixdata
				if ( (*this)[i][k].getNoOfRows() > 0 )
					fResult += (*this)[i][k][j] * y[k];
			}// for

			x[i][j] = fResult;
		}// for
    }// for
}

TBlockVector<TFloat> 
TBlockMatrix::operator * ( const TBlockVector<TFloat>& vec ) const
{
    TBlockVector<TFloat> vecTmp;
    
    matmul( vecTmp, vec );

    return vecTmp;
}

// get tranposed j'th row of i'th matrix block row 
// for normal blockvectors
TBlockVector<TFloat> 
TBlockMatrix::getTransposedRow( uint i, uint j, uint *size) 
{
	TBlockVector<TFloat> vecTmp(n, size);

	for (uint k = 0; k < n; k++ ) 
		// get transposed row of size[k]
		if ( (*pRows[i])[k].getNoOfRows() > 0 )
			vecTmp[k] = (*pRows[i])[k].getTransposedRow( j, size[k]);

	return (vecTmp);
}
// for sparsevectors with copy
TSparseBlockVector 
TBlockMatrix::getTransposedRow( uint i, uint j ) 
{
	TSparseBlockVector vecTmp( n );
	
	for (uint k = 0; k < n; k++ ) 
		// check whether the block has elements
		if ( (*pRows[i])[k].getNoOfRows() > 0 )
			// and assign the needed matrixdata
			vecTmp[k] = (*pRows[i])[k][j];
        else vecTmp[k].set( 0 );
	
	return (vecTmp);
}
// for sparsevectors without copy
void 
TBlockMatrix::getTransposedRow(TSparseBlockVector& vec, uint i, uint j) 
{
	vec.setNoOfBlocks( n );
	
	for (uint k = 0; k < n; k++ ) 
		// check whether the block has elements
		if ( (*pRows[i])[k].getNoOfRows() > 0 )
			// and assign the needed matrixdata
			vec[k] = (*pRows[i])[k][j];
        else vec[k].set( 0 );
}

//
// multiplication with transpose matrix
//
TBlockVector<TFloat> 
TBlockMatrix::transposeMult ( const TBlockVector<TFloat>& vec ) const
{
	TBlockVector<TFloat> vecTmp1(1,n);
	TBlockVector<TFloat> vecTmp2(1,n);
	// we assume n x n blocks in matrix
    uint row, col;
	
	// set the size of each block in vecTmp to the size in vec
	for (row = 0; row < n; row++) 
	{
		vecTmp1[row].setSize( vec[row].getSize() );
        vecTmp1[row].set( 0.0 );
		vecTmp2[row].setSize( vec[row].getSize() );
        vecTmp2[row].set( 0.0 );
	}// for

    for (col = 0; col < n; col++)
	{
        for (row = 0; row < n; row++) 
		{
		    // check whether the block has elements
		    if ( (*pRows[row])[col].getNoOfRows() > 0 ) 
			{
                (*pRows[row])[col].transposeMult( vec[row], vecTmp1[col]);
                vecTmp2[col] += vecTmp1[col];
			}// if
		}// for
	}// for

    return (vecTmp2);
}
	
//
// diag - vector multiplication 
// (we assume, that all blocks on the blockdiag are mxm - mat.)
//
TBlockVector<TFloat> 
TBlockMatrix::diagMult ( const TBlockVector<TFloat>& vec ) const 
{
	uint         i, j;
	TBlockVector<TFloat> vecTmp(1,n);
	
	// set the size of each block in vecTmp to the size in vec
	for (i = 0; i < n; i++)
		vecTmp[i].setSize( vec[i].getSize() );
	
	// and now perform diagmatrix * vector with each block
	for (i = 0; i < n; i++)
		// for all elements in a vectorblock
		for (j = 0; j < vec[i].getSize(); j++)
			vecTmp[i][j] = (*this)[i][i][j].getValue(0) * vec[i][j];
	
	return (vecTmp); 
}

//
// invDiag - vector multiplication
//
TBlockVector<TFloat>
TBlockMatrix::invDiagMult ( const TBlockVector<TFloat> &vec ) const
{
	uint    i, j;
	TBlockVector<TFloat> vecTmp(vec);

	// and now perform diagmatrix * vector with each block
	for (i = 0; i < n; i++) 
		// for all elements in a vectorblock
		for (j = 0; j < vec[i].getSize(); j++) 
			vecTmp[i][j] = vec[i][j] / (*this)[i][i][j].getValue(0);

	return (vecTmp); 
}

//
// blockDiag   - vector multiplication
//
TBlockVector<TFloat> 
TBlockMatrix::blockDiagMult ( const TBlockVector<TFloat>& vec ) const
{
	uint    i, j;
	TBlockVector<TFloat> vecTmp(1,n);

	// set the size of each block in vecTmp to the size in vec
	for (i = 0; i < n; i++)
		vecTmp[i].setSize( vec[i].getSize() );

	// and now perform matrix * vector with each block
    if (n == 1) vecTmp[0] = (*this)[0][0].blockDiagMult( vec[0] );
    else
	    for (i = 0; i < n; i++)
		    // for all elements in a vectorblock
		    for (j = 0; j < vec[i].getSize(); j++)
		    	// and mult. with all indices in a block
		    	vecTmp[i][j] = (*this)[i][i][j] * vec[i]; 

	return (vecTmp); 
}

// blockBlockDiag   - vector multiplication
// performs blockiterations on the matrixblocks
TBlockVector<TFloat>
TBlockMatrix:: blockBlockDiagMult ( const TBlockVector<TFloat>& vec ) const {
  uint    i;
  TBlockVector<TFloat> vecTmp(1,n);
  
  // set the size of each block in vecTmp to the size in vec
  // an perform blockDiagMult on the single matrixblocks
  for (i = 0; i < n; i++){
    vecTmp[i].setSize( vec[i].getSize() );
    vecTmp[i] =(*this)[i][i].blockDiagMult( vec[i] );
  }

return (vecTmp);


}

//
// invLower - vector multiplication
//
TBlockVector<TFloat> 
TBlockMatrix::invLowerMult ( const TBlockVector<TFloat>& vec ) const
{
	uint    j, k;
	TBlockVector<TFloat> vecTmp( vec );

	// set the size of each block in vecTmp to the size in vec
	//for (j = 0; j < n; j++)
	//	vecTmp[j].setSize( vec[j].getSize() );

	// and now perform matrix * vector with each row
	for (j = 0; j < n; j++)
    {
	    for (k = 0; k < j; k++ )
        {
            if ( (*pRows[j])[k].getNoOfRows() > 0 )
                vecTmp[j] -= (*pRows[j])[k] * vecTmp[k];
        }// for

	    // and use invUpperMult in matrixdata for diagonal block
	    vecTmp[j] = (*pRows[j])[j].invLowerMult( vecTmp[j] );
	}// for
	
	return (vecTmp); 
}

TBlockVector<TFloat> 
TBlockMatrix::invUpperMult ( const TBlockVector<TFloat>& vec ) const
{
	uint    j, k;
	TBlockVector<TFloat> vecTmp(vec);

	// set the size of each block in vecTmp to the size in vec
	//for (j = 0; j < n; j++)
	//	vecTmp[j].setSize( vec[j].getSize() );

	// and now perform matrix * vector with each row
	for (j = n; j > 0; j--) {
	    for (k = n; k > j; k-- )
            if ( (*pRows[j-1])[k-1].getNoOfRows() > 0 )
                vecTmp[j-1] -= (*pRows[j-1])[k-1] * vecTmp[k-1];

	    // and use invUpperMult diagonal block
	    vecTmp[j-1] = (*pRows[j-1])[j-1].invUpperMult( vecTmp[j-1] );

	}// for

	return (vecTmp); 
}

//
// blockUpper - vector multiplication
//
TBlockVector<TFloat> 
TBlockMatrix::blockUpperMult ( const TBlockVector<TFloat> &vec ) const 
{
	uint         i, j, k;
	TFloat       fResult;
	TBlockVector<TFloat> vecTmp( 1, n );

	// set the size of each block in vecTmp to the size in vec
	for (i = 0; i < n; i++)
		vecTmp[i].setSize( vec[i].getSize() );

	// and now perform matrix * vector with each block
	for (i = 0; i < n; i++)
		// for all elements in a vectorblock
		for (j = 0; j < vec[i].getSize(); j++)
		{
			fResult = 0.0;
				
			// go over all blocks
			for (k = i; k < n; k++)
				// and use * operator in matrixdata
				fResult += (*this)[i][k][j] * vec[k];

			vecTmp[i][j] = fResult;
		}// for

	return (vecTmp); 
}

//
// blockLower - vector multiplication
//
TBlockVector<TFloat> 
TBlockMatrix::blockLowerMult ( const TBlockVector<TFloat> &vec ) const 
{
	uint         i, j, k;
	TFloat       fResult;
	TBlockVector<TFloat> vecTmp(1, n);

	// set the size of each block in vecTmp to the size in vec
	for (i = 0; i < n; i++)
		vecTmp[i].setSize( vec[i].getSize() );

	// and now perform matrix * vector with each block
	for (i = 0; i < n; i++)
		// for all elements in a vectorblock
		for (j = 0; j < vec[i].getSize(); j++)
		{
			fResult = 0.0;
				
			// go over all blocks
			for (k = 0; k <= i; k++)
				// and use * operator in matrixdata
				fResult += (*this)[i][k][j] * vec[k];

			vecTmp[i][j] = fResult;
		}// for

	return (vecTmp); 
}

//
//  invBlockLowerMult: only for scalar problems, based on concentric
//                     cycle ordering and decomposition 
//                     
//
TBlockVector<TFloat> 
TBlockMatrix::invBlockLowerMult ( const TBlockVector<TFloat> &vec ) const 
{
	TBlockVector<TFloat> vecTmp(1, n);

    if ( n != 1 ) {
        cerr << "TBlockMatrix::invBlockLowerMult: only defined for\n";
	    cerr << "matrices with 1 block (scalar problems).";
	}
    else 
        vecTmp[0] = (*this)[0][0].invBlockLowerMult( vec[0] );
    return (vecTmp);
}

//
//  invBlockUpperMult: only for scalar problems, based on concentric
//                     cycle ordering and decomposition 
//                     
//
TBlockVector<TFloat> 
TBlockMatrix::invBlockUpperMult ( const TBlockVector<TFloat> &vec ) const 
{
	TBlockVector<TFloat> vecTmp(1, n);

    if ( n != 1 ) {
        cerr << "TBlockMatrix::invBlockUpperMult: only defined for\n";
	    cerr << "matrices with 1 block (scalar problems).";
	}
    else 
        vecTmp[0] = (*this)[0][0].invBlockUpperMult( vec[0] );
    return (vecTmp);
}

//
// add or subtract factor times identity matrix to (n-1) blocks
// of the blockmatrix
//
void
TBlockMatrix::addIdentity ( const TFloat factor ) 
{
    uint i, j;
	for (i = 0; i < n-1; i++)
		for (j = 0; j < (*this)[i][i].getNoOfRows(); j++)
            (*this)[i][i][j].setValue(0, (*this)[i][i][j].getValue(0) + factor );
}
void
TBlockMatrix::subtractIdentity ( const TFloat factor ) 
{
    uint i, j;
	for (i = 0; i < n-1; i++)
		for (j = 0; j < (*this)[i][i].getNoOfRows(); j++)
            (*this)[i][i][j].setValue(0, (*this)[i][i][j].getValue(0) - factor );
}

//
// permutation of the matrix 
// (p: row permutation, q: column permutation)
//
void 
TBlockMatrix::permute ( const TVectorBlock<uint> *p ) 
{
    if (n == 1) 
		(*pRows[0])[0].permute( p, p ); 
	else
	{
		for (uint i = 0; i < ( n-1 ); i++) 
		{
			// permute rows and cols of diagonal blocks
			(*pRows[i])[i].permute( p, p ); 
			// permute rows of last column
			(*pRows[i])[n-1].permute( p, NULL );
			// permute columns of last row 
			(*pRows[n-1])[i].permute( NULL, p );
		}// for
	}// for
}

void
TBlockMatrix::permute ( const TPermutation * pPerm )
{
    if (pPerm == NULL)
    {
        cout << "TBlockMatrix::permute (...) : got NULL arg" << endl;
        return;
    }// if
    
    if (n == 1) 
		(*pRows[0])[0].permute( pPerm, pPerm ); 
	else
	{
		for (uint i = 0; i < ( n-1 ); i++) 
		{
			// permute rows and cols of diagonal blocks
			(*pRows[i])[i].permute( pPerm, pPerm ); 

			// permute rows of last column
			(*pRows[i])[n-1].permute( pPerm, NULL );

			// permute columns of last row 
			(*pRows[n-1])[i].permute( NULL, pPerm );
		}// for
	}// for
}

//
// energynorm of a given vector
//
TFloat 
TBlockMatrix::energyNorm ( const TBlockVector<TFloat>& vec ) const
{
	TBlockVector<TFloat> vecTmp( (*this) * vec );
	return (vecTmp * vec);
}
//
// row sum norm (infinity norm) of blockmatrix (with scaled rows/cols)
//
TFloat 
TBlockMatrix::rowSumNorm ( TFloat *scalRows, TFloat *scalCols ) const
{
    TFloat max = 0.0;
    TFloat rowSum = 0.0;
    TFloat kmax;
    uint i,j,s;
    uint k;

    for (i = 0; i < n; i++) {
        kmax = (*pRows[i])[i].getNoOfRows();
        if ( kmax == 0 ) kmax = (*pRows[i])[0].getNoOfRows();
		for (k = 0; k < kmax; k++) {
            rowSum = 0.0;
            for (j = 0; j < n; j++)
                if ( (*pRows[i])[j].getNoOfRows() > 0 )
			    {
                    for (s = 0; s < (*pRows[i])[j][k].getNoOfEntries(); s++)
                        if (scalRows != NULL)
                            rowSum += ( scalRows[i] * scalCols[j] * fabs((*pRows[i])[j][k].getValue( s )) );
                        else rowSum += fabs((*pRows[i])[j][k].getValue( s ));
				}
            if (rowSum > max) max = rowSum;
		}
	}

    return (max);

}

//
// check if decomposition exists already
//
bool 
TBlockMatrix::hasLUdecomp() 
{ 
    if (pLUdecomposition != NULL)
        return (TRUE);
	else
        return (FALSE); 
}

//
// delete LU decomposition
//
void
TBlockMatrix::deleteLUdecomp()
{
    if (pLUdecomposition != NULL) 
	{
        delete (pLUdecomposition);
        pLUdecomposition = NULL;
	}
}

//
// calculate (exact) LU decomposition
//
void
TBlockMatrix::calculateLUdecomp()
{
	cout << "calc LU " << flush;
    uint row, col, smRow;
    uint i, j, k;
    TFloat value;
    uint dim[n];
    for (i = 0; i < n-1; i++) 
        dim[i] = (*pRows[i])[i].getNoOfRows();
    dim[n-1] = (*pRows[n-1])[0].getNoOfRows();

	// initialization
    pLUdecomposition = new TBlockMatrix( *this );
    for (row = 0; row < n; row++) 
        for (col = 0; col < n; col++) {
            (*pLUdecomposition)[row][col].setNoOfRows( dim[row] );
            for (i = 0; i < dim[row]; i++)
                (*pLUdecomposition)[row][col][i].set( dim[col] );
		}

	// calculate LU decomposition
    for (row = 0; row < n; row++)
        for (i = 0; i < dim[row]; i++)    
            for (col = 0; col < n; col++)
                for (k = 0; k < dim[col]; k++)
				{
 
                   if ( ((*this)[row][col].getNoOfRows() > 0)  && 
                        ((*this)[row][col][i].hasIndex(k, dim[col])) )
                       value = (*this)[row][col][i][k];
                   else value = 0.0;

                   if ( (col < row) || ((col == row) && (k < i)) )
				   {
					   // calculate L[row][col][i][k]

                       for (smRow = 0; smRow < col; smRow++)
                           for (j = 0; j < dim[smRow]; j++)
                               value -= (*pLUdecomposition)[row][smRow][i][j] * (*pLUdecomposition)[smRow][col][j][k];
                       for (j = 0; j < k; j++)
						       value -= (*pLUdecomposition)[row][col][i][j] * (*pLUdecomposition)[col][col][j][k];
                       value /= (*pLUdecomposition)[col][col][k][k];
				   }
                   else
				   {
					   // calculate U[row][col][i][k]

                       for (smRow = 0; smRow < row; smRow++)
                           for (j = 0; j < dim[smRow]; j++)
                               value -= (*pLUdecomposition)[row][smRow][i][j] * (*pLUdecomposition)[smRow][col][j][k];
                       for (j = 0; j < i; j++)
                           value -= (*pLUdecomposition)[row][row][i][j] * (*pLUdecomposition)[row][col][j][k];
                       if (fabs(value) < 1e-10 && row == col && i == k ) {
                           if ( row < (n-1) || i < dim[n-1]-1 ) {
                               cout << "warning: decomposition divides by zero  ";
						       cout << row << " " << i << " " << value << endl;
						   }
                           value = 1.0;
					   }
				   }//else
                   (*pLUdecomposition)[row][col][i].setIndex(k,k,value);
				}//for

	// switch so that diagonal entry is at the beginning
    for (row = 0; row < n; row++)
        for ( i = 0; i < dim[row]; i++ ) {
            value = (*pLUdecomposition)[row][row][i].getValue( i );
            (*pLUdecomposition)[row][row][i].setIndex(i, 0, (*pLUdecomposition)[row][row][i].getValue(0) );
            (*pLUdecomposition)[row][row][i].setIndex(0, i, value);
		}

}

//
// calculate incomplete LU decomposition
// with sparsity structure of *this (plus diagonal!)
//
void
TBlockMatrix::calculateILUdecomp()
{
	cout << "calc ILU " << flush;
    
    uint row, col, lli, ll;    //br: blockrow
    uint i, j;
    TFloat quot;
    uint dim[n];
    for (i = 0; i < n-1; i++) 
        dim[i] = (*pRows[i])[i].getNoOfRows();
    dim[n-1] = (*pRows[n-1])[0].getNoOfRows();

	// initialization
    pLUdecomposition = new TBlockMatrix( *this );
    for (row = 0; row < n; row++) 
        for (col = 0; col < n; col++) {
            (*pLUdecomposition)[row][col].setNoOfRows( (*this)[row][col].getNoOfRows() );
            for (i = 0; i < (*pLUdecomposition)[row][col].getNoOfRows(); i++)
                (*pLUdecomposition)[row][col][i] = (*this)[row][col][i];
		}
    (*pLUdecomposition)[n-1][n-1].setNoOfRows( dim[n-1] );
	// approximate by diagonal
    for (i = 0; i < dim[n-1]; i++) {
        (*pLUdecomposition)[n-1][n-1][i].set( 1 );
        (*pLUdecomposition)[n-1][n-1][i].setIndex( 0, i, 0.0 );
	}

	// approximate by sparsity structure of coarse grid (even more)
	/*
    uint maxSize = 100;
    uint pressure = dim[n-1];
    uint noOfEntries[pressure];
	uint  *newIndex = new uint[maxSize];
	TFloat *newValue = new TFloat[maxSize];
    memset( newValue, 0, sizeof(TFloat)*maxSize );
    for (i = 0; i < dim[n-1]; i++) 
	{
        (*pLUdecomposition)[n-1][n-1][i].set( maxSize, newIndex, newValue);
        (*pLUdecomposition)[n-1][n-1][i].setIndex(0, i);
		noOfEntries[i] = 1;
	}
    for (i = 0; i < dim[0]; i++) 
        // if ((*pLUdecomposition)[0][n-1][i].getNoOfEntries() < 5) 
        for ( j = 0; j < (*pLUdecomposition)[0][n-1][i].getNoOfEntries(); j++)
		{
            row = (*pLUdecomposition)[0][n-1][i].getIndex( j );
            for ( k = j+1; k < (*pLUdecomposition)[0][n-1][i].getNoOfEntries(); k++ )
			{
                col = (*pLUdecomposition)[0][n-1][i].getIndex( k );
                if ((*pLUdecomposition)[n-1][n-1][row].hasIndex( col, noOfEntries[row] ) == FALSE)
				{
                    (*pLUdecomposition)[n-1][n-1][row].setIndex( noOfEntries[row]++, col );                
                    (*pLUdecomposition)[n-1][n-1][col].setIndex( noOfEntries[col]++, row );  
				}              
			}
		}

	// adjust sizes of matrixData
    for (i = 0; i < dim[n-1]; i++)
	{
        //cout << "i: " << noOfEntries[i] << endl;
		(*pLUdecomposition)[n-1][n-1][i].set( noOfEntries[i] );
	}
    delete[] newIndex;
    delete[] newValue;
	*/

	// set all entries below the diagonal to zero in blocks 0,1,(2)
	/*
    for (row = 0; row < n-1; row++) 
        for (i = 0; i < (*pLUdecomposition)[row][row].getNoOfRows(); i++)
            for (j = 1; j < (*pLUdecomposition)[row][row][i].getNoOfEntries(); j++) 
		    {
                col = (*pLUdecomposition)[row][row][i].getIndex(j);
                if ( col < i )
                    (*pLUdecomposition)[row][row][i].setValue(j, 0.0);
			}
*/
	// first n-1 block rows
    for ( i = 0; i < n; i++ )
	{
        for (row = 0; row < dim[i]; row++) 
		{
            for (j = 1; j < (*pLUdecomposition)[i][i][row].getNoOfEntries(); j++) 
		    {
                col = (*pLUdecomposition)[i][i][row].getIndex(j);
                if ( col > row ) {
					// ILU for diagonal block
                    quot = (*pLUdecomposition)[i][i][col][row] / (*pLUdecomposition)[i][i][row][row];
				    for (ll = 0; ll < (*pLUdecomposition)[i][i][row].getNoOfEntries(); ll++)
				    {
                        lli = (*pLUdecomposition)[i][i][row].getIndex(ll);
                        if ( (lli >= row) && (*pLUdecomposition)[i][i][col].hasIndex(lli, (*pLUdecomposition)[i][i][col].getNoOfEntries())) 
                            (*pLUdecomposition)[i][i][col][lli] -= quot * (*pLUdecomposition)[i][i][row][lli];
				    }
                    (*pLUdecomposition)[i][i][col][row] = quot;

					// ILU for blocks i,n-1
                    if ( i < n-1)
				        for (ll = 0; ll < (*this)[i][n-1][row].getNoOfEntries(); ll++)
						{
                            lli = (*this)[i][n-1][row].getIndex(ll);
                            if ( (*this)[i][n-1][col].hasIndex(lli, (*this)[i][n-1][col].getNoOfEntries()))     
                                (*pLUdecomposition)[i][n-1][col][lli] -= quot * (*pLUdecomposition)[i][n-1][row][lli];
							
						}
			    }
		    }
			// now for last block row
            if ( i < n-1)
			{
                for (j = 0; j < (*this)[i][n-1][row].getNoOfEntries();j++)
				{
                    col = (*this)[i][n-1][row].getIndex(j);
                    quot = (*pLUdecomposition)[n-1][i][col][row] / (*pLUdecomposition)[i][i][row][row];
				    for (ll = 0; ll < (*this)[i][i][row].getNoOfEntries(); ll++)
					{
					    // ILU for blocks n-1,i
                        lli = (*this)[i][i][row].getIndex(ll);
                        if ( (lli >= row) && (*this)[n-1][i][col].hasIndex(lli, (*this)[n-1][i][col].getNoOfEntries())) 
					        (*pLUdecomposition)[n-1][i][col][lli] -= quot * (*pLUdecomposition)[i][i][row][lli];
					
				    }
				    for (ll = 0; ll < (*this)[i][n-1][row].getNoOfEntries(); ll++)
				    {
						// ILU for block n-1,n-1
                        lli = (*this)[i][n-1][row].getIndex(ll);
                        if ( (*pLUdecomposition)[n-1][n-1][col].hasIndex(lli, (*pLUdecomposition)[n-1][n-1][col].getNoOfEntries())) 
                            (*pLUdecomposition)[n-1][n-1][col][lli] -= quot * (*pLUdecomposition)[i][n-1][row][lli];
						
				    }
                    (*pLUdecomposition)[n-1][i][col][row] = quot;
			    }
		    }
		}
	}

	//cout << (*pLUdecomposition)[n-1][n-1] << endl;
    //cout << endl;
    //for (i = 0; i < dim[n-1]; i++)
	//    cout << i << "  " << (*pLUdecomposition)[n-1][n-1][i][i] << endl;
	cout << "done (dim ";
    for ( i = 0; i < n; i++ )
		cout << dim[i] << ",";
    cout << ")" << endl;
}


//
// calculate (exact) Stokes LU decomposition
// (use that 2 blocks remain zero)
//
void
TBlockMatrix::calcStokesLUdecomp1()
{
	cout << "calc Stokes LU ";
    
    uint row, col, smRow;
    uint i, j, k;
    TFloat value;
    uint dim[n];
    for (i = 0; i < n-1; i++) 
        dim[i] = (*pRows[i])[i].getNoOfRows();
    dim[n-1] = (*pRows[n-1])[0].getNoOfRows();

	// initialization
    pLUdecomposition = new TBlockMatrix( *this );
    for (row = 0; row < n-1; row++) {
        (*pLUdecomposition)[row][row].setNoOfRows( dim[row] );
        (*pLUdecomposition)[row][n-1].setNoOfRows( dim[row] );
        for (i = 0; i < dim[row]; i++) {
            (*pLUdecomposition)[row][row][i].set( dim[row] );
            (*pLUdecomposition)[row][n-1][i].set( dim[n-1] );
		}
	}
    for (col = 0; col < n; col++) {
        (*pLUdecomposition)[n-1][col].setNoOfRows( dim[n-1] );
        for (i = 0; i < dim[n-1]; i++) 
            (*pLUdecomposition)[n-1][col][i].set( dim[col] );
	}

	// calculate LU decomposition
    for (row = 0; row < n; row++)
        for (col = 0; col < n; col++)
           if ( (*pLUdecomposition)[row][col].getNoOfRows() > 0 )
		   {
               if ( (row == col) && (row < n-1) && (row > 0) ) 
                   for (i = 0; i < dim[row]; i++ ) 
                      (*pLUdecomposition)[row][col][i] = (*pLUdecomposition)[0][0][i];
               else
               for (i = 0; i < dim[row]; i++)    
                    for (k = 0; k < dim[col]; k++)
					{

                   if ( ((*this)[row][col].getNoOfRows() > 0)  && 
                        ((*this)[row][col][i].hasIndex(k, dim[col])) )
                       value = (*this)[row][col][i][k];
                   else value = 0.0;

                   if ( (col < row) || ((col == row) && (k < i)) )
				   {
					   // calculate L[row][col][i][k]
                       for (smRow = 0; smRow < col; smRow++)
                           if ( (*pLUdecomposition)[row][smRow].getNoOfRows() > 0 && (*pLUdecomposition)[smRow][col].getNoOfRows() > 0 )
                               for (j = 0; j < dim[smRow]; j++)
                                   value -= (*pLUdecomposition)[row][smRow][i][j] * (*pLUdecomposition)[smRow][col][j][k];
                           for (j = 0; j < k; j++)
						           value -= (*pLUdecomposition)[row][col][i][j] * (*pLUdecomposition)[col][col][j][k];
                       value /= (*pLUdecomposition)[col][col][k][k];
				   }
                   else
				   {
					   // calculate U[row][col][i][k]

                       for (smRow = 0; smRow < row; smRow++)
                           if ( (*pLUdecomposition)[row][smRow].getNoOfRows() > 0 && (*pLUdecomposition)[smRow][col].getNoOfRows() > 0 )
                               for (j = 0; j < dim[smRow]; j++)
                                   value -= (*pLUdecomposition)[row][smRow][i][j] * (*pLUdecomposition)[smRow][col][j][k];
                           for (j = 0; j < i; j++)
                               value -= (*pLUdecomposition)[row][row][i][j] * (*pLUdecomposition)[row][col][j][k];
						   if (fabs(value) < 1e-10 && row == col && i == k ) 
						   {
                               if ( row < (n-1) || i < dim[n-1]-1 ) {
                                   cout << "warning: decomposition divides by zero  ";
						           cout << row << " " << i << " " << value << endl;
							   }
                               value = (*pLUdecomposition)[row][col][i-1][k-1];
						   }
				   }//else

                   (*pLUdecomposition)[row][col][i].setIndex(k, k, value);
				}//for
		   }//if

	// switch so that diagonal entry is at the beginning
    for (row = 0; row < n; row++)
        for ( i = 0; i < dim[row]; i++ ) {
            value = (*pLUdecomposition)[row][row][i].getValue( i );
            (*pLUdecomposition)[row][row][i].setIndex(i, 0, (*pLUdecomposition)[row][row][i].getValue(0) );
            (*pLUdecomposition)[row][row][i].setIndex(0, i, value);
		}
	cout << "done" << endl;

	//cout << (*pLUdecomposition)[0][0] << endl;
    //cout << endl;
}


//
// calculate (exact) Stokes LU decomposition
// (use that 2 blocks remain zero, some more insignificant optimization)
//
void
TBlockMatrix::calcStokesLUdecomp2()
{
	//cout << "calc Stokes LU\n";
    uint row, col;
    uint i, j, k;
    TFloat value, value02, value12, value20, value21;
    uint dim[n];
    for (i = 0; i < n-1; i++) 
        dim[i] = (*pRows[i])[i].getNoOfRows();
    dim[n-1] = (*pRows[n-1])[0].getNoOfRows();

	// initialization
    pLUdecomposition = new TBlockMatrix( *this );
    for (row = 0; row < n-1; row++) {
        (*pLUdecomposition)[row][row].setNoOfRows( dim[row] );
        (*pLUdecomposition)[row][n-1].setNoOfRows( dim[row] );
        for (i = 0; i < dim[row]; i++) {
            (*pLUdecomposition)[row][row][i].set( dim[row] );
            (*pLUdecomposition)[row][n-1][i].set( dim[n-1] );
		}
	}
    for (col = 0; col < n; col++) {
        (*pLUdecomposition)[n-1][col].setNoOfRows( dim[n-1] );
        for (i = 0; i < dim[n-1]; i++) 
            (*pLUdecomposition)[n-1][col][i].set( dim[col] );
	}

	// calculate decomposition of block (0,0), copy result to block (1,1)
    for (i = 0; i < dim[0]; i++)    
        for (k = 0; k < dim[0]; k++)
		{
            if ( (*this)[0][0][i].hasIndex(k, dim[0]) )
                value = (*this)[0][0][i][k];
            else value = 0.0;

             if ( k < i )
			 {
				// calculate L[row][col][i][k]
                for (j = 0; j < k; j++)
				    value -= (*pLUdecomposition)[0][0][i][j] * (*pLUdecomposition)[0][0][j][k];
                    value /= (*pLUdecomposition)[0][0][k][k];
			 }
             else
			     // calculate U[row][col][i][k]
                 for (j = 0; j < i; j++)
                      value -= (*pLUdecomposition)[0][0][i][j] * (*pLUdecomposition)[0][0][j][k];
			 
             (*pLUdecomposition)[0][0][i].setIndex(k, k, value);
             (*pLUdecomposition)[1][1][i].setIndex(k, k, value);
		}

	// calculate entries for blocks (0,2), (1,2), (2,0), (2,1)
    for (i = 0; i < dim[n-1]; i++)    
        for (k = 0; k < dim[0]; k++)
		{
            if ( ((*this)[0][2][i].hasIndex(k, dim[n-1])) )
			{
                 value02 = (*this)[0][2][i][k];
                 value12 = (*this)[1][2][i][k];
                 value20 = (*this)[2][0][k][i];
                 value21 = (*this)[2][1][k][i];
			}
            else 
			{
                value02 = 0.0;
                value12 = 0.0;
                value20 = 0.0;
                value21 = 0.0;
			}
            // calculate blocks (2,0), (2,1)
            for (j = 0; j < k; j++) {
				value20 -= (*pLUdecomposition)[2][0][i][j] * (*pLUdecomposition)[0][0][j][k];
				value21 -= (*pLUdecomposition)[2][1][i][j] * (*pLUdecomposition)[1][1][j][k];
			}
            value20 /= (*pLUdecomposition)[0][0][k][k];
            value21 /= (*pLUdecomposition)[1][1][k][k];
            (*pLUdecomposition)[2][0][i].setIndex(k,k,value20);
            (*pLUdecomposition)[2][1][i].setIndex(k,k,value21);

		    // calculate blocks (0,2), (1,2)
            for (j = 0; j < i; j++)
		    {
                value02 -= (*pLUdecomposition)[0][0][i][j] * (*pLUdecomposition)[0][2][j][k];
                value12 -= (*pLUdecomposition)[1][1][i][j] * (*pLUdecomposition)[1][2][j][k];
		    }
            (*pLUdecomposition)[0][2][i].setIndex(k,k,value02);
            (*pLUdecomposition)[1][2][i].setIndex(k,k,value12);
		}

	// calculate entries for block (2,2)
		
    for (i = 0; i < dim[n-1]; i++)    
        for (k = 0; k < dim[n-1]; k++)
		{
            if ( (*this)[2][2][i].hasIndex(k, dim[2]) )
                value = (*this)[2][2][i][k];
            else value = 0.0;

            if ( k < i )
			{
				// calculate L[2][2][i][k]
                for (j = 0; j < dim[0]; j++) {
                    value -= (*pLUdecomposition)[2][0][i][j] * (*pLUdecomposition)[0][2][j][k];
                    value -= (*pLUdecomposition)[2][1][i][j] * (*pLUdecomposition)[1][2][j][k];
				}
                for (j = 0; j < k; j++)
					value -= (*pLUdecomposition)[2][2][i][j] * (*pLUdecomposition)[2][2][j][k];
                value /= (*pLUdecomposition)[2][2][k][k];
			}
            else
		    {
				// calculate U[2][2][i][k]
                for (j = 0; j < dim[0]; j++) {
                    value -= (*pLUdecomposition)[2][0][i][j] * (*pLUdecomposition)[0][2][j][k];
                    value -= (*pLUdecomposition)[2][1][i][j] * (*pLUdecomposition)[1][2][j][k];
				}
                for (j = 0; j < i; j++)
                    value -= (*pLUdecomposition)[2][2][i][j] * (*pLUdecomposition)[2][2][j][k];
                 if (fabs(value) < 1e-10 ) {
                     if ( i < dim[n-1]-1 ) {
                         cout << "warning: decomposition divides by zero  ";
				         cout << row << " " << i << " " << value << endl;
					 }
                     value = 1.0;
				 }
		    }//else
            (*pLUdecomposition)[2][2][i].setIndex(k, k, value);
		}//for

	// switch so that diagonal entry is at the beginning
    for (row = 0; row < n; row++)
        for ( i = 0; i < dim[row]; i++ ) {
            value = (*pLUdecomposition)[row][row][i].getValue( i );
            (*pLUdecomposition)[row][row][i].setIndex(i, 0, (*pLUdecomposition)[row][row][i].getValue(0) );
            (*pLUdecomposition)[row][row][i].setIndex(0, i, value);
		}
}

//
// exact solver based on LU decomposition
//
TBlockVector<TFloat>
TBlockMatrix::solveLUdecomp( const TBlockVector<TFloat> &vec ) const
{
	//cout << "solve LU\n";
    uint i, j, size;
	TBlockVector<TFloat> vecTmp(1,n);
	TBlockVector<TFloat> diag(1,n);
	for (i = 0; i < n; i++) {
		vecTmp[i].setSize( vec[i].getSize() );
		diag[i].setSize( vec[i].getSize() );
	}

	for (i = 0; i < n; i++)
	{
		size = vec[i].getSize();
		for (j = 0; j < size; j++)
		{
            diag[i][j] = (*pLUdecomposition)[i][i][j].getValue( 0 );
            (*pLUdecomposition)[i][i][j].setValue( 0, 1.0 );
		}
	}
    
    vecTmp = pLUdecomposition->invLowerMult( vec );

	for (i = 0; i < n; i++)
	{
		size = vec[i].getSize();
		for (j = 0; j < size; j++)
            (*pLUdecomposition)[i][i][j].setValue( 0, diag[i][j] );
	}

    vecTmp = pLUdecomposition->invUpperMult( vecTmp );

    return (vecTmp);
}

