//
//  Project     : libnet
//  File        : matrixblock.cc
//  Author      : Sabine Gutsch / Ronald Kriemann / Thomas Probst
//  Purpose     : a matrixblock is a TDynArray with type TMatrixData
//
// arch-tag: 98627bba-711f-418e-a106-1ed355f73f98
//

#include <math.h>

#include "TPermutation.hh"

#include "matrixblock.hh"

//////////////////////////////////////////////////////
//
//  TMatrixBlock
//

//
// constructors
//
TMatrixBlock::TMatrixBlock () 
{
    pDecomp          = NULL; 
    decompType       = NONE; 
    decompBlocks     = NULL;
    noOfDecompBlocks = 0;
    outputPrecision  = 4;	
}

//
// init. with an given number of rows
//
TMatrixBlock::TMatrixBlock ( uint rows ) 
{ 
    setNoOfRows( rows ); 

    pDecomp          = NULL; 
    decompType       = NONE; 
    decompBlocks     = NULL;
    noOfDecompBlocks = 0;
    outputPrecision  = 4;	
}

//
//
//
TMatrixBlock::TMatrixBlock ( const TMatrixBlock &m ) 
{
    uint blockSize = m.getNoOfRows(); 

    setNoOfRows( blockSize );

    for (uint i = 0; i < blockSize; i++)
        (*this)[i].set( m[i] );

    pDecomp          = NULL; 
    decompType       = NONE; 
    decompBlocks     = NULL;
    noOfDecompBlocks = 0;
    outputPrecision  = m.getOutputPrecision();	
}

//
// destructor
//
TMatrixBlock::~TMatrixBlock () 
{ 
    if (pDecomp) 
       pDecomp->free(); 

    for (uint i = 0; i < noOfDecompBlocks; i++)
    {
        if (diagBlocks[i] != NULL)
            diagBlocks[i]->free();
    }// for
}

//
// copy operator =
//
TMatrixBlock& 
TMatrixBlock::operator= ( const TMatrixBlock &m ) 
{
    uint blockSize = m.getNoOfRows(); 

    setNoOfRows( blockSize );

    for (uint i = 0; i < blockSize; i++)
        (*this)[i].set( m[i] );

    pDecomp = NULL; 
    decompType = NONE; 

    return *this;
}

//
// return number of rows
//
const uint 
TMatrixBlock::getNoOfRows () const 
{ 
    return getSize();
}

void
TMatrixBlock::setNoOfRows ( uint i ) 
{ 
    setSize(i);
}

//
// return DecompType
//
TDecompType 
TMatrixBlock::getDecompType ( ) const  
{ 
    return decompType;
}
	
//
// set DecompType
//
void 
TMatrixBlock::setDecompType (TDecompType aDecompType) 
{ 
    decompType = aDecompType;
}
	
//
// return Pointer to decompBlocks
//
TVectorBlock<uint>* 
TMatrixBlock::getDecompBlocksPointer() const 
{ 
    return decompBlocks;
}
 
//
// return  entry of *decompBlocks:   *decompBlocks[i] 
//
uint 
TMatrixBlock::getDecompBlocksEntry(const uint i) const 
{ 
    return (*decompBlocks)[i];
}


//
// return number of DecompBlocks
//
uint 
TMatrixBlock::getNoOfDecompBlocks() const 
{ 
    return noOfDecompBlocks;
}

//
// return Pointer to Decomposition
//
TMatrixBlock* 
TMatrixBlock::getDecomposition() const
{ 
    if (decompType == NONE) 
        cerr << " Warning: Matrix has no decomposition " << endl;

    return pDecomp; 
}

//
//set Pointer to Decomposition
//
void 
TMatrixBlock::setDecomposition (TMatrixBlock* apDecomp ) 
{ 
    if (decompType == NONE) 
        cerr << "TMatrixBlock::setDecomposition Warning: Matrix has no "
             << "decompositiontype assigned " << endl;

    if (apDecomp->getNoOfRows() != getNoOfRows())
        cerr << "TMatrixBlock::setDecomposition Warning: Decompostition has "
             << "other size than original matrix " << endl;

    pDecomp = apDecomp; 
}

//
// block - vectorblock multiplication  with transpose matrix
//
void 
TMatrixBlock::transposeMult ( const TVectorBlock<TFloat>& v, 
                              TVectorBlock<TFloat>& w ) const 
{
    uint  i, rowlength;
    uint   blocksize = getNoOfRows();

    w.set( 0.0 );
    if ( v.getSize() == blocksize) {
        for (uint j = 0; j < blocksize; j++) {
            // number of entries in this row
            rowlength = (*this)[j].getNoOfEntries(); 
			   
            for ( i = 0; i < rowlength; i++ ) 
                w[(*this)[j].getIndex(i)] += (*this)[j].getValue(i) * v[j];
				//cout << "(*this)[j].getIndex(i)" << (*this)[j].getIndex(i)<<endl;
        }
    } 
    else
        cerr << "TMatrixBlock::transposeMult: argument vector has wrong size  (" 
				 << v.getSize() << ", " << blocksize << ")" << endl;
}

//
// multiply with diagonal of block (for full vectors)
//
TVectorBlock<TFloat> 
TMatrixBlock::diagMult ( const TVectorBlock<TFloat> &v ) const
{
    TVectorBlock<TFloat> vecTmp ( v );

    uint size = v.getSize();
    for (uint i = 0; i < size; i++ )
        vecTmp[i] = v[i] * (*this)[i].getValue(0);

    return (vecTmp);
}

//
// multiply with inverse of diagonal part of block (for full vectors)
// (only meaningful for quadratic matrices)
//
TVectorBlock<TFloat> 
TMatrixBlock::invDiagMult ( const TVectorBlock<TFloat> &v )	const 
{
    TVectorBlock<TFloat> vecTmp ( v );

    uint size = v.getSize();
    for (uint i = 0; i < size; i++ )
        vecTmp[i] = v[i] / (*this)[i].getValue(0);

    return (vecTmp);
}

//
// for sparse vectors
//
TMatrixData 
TMatrixBlock::invDiagMult ( const TMatrixData& data ) const
{
    TMatrixData dataTmp( data );

    uint noOfEntries = dataTmp.getNoOfEntries();
    uint  *indices  = dataTmp.getIndex();
    TFloat *values   = dataTmp.getValue();

    for (uint i = 0; i < noOfEntries; i++)
        // this is really dirty, because we operate with the pointers
        // not with the object
        values[i] /= (*this)[indices[i]].getValue(0);

    return (dataTmp);
}

//
// blockDiagMult - multiplication with block diagonal (block structure
//                 as given by decompBlocks)
//
TVectorBlock<TFloat> 
TMatrixBlock::blockDiagMult ( const TVectorBlock<TFloat> &v ) const 
{
    TVectorBlock<TFloat> vecTmp ( v );
    uint i, j;
  
    if ( noOfDecompBlocks == 0) vecTmp = diagMult( v );
    else 
    {
        uint minBlockIndex = 0, maxBlockIndex = (*decompBlocks)[0];
      
        for ( i = 0; i < noOfDecompBlocks; i++ ) {
            for ( j = minBlockIndex; j < maxBlockIndex; j++ )
                vecTmp[j] = (*this)[j].mult( v, minBlockIndex, maxBlockIndex);
        
            minBlockIndex = maxBlockIndex;
            maxBlockIndex = (*decompBlocks)[2*i+1] + 1;
            for ( j = minBlockIndex; j < maxBlockIndex; j++ )
                vecTmp[j] = (*this)[j].mult( v, minBlockIndex, maxBlockIndex);
            minBlockIndex = maxBlockIndex;
            if ( i < (noOfDecompBlocks-1) ) 
                maxBlockIndex = (*decompBlocks)[2*(i+1)];
        }
      
        maxBlockIndex = v.getSize();
        for ( j = minBlockIndex; j < maxBlockIndex; j++ )
            vecTmp[j] = (*this)[j].mult( v, minBlockIndex, maxBlockIndex);
    }
  
    return (vecTmp);
}

//
// multiply with inverse of lower triangular part of block
// (only meaningful for quadratic matrices)
//
TVectorBlock<TFloat> 
TMatrixBlock::invLowerMult ( const TVectorBlock<TFloat> &v ) const
{
    TVectorBlock<TFloat> vecTmp;
    uint                 i;
    uint                 size = v.getSize();

    vecTmp = v;
    
    for ( i = 0; i < size; i++ )
        vecTmp[i] = (*this)[i].invLowerMult( vecTmp );

    return vecTmp;
}

TVectorBlock<TFloat>
TMatrixBlock::invLowerMult ( const TVectorBlock<TFloat> &v,
                             const TFloat & diagDamp ) const
{
    TVectorBlock<TFloat>  vecTmp;
    uint                  i;
    uint                  size = v.getSize();

    vecTmp = v;
    
    for ( i = 0; i < size; i++ )
        vecTmp[i] = (*this)[i].invLowerMult( vecTmp, diagDamp );

    return vecTmp;
}

//
// for sparse vectors (tpr)
//
TVectorBlock<TFloat> 
TMatrixBlock::invLowerMult ( const TMatrixData& data ) const
{ 
    uint                 rows = getNoOfRows(); 
    TVectorBlock<TFloat> vecTmp(0.0, rows);
    TFloat               fResult;

    for (uint i = 0; i < rows; i++)
    {
        // vecTmp[i] = contents[i].invLowerMult(data)
        uint    MnoOfEntries = (*this)[i].getNoOfEntries();
        uint  *Mindices     = (*this)[i].getIndex();
        TFloat *Mvalues      = (*this)[i].getValue();
        uint    diag         = Mindices[0];
    
        uint    noOfEntries = data.getNoOfEntries();
        uint  *indices     = data.getIndex();
        TFloat *values      = data.getValue();
            
		fResult=0;

		for (uint j = 0; j < MnoOfEntries; j++)
			if ( Mindices[j] < diag ) 
				for (uint l=0; l <noOfEntries;l++)
					if (Mindices[j]==indices[l])
						fResult -= Mvalues[j] * values[j];

		fResult *= (1 / Mvalues[0]);
		vecTmp[i]= fResult;
	}
	return (vecTmp);
}

//
// multiply with inverse of lower triangular part of block
// (only meaningful for quadratic matrices)
// assuming 1 in diagonal
TVectorBlock<TFloat> 
TMatrixBlock::invUnitLowerMult ( const TVectorBlock<TFloat> &v )	const 
{
	TVectorBlock<TFloat> vecTmp ( v );
    uint i;

	uint size = v.getSize();
	for (i = 0; i < size; i++ )
		vecTmp[i] = (*this)[i].invUnitLowerMult( vecTmp );

    return (vecTmp);
}


//
// invBlockLowerMult - assumes decomposition for some cyclic bidiagonal
//                     blocks on diagonal, solves pointwise (upper)GS 
//                     for all entries not in these blocks
//
TVectorBlock<TFloat> 
TMatrixBlock::invBlockLowerMult ( const TVectorBlock<TFloat> &v ) const
{
    TVectorBlock<TFloat> vecTmp ( v );
    TVectorBlock<TFloat> vecBlock ( 0, 1);

    uint i;
    uint j;
    uint startIndex;
    uint endIndex;
    uint blockSize;

    if ( noOfDecompBlocks == 0 ) vecTmp = invLowerMult( v );
    else {
        // before the first  block
        startIndex = 0;
        endIndex   = (*decompBlocks)[0];

        for ( i = 0; i < noOfDecompBlocks; i++ ) {
            // first pointwise Gauss-Seidel for entries before
            // current block
            //     for ( j = startIndex; j < endIndex; j++ )
            // vecTmp[j] = (*this)[j].invLowerMult( vecTmp);
            for ( j = startIndex; j < endIndex; j++ )
                vecTmp[j] -= (*this)[j].lowerMult( vecTmp, startIndex );

            for ( j = endIndex; j > startIndex; j-- )
                vecTmp[j-1] = (*this)[j-1].invUpperMult( vecTmp, endIndex );

            // and now treat the block
            // blocksize of current block
            startIndex = (*decompBlocks)[2*i];
            endIndex   = (*decompBlocks)[2*i+1];
            blockSize = endIndex - startIndex + 1;

            //first step in block step( b - \sum ...)
            for ( j = startIndex; j <= endIndex; j++ )
                vecTmp[j] -= (*this)[j].lowerMult( vecTmp, startIndex );

            // apply inverse of block diagonal
            vecBlock.setSize( blockSize );

            for ( j = 0; j < blockSize; j++ )
                vecBlock[j] = vecTmp[startIndex + j];
            vecBlock = diagBlocks[i]->solveDecomp( vecBlock );

            for ( j = 0; j < blockSize; j++ )
                vecTmp[startIndex + j] = vecBlock[j];

            startIndex = endIndex + 1;
            if ( i < (noOfDecompBlocks - 1) ) 
                endIndex = (*decompBlocks)[2*(i+1)];
        }    

        // remaining entries: pointwise Gauss-Seidel
        //for ( i = startIndex; i < getNoOfRows(); i++ )
        //vecTmp[i] = (*this)[i].invLowerMult( vecTmp);


        for ( i = startIndex; i < getNoOfRows(); i++ )
            vecTmp[i] -= (*this)[i].lowerMult( vecTmp, startIndex );
        for ( i = getNoOfRows(); i > startIndex; i-- )
            vecTmp[i-1] = (*this)[i-1].invUpperMult( vecTmp );      
    }
    return (vecTmp);
}

//
// multiply with inverse of upper triangular part of block
// (only meaningful for quadratic matrices)
//
TVectorBlock<TFloat> 
TMatrixBlock::invUpperMult ( const TVectorBlock<TFloat> &v )	const 
{
    TVectorBlock<TFloat> vecTmp ( v );

    uint size = v.getSize();
    for (uint i = size; i > 0; i-- )
        vecTmp[i-1] = (*this)[i-1].invUpperMult( vecTmp );

    return (vecTmp);
}

TVectorBlock<TFloat>
TMatrixBlock::invUpperMult ( const TVectorBlock<TFloat> &v,
                             const TFloat & diagDamp ) const
{
    TVectorBlock<TFloat>  vecTmp;
    uint                  i;
    uint                  size = v.getSize();
    
    vecTmp = v;

    for ( i = size; i > 0; i-- )
        vecTmp[i-1] = (*this)[i-1].invUpperMult( vecTmp, diagDamp );

    return vecTmp;
}

//
// invBlockUpperMult - assumes decomposition for some cyclic bidiagonal
//                     blocks on diagonal, solves pointwise GS for
//                     all entries not in these blocks
//
TVectorBlock<TFloat> 
TMatrixBlock::invBlockUpperMult ( const TVectorBlock<TFloat> &v ) const
{
    TVectorBlock<TFloat> vecTmp ( v );
    TVectorBlock<TFloat> vecBlock ( 0, 1 );

    uint i;
    uint j;
    uint startIndex;
    uint endIndex;
    uint blockSize;

    if ( noOfDecompBlocks == 0 ) vecTmp = invUpperMult( v );
    else {
        // blocksize of current block
        startIndex = getNoOfRows() - 1;
        endIndex   = (*decompBlocks)[2*(noOfDecompBlocks-1)+1];

        for ( i = noOfDecompBlocks; i > 0; i-- ) {
            // first pointwise Gauss-Seidel for entries after 
            // current block
            for ( j = startIndex; j > endIndex; j-- )
                vecTmp[j] = (*this)[j].invUpperMult( vecTmp );
 
           
            // and now treat the block
            startIndex = (*decompBlocks)[2*(i-1)+1];
            endIndex   = (*decompBlocks)[2*(i-1)];
            blockSize = startIndex - endIndex + 1;

            //first step in block step( b - \sum ...)
            for ( j = startIndex+1; j > endIndex; j-- )
                vecTmp[j-1] -= (*this)[j-1].upperMult( vecTmp, startIndex );

            // apply inverse of block diagonal
            vecBlock.setSize( blockSize );

            for ( j = 0; j < blockSize; j++ )
                vecBlock[j] = vecTmp[endIndex + j];

            vecBlock = diagBlocks[i-1]->solveDecomp( vecBlock );

            for ( j = 0; j < blockSize; j++ )
                vecTmp[endIndex + j] = vecBlock[j];

            startIndex = endIndex - 1;
            if ( i > 1 ) endIndex = (*decompBlocks)[2*(i-2)+1];
        }    

        for ( i = (*decompBlocks)[0]; i > 0; i-- )
            vecTmp[i-1] = (*this)[i-1].invUpperMult( vecTmp );
    }

    return (vecTmp);
}

//
// get transpose of j'th row
//
TVectorBlock<TFloat> 
TMatrixBlock::getTransposedRow ( uint j, uint size ) const 
{
    TVectorBlock<TFloat> vecTmp( 0.0, size );

    uint noOfEntries = (*this)[j].getNoOfEntries();
    for (uint i = 0; i < noOfEntries; i++ )
        vecTmp[(*this)[j].getIndex(i)] = (*this)[j].getValue(i);
		
    return (vecTmp);
}

//
// get column as vectorblock
//
TVectorBlock<TFloat> 
TMatrixBlock::getColumn ( uint j ) const 
{
    TVectorBlock<TFloat> vecTmp( 0.0, getNoOfRows() );

    uint noOfEntries = (*this)[j].getNoOfEntries();
    for (uint i = 0; i < noOfEntries; i++ )
        vecTmp[(*this)[j].getIndex(i)] = (*this)[(*this)[j].getIndex(i)][j];
		
    return (vecTmp);
}

void 
TMatrixBlock::clearDecomposition(){

    if (pDecomp) delete (pDecomp);
    for (uint i = 0; i < noOfDecompBlocks; i++)
    {
        if (diagBlocks[i] != NULL)
            diagBlocks[i]->free();
    }// for
    noOfDecompBlocks = 0;
    decompType = NONE;
}



void 
TMatrixBlock::decompose ( TDecompType Type, 
                          TVectorBlock<uint> *aDecompBlocks ) 
{
    if ( (decompType != NONE) && (decompType != Type) ) {
        cout << "Waring: Current type of decomposition: ";
        printDecompTypeName();
        cout << "\nRequested type of decomposition: ";
        printDecompTypeName( Type );
        clearDecomposition();
    }

    if ( (decompType != Type) ) {
        // new decomposition
        if ( (aDecompBlocks != NULL) && (Type == BLOCK_CYC_LU)) {
            // clear old decompostion
            if (pDecomp) delete (pDecomp);
            for (uint i = 0; i < noOfDecompBlocks; i++)
            {
                if (diagBlocks[i] != NULL)
                    diagBlocks[i]->free();
            }// for
            noOfDecompBlocks = 0;

            // block decomposition
            noOfDecompBlocks  = aDecompBlocks->getSize() / 2;
            if ( noOfDecompBlocks > 0 ) {
                decompType        = BLOCK_CYC_LU;
                decompBlocks      = aDecompBlocks;
             
                // initialize array of noOfDiagBlocks pointers 
                // to matrixblocks
                diagBlocks = new TMatrixBlock * [noOfDecompBlocks];  
                for ( uint i = 0; i < noOfDecompBlocks; i++ ) {
                    createBidiagBlocks( i, (*aDecompBlocks)[2*i], 
                                        (*aDecompBlocks)[2*i + 1] );
                    diagBlocks[i]->decompose( CYC_LU );
                }
            }
            else 
                decompType = NONE;   
        }
        else
            if ( (aDecompBlocks != NULL) && (Type == BLOCK_DIAG)) {
                // clear old decompostion
                if (pDecomp) delete (pDecomp);
                for (uint i = 0; i < noOfDecompBlocks; i++)
                {
                    if (diagBlocks[i] != NULL)
                        diagBlocks[i]->free();
                }// for
                noOfDecompBlocks = 0;

                // block decomposition
                noOfDecompBlocks  = aDecompBlocks->getSize() / 2;
                if ( noOfDecompBlocks > 0 ) {
                    decompType        = BLOCK_DIAG;
                    decompBlocks      = aDecompBlocks;
             
                    // initialize array of noOfDiagBlocks pointers 
                    // to matrixblocks
                    diagBlocks = new TMatrixBlock * [noOfDecompBlocks];  
                    for ( uint i = 0; i < noOfDecompBlocks; i++ ) 
                        createDiagBlock( i, (*aDecompBlocks)[2*i], 
                                         (*aDecompBlocks)[2*i + 1] );
                }
                else 
                    decompType = NONE;   
            }
            else
                // decomposition of the whole matrix block (no subblocks) 
                switch (Type) {
                case NONE: ; break;
                case ILU: {
                    // ILU-Elimination:
                    decompType = ILU;

                    uint ju = 0;
                    uint lli;
                    TFloat quot; 
			
                    pDecomp = new TMatrixBlock(*this);

                    uint blocksize = getNoOfRows();

                    for (uint i = 0; i < blocksize; i++) { 
                        // for each line of *this
                        for (uint l = 1; l < (*this)[i].getNoOfEntries(); l++) {
                            if ((*this)[i].getIndex(l)>(*this)[i].getIndex(0)) {
                                ju = (*this)[i].getIndex(l);  // found!
                                quot = (*pDecomp)[ju][i] / (*pDecomp)[i][i];
                                for (uint ll = 0; ll < (*this)[i].getNoOfEntries(); ll++ ) {
                                    if ( (*this)[i].getIndex(ll) >= (*this)[i].getIndex(0) ) {
                                        lli = (*this)[i].getIndex(ll);
                                        if ((*this)[ju].hasIndex(lli, (*this)[ju].getNoOfEntries()) )
                                            (*pDecomp)[ju][lli] -= quot * ((*pDecomp)[i][lli]);
                                    }
                                }
                                (*pDecomp)[ju][i] = quot;
                            }
                        }
                    }
                    break; 
                }
                case LU: {
                    pDecomp = new TMatrixBlock(*this);
                    decompType = LU;
                    cout << "to be done...\n";

                    break;
                }
                case CYC_LU: {
                    pDecomp = new TMatrixBlock(*this);
                    decompType = CYC_LU;

                    uint dim = getNoOfRows();
                    uint i;
                    TFloat lni;

                    (*pDecomp)[dim - 1].set( dim );

                    lni = (*this)[dim - 1].getValue( 1 ) / (*this)[0].getValue( 0 ); 
                    (*pDecomp)[dim - 1].setIndex( 1, 0, lni );

                    for ( i = 1; i < dim-1; i++ ) {
                        lni *= (-(*this)[i-1].getValue(1) / (*this)[i].getValue(0) );
                        (*pDecomp)[dim - 1].setIndex( i+1, i, lni );
                    }
                    lni *= (*this)[dim - 2].getValue( 1 );
                    lni -= (*this)[dim - 1].getValue( 0 );
                    (*pDecomp)[dim - 1].setIndex( 0, dim-1, -lni );
                    break; 
                }
                case CYC_SCHUR: {
                    decompType = CYC_SCHUR;
                    pDecomp = new TMatrixBlock(1);

                    // calculate S = A22 - A21 A11^{-1} A12
                    uint n = getNoOfRows();

                    // b := A12 e_1 
                    TVectorBlock<TFloat> *b;
                    b = new TVectorBlock<TFloat>( 0.0, n-1);
                    (*b)[0] = (*this)[0].getValue(1);
      
                    // b := A11^{-1} b = A11^{-1} A12 e_1
                    for (uint i = 0; i < n-1; i++ )
                        (*b)[i] = (*this)[i].invLowerMult( *b );
       
                    //S = (*Decomp)[0][0]: = A21 b
                    TFloat s[1];
                    uint u[1];
                    s[0] = (*this)[n-1].getValue(0) - (*this)[n-1].getValue(1) * (*b)[n-2];
                    u[0] = 0;
                    (*pDecomp)[0].set(1, u, s); 
                    break; 
                }
                default: break;
                }
    }
}

//
// solve for decomposition 
//
TVectorBlock<TFloat> 
TMatrixBlock::solveDecomp ( const TVectorBlock<TFloat> &v )
{
    TVectorBlock<TFloat> vecTmp( v );

    if (decompType == NONE) 
        cout << "Matrix has no decomposition ";
    else {
        //cout << "solveDecomp with "; printDecompTypeName();
		
        switch (decompType) {
        case NONE: ; break;
        case ILU: cout << "Sorry: to be done" << endl; break;
            // ILU-Elimination:
        case LU: cout << "Sorry: to be done" << endl; break;
        case CYC_LU: {
            uint i, rows;
            rows = v.getSize();
           
            // L vecTmp = v
            for (i = 0; i < rows-1; i++ ) 
                vecTmp[rows-1] -= 
                                  ( (*pDecomp)[rows-1].getValue( i+1 ) * v[i] );
           
            // U x = vecTmp (store x in vecTmp)
            vecTmp[rows-1] /= (*pDecomp)[rows-1].getValue( 0 );
            for (i = 1; i < rows; i++ ) {
                vecTmp[rows-1-i] -=( (*pDecomp)[rows-1-i].getValue( 1 ) 
                                     * vecTmp[rows-i] );
                vecTmp[rows-1-i] /=(*pDecomp)[rows-1-i].getValue( 0 ); 
            }
            return (vecTmp);
                
        }
        case CYC_SCHUR: {
            TVectorBlock<TFloat> h ( v );
            uint n = v.getSize();
            // vecTemp_1 = v_1
            // h= v1; vecTmp_2= v_2-A21 A11^{-1} h
            // vecTmp= [ v_1, v_2-A21 A11^{-1} v1]
            
            for (uint i = 0; i < n-1; i++ )
                h[i] = (*this)[i].invLowerMult(h);
            vecTmp[n-1] = v[n-1] - (*this)[n-1].getValue(1)*h[n-2];
            
            delete(&h);
         
            // vecTmp_2 = S^-1 vecTmp_2
            vecTmp[n-1] /= (*pDecomp)[0][0];
            // vecTmp_1 = vecTmp_1 - A12 vecTmp_2
            vecTmp[0] -= (*this)[0].getValue(1)*vecTmp[n-1];

            // vecTmp_1 = A11^{-1}vecTmp_1
            for (uint i = 0; i < n-1; i++ )
                vecTmp[i] = (*this)[i].invLowerMult(vecTmp);
            return (vecTmp);
        }           
        default: break;
        }
    }
    return (vecTmp);
}

//
// set sparsity pattern to cyclic (upper) bidiagonal matrixblock
//
void
TMatrixBlock::setCyclicPattern ()
{
    uint rows = getNoOfRows ();
    uint  newArrIndex[2];
    TFloat newArrValue[2] = {0.0, 0.0}; 

    (*this).setSize(rows); 

    for (uint i = 0; i < rows-1; i++) {
        newArrIndex[0] = i;
        newArrIndex[1] = i+1;
        (*this)[i].set( 2, newArrIndex, newArrValue );
    }
    newArrIndex[0] = rows-1;
    newArrIndex[1] = 0;
    (*this)[rows-1].set( 2, newArrIndex, newArrValue );
}

//
// create diagonal blocks:
// clopy block diag matrix 
void
TMatrixBlock::createDiagBlock( uint blockNo, uint startIndex, 
                               uint endIndex ){
    
    uint blockSize;
    uint j;
    
    // create matrixblocks and copy corresponding values from *this
    blockSize = endIndex - startIndex + 1;

    diagBlocks[blockNo] = new TMatrixBlock( blockSize );
    
    for ( j = 0 ;j <blockSize;  j++ ) {
        uint noOfEntries= (*this)[j+startIndex].getNoOfEntries();
        uint  *arrIndex = (*this)[j+startIndex].getIndex();
        TFloat *arrValue = (*this)[j+startIndex].getValue();
        
        uint  *newIndex = new uint[noOfEntries+1];
        TFloat *newValue = new TFloat[noOfEntries+1];
        
        uint noOfBlockEntries=0;
        for (uint i=0; i<noOfEntries;i++){
            uint index=arrIndex[i];
            if ((index >=startIndex) && ( index <= endIndex)){
                // copy matrix entry if in diag block
                newIndex[noOfBlockEntries]=index-startIndex;
                newValue[noOfBlockEntries]=arrValue[i];
                noOfBlockEntries++;
            }
        }
        (*diagBlocks[blockNo])[j].set(noOfBlockEntries,newIndex,newValue);
        delete[] newIndex;   delete[] newValue;
    }
   
    //cout << "block:\n" << (*diagBlocks[blockNo]) << endl;
}

//
// create diagonal blocks, each with an bidiag pattern
//
void
TMatrixBlock::createBidiagBlocks ( uint blockNo, uint startIndex, 
                                   uint endIndex )
{
    uint blockSize;
    uint j;

    // create matrixblocks and copy corresponding values from *this
    blockSize = endIndex - startIndex + 1;

    diagBlocks[blockNo] = new TMatrixBlock( blockSize );
    diagBlocks[blockNo]->setCyclicPattern();

    for ( j = 0; j < blockSize-1; j++ ) {
        (*diagBlocks[blockNo])[j].setIndex(0, j, 
                                           (*this)[startIndex + j].getValue(0) );
        (*diagBlocks[blockNo])[j].setIndex(1, j+1, 
                                           (*this)[startIndex + j][startIndex+j+1] );
    }
    // and the very last row
    (*diagBlocks[blockNo])[blockSize-1].setIndex(0, blockSize-1, 
                                                 (*this)[startIndex + blockSize-1].getValue(0) );
    (*diagBlocks[blockNo])[blockSize-1].setIndex(1, 0, 
                                                 (*this)[startIndex + blockSize-1][startIndex] );

    //cout << "block:\n" << (*diagBlocks[blockNo]) << endl;
}

//
// permutation for matrix block
//
void 
TMatrixBlock::permute ( const TVectorBlock<uint> *p, 
                        const TVectorBlock<uint> *q) 
{
    TMatrixBlock blockTmp( *this );
    uint blocksize = getNoOfRows();

    blockTmp.delRef();
    
    if ( ( p != NULL) && ( q != NULL) )
        // row and column permutation
        for (uint i = 0; i < blocksize; i++)  
            (*this)[(*p)[i]].permute( q, blockTmp[i] ); 
	
    if ( ( p != NULL) && ( q == NULL) )
        // only row permutation
        for (uint i = 0; i < blocksize; i++) 
            (*this)[(*p)[i]].set( blockTmp[i] ); 
	
    if ( ( p == NULL) && ( q != NULL) )
        // only column permutation
        for (uint i = 0; i < blocksize; i++)  
            (*this)[i].permute( q, blockTmp[i] ); 
}

void
TMatrixBlock::permute ( const TPermutation * pPerm1,
                        const TPermutation * pPerm2 )
{
	TMatrixBlock blockTmp( *this );
	uint         blocksize = getNoOfRows();
	
    blockTmp.delRef();
    
	if (( pPerm1 != NULL) && ( pPerm2 != NULL))
    {
		// row and column permutation
		for (uint i = 0; i < blocksize; i++)  
			(*this)[ (*pPerm1)[i] ].permute( pPerm2, blockTmp[i] );
    }// if
	
	if (( pPerm1 != NULL) && ( pPerm2 == NULL))
    {
		// only row permutation
		for (uint i = 0; i < blocksize; i++) 
			(*this)[ (*pPerm1)[i] ].set( blockTmp[i] );
    }// if
	
	if (( pPerm1 == NULL) && ( pPerm2 != NULL))
    {
		// only column permutation
		for (uint i = 0; i < blocksize; i++)  
			(*this)[i].permute( pPerm2, blockTmp[i] );
    }// if
}

//
// return DecompTypeName
//
void 
TMatrixBlock::printDecompTypeName () 
{ 
    cout <<"Decomposition type: ";
    switch (decompType) 
    {
    case NONE:         cout << "NONE"     << endl; break;
    case ILU:          cout << "ILU"      << endl; break;
    case LU:           cout << "LU"       << endl; break;
    case CYC_LU:       cout << "CYC_LU"   << endl; break;
    case CYC_SCHUR:    cout << "CYC_SCHUR"<< endl; break;
    case SCHUR:        cout << "SCHUR"    << endl; break;
    default:           cout << "NONE"     << endl; break;
    }
}

//
// prints DecompTypename of Type
//
void 
TMatrixBlock::printDecompTypeName( TDecompType Type ) 
{ 
    switch (Type) 
    {
    case NONE:         cout << "NONE"     << endl; break;
    case ILU:          cout << "ILU"      << endl; break;
    case LU:           cout << "LU"       << endl; break;
    case CYC_LU:       cout << "CYC_LU"   << endl; break;
    case CYC_SCHUR:    cout << "CYC_SCHUR"<< endl; break;
    case SCHUR:        cout << "SCHUR"    << endl; break;
    default:           cout << "NONE"     << endl; break;
    }
}

//
// block addition (we assume, that the other block has the same
// index structure and same size of (*this))
//
TMatrixBlock& 
TMatrixBlock::operator+= ( const TMatrixBlock& block ) 
{
    uint blocksize = getNoOfRows();
    for (uint i = 0; i < blocksize; i++)
        (*this)[i] += block[i];
    return (*this);
}

TMatrixBlock 
TMatrixBlock::operator+ ( const TMatrixBlock& block ) const 
{
    TMatrixBlock blockTmp( *this );

    return blockTmp += block;
}

//
// block subtraction (we assume, that the other block has the same
// index structure and same size of (*this))
//
TMatrixBlock& 
TMatrixBlock::operator-= ( const TMatrixBlock& block ) 
{
    uint blocksize = getNoOfRows();

    for (uint i = 0; i < blocksize; i++)
        (*this)[i] -= block[i];
    return (*this);
}

TMatrixBlock 
TMatrixBlock::operator- ( const TMatrixBlock& block ) const 
{
    TMatrixBlock blockTmp( *this );

    return blockTmp -= block;
}

//
// scalar multiplication
//
TMatrixBlock& 
TMatrixBlock::operator*= ( TFloat scalar ) 
{
    uint blocksize = getNoOfRows();

    for (uint i = 0; i < blocksize; i++)
        (*this)[i] *= scalar;

    return *this;
}

TMatrixBlock 
TMatrixBlock::operator* ( TFloat scalar ) const 
{
    TMatrixBlock  matrixblock(*this);

    return matrixblock *= scalar;
}


TMatrixBlock 
operator* ( TFloat scalar, const TMatrixBlock& block ) 
{
    TMatrixBlock  matrixblock(block);

    return matrixblock *= scalar;
}

//
// block - vectorblock multiplication 
//
TVectorBlock<TFloat> 
TMatrixBlock::operator* ( const TVectorBlock<TFloat>& v ) const 
{ 
    uint  blocksize = getNoOfRows();
    uint  j;

    TVectorBlock<TFloat> vecTmp ( 0.0,blocksize);

    // for all elements in a vectorblock
    for (j = 0; j < blocksize; j++)
        vecTmp[j] = (*this)[j] * v;

    return vecTmp;
}

//
// row sum norm 
//
TFloat 
TMatrixBlock::rowSumNorm ( ) const
{
    TFloat max = 0.0;
    TFloat rowSum = 0.0;
    uint i,k,l;
    uint j;

    k = getNoOfRows();

    for (i = 0; i < k; i++) {
        rowSum = 0.0;
        l = (*this)[i].getNoOfEntries();
        for (j = 0; j < l; j++)
            rowSum += fabs( (*this)[i].getValue(j) );
				
        if (rowSum > max) max = rowSum;
		
    }
    return (max);
}

//
// return Pointer to diagonal subblock
//
TMatrixBlock* 
TMatrixBlock::getDiagBlock( uint i ) const
{ 
    if ( i >= noOfDecompBlocks) {
        cerr << " Warning: Matrix has no decomposition " << endl;
        return (NULL);
    }

    return diagBlocks[i]; 
}

//
// output
//
ostream& 
operator<< (ostream& os, TMatrixBlock& matrixblock)
{
    uint  prec = matrixblock.getOutputPrecision();

    os.setf( ios::scientific, ios::floatfield);
    os.precision(prec);
  
    for (uint i = 0; i < matrixblock.getNoOfRows(); i++) 
        os << "(" << i << ") : " << matrixblock[i] << "\n";

    os.precision(4);

    return os;
}

