//
//  Project   : Navier-Stokes
//  File      : cmdline.cc
//  Author    : Ronald Kriemann / Sabine Gutsch
//  Purpose   : evaluation of commandline parameters
//
// arch-tag: eea0150a-7158-4167-bc13-8e555611a852
//

extern "C" {
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
}

#include "defines.hh"
#include "graphtool.hh"
#include "blockinfo.hh"

#include "cmdline.hh"

// default object for cmdline parsing
TParameter param;

//
// constructor
//
TParameter::TParameter()
{
    startLevel   = endLevel = 0;
    noOfNets     = steps    = 0;
    cycle        = 1;
    preSmoothing = postSmoothing = 1;
    ordering     = 0;
    epsilon      = 1e-5;
    xi           = 1.0;
    criterion    = 0;
    draw         = FALSE;
    write_matrix = FALSE;
    mgSort       = MGSORT_INDEP;
    searchType   = SEARCH_DFS;
    pblock       = FALSE;

    fsAlpha = 150.0 * PI / 180.0;
    fsBeta  = PI/8;
    fsKappa = PI/16;
    fsGamma = PI/8;

    sigma   = 0.0;

    global     = FALSE;
    sortOFD    = FALSE;
    
    // red-black 
    MaxLevel=0;
    MaxNoOfBlacks=0;
    lumping=FALSE;

    reactionTerm = FALSE;
    
    starttri = "";
    convdir  = "";
}

void TParameter::print()
{
    cout << *this;
}

ostream& operator<< (ostream& os, const TParameter& param)
{
    os << "Start/Endlevel:       " << param.startLevel << " / " << param.endLevel << endl;
    os << "Cycle(s) in MG:       " << param.cycle << " | ";
    os << "smoothing: "            << param.preSmoothing << "/" << param.postSmoothing << endl;
    os << "Number of Nets:       " << param.noOfNets << endl;
    os << "Number of iterations: " << param.steps << endl;
    os << "Solver:               " << param.solver << endl;
    os << "Preconditioner:       " << param.prec << endl;

    os << "ordering:             ";
    switch (param.ordering)
    {
        case NONE           :  os << "none"                 << endl; break;
        case PLANAR_FVS     :  os << "planar FVS"           << endl; break;
        case HEURISTIC_FVS  :  os << "heuristic FVS"        << endl; break;
        case CONCENTRIC_MIN :  os << "CONCENTRIC_MIN"       << endl; break;
        case CONCENTRIC_MAX :  os << "CONCENTRIC_MAX"       << endl; break;
        case CROSSWIND_WO   :  os << "Crosswind w/o inner"  << endl; break;
        case CROSSWIND_WI   :  os << "Crosswind with inner" << endl; break;
        case RCM            :  os << "RCM"                  << endl; break;
        case ORDER_ALL      :  os << "all techniques"       << endl; break;
    }// switch

    os << "search in graphs:     ";
    if (param.searchType == SEARCH_DFS)
        os << "dfs" << endl;
    if (param.searchType == SEARCH_BFS)
        os << "bfs" << endl;
    
    os << "graph criterion / xi: ";
    switch(param.criterion) 
    {
    case 0 : os << "E0"; break;
    case 1 : os << "E1"; break;
    case 2 : os << "E2"; break;
    case 3 : os << "E3"; break;
    case 4 : os << "E4"; break;
    case 5 : os << "E5"; break;
    case 6 : os << "E6"; break;
    case 7 : os << "E7"; break;
    case 8 : os << "E8"; break;
    default: os << "TParameter::operator<< (...) : no criterion " << endl;  
    }// switch
    os << " / " << param.xi;
    if (param.sortOFD)
        cout << " / OFD";
    else
        cout << " / edge-weight";

    if (param.pblock)
        cout << " (planar blocking)" << endl;
    else
        cout << endl;
    
    os << "Sort on Levels:       ";
    if ( param.mgSort == MGSORT_PROJ)   os << "projected" << endl;
    if ( param.mgSort == MGSORT_INDEP)  os << "independent" << endl;
    if ( param.mgSort == MGSORT_NESTED) os << "nested" << endl;

    os << "epsilon:              " << param.epsilon << endl;
    os << "convection direction: " << param.convdir << endl;
    os << "reaction term:        ";
    if (param.reactionTerm)
        os << "yes" <<endl;
    else
        os << "no" << endl;
    os << "triangulation file:   " << param.starttri << endl;
    os << "sigma =               " << param.sigma << endl;
    os << "parameters for fs :   "
       << "alpha = " << param.fsAlpha * 180.0 / PI << "  |  "
       << "gamma = " << param.fsGamma * 180.0 / PI << "  |  "
       << "kappa = " << param.fsKappa * 180.0 / PI << endl;

    // red-black 
    os << "MaxLevel in RedBlack: " << param.MaxLevel << endl;
    os << "MaxNoOfBlacks:        " << param.MaxNoOfBlacks << endl;
    os << "Lumping:              ";
    if (param.lumping) os << "yes" <<endl;
    else  os << "no" << endl;
   
    //    os << "max level = " << param.maxLevel << endl;

    return os;
}

//
// print usage information
//
void
usage ( int status = SUCCESS ) 
{
    if (status != SUCCESS)
        cerr << "error while parsing command line" << endl << endl;

    cout << "usage: stokes [ options ... ]" << endl
         << "  where options include:"                                   << endl << endl
         << "  -c <name>         name of convection direction"           << endl
         << "  -d                draw"                                   << endl
         << "  -e <epsilon>      set epsilon "                           << endl
         << "  -g <number> <xi>  criterion and xi for graph definition"  << endl
         << "                    i - Ei  (i=0..5,7)  "                   << endl  
         << "  -h                to show this message"                   << endl
         << "  -i <filename>     start triangulation"                    << endl
         << "  -l <number>       set minimal level"                      << endl
         << "  -m [sort]         multigrid sort method where sort is"    << endl
         << "                        p - projected"                      << endl
         << "                        n - nested"                         << endl
         << "                        i - independent"                    << endl
         << "  -n <number>       set number of nets."                    << endl
         << "  -o <order>        ordering technique where order is"      << endl
         << "                        n     - none"                       << endl
         << "                        c     - concentric, minimal cycles" << endl
         << "                        C     - concentric, maximal cycles" << endl
         << "                        pfvs  - planar fvs"                 << endl
         << "                        hfvs  - heuristic fvs"              << endl
         << "                        ghfvs - global heuristic fvs"       << endl
         << "                        r     - rcm"                        << endl
         << "                        cwwi  - crosswind with"             << endl
         << "                        cwwo  - crosswind without"          << endl
         << "                        all   - all techniques"             << endl
         << "  -r                use reaction term (\"onereac\") "       << endl
         << "  -s <number>       set number of iteration steps."         << endl
         << "  -v <number>       verbosity level."                       << endl
         << "  -S                only smoothing  in MG "                 << endl
         << "  -V                V cycle in MG "                         << endl
         << "  -W                W cycle in MG "                         << endl
         << "  -w                write matrix and vertices to file and exit" << endl
         << "  -M  <number>      MaxLevel in RedBlack "                  << endl
         << "  -B <number>       MaxNoOfBlacks in RedBlack "              << endl
         << "  -L \t\t\t Lumping "                                       << endl
         << "  -T                cycle = 100 => TGM "                    << endl
         << "  --alpha <deg>     lower bound for angle between flow surfaces"    << endl
         << "  --gamma <deg>     upper bound for angle between phi and flow"     << endl
         << "  --kappa <deg>     lower bound for angle between flow and edgedir" << endl
         << "  --dfs, --bfs      searchtype in graphs"                   << endl
         << "  --sigma           used to define strong edges"            << endl
         << "  --global          build global fvs"                       << endl
         << "  --maxlvl          maximal level for cyclic reduction"     << endl
         << "  --ofd             sort edges by OFD"                      << endl
         << "  --sm <pre> <post> number of pre- and post-smoothing steps" << endl
         << "  --solver <name>   set solver"                             << endl
         << "  --prec <name>     set preconditioner"                     << endl
         << endl;

    if (status != SUCCESS)
        exit( status );
}

//
// read every option and set the parameter
//
int
TParameter::evalCmdLine ( int argc, char **argv )
{
    char  *subStr;
    
    if (argc > 0)
    {
        for (int i = 1; i < argc; i++)
        {
            if ((argv[i][0] == '-') && (argv[i][1] != '-'))
            {
                switch (argv[i][1])
                {
                case 'h': // help
                    usage( SUCCESS );
                    exit( 0 );
                    break;
              
                case 'c': // convection direction
                    if (i+1 < argc)
                        convdir = argv[++i];
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected filename after -c (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                    break;

                case 'd': // drawing?
                    draw=TRUE;
                    break;

                case 'e' : // set epsilon
                    readFloat( i, argc, argv, epsilon );
                    break;

                case 'g' : // set criterion
                    readInt( i, argc, argv, criterion );
                    readFloat( i, argc, argv, xi );
                    break;

                case 'i': // init triangulation
                    if (i+1 < argc)
                    {
                        starttri = argv[i+1];
                        i++;
                    }// if
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected filename after -i (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                    break;

                case 'l': // minimal  Level
                    readInt( i, argc, argv, startLevel );
                    break;

                case 'n': // noOfNets
                    readInt( i, argc, argv, noOfNets );
                    break;
               
                case 'r': // reaction term
                    reactionTerm=TRUE;
                    break;
                    
                case 'o': // ordering technique
                    if (i+1 < argc)
                    {
                        if (strcmp( "pfvs", argv[i+1] ) == 0)
                        {
                            ordering = PLANAR_FVS;
                            sortOFD  = true;
                        }// if
                        else if (strcmp( "hfvs", argv[i+1] ) == 0)
                            ordering = HEURISTIC_FVS;
                        else if (strcmp( "ghfvs", argv[i+1] ) == 0)
                        {
                            ordering = HEURISTIC_FVS;
                            global   = TRUE;
                        }// else if
                        else if (strcmp( "cwwi", argv[i+1] ) == 0)
                            ordering = CROSSWIND_WI;
                        else if (strcmp( "cwwo", argv[i+1] ) == 0)
                            ordering = CROSSWIND_WO;
                        else if (strcmp( "all", argv[i+1] ) == 0)
                            ordering = ORDER_ALL;
                        else
                        {
                            switch (argv[i+1][0])
                            {
                            case 'n' : ordering = NONE; break;
                            case 'c' : ordering = CONCENTRIC_MIN; break;
                            case 'C' : ordering = CONCENTRIC_MAX;  break;
                            case 'r' : ordering = RCM; break;
                            default  :
                                cerr << "TParameter::evalCmdLine (...) : "
                                     << "wrong ordering technique (try -h)." 
                                     << endl;
                                exit (ERROR);
                            }// switch
                        }// else
                        i++;
                    }// if
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected ordering technique after -o (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                    break;

                case 'm': // multigrid sort
                    if (i+1 < argc)
                    {
                        switch (argv[i+1][0])
                        {
                        case 'p' : mgSort = MGSORT_PROJ;   break;
                        case 'n' : mgSort = MGSORT_NESTED; break;
                        case 'i' : mgSort = MGSORT_INDEP;  break;
                        default  :
                            cerr << "TParameter::evalCmdLine (...) : "
                                 << "wrong multigrid sort method (try -h)." 
                                 << endl;
                            exit (ERROR);
                        }// switch
                        i++;
                    }// if
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected multigrid sort method (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                    break;

                case 's': // steps
                    readInt( i, argc, argv, steps );
                    break;

                case 'v': // verbosity
                    readInt( i, argc, argv, verbosity );
                    break;

                case 'S': // only smoothing
                    cycle=0;
                    break;

                case 'V': // V cycle
                    cycle=1;
                    break;

                case 'W': // W cycle
                    cycle=2;
                    break;

                case 'w': // write matrix to file
                    write_matrix=TRUE;
                    break;

                case 'T': // TGM
                    cycle=100;
                    break;

                case 'M': //  MaxLevel
                    readInt( i, argc, argv,MaxLevel );
                    break;  

                case 'B': //MaxNoOfBlacks
                    readInt( i, argc, argv,MaxNoOfBlacks );
                    break;  
                        
                case 'L': // Lumping
                    lumping=TRUE;
                    break;
				
                default : // error
                    cerr << "TParameter::evalCmdLine (...) : "
                         << "wrong argument (try -h)." 
                         << endl;
                    usage(ERROR);
                    break;
                }// switch
            }// if
            else if ((argv[i][0] == '-') && (argv[i][1] == '-'))
            {
                subStr = argv[i];
                subStr += 2;

                if (strcmp( subStr, "dfs" ) == 0)
                {
                    searchType = SEARCH_DFS;
                }// if
                else if (strcmp( subStr, "bfs" ) == 0)
                {
                    searchType = SEARCH_BFS;
                }// if
                else if (strcmp( subStr, "alpha" ) == 0)
                {
                    readFloat( i, argc, argv, fsAlpha );
                    fsAlpha *= PI / 180;
                }// if
                else if (strcmp( subStr, "beta" ) == 0)
                {
                    readFloat( i, argc, argv, fsBeta );
                    fsBeta *= PI / 180;
                }// if
                else if (strcmp( subStr, "gamma" ) == 0)
                {
                    readFloat( i, argc, argv, fsGamma );
                    fsGamma *= PI / 180;
                }// if
                else if (strcmp( subStr, "kappa" ) == 0)
                {
                    readFloat( i, argc, argv, fsKappa );
                    fsKappa *= PI / 180;
                }// if
                else if (strcmp( subStr, "sigma" ) == 0)
                {
                    readFloat( i, argc, argv, sigma );
                }// if
                else if (strcmp( subStr, "global" ) == 0)
                {
                    global = TRUE;
                }// if
                else if (strcmp( subStr, "maxlvl" ) == 0)
                {
                    readInt( i, argc, argv, maxLevel );
                }// if
                else if (strcmp( subStr, "ofd" ) == 0)
                {
                    sortOFD = TRUE;
                }// if
                else if (strcmp( subStr, "sm" ) == 0)
                {
                    readInt( i, argc, argv, preSmoothing );
                    readInt( i, argc, argv, postSmoothing );
                }// if
                else if (strcmp( subStr, "solver" ) == 0)
                {
                    if (i+1 < argc)
                        solver = argv[++i];
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected solvername (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                }// if
                else if (strcmp( subStr, "prec" ) == 0)
                {
                    if (i+1 < argc)
                        prec = argv[++i];
                    else
                    {
                        cerr << "TParameter::evalCmdLine (...) : "
                             << "expected prec-name (try -h)." 
                             << endl;
                        exit (ERROR);
                    }// else
                }// if
                else if (strcmp( subStr, "pblock" ) == 0)
                {
                    pblock = TRUE;
                }// if
                else  
                    cerr << "TParameter::evalCmdLine (...) : "
                         << "wrong argument (try -h)." 
                         << endl;
            }// else
        }// for
    }// if

    return (SUCCESS);
}

////////////////////////////////////////////////////
//
// access of member variables in an easier way
//

//
// read int/float with error handling
//
void
TParameter::readInt ( int & pos, int argc, char **argv, int & n )
{
    if ((pos++ < argc) && (isdigit(argv[pos][0])))
        sscanf( argv[pos], "%d", & n );
    else
    {
        cerr << "TParameter::evalCmdLine (...) : "
             << "expected number after " << argv[pos-1] << " (try -h)." 
             << endl;
        exit (ERROR);
    }// else
}

void
TParameter::readInt ( int & pos, int argc, char **argv, uint & n )
{
    if ((++pos < argc) && (isdigit(argv[pos][0])))
        sscanf( argv[pos], "%d", & n );
    else
    {
        cerr << "TParameter::evalCmdLine (...) : "
             << "expected number after " << argv[pos-1] << " (try -h)." 
             << endl;
        exit (ERROR);
    }// else
}

void
TParameter::readFloat ( int & pos, int argc, char **argv, TFloat & f )
{
    double  d;
    
    if ((++pos < argc) && (isdigit(argv[pos][0])))
    {
        sscanf( argv[pos], "%lf", & d );
        f = d;
    }// if
    else
    {
        cerr << "TParameter::evalCmdLine (...) : "
             << "expected number after " << argv[pos-1] << " (try -h)." 
             << endl;
        exit (ERROR);
    }// else
}
