//
//  Project  : libmisc
//  File     : matlab.cc
//  Author   : Ronald Kriemann / Sabine Gutsch
//  Purpose  : functions to write matrix and vector in matlab format
//
// arch-tag: 6528f704-73a3-4f85-9be5-4b7663d3ddb5
//


#ifdef MATLAB

#include "matlab.hh"

int
matWriteMatrixBlock( char *filename, const TMatrixBlock& block, uint noOfColumns,
                     char *name )
{
    uint i,j;

    uint noOfRows = block.getNoOfRows();
    int  nzmax = 0;
    // number of non-zero elements in a column
    int  *arrNZCol = new int[noOfColumns];
    int  *idx;

    for (i = 0; i < noOfColumns; i++)
        arrNZCol[i] = 0;

    MATFile  *fp = matOpen( filename, "w" );
    Matrix   *mat;

    // go through all rows and compute number of entries
    uint *arrIndex;
    uint   noOfEntries;
    for (i = 0; i < noOfRows; i++)
    {
        arrIndex = block[i].getIndex();
        noOfEntries = block[i].getNoOfEntries();

        // increase counter for every used column
        for (j = 0; j < noOfEntries; j++)
            arrNZCol[arrIndex[j]]++;

        nzmax += noOfEntries;
    }// for

    // create matlab matrix
    mat = mxCreateSparse( noOfRows, noOfColumns, nzmax, REAL );
    mxSetName( mat, name );

    // get the needed pointers
    // mxGetPr  is double* ; cast for the case that  TFloat is not "double"
    TFloat *pr = (TFloat *) mxGetPr( mat );  
    int    *ir = mxGetIr( mat );
    int    *jc = mxGetJc( mat );

    // create indices for the columns
    jc[0] = 0;
    for (i = 1; i <= noOfColumns; i++)
        jc[i] = jc[i-1] + arrNZCol[i-1];

    // now set the beginning of every column in an array
    idx = arrNZCol;
    for(i = 0; i < noOfColumns; i++)
        idx[i] = jc[i];

    // now fill the data and row indices
    TFloat  *arrValue;
    uint     col;
    for (i = 0; i < noOfRows; i++)
    {
        arrIndex = block[i].getIndex();
        arrValue = block[i].getValue();
        noOfEntries = block[i].getNoOfEntries();

        for (j = 0; j < noOfEntries; j++)
        {
            col = arrIndex[j];
            pr[idx[col]] = arrValue[j];
            ir[idx[col]] = i;
            idx[col]++;
        }// for
    }// for
    
    matPutMatrix( fp, mat );
    matClose( fp );
    mxFreeMatrix( mat );

    delete idx;

    return (SUCCESS);
}

//
//
//
int
matWriteBlockMatrix( char *filename, const TBlockMatrix& mat, char *name )
{
    uint i;
    uint blockdim = mat.getNoOfRows();
    char blockfilename[256], blockname[256];

    // write first n-1 block-diag-blocks
    for (i = 0; i < blockdim-1; i++)
    {
        sprintf( blockfilename, "%s%d%d.mat", filename, i, i );
        sprintf( blockname, "%s%d%d", name, i, i );
        matWriteMatrixBlock( blockfilename, mat[i][i], 
                             mat[i][i].getNoOfRows(), blockname );
        cout << " writing file " << blockfilename << endl;

    }// for

    // and now the last column and last row
    for (i = 0; i < blockdim; i++)
    {
        // last column
        sprintf( blockfilename, "%s%d%d.mat", filename, i, blockdim-1 );
        sprintf( blockname, "%s%d%d", name, i, blockdim-1 );
        matWriteMatrixBlock( blockfilename, mat[i][blockdim-1], 
                             mat[blockdim-1][i].getNoOfRows(), blockname );
        cout << " writing file " << blockfilename << endl;

        // last row
        if (blockdim-1 != i){
            sprintf( blockfilename, "%s%d%d.mat", filename, blockdim-1, i );
            sprintf( blockname, "%s%d%d", name, blockdim-1, i );
            matWriteMatrixBlock( blockfilename, mat[blockdim-1][i], 
                                 mat[i][blockdim-1].getNoOfRows(), blockname );
            cout << " writing file " << blockfilename << endl;
        }
    }// for

    return (SUCCESS);
}

//
//
//
int
matWriteCompleteMatrix ( char *filename, const TBlockMatrix &mat, char *name )
{
    uint          i,j,k,l;
    TMatrixBlock  tempMat;
    uint         noOfRows, rowOffs, offset, entries, noOfEntries;
    uint         arrRows[mat.getNoOfRows()];
    uint         arrIndex[1024];
    TFloat        arrValue[1024];
    uint        *pIndex;
    TFloat       *pValue;

    // check for 1-d case
    if (mat.getNoOfRows() == 1)
        matWriteMatrixBlock( filename, mat[0][0], mat[0][0].getNoOfRows(), name );

    //
    // create new block which holds all elements from mat
    //

    // get no. of rows
    noOfRows = 0;
    for (i = 0; i < mat.getNoOfRows(); i++)
    {
        j = 0;
        while (mat[i][j].getNoOfRows() == 0)
            j++;

        noOfRows   += mat[i][j].getNoOfRows();
        arrRows[i]  = mat[i][j].getNoOfRows();
    }// for

    tempMat.setNoOfRows( noOfRows );

    // create matrixdata for each row by using matrixdata 
    // from all columns
    rowOffs = 0;
    for (i = 0; i < mat.getNoOfRows(); i++)
    {
        for (l = 0; l < arrRows[i]; l++)
        {
            noOfEntries = 0;
            offset = 0;

            for (j = 0; j < mat[i].getNoOfBlocks(); j++)
            {
                if (mat[i][j].getNoOfRows() != 0)
                {
                    pIndex  = mat[i][j][l].getIndex();
                    pValue  = mat[i][j][l].getValue();
                    entries = mat[i][j][l].getNoOfEntries();

                    for (k = 0; k < entries; k++)
                    {
                        arrIndex[noOfEntries] = pIndex[k] + offset;
                        arrValue[noOfEntries] = pValue[k];
                        noOfEntries++;
                    }// for
                }// if
                
                offset += arrRows[j];
            }// for

            tempMat[l+rowOffs].set( noOfEntries, arrIndex, arrValue );
        }// for

        rowOffs += arrRows[i];
    }// for

    return matWriteMatrixBlock( filename, tempMat, noOfRows, name );
}

//
//
//

#if DIMENSION == 3 

int
matWriteBlockProlongation( char *filename, const TProlongation<3> & mat, char *name )
{
    uint i;
    uint blockdim = mat.getNoOfBlocks();
    char blockfilename[256], blockname[256];

    // write prolongation  matrix blocks
    for (i = 0; i < blockdim; i++)
    {
        sprintf( blockfilename, "%s%d.mat", filename, i );
        sprintf( blockname, "%s%d", name, i );
        matWriteMatrixBlock( blockfilename, *(mat.pBlockProlong[i]), 
                             (mat.pBlockProlong[i])->getcDim(), blockname );
        cout << " writing file " << blockfilename << endl;

    }// for
   
    return (SUCCESS);
}
#endif

#if DIMENSION == 2

int
matWriteBlockProlongation( char *filename, const TProlongation & mat, char *name )
{
    uint i;
    uint blockdim = mat.getNoOfBlocks();
    char blockfilename[256], blockname[256];

    // write prolongation  matrix blocks
    for (i = 0; i < blockdim; i++)
    {
        sprintf( blockfilename, "%s%d.mat", filename, i );
        sprintf( blockname, "%s%d", name, i );
        matWriteMatrixBlock( blockfilename, *(mat.pBlockProlong[i]), 
                             (mat.pBlockProlong[i])->getcDim(), blockname );
        cout << " writing file " << blockfilename << endl;

    }// for
   
    return (SUCCESS);
}
#endif

//
//
//

template <class T>
int
matWriteVectorBlock( char *filename, const TVectorBlock<T>& block, char *name )
{
    uint dim = block.getSize();
    MATFile  *fp = matOpen( filename, "w" );
    Matrix   *mat = mxCreateFull( dim, 1, REAL );
    mxSetName( mat, name );

    // mxGetPr  is double* ; cast for the case that  TFloat is not "double"
    TFloat *pr = (TFloat*) mxGetPr( mat );

    for (uint i = 0; i < dim; i++)
        pr[i] = block[i];

    matPutMatrix( fp, mat );
    matClose( fp );
    mxFreeMatrix( mat );

    for (uint i = 0; i < dim; i++)
        pr[i] = block[i];

    matPutMatrix( fp, mat );
    matClose( fp );
    mxFreeMatrix( mat );

    return (SUCCESS);
}

//
//
//
template <class T>
int
matWriteBlockVector( char *filename, const TBlockVector<T>& vec, char *name )
{
    uint noOfBlocks = vec.getNoOfBlocks();
    char  blockfilename[256], blockname[256];

    for (uint i = 0; i < noOfBlocks; i++)
    {
        sprintf( blockfilename, "%s%d.mat", filename, i );
        sprintf( blockname, "%s%d", name, i );
        matWriteVectorBlock( blockfilename, vec[i], blockname );
    }// for

    return (SUCCESS);
}

int
matWriteCompleteVector( char *filename, const TBlockVector<TFloat>& vec, char *name )
{
    uint  i,j;
    uint  noOfBlocks = vec.getNoOfBlocks();
    uint vecSize, offset;
    TVectorBlock<TFloat>  tempVec;

    if (noOfBlocks == 1)
        matWriteVectorBlock( filename, vec[0], name );

    vecSize = 0;
    for (i = 0; i < noOfBlocks; i++)
        vecSize += vec[i].getSize();

    tempVec.setSize( vecSize );

    offset = 0;
    for (i = 0; i < noOfBlocks; i++)
    {
        for (j = 0; j < vec[i].getSize(); j++)
        {
            tempVec[j + offset] = vec[i][j];
        }// for

        offset += vec[i].getSize();
    }// for

    return matWriteVectorBlock( filename, tempVec, name );
}

//
//
//
int
matReadVectorBlock( char *filename, TVectorBlock<TFloat>& block, char *name )
{
    uint dim = block.getSize(); cout <<" dim: "<<dim<<endl;
    MATFile  *fp;
    fp = matOpen( filename, "r" );
    if  (!fp){
        cerr<< "matReadVectorBlock:  Can't open file "<<filename<<endl;exit(1);}

    Matrix   *mat;
    mat = matGetMatrix(fp,name);
    if (!mat){ cerr<<"matReadVectorBlock:  no matrix with name "<<name<<" found in file "<<filename<<endl;exit(1);}
    //int m =mxGetM(mat); cout<<" m: "<<m;
    //int n =mxGetN(mat); cout<<" n: "<<n<<endl;

    // mxGetPr  is double* ; cast for the case that  TFloat is not "double"
    TFloat *pr = (TFloat *)mxGetPr( mat );

    for (uint i = 0; i < dim; i++)
        block[i]= pr[i];
    matClose( fp );
    mxFreeMatrix( mat );

    return (SUCCESS);
}

//
//
//
int
matReadBlockVector( char *filename,  TBlockVector<TFloat>& vec, char *name )
{
    uint noOfBlocks = vec.getNoOfBlocks();
    char  blockfilename[256], blockname[256];

    for (uint i = 0; i < noOfBlocks; i++)
    {
        sprintf( blockfilename, "%s%d.mat", filename, i );
        sprintf( blockname, "%s%d", name, i );
        matReadVectorBlock( blockfilename, vec[i], blockname );
    }// for

    return (SUCCESS);
}

// to initialize the uses templates
template int matWriteBlockVector( char *filename, const TBlockVector<TFloat>& vec, char *name );
template int matWriteBlockVector( char *filename, const TBlockVector<uint>& vec, char *name );

//#if DIMENSION == 3 
//template int matWriteBlockProlongation( char *filename, const TProlongation<3> &mat, char *name )
//#endif

#endif  // MATLAB
