//
//  Project   : libnet
//  File      : T3DCellEdge.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for an edge of a cell (in 3D)
//
// arch-tag: a705d1b9-1b67-454c-bd63-dd111e26229a
//

#include "defines.hh"

#include "vertex.hh"

#include "T3DCellEdge.hh"

//
// constructor
//
T3DCellEdge::T3DCellEdge ( TVertex<3> *node0, TVertex<3> *node2,
                           TBoundary *boundary )
{
    arrNodes[0] = node0;
    arrNodes[1] = NULL;
    arrNodes[2] = node2;

    if (node0 != NULL)
        node0->addReference();

    if (node2 != NULL)
        node2->addReference();

    arrChildren[0] = NULL;
    arrChildren[1] = NULL;
    noOfChildren   = 0;

    pBoundary = boundary;

    _data = (data_t) NULL;
}

//
// destructor
//
T3DCellEdge::~T3DCellEdge ()
{
    uint  i;

    for (i = 0; i < 3; i++)
    {
        if ((arrNodes[i] != NULL) && (arrNodes[i]->delReference()))
            delete arrNodes[i];
    }// for

    for (i = 0; i < noOfChildren; i++)
    {
        if ((arrChildren[i] != NULL) && (arrChildren[i]->delReference()) &&
            (arrChildren[i] != this))
            delete arrChildren[i];
    }// if
}

//
// access nodes (absolute and with change of direction)
//
TVertex<3> * 
T3DCellEdge::getNode ( uint i ) const
{
    if (i < 3)
        return arrNodes[i];
    else
    {
        cerr << "T3DCellEdge::getNode (" << i << ") : no such node" << endl;
        exit( -1 );
    }// else

    return NULL;
}

TVertex<3> * 
T3DCellEdge::getNode ( uint i, const bool bChangeDir ) const
{
    if (i < 3)
    {
        if (bChangeDir)
            return arrNodes[2-i];
        else
            return arrNodes[i];
    }// if
    else
    {
        cerr << "T3DCellEdge::getNode (" << i << ",bool) : no such node" << endl;
        exit( -1 );
    }// else

    return NULL;
}

//
// build midnode
//
void 
T3DCellEdge::buildMidNode ()
{
    if (hasMidNode())
    {
        cout << "T3DCellEdge::buildMidNode () : already have midnode" << endl;
        return;
    }// if

    arrNodes[1] = new TVertex<3>( (*arrNodes[0]) + (((*arrNodes[2]) - (*arrNodes[0]))*0.5),
                                  pBoundary );
}

//
// return midnode info
//
const bool 
T3DCellEdge::hasMidNode () const
{
    if (arrNodes[1] != NULL)
        return TRUE;
    else
        return FALSE;
}

//
// access children (absolute and with change of direction)
//
T3DCellEdge * 
T3DCellEdge::getChild ( uint i ) const
{
    if (i < noOfChildren)
        return arrChildren[i];
    else
    {
        cerr << "T3DCellEdge::getChild (" << i << ") : no such child" << endl;
        exit( -1 );
    }// else
}

T3DCellEdge * 
T3DCellEdge::getChild ( uint i, bool bChangeDir ) const
{
    if (i < noOfChildren)
    {
        if (bChangeDir)
            return arrChildren[1-i];
        else
            return arrChildren[i];
    }// if
    else
    {
        cerr << "T3DCellEdge::getChild (" << i << ",bool) : no such child" << endl;
        exit( -1 );
    }// else
}

//
// build children
//
void 
T3DCellEdge::buildChildren ()
{
    if (noOfChildren == 0)
    {
        if (hasMidNode())
        {
            arrChildren[0] = new T3DCellEdge( arrNodes[0], arrNodes[1], pBoundary );
            arrChildren[1] = new T3DCellEdge( arrNodes[1], arrNodes[2], pBoundary );

            noOfChildren = 2;
        }// if
        else
        {
            arrChildren[1] = this;
            this->addReference();

            noOfChildren = 1;
        }// else
    }// if
    else
    {
        cerr << "T3DCellEdge::buildChildren () : already have children" << endl;
        exit( -1 );
    }// else
}

//
// return children info
//
const bool
T3DCellEdge::hasChildren () const
{
    if (noOfChildren > 0)
        return TRUE;
    else
        return FALSE;
}

//
// access boundary
//
TBoundary *
T3DCellEdge::getBoundary () const
{
    return pBoundary;
}

//
// access data
//
T3DCellEdge::data_t
T3DCellEdge::data () const
{
    return _data;
}

void
T3DCellEdge::setData ( T3DCellEdge::data_t d )
{
    _data = d;
}

//
// output
//
ostream &
operator << ( ostream &os, const T3DCellEdge &edge )
{
    if (edge.hasMidNode())
        os << *edge.getNode(0) << " -> "
           << *edge.getNode(1) << " -> "
           << *edge.getNode(2) << endl;
    else
        os << *edge.getNode(0) << " -> "
           << *edge.getNode(2) << endl;

    return os;
}
