//
//  Project   : libnet
//  File      : T3DFrontier.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for a frontier (separates two cells) in 3d
//
// arch-tag: 388aa27b-2d47-4696-a16a-1e40e717a55f
//

#include "T3DFrontier.hh"

//
// constructor
//
T3DFrontier::T3DFrontier ( T3DFrontier *parent,
                           TBoundary   *boundary )
        : TFrontier<3>( parent, boundary )
{
    arrEdges   = NULL;
    noOfEdges  = 0;

    _data = (data_t) NULL;
}

//
// destructor
//
T3DFrontier::~T3DFrontier ()
{
    uint  i;

    for (i = 0; i < noOfEdges; i++)
    {
        if (arrEdges[i] != NULL)
            arrEdges[i]->free();
    }// for
}

//
// init memory for used arrays
//
void 
T3DFrontier::initEdgePtr ( uint n )
{
    uint  i;

    if (n == noOfEdges) 
        return;

    if (arrEdges != NULL) 
        delete arrEdges;

    if (n == 0)
    {
        arrEdges  = NULL;
        noOfEdges = 0;
    }// if
    else
    {
        arrEdges  = new T3DCellEdge * [n];
        noOfEdges = n;

        for (i = 0; i < noOfEdges; i++)
            arrEdges[i] = NULL;
    }// else
}

//
// set direction of each edge for continous access
//
const bool
T3DFrontier::edgeDir ( uint i ) const
{
    return hasFlag( 1 << i );
}

void
T3DFrontier::setEdgeDir ( uint i, const bool b )
{
    if (b == TRUE)
        setFlag( 1 << i );
    else
        clearFlag( 1 << i );
}

void 
T3DFrontier::initEdgeDir ()
{
    uint  i;

    if (noOfEdges == 0)
        return;

    if (arrEdges[0] == NULL)
    {
        cerr << "T3DFrontier::setEdgeDir () : we have a NULL edge" << endl;
        exit( -1 );
    }// if

    // we consider edge 0 as base for further direction settings
    for (i = 1; i < noOfEdges; i++)
    {
        if (arrEdges[i] == NULL)
        {
            cerr << "T3DFrontier::setEdgeDir () : we have a NULL edge" << endl;
            exit( -1 );
        }// if

        if ((arrEdges[i]->getNode( 0 ) == arrEdges[i-1]->getNode( 0, edgeDir(i-1) )) ||
            (arrEdges[i]->getNode( 2 ) == arrEdges[i-1]->getNode( 2, edgeDir(i-1) )))
        {
            setEdgeDir( i, TRUE );
        }// if
        else if ((arrEdges[i]->getNode( 0 ) == arrEdges[i-1]->getNode( 2, edgeDir(i-1) )) ||
                 (arrEdges[i]->getNode( 2 ) == arrEdges[i-1]->getNode( 0, edgeDir(i-1) )))
        {
            setEdgeDir( i, FALSE );
        }// else if
        else
        {
            cerr << "T3DFrontier::setEdgeDir () : edges don't have common node" << endl;
            exit( -1 );
        }// else
    }// for
}

//
// return number of nodes
//
const uint 
T3DFrontier::getNoOfNodes () const
{
    return noOfEdges*2;
}

//
// return i'th node
//
TVertex<3> * 
T3DFrontier::getNode ( uint i ) const
{
    uint edgeIdx = i / 2;

    if ((edgeIdx < noOfEdges) && (arrEdges[edgeIdx] != NULL))
    {
        return arrEdges[edgeIdx]->getNode( i % 2, edgeDir(edgeIdx) );
    }// if
    else
    {
        cerr << "T3DFrontier::getNode (" << i << ") : no such node" << endl;
        exit( -1 );
    }// else
}

//
// return i'th node (in 3D there is no ccw or cw)
//
TVertex<3> * 
T3DFrontier::getNode ( const TCell<3> *, uint i ) const
{
    return getNode( i );
}

//
// return edges
//
T3DCellEdge *
T3DFrontier::getEdge ( uint i ) const
{
    if (i < noOfEdges)
        return arrEdges[i];
    else
    {
        cerr << "T3DFrontier::getEdge (" << i << ") : no such edge" << endl;
        exit( -1 );
    }// else
}

//
// return the number of edges
//
const uint
T3DFrontier::getNoOfEdges () const
{
    return noOfEdges;
}

//
// access data field
//
T3DFrontier::data_t
T3DFrontier::data () const
{
    return _data;
}

void
T3DFrontier::setData ( data_t d )
{
    _data = d;
}
