//
//  Project   : libnet
//  File      : T3DTriangle.hh
//  Author    : Ronald Kriemann
//  Purpose   : class for a triangle (separates two cells) in 3d
//
// arch-tag: 8703e0ea-cf73-4c9e-9204-53841cc2832f
//

#include "T3DTriangle.hh"

//
// constructor
//
T3DTriangle::T3DTriangle ( T3DFrontier *parent,
                           TBoundary *boundary )
    : T3DFrontier( parent, boundary )
{
    initEdgePtr( 3 );

    arrEdges[0] = NULL;
    arrEdges[1] = NULL;
    arrEdges[2] = NULL;
}

T3DTriangle::T3DTriangle ( T3DFrontier *parent,
                           T3DCellEdge *e0, 
                           T3DCellEdge *e1, 
                           T3DCellEdge *e2, 
                           TBoundary *boundary )
    : T3DFrontier( parent, boundary )
{
    initEdgePtr( 3 );

    arrEdges[0] = e0;
    arrEdges[1] = e1;
    arrEdges[2] = e2;

    e0->addReference();
    e1->addReference();
    e2->addReference();

    initEdgeDir();
}

T3DTriangle::T3DTriangle ( T3DFrontier *parent,
                           T3DCellEdge *e0, 
                           T3DCellEdge *e1, 
                           T3DCellEdge *e2,
                           const bool dir0,
                           const bool dir1,
                           const bool dir2,
                           TBoundary *boundary )
    : T3DFrontier( parent, boundary )
{
    initEdgePtr( 3 );

    arrEdges[0] = e0;
    arrEdges[1] = e1;
    arrEdges[2] = e2;

    e0->addReference();
    e1->addReference();
    e2->addReference();

    setEdgeDir( 0, dir0 );
    setEdgeDir( 1, dir1 );
    setEdgeDir( 2, dir2 );
}

//
// set direction of each edge for continous access
//
void 
T3DTriangle::initEdgeDir ()
{
    if (noOfEdges == 0)
        return;

    if (arrEdges[0] == NULL)
    {
        cerr << "T3DTriangle::setEdgeDir () : we have a NULL edge" << endl;
        exit( -1 );
    }// if

    //
    // we consider edge 0 as base for further direction settings
    //
    setEdgeDir( 0, FALSE );

    //
    // compare edge 0 and edge 1
    //
    if (arrEdges[0]->getNode( 2 ) == arrEdges[1]->getNode( 0 ))
    {
        setEdgeDir( 1, FALSE );
    }// if
    else if (arrEdges[0]->getNode( 2 ) == arrEdges[1]->getNode( 2 ))
    {
        setEdgeDir( 1, TRUE );
    }// if
    else if (arrEdges[0]->getNode( 0 ) == arrEdges[1]->getNode( 2 ))
    {
        setEdgeDir( 0, TRUE );
        setEdgeDir( 1, TRUE );
    }// if
    else if (arrEdges[0]->getNode( 0 ) == arrEdges[1]->getNode( 0 ))
    {
        setEdgeDir( 0, TRUE );
        setEdgeDir( 1, FALSE );
    }// else if
    else
    {
        cerr << "T3DTriangle::setEdgeDir () : edges don't have common node" << endl;
        exit( -1 );
    }// else

    //
    // compare edge 1 and edge 2
    //
    if (arrEdges[1]->getNode( 2, edgeDir(1) ) == arrEdges[2]->getNode( 0 ))
    {
        setEdgeDir( 2, FALSE );
    }// if
    else if (arrEdges[1]->getNode( 2, edgeDir(1) ) == arrEdges[2]->getNode( 2 ))
    {
        setEdgeDir( 2, TRUE );
    }// else if
    else
    {
        cerr << "T3DTriangle::setEdgeDir () : edges don't have common node" << endl;
        exit( -1 );
    }// else

    // test
    if (arrEdges[2]->getNode( 2, edgeDir(2) ) != arrEdges[0]->getNode( 0, edgeDir(0) ))
    {
        cerr << "T3DTriangle::setEdgeDir : first and last node not correct" << endl;
        cout << "T3DTriangle::setEdgeDir : e0 = "
             << * arrEdges[0]->getNode( 0, edgeDir(0) ) << " -> "
             << * arrEdges[0]->getNode( 2, edgeDir(0) ) << endl;
        cout << "T3DTriangle::setEdgeDir : e1 = "
             << * arrEdges[1]->getNode( 0, edgeDir(1) ) << " -> "
             << * arrEdges[1]->getNode( 2, edgeDir(1) ) << endl;
        cout << "T3DTriangle::setEdgeDir : e2 = "
             << * arrEdges[2]->getNode( 0, edgeDir(2) ) << " -> "
             << * arrEdges[2]->getNode( 2, edgeDir(2) ) << endl;
        exit( -1 );
    }// if
}

//
// return info about built midnodes
//
const bool 
T3DTriangle::hasMidNode () const
{
    if (noOfEdges > 0)
    {
        if ((arrEdges[0] != NULL) && !arrEdges[0]->hasMidNode())
            return FALSE;
    
        if ((arrEdges[1] != NULL) && !arrEdges[1]->hasMidNode())
            return FALSE;

        if ((arrEdges[2] != NULL) && !arrEdges[2]->hasMidNode())
            return FALSE;

        return TRUE;
    }// if
    
    return FALSE;
}

//
// make the midnode(s) of the frontier
//
void 
T3DTriangle::buildMidNode ()
{
    uint  i;

    if (noOfEdges == 0)
        return;

    if (hasMidNode())
    {
        cerr << "T3DTriangle::buildMidNode () : already have midnodes" << endl;
        exit( -1 );
    }// if

    for (i = 0; i < 3; i++)
    {
        if ((arrEdges[i] != NULL) && (!arrEdges[i]->hasMidNode()))
            arrEdges[i]->buildMidNode();
    }// for
}

//
// build the child(ren)
//
void 
T3DTriangle::buildChild ()
{
    if (noOfChildren != 0)
    {
        cerr << "T3DTriangle::buildChild () : children exists" << endl;
        return;
    }// if
    else if (hasMidNode())
    {
        T3DCellEdge  *arrInnerEdges[3];

        initChildPtr( 4 );

        // build children of edges
        if ((arrEdges[0] != NULL) && (!arrEdges[0]->hasChildren()))
            arrEdges[0]->buildChildren();

        if ((arrEdges[1] != NULL) && (!arrEdges[1]->hasChildren()))
            arrEdges[1]->buildChildren();

        if ((arrEdges[2] != NULL) && (!arrEdges[2]->hasChildren()))
            arrEdges[2]->buildChildren();

        // and inner edges of triangle
        arrInnerEdges[0] = new T3DCellEdge( getNode( 1 ), getNode( 5 ), pBoundary );
        arrInnerEdges[1] = new T3DCellEdge( getNode( 3 ), getNode( 1 ), pBoundary );
        arrInnerEdges[2] = new T3DCellEdge( getNode( 5 ), getNode( 3 ), pBoundary );

        arrInnerEdges[0]->delReference();
        arrInnerEdges[1]->delReference();
        arrInnerEdges[2]->delReference();

        // finally the children
        arrChildren[0] = new T3DTriangle( this,
                                          arrEdges[2]->getChild( 1, edgeDir(2) ),
                                          arrEdges[0]->getChild( 0, edgeDir(0) ),
                                          arrInnerEdges[0],
                                          edgeDir(2), edgeDir(0), FALSE,
                                          pBoundary );

        arrChildren[1] = new T3DTriangle( this,
                                          arrEdges[0]->getChild( 1, edgeDir(0) ),
                                          arrEdges[1]->getChild( 0, edgeDir(1) ),
                                          arrInnerEdges[1],
                                          edgeDir(0), edgeDir(1), FALSE,
                                          pBoundary );

        arrChildren[2] = new T3DTriangle( this,
                                          arrEdges[1]->getChild( 1, edgeDir(1) ),
                                          arrEdges[2]->getChild( 0, edgeDir(2) ),
                                          arrInnerEdges[2],
                                          edgeDir(1), edgeDir(2), FALSE,
                                          pBoundary );

        arrChildren[3] = new T3DTriangle( this,
                                          arrInnerEdges[0],
                                          arrInnerEdges[2],
                                          arrInnerEdges[1],
                                          FALSE, FALSE, FALSE,
                                          pBoundary );
    }// else if
    else
    {
        initChildPtr( 1 );
        arrChildren[0] = this;
        addReference();
    }// else
}

//
// return i'th child (no direction control)
//
TFrontier<3>* 
T3DTriangle::getChild ( const TCell<3> *, uint i ) const
{
    if (i < noOfChildren)
        return arrChildren[i];
    else
    {
        cerr << "T3DTriangle::getChild (TCell, " << i << ") : no such child" << endl;
        return NULL;
    }// else
}
