//
//  Project   : libnet
//  File      : baensch.cc
//  Author    : Sabine Gutsch / Ronald Kriemann
//  Purpose   : class of cell which implements the Baensch-triangulation
//
// arch-tag: 0e62128e-70a6-4c69-90fe-f092c517018e
//

#include "baensch.hh"

//
// constructor
//
T2DBaenschCell::T2DBaenschCell ( TCell<2> *parent, 
                                 T2DFrontier *frontier0, 
                                 T2DFrontier *frontier1, 
                                 T2DFrontier *frontier2 )    
    : TCell<2>( parent ) 
{
    initFrontierPtr( 3 );
    
    arrFrontiers[0] = frontier0;
    arrFrontiers[1] = frontier1;
    arrFrontiers[2] = frontier2;

    frontier0->addReference();
    frontier1->addReference();
    frontier2->addReference();

    frontier0->setCell( this );
    frontier1->setCell( this );
    frontier2->setCell( this );
    
    
    nMark = -1;
    fSize = calcSize();
}

T2DBaenschCell::T2DBaenschCell () : TCell<2>(NULL) 
{}

//
// divide all marked frontiers before real refinement
//
void 
T2DBaenschCell::prepareRefine ()
{
    if (nMark < 0)
    {
        cerr << "T2DBaenschCell::prepareRefine() : no mark set" << endl;
        return;
    }// if

    if (!arrFrontiers[nMark]->hasMidNode()) arrFrontiers[nMark]->buildMidNode();
}

//
// return wich frontier is marked
//
const uint 
T2DBaenschCell::getMark () const 
{ 
    return nMark;
}

//
// set mark of frontier for refinement
//
void 
T2DBaenschCell::setMark ( uint i ) 
{ 
    nMark = i; 
}

//
// set mark automagically (at start -> the biggest frontier :) )
//
void 
T2DBaenschCell::setMark () 
{
    if (arrFrontiers[0]->size() > arrFrontiers[1]->size())
    {
        if (arrFrontiers[0]->size() > arrFrontiers[2]->size())
            nMark = 0;
        else
            nMark = 2;
    }// if
    else
    {
        if (arrFrontiers[1]->size() > arrFrontiers[2]->size())
            nMark = 1;
        else
            nMark = 2;
    }// else
}

//
// refine yourself (build only the children)
//
void 
T2DBaenschCell::refine () 
{
    if (nMark < 0)
    {
        cerr << "T2DBaenschCell::prepareRefine() : no mark set" << endl;
        return;
    }// if

    // first tile marked frontier
    if (!arrFrontiers[nMark]->hasMidNode()) arrFrontiers[nMark]->buildMidNode();

    // then create children of all frontiers
    if (!arrFrontiers[0]->hasChild()) arrFrontiers[0]->buildChild();
    if (!arrFrontiers[1]->hasChild()) arrFrontiers[1]->buildChild();
    if (!arrFrontiers[2]->hasChild()) arrFrontiers[2]->buildChild();

    // count the numbers of divided frontiers
    int  nCount = 0;
    for (int i = 0; i < 3; i++)
    {
        if (arrFrontiers[i]->hasMidNode()) nCount++;
    }// for

    // update field noOfChildren
    initChildPtr( nCount+1 );
    
    // get the opposite node of the marked frontier
    T2DVertex  *pOppositeNode = ((T2DFrontier*)arrFrontiers[(nMark+1)%3])->getNode( this, 2 );
    
    // create middlefrontier
    T2DFrontier *pMidFrontier = new T2DFrontier( NULL, pOppositeNode, 
                                                 ((T2DFrontier*)arrFrontiers[nMark])->getNode(this,1));
    pMidFrontier->delReference();

    // decide how to divide the cell by the number of marked edges
    if (nCount == 1)
    {
#ifdef DEBUG
        //            cout << "one divided frontier" << endl;
#endif
        // one divided frontier -> create two children
        // - 00 is the first child of the marked frontier
        //   and 01 is the second child
        // - 1 is the (child of the) frontier after the marked one
        //   and 2 is the frontier before the marked
        T2DFrontier  *p00 = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,0);
        T2DFrontier  *p01 = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,1);
        T2DFrontier  *p1  = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,0);
        T2DFrontier  *p2  = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,0);
            
        arrChildren[0] = new T2DBaenschCell( this, p01, p1, pMidFrontier );
        arrChildren[0]->setMark( 1 );
        arrChildren[1] = new T2DBaenschCell( this, p00, pMidFrontier, p2 );
        arrChildren[1]->setMark( 2 );
    }// if
    else if (nCount == 2)
    {
        // two divided frontier -> create three children
        // - for 00 and 01 see case 1
        T2DFrontier  *p00 = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,0);
        T2DFrontier  *p01 = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,1);
            
        // check if the other divided frontier is next or prev
        if (arrFrontiers[(nMark+1)%3]->hasMidNode())
        {
#ifdef DEBUG
            //            cout << "two divided frontiers (case 1)" << endl;
#endif
            // the frontier after the marked frontier is divided
            // for 10, 11, mid1 see case 3, for 2 see case 1
            T2DFrontier *p10  = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,0);
            T2DFrontier *p11  = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,1);
            T2DFrontier *pMid1= new T2DFrontier(NULL,
                                                arrFrontiers[(nMark+1)%3]->getNode(this,1),
                                                arrFrontiers[nMark]->getNode(this,1) );
            pMid1->delReference();
            T2DFrontier *p2   = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,0);

            arrChildren[0] = new T2DBaenschCell( this, p01, p10, pMid1 );
            arrChildren[0]->setMark(0);
            arrChildren[1] = new T2DBaenschCell( this, pMid1, p11, pMidFrontier );
            arrChildren[1]->setMark(2);
            arrChildren[2] = new T2DBaenschCell( this, pMidFrontier, p2, p00 );
            arrChildren[2]->setMark(1);
        }// if
        else
        {
#ifdef DEBUG
            //            cout << "two divided frontiers (case 2)" << endl;
#endif
            // the frontier before the marked frontier is divided
            // for 20, 21, mid2 see case 3, for 1 see case 1
            T2DFrontier *p1   = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,0);
            T2DFrontier *p20  = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,0);
            T2DFrontier *p21  = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,1);
            T2DFrontier *pMid2= new T2DFrontier(NULL,
                                                arrFrontiers[(nMark+2)%3]->getNode(this,1),
                                                arrFrontiers[nMark]->getNode(this,1) );
            pMid2->delReference();

            arrChildren[0] = new T2DBaenschCell( this, p01, p1, pMidFrontier );
            arrChildren[0]->setMark(1);
            arrChildren[1] = new T2DBaenschCell( this, pMidFrontier, p20, pMid2 );
            arrChildren[1]->setMark(0);
            arrChildren[2] = new T2DBaenschCell( this, pMid2, p21, p00 );
            arrChildren[2]->setMark(2);
        }// else
    }// if
    else if (nCount == 3)
    {
#ifdef DEBUG
        //            cout << "three divided frontiers" << endl;
#endif
        // three divided frontier -> create four children
        // - the marked frontier is numbered with 0, the next frontier is 1
        //   and the prev is 2
        // - children of this frontiers are numbered like 00 for the first
        //   children of the marked frontier or 11 for the second children of
        //   frontier 1
        // - mid1 is the frontier from midnode of frontier 1 to midnode of
        //   the marked frontier
        // - mid2 is the frontier from midnode of frontier 2 to midnode of
        //   the marked frontier
        T2DFrontier *p00  = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,0);
        T2DFrontier *p01  = (T2DFrontier*)arrFrontiers[nMark]->getChild(this,1);
        T2DFrontier *p10  = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,0);
        T2DFrontier *p11  = (T2DFrontier*)arrFrontiers[(nMark+1)%3]->getChild(this,1);
        T2DFrontier *p20  = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,0);
        T2DFrontier *p21  = (T2DFrontier*)arrFrontiers[(nMark+2)%3]->getChild(this,1);
        T2DFrontier *pMid1= new T2DFrontier(NULL,
                                            arrFrontiers[(nMark+1)%3]->getNode(this,1),
                                            arrFrontiers[nMark]->getNode(this,1) );
        pMid1->delReference();
        T2DFrontier *pMid2= new T2DFrontier(NULL,
                                            arrFrontiers[(nMark+2)%3]->getNode(this,1),
                                            arrFrontiers[nMark]->getNode(this,1) );
        pMid2->delReference();

        arrChildren[0] = new T2DBaenschCell( this, p01, p10, pMid1 );
        arrChildren[0]->setMark(0);
        arrChildren[1] = new T2DBaenschCell( this, pMid1, p11, pMidFrontier );
        arrChildren[1]->setMark(2);
        arrChildren[2] = new T2DBaenschCell( this, pMidFrontier, p20, pMid2 );
        arrChildren[2]->setMark(0);
        arrChildren[3] = new T2DBaenschCell( this, pMid2, p21, p00 );
        arrChildren[3]->setMark(2);
    }// if
    else
        cerr << "T2DBaenschCell::calcRefinement() : no frontier divided" << endl;
}

//
// output for easier debugging
//
ostream& 
operator<< (ostream& os, const T2DBaenschCell& cell) 
{
    os << *(cell.getFrontier(0)->getNode(&cell,0))
       << " --> " 
       << *(cell.getFrontier(0)->getNode(&cell,2)) << endl;
    os << *(cell.getFrontier(1)->getNode(&cell,0))
       << " --> " 
       << *(cell.getFrontier(1)->getNode(&cell,2)) << endl;
    os << *(cell.getFrontier(2)->getNode(&cell,0))
       << " --> " 
       << *(cell.getFrontier(2)->getNode(&cell,2)) << endl;
    
    return (os);
}

//
// and the opposite : coarse
//
void 
T2DBaenschCell::coarsen () 
{}
