%{
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "dynarray.hh"
#include "list.hh"
#include "TString.hh"
#include "var.hh"

#include "vertex.hh"
#include "T3DCellEdge.hh"
#include "frontier.hh"
#include "T2DFrontier.hh"
#include "T3DFrontier.hh"
#include "T3DTriangle.hh"
#include "cell.hh"
#include "baensch.hh"
#include "regular.hh"
#include "T3DTetrahedron.hh"
#include "TBoundary.hh"
#include "TDomain.hh"
#include "T3DCubeDomain.hh"
    
//
// decl for parser and scanner
//
extern FILE* yyin;
extern void yyrestart(FILE *input_file);
extern int yylex();
extern int yyparse();
void yyerror( const char *s );
void yywarn( const char *s );

//
// some messages
//
#define MSG_HAS_BOUNDNAME    "already has a boundary name"
#define MSG_IS_INTERIOR      "already declared as interior"
#define MSG_WRONG_DIM        "wrong dimension defined"
#define MSG_WRONG_TYPE       "wrong type was given"
#define MSG_WRONG_REFINE     "wrong refinement strategy defined"
#define MSG_HAVE_NAME        "already have this name"
#define MSG_NO_TYPE          "no type defined"
#define MSG_NULL_PTR         "have null pointer (given lists ?)"
#define MSG_HAVE_DOMAIN      "already have a domain"
#define MSG_NOT_IMPLEMENTED  "not implemented"

//
// other forward decl
//
void init ();
void cleanup ();
void resetData ();
int idxOfName ( const TDynArray<TString> * pArray, const TString & name );

#define NO_IDX -1

//
// act. pos in file for debugging
//
int nLine;
int nPos;

//
// dimension of the problem
//
int dim;

//
// names of the red objects
//
TDynArray<TString>  *pArrNodeNames;
TDynArray<TString>  *pArrEdgeNames;
TDynArray<TString>  *pArrFrontierNames;
TDynArray<TString>  *pArrCellNames;
TDynArray<TString>  *pArrBoundNames;
TDynArray<TString>  *pArrNames;

//
// the red objects (void-pointer to be dim-independent)
//
TDynArray<void*>         *pArrNodes;
TDynArray<T3DCellEdge*>  *pArrEdges;           // no edges in 2D
TDynArray<void*>         *pArrFrontiers;
TDynArray<TBoundary*>    *pArrBoundaries;

//
// celllists for 2D and 3D
//
TList< TCell<2> >  *pList2DCells;
TList< TCell<3> >  *pList3DCells;

//
// working variables (stored parsed info)
//
TVector<2>  *pVec2D;
TVector<3>  *pVec3D;

bool     hasBoundName, isInterior, hasCoord, hasType, hasRefine, hasDomain;
TString  *pszName, *pszBoundName, *pszType, *pszRefine;
%}

%union{
	int    nInt;
	double fFloat;
    char   szString[256];
}

%token DIM STRUCTNODE STRUCTEDGE STRUCTFRONTIER STRUCTCELL STRUCTBOUND STRUCTDOMAIN
%token STRING FLOAT INT
%token COORD NODEFIELD EDGEFIELD FRONTIERFIELD TYPEFIELD REFINEFIELD
%token BOUNDFIELD INTERIORFIELD

%token <fFloat> FLOAT
%token <nInt> INT
%token <szString> STRING

%start START

%%

START: EXP1
     ;

EXP1:
    | EXP1 EXP
    ;

EXP: DIM '=' INT
    {
        if ((dim != 1) && (dim != $3))
            yyerror( "gave a second (different) def. of dim" );

        if ((dim != 1) && (dim == $3))
            yywarn( "gave a second (equal) def. of dim" );
        
        dim = $3;
        
        if ((dim != 2) && (dim != 3))
            yyerror( "support's only dim \\in {2,3}" );
    }
   | NODE
   | EDGE
   | FRONTIER
   | CELL
   | BOUNDARY
   | DOMAIN
;

NODE: NODEDECL '{' NODEBODY '}'
    {
        if (hasCoord && (hasBoundName || isInterior))
        {
            int  idx;
            
            //
            // build node
            //
            
            idx = pArrNodes->getSize();
            
            pArrNodes->setSize( idx + 1 );
            pArrNodeNames->setSize( idx + 1 );
            
            (*pArrNodeNames)[idx] = *pszName;
            
            if (dim == 2)
            {
                TVertex<2>  *pVertex;
                
                if (isInterior)
                    pVertex = new TVertex<2>( *pVec2D, INTERIOR );
                else
                    pVertex = new TVertex<2>( *pVec2D, 1 );

                (*pArrNodes)[idx] = pVertex;

                if (pVertex != NULL)
                    pVertex->delRef();
            }// if
            else if (dim == 3)
            {
                TVertex<3>  *pVertex;
                
                if (isInterior)
                    pVertex = new TVertex<3>( *pVec3D, INTERIOR );
                else
                {
                    TBoundary *pBound;

                    pBound = (*pArrBoundaries)[idxOfName(pArrBoundNames,
                                                         *pszBoundName )];
                    pVertex = new TVertex<3>( *pVec3D, pBound );
                }// else

                (*pArrNodes)[idx] = pVertex;

                if (pVertex != NULL)
                    pVertex->delRef();
            }// else if
            else
                yyerror( MSG_WRONG_DIM );

            //
            // reset all data
            //
            resetData();
        }// if
        else
            yyerror( "do not have all data" );
    }
    ;

NODEDECL: STRUCTNODE STRING
        {
            *pszName = $2;

            if (idxOfName( pArrNodeNames, *pszName ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
        }
        ;

NODEBODY:
        | NODEBODY COORD_ND
        {
            if (hasCoord)
                yywarn( "already have coord for this node" );

            hasCoord = TRUE;
        }
        | NODEBODY BOUNDNAME
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );

            isInterior   = FALSE;
            hasBoundName = TRUE;
        }
        | NODEBODY INTERIORFIELD
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );

            isInterior   = TRUE;
            hasBoundName = FALSE;
        }
;

COORD_ND: COORD_2D
        {
            if (dim != 2)
                yyerror( "dim != 2 but 2D coord was given" );
        }
        | COORD_3D
        {
            if (dim != 3)
                yyerror( "dim != 3 but 3D coord was given" );
        }
        ;

COORD_2D: COORD FLOAT FLOAT
        {
            (*pVec2D)[X] = $2;
            (*pVec2D)[Y] = $3;
        }
        ;

COORD_3D: COORD FLOAT FLOAT FLOAT
        {
            (*pVec3D)[X] = $2;
            (*pVec3D)[Y] = $3;
            (*pVec3D)[Z] = $4;
        }
        ;

EDGE: EDGEDECL '{' EDGEBODY '}'
        {
            int           idx;
            int           nodeIdx1, nodeIdx2;
            T3DCellEdge  *pEdge;
            
            if (dim == 2)
                yyerror( "no edges in 2D" );

            if (dim != 3)
                yyerror( MSG_WRONG_DIM );

            if (!hasBoundName && !isInterior)
                yyerror( "no type (boundary/interior) was specified" );

            idx = pArrEdges->getSize();
            pArrEdgeNames->setSize( idx+1 );
            pArrEdges->setSize( idx+1 );

            (*pArrEdgeNames)[idx] = *pszName;

            nodeIdx1 = idxOfName( pArrNodeNames, (*pArrNames)[0] );
            nodeIdx2 = idxOfName( pArrNodeNames, (*pArrNames)[1] );
            
            if (isInterior)
            {
                pEdge = new T3DCellEdge( (TVertex<3>*) (*pArrNodes)[nodeIdx1],
                                         (TVertex<3>*) (*pArrNodes)[nodeIdx2],
                                         INTERIOR );
            }// if
            else
            {
                TBoundary  *pBound;

                pBound = (*pArrBoundaries)[idxOfName( pArrBoundNames,
                                                      *pszBoundName )];
                pEdge = new T3DCellEdge( (TVertex<3>*) (*pArrNodes)[nodeIdx1],
                                         (TVertex<3>*) (*pArrNodes)[nodeIdx2],
                                         pBound );
            }// else

            (*pArrEdges)[idx] = pEdge;
            
            if (pEdge != NULL)
                pEdge->delRef();
            
            //
            // reset all data
            //
            resetData();
        }
        ;

EDGEDECL: STRUCTEDGE STRING
        {
            *pszName = $2;

            if (idxOfName( pArrEdgeNames, *pszName ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
        }
        ;

EDGEBODY:                     
        | EDGEBODY NODELIST
        {
            if (pArrNames->getSize() != 2)
                yyerror( "did not gave correct number of nodes (2)" );
        }    
        | EDGEBODY BOUNDNAME
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );

            isInterior   = FALSE;
            hasBoundName = TRUE;
        }
        | EDGEBODY INTERIORFIELD
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );

            isInterior   = TRUE;
            hasBoundName = FALSE;
        }
        ;

FRONTIER: FRONTIERDECL '{' FRONTIERBODY '}'
        {
            int  idx;
            
            if ((dim != 2) && (dim != 3))
                yyerror( MSG_WRONG_DIM );

            if (!hasBoundName && !isInterior)
                yyerror( "no type (boundary/interior) was specified" );

            if (!hasType)
                yyerror( MSG_NO_TYPE );

            if ((*pszType == "normal") && (pArrNames->getSize() != 2))
                yyerror( "incorrect number of nodes was given (must be 2)" );
            
            if ((*pszType == "triangle") && (pArrNames->getSize() != 3))
                yyerror( "incorrect number of edges was given (must be 3)" );

            idx = pArrFrontiers->getSize();
            pArrFrontierNames->setSize( idx+1 );
            pArrFrontiers->setSize( idx+1 );

            (*pArrFrontierNames)[idx] = *pszName;

            if (dim == 2)
            {
                TVertex<2>   *pNode[2];
                TFrontier<2> *pFrontier = NULL;
                
                if (*pszType == "normal")
                {
                    pNode[0] = (TVertex<2>*) (*pArrNodes)[idxOfName( pArrNodeNames,
                                                                    (*pArrNames)[0])];
                    pNode[1] = (TVertex<2>*) (*pArrNodes)[idxOfName( pArrNodeNames,
                                                                    (*pArrNames)[1])];
                    
                    if (isInterior)
                        pFrontier = new T2DFrontier( NULL, pNode[0], pNode[1],
                                                                 INTERIOR );
                    else
                        pFrontier = new T2DFrontier( NULL, pNode[0], pNode[1], 1 );
                }// if

                (*pArrFrontiers)[idx] = pFrontier;

                if (pFrontier != NULL)
                    pFrontier->delRef();
            }// if
            else
            {
                T3DCellEdge  *pEdge[3];
                TFrontier<3> *pFrontier = NULL;
                
                if (*pszType == "triangle")
                {
                    pEdge[0] = (*pArrEdges)[idxOfName( pArrEdgeNames, (*pArrNames)[0] )];
                    pEdge[1] = (*pArrEdges)[idxOfName( pArrEdgeNames, (*pArrNames)[1] )];
                    pEdge[2] = (*pArrEdges)[idxOfName( pArrEdgeNames, (*pArrNames)[2] )];
                    
                    if (isInterior)
                    {
                        pFrontier = new T3DTriangle( NULL,
                                                     pEdge[0], pEdge[1], pEdge[2],
                                                     INTERIOR );
                    }// if
                    else
                    {
                        TBoundary *pBound;
                        int           boundIdx;

                        boundIdx = idxOfName(pArrBoundNames,*pszBoundName);
                        pBound = (*pArrBoundaries)[boundIdx];
                        pFrontier = new T3DTriangle( NULL,
                                                     pEdge[0], pEdge[1], pEdge[2],
                                                     pBound );
                    }// else
                }// if

                (*pArrFrontiers)[idx] = pFrontier;

                if (pFrontier != NULL)
                    pFrontier->delRef();
            }// else
            
            //
            // reset all data regarding nodes
            //
            resetData();
        }
        ;

FRONTIERDECL: STRUCTFRONTIER STRING
        {
            *pszName = $2;

            if (idxOfName( pArrFrontierNames, *pszName ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
        }
        ;

FRONTIERBODY:
        | FRONTIERBODY EDGELIST
        {
            if (dim == 2)
                yyerror( "must use nodes in 2D" );
            
            if (dim != 3)
                yyerror( MSG_WRONG_DIM );
        }
        | FRONTIERBODY NODELIST
        {
            if (dim == 3)
                yyerror( "must use edges in 3D" );
            
            if (dim != 2)
                yyerror( MSG_WRONG_DIM );
        }
        | FRONTIERBODY BOUNDNAME
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );
            
            isInterior   = FALSE;
            hasBoundName = TRUE;
        }
        | FRONTIERBODY INTERIORFIELD
        {
            if (isInterior)
                yywarn( MSG_IS_INTERIOR );
            
            if (hasBoundName)
                yywarn( MSG_HAS_BOUNDNAME );
            
            isInterior   = TRUE;
            hasBoundName = FALSE;
        }
        | FRONTIERBODY TYPE
        {
            if (hasType)
                yywarn( "type already defined" );

            if (dim == 2)
            {
                if (*pszType != "normal")
                    yyerror( MSG_WRONG_TYPE );
            }// if
            else if (dim == 3)
            {
                if (*pszType != "triangle")
                    yyerror( MSG_WRONG_TYPE );
            }// else
            else
                yyerror( MSG_WRONG_DIM );

            hasType = TRUE;
        }
        ;

CELL: CELLDECL '{' CELLBODY '}'
        {
            int  idx;
            
            if ((dim != 2) && (dim != 3))
                yyerror( MSG_WRONG_DIM );

            if (!hasType)
                yyerror( MSG_NO_TYPE );

            if ((*pszType == "triangle") &&
                (pArrNames->getSize() != 3))
                yyerror( "incorrect number of frontiers was given" );

            if ((*pszType == "tetrahedron") &&
                (pArrNames->getSize() != 4))
                yyerror( "incorrect number of frontiers was given" );

            idx = pArrCellNames->getSize();
            pArrCellNames->setSize( idx+1 );

            (*pArrCellNames)[idx] = *pszName;

            if (dim == 2)
            {
                TCell<2>  *pCell = NULL;
                
                if (pList2DCells == NULL)
                    yyerror( MSG_NULL_PTR );
                        
                if (*pszType == "triangle")
                {
                    int          pArrIdx[3];
                    T2DFrontier *pArrFront[3];
                    
                    pArrIdx[0] = idxOfName( pArrFrontierNames, (*pArrNames)[0] );
                    pArrIdx[1] = idxOfName( pArrFrontierNames, (*pArrNames)[1] );
                    pArrIdx[2] = idxOfName( pArrFrontierNames, (*pArrNames)[2] );
                    
                    pArrFront[0] = (T2DFrontier*) (*pArrFrontiers)[pArrIdx[0]];
                    pArrFront[1] = (T2DFrontier*) (*pArrFrontiers)[pArrIdx[1]];
                    pArrFront[2] = (T2DFrontier*) (*pArrFrontiers)[pArrIdx[2]];

                    if (*pszRefine == "baensch")
                        pCell = new T2DBaenschCell( NULL,
                                                    pArrFront[0],
                                                    pArrFront[1],
                                                    pArrFront[2] );
                    else if (*pszRefine == "regular")
                        pCell = new T2DRegularCell( NULL,
                                                    pArrFront[0],
                                                    pArrFront[1],
                                                    pArrFront[2] );
                }// if

                if (pCell != NULL)
                {
                    pCell->delRef();
                    pList2DCells->add( *pCell );
                }// if
            }// if
            else // dim == 3
            {
                TCell<3>  *pCell = NULL;
                
                if (pList3DCells == NULL)
                    yyerror( MSG_NULL_PTR );
                        
                if (*pszType == "tetrahedron")
                {
                    int  pArrIdx[4];

                    pArrIdx[0] = idxOfName( pArrFrontierNames, (*pArrNames)[0] );
                    pArrIdx[1] = idxOfName( pArrFrontierNames, (*pArrNames)[1] );
                    pArrIdx[2] = idxOfName( pArrFrontierNames, (*pArrNames)[2] );
                    pArrIdx[3] = idxOfName( pArrFrontierNames, (*pArrNames)[3] );
                    
                    if (*pszRefine == "bey")
                    {
                        T3DTriangle *pArrFront[4];
                    
                        pArrFront[0] = (T3DTriangle*) (*pArrFrontiers)[pArrIdx[0]];
                        pArrFront[1] = (T3DTriangle*) (*pArrFrontiers)[pArrIdx[1]];
                        pArrFront[2] = (T3DTriangle*) (*pArrFrontiers)[pArrIdx[2]];
                        pArrFront[3] = (T3DTriangle*) (*pArrFrontiers)[pArrIdx[3]];

                        pCell = new T3DTetrahedron( NULL,
                                                    pArrFront[0],
                                                    pArrFront[1],
                                                    pArrFront[2],
                                                    pArrFront[3] );
                    }// if
                }// if

                if (pCell != NULL)
                {
                    pCell->delRef();
                    pList3DCells->add( *pCell );
                }// if
            }// else

            //
            // reset all changed data
            //
            resetData();
        }
        ;

CELLDECL: STRUCTCELL STRING
        {
            *pszName = $2;

            if (idxOfName( pArrCellNames, *pszName ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
        }
        ;

CELLBODY:
        | CELLBODY FRONTIERLIST
        | CELLBODY TYPE
        {
            if (hasType)
                yywarn( "type already defined" );

            if (dim == 2)
            {
                if (*pszType != "triangle")
                    yyerror( MSG_WRONG_TYPE );
            }// if
            else if (dim == 3)
            {
                if (*pszType != "tetrahedron")
                    yyerror( MSG_WRONG_TYPE );
            }// else
            else
                yyerror( MSG_WRONG_DIM );

            hasType = TRUE;
        }
        | CELLBODY REFINEMENT
        {
            if (hasRefine)
                yywarn( "refinement already defined" );

            if (dim == 2)
            {
                if ((*pszRefine != "baensch") &&
                    (*pszRefine != "regular"))
                    yyerror( MSG_WRONG_REFINE );
            }// if
            else if (dim == 3)
            {
                if (*pszRefine != "bey")
                    yyerror( MSG_WRONG_REFINE );
            }// else if
            else
                yyerror( MSG_WRONG_DIM );

            hasRefine = TRUE;
        }
        ;

NODELIST: NODEFIELD NAMES
        {
            for (uint i = 0; i < pArrNames->getSize(); i++)
            {
                if (idxOfName( pArrNodeNames, (*pArrNames)[i] ) == NO_IDX)
                    yyerror( "did not found a node with given name" );
            }// if
        }
        ;

EDGELIST: EDGEFIELD NAMES
        {
            for (uint i = 0; i < pArrNames->getSize(); i++)
            {
                if (idxOfName( pArrEdgeNames, (*pArrNames)[i] ) == NO_IDX)
                    yyerror( "did not found an edge with this name" );
            }// if
        }
        ;

FRONTIERLIST: FRONTIERFIELD NAMES
        {
            for (uint i = 0; i < pArrNames->getSize(); i++)
            {
                if (idxOfName( pArrFrontierNames, (*pArrNames)[i] ) == NO_IDX)
                    yyerror( "did not found a frontier with this name" );
            }// if
        }
        ;

BOUNDLIST: BOUNDFIELD NAMES
        {
            for (uint i = 0; i < pArrNames->getSize(); i++)
            {
                if (idxOfName( pArrBoundNames, (*pArrNames)[i] ) == NO_IDX)
                    yyerror( "did not found a boundary with this name" );
            }// if
        }
        ;

NAMES:
        | NAMES STRING
        {
            int  idx;
            
            if (idxOfName( pArrNames, $2 ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
            
            idx = pArrNames->getSize();
            pArrNames->setSize( idx+1 );
            (*pArrNames)[idx] = $2;
        }
        ;

BOUNDNAME: STRUCTBOUND '=' STRING
        {
            *pszBoundName = $3;

            if (idxOfName( pArrBoundNames, *pszBoundName ) == NO_IDX)
                yyerror( "no boundary with this name defined" );
        }
        ;

TYPE: TYPEFIELD '=' STRING
        {
            *pszType = $3;
        }
        ;

REFINEMENT: REFINEFIELD '=' STRING
        {
            *pszRefine = $3;
        }
        ;

BOUNDARY: BOUNDDECL '{' '}'
        {
            int  idx;
            
            idx = pArrBoundNames->getSize();
            pArrBoundNames->setSize( idx + 1 );
            pArrBoundaries->setSize( idx + 1 );
                
            (*pArrBoundNames)[idx] = *pszName;
            
            if (dim == 2)
                (*pArrBoundaries)[idx] = new TBoundary;
            else if (dim == 3)
                (*pArrBoundaries)[idx] = new TBoundary;
            else
                yyerror( MSG_WRONG_DIM );

            (*pArrBoundaries)[idx]->setName( *pszName );
        }
        ;

BOUNDDECL: STRUCTBOUND STRING
        {
            *pszName = $2;

            if (idxOfName( pArrBoundNames, *pszName ) != NO_IDX)
                yyerror( MSG_HAVE_NAME );
        }
        ;

DOMAIN: DOMAINDECL '{' DOMAINBODY '}'
        {
            if (hasDomain)
                yyerror( MSG_HAVE_DOMAIN );
            
            if (dim == 2)
            {
                yyerror( MSG_NOT_IMPLEMENTED );
            }// if

            if (dim == 3)
            {
                uint         i;
                uint         noOfBound = pArrNames->getSize();
                TBoundary  * pBoundary;

                if ((*pszType) == "cube" )
                    pGlobalDomain = new T3DCubeDomain;
                else
                    yyerror( MSG_WRONG_TYPE );
                
                for (i = 0; i < noOfBound; i++)
                {
                    pBoundary = (*pArrBoundaries)[i];

                    pGlobalDomain->addBoundary( pBoundary );
                }// for
            }// if

            hasDomain = TRUE;

            //
            // reset all data
            //
            resetData();
        }
        ;

DOMAINDECL: STRUCTDOMAIN
        {
            if (hasDomain)
                yyerror( MSG_HAVE_DOMAIN );
        }
        ;

DOMAINBODY:
        | DOMAINBODY BOUNDLIST
        | DOMAINBODY TYPE
        {
            if (hasType)
                yywarn( "type already defined" );
            
            if (dim == 2)
                yyerror( MSG_NOT_IMPLEMENTED );
            else if (dim == 3)
            {
                if (*pszType != "cube")
                    yyerror( MSG_WRONG_TYPE );
            }// else
            else
                yyerror( MSG_WRONG_DIM );
            
            hasType = TRUE;
        }
        ;
%%

//
// call parser
//
int 
parseFile( FILE              * pFile,
           TList< TCell<2> > * pList2D,
           TList< TCell<3> > * pList3D )
{
    nLine = 1;
    dim   = 1;
    
    pList2DCells = pList2D;
    pList3DCells = pList3D;

    init();
    
    yyin = pFile;
    yyparse();

    cleanup();

    return dim;
}

//
// init all variables
//
void
init ()
{
    hasBoundName = FALSE;
    isInterior   = FALSE;
    hasCoord     = FALSE;
    hasDomain    = FALSE;

    pArrNodeNames     = new TDynArray<TString>;
    pArrEdgeNames     = new TDynArray<TString>;
    pArrFrontierNames = new TDynArray<TString>;
    pArrCellNames     = new TDynArray<TString>;
    pArrBoundNames    = new TDynArray<TString>;
    pArrNames         = new TDynArray<TString>;

    pArrNodes      = new TDynArray<void*>;
    pArrEdges      = new TDynArray<T3DCellEdge*>;
    pArrFrontiers  = new TDynArray<void*>;
    pArrBoundaries = new TDynArray<TBoundary*>;

    pVec2D = new TVector<2>;
    pVec3D = new TVector<3>;

    pGlobalDomain = NULL;
    
    pszName      = new TString;
    pszBoundName = new TString;
    pszType      = new TString;
    pszRefine    = new TString;
}

//
// clean up after parsing
//
void
cleanup ()
{
    delete pArrNodeNames;
    delete pArrEdgeNames;
    delete pArrFrontierNames;
    delete pArrCellNames;
    delete pArrBoundNames;
    delete pArrNames;

    delete pArrNodes;
    delete pArrEdges;
    delete pArrFrontiers;
    delete pArrBoundaries;

    delete pVec2D;
    delete pVec3D;

    delete pszName;
    delete pszBoundName;
    delete pszType;
    delete pszRefine;
}

//
// reset all data
//
void
resetData ()
{
    hasBoundName = FALSE;
    isInterior   = FALSE;
    hasCoord     = FALSE;
    hasType      = FALSE;
    hasRefine    = FALSE;

    *pszName      = "";
    *pszBoundName = "";
    *pszType      = "";
    *pszRefine    = "";
    
    pArrNames->setSize( 0 );
}

//
// look in name-pArray and return idx of name
// (if name is not in pArray, return NO_IDX)
//
int
idxOfName ( const TDynArray<TString> * pArray, const TString & name )
{
    uint  i;

    if (pArray == NULL)
        return NO_IDX;
    
    for (i = 0; i < pArray->getSize(); i++)
    {
        if (name == (*pArray)[i])
            return i;
    }// for

    return NO_IDX;
}

//
// Errorhandling
//
void 
yyerror ( const char *s )
{
    cerr << "TParser::buildList () [error] : "
         << s
         << " (line = " << nLine << ", row = " << nPos << ')' << endl;

    exit( -1 );
}

//
// warnings
//
void 
yywarn ( const char *s )
{
    cout << "TParser::buildList () [warning] : "
         << s
         << " (line = " << nLine << ", row = " << nPos << ')' << endl;
}
