//
//  Project  : libnet
//  File     : regular.cc
//  Author   : Sabine Gutsch / Ronald Kriemann
//  Purpose  : class of cell which implements a regular triangulation
//
// arch-tag: d0dac967-62ac-40ff-85ec-765876e8903a
//

#include "regular.hh"

//
// constructors
//
T2DRegularCell::T2DRegularCell ( TCell<2> *parent, 
                                 T2DFrontier *frontier0, 
                                 T2DFrontier *frontier1, 
                                 T2DFrontier *frontier2 ) 
    : TCell<2>( parent ) 
{
    initFrontierPtr( 3 );
    
    arrFrontiers[0] = frontier0;
    arrFrontiers[1] = frontier1;
    arrFrontiers[2] = frontier2;

    frontier0->addReference();
    frontier1->addReference();
    frontier2->addReference();

    frontier0->setCell( this );
    frontier1->setCell( this );
    frontier2->setCell( this );
    
    fSize = calcSize();
}

T2DRegularCell::T2DRegularCell () : TCell<2>(NULL) 
{}

//
// divide all marked frontiers before real refinement
//
void 
T2DRegularCell::prepareRefine () 
{
    uint  i;

    for (i = 0; i < noOfFrontiers; i++) 
    {
        if (!arrFrontiers[i]->hasMidNode()) 
            arrFrontiers[i]->buildMidNode();
    }// for
}

//
// refine yourself (build only the children)
//
void 
T2DRegularCell::refine ()
{
    // create children of all frontiers
    if (!arrFrontiers[0]->hasChild()) arrFrontiers[0]->buildChild();
    if (!arrFrontiers[1]->hasChild()) arrFrontiers[1]->buildChild();
    if (!arrFrontiers[2]->hasChild()) arrFrontiers[2]->buildChild();
    
    // update field noOfChildren
    initChildPtr( 4 );
    
    // three divided frontiers -> create four children
    // - the frontiers of the cells are numbered with 0, 1, 2
    // - children frontiers are numbered 00, 01, 10, 11, 20, 21
    // - mid0 is the frontier from midnode of frontier 0 to 
    //   midnode of frontier 1
    // - mid1 is the frontier from midnode of frontier 1 to 
    //   midnode of frontier 2
    // - mid2 is the frontier from midnode of frontier 2 to 
    //   midnode of frontier 0
    
    T2DFrontier *p00  = (T2DFrontier*)arrFrontiers[0]->getChild(this,0);
    T2DFrontier *p01  = (T2DFrontier*)arrFrontiers[0]->getChild(this,1);
    T2DFrontier *p10  = (T2DFrontier*)arrFrontiers[1]->getChild(this,0);
    T2DFrontier *p11  = (T2DFrontier*)arrFrontiers[1]->getChild(this,1);
    T2DFrontier *p20  = (T2DFrontier*)arrFrontiers[2]->getChild(this,0);
    T2DFrontier *p21  = (T2DFrontier*)arrFrontiers[2]->getChild(this,1);
    T2DFrontier *pMid0= new T2DFrontier(NULL,
                                        arrFrontiers[1]->getNode(this,1),
                                        arrFrontiers[0]->getNode(this,1) );
    pMid0->delReference();
    T2DFrontier *pMid1= new T2DFrontier(NULL,
                                        arrFrontiers[2]->getNode(this,1),
                                        arrFrontiers[1]->getNode(this,1) );
    pMid1->delReference();
    T2DFrontier *pMid2= new T2DFrontier(NULL,
                                        arrFrontiers[0]->getNode(this,1),
                                        arrFrontiers[2]->getNode(this,1) );
    pMid2->delReference();
    
    // Note: direction of frontiers and 
    arrChildren[0] = new T2DRegularCell( this, p00, pMid2, p21 );
    arrChildren[1] = new T2DRegularCell( this, p01, p10, pMid0 );
    arrChildren[2] = new T2DRegularCell( this, p11, p20, pMid1 );
    arrChildren[3] = new T2DRegularCell( this, pMid0, pMid1, pMid2 );
}

//
// and the opposite : coarse
//
void
T2DRegularCell::coarsen () 
{}

//
// output for easier debugging
//
ostream& operator<< (ostream& os, const T2DRegularCell& cell) 
{
    os << cell.getFrontier(0)->getNode(&cell,0)->getName() << " "
       << *(cell.getFrontier(0)->getNode(&cell,0))
       << " --> " 
       << cell.getFrontier(0)->getNode(&cell,2)->getName() << " "
       << *(cell.getFrontier(0)->getNode(&cell,2)) << endl;
    os << cell.getFrontier(1)->getNode(&cell,0)->getName() << " "
       << *(cell.getFrontier(1)->getNode(&cell,0))
       << " --> " 
       << cell.getFrontier(1)->getNode(&cell,2)->getName() << " "
       << *(cell.getFrontier(1)->getNode(&cell,2)) << endl;
    os << cell.getFrontier(2)->getNode(&cell,0)->getName() << " "
       << *(cell.getFrontier(2)->getNode(&cell,0))
       << " --> " 
       << cell.getFrontier(2)->getNode(&cell,2)->getName() << " "
       << *(cell.getFrontier(2)->getNode(&cell,2)) << endl;
    
    return (os);
}
