//
//  Project : libnet
//  File    : vertex.cc
//  Author  : Sabine Gutsch / Ronald Kriemann 
//  Purpose : classes for 2d and 3d vertices
//
// arch-tag: 483c71aa-1788-4835-840f-d2d97b95794a
//

#include "vertex.hh"

// definition of the static members
static uint maxInteriorName = 0;
static uint minBoundaryName = ULONG_MAX;

//
// constructors
//
TVertex<2>::TVertex ( TFloat x, TFloat y, uint vtype ) 
    : TVector<2>( x, y ) 
{
    if ( vtype == 0 ) 
        name = maxInteriorName++; 
    else 
        name = minBoundaryName--;

    pDoFDesc = NULL;
}

TVertex<3>::TVertex ( TFloat x, TFloat y, TFloat z, TBoundary *boundary ) 
    : TVector<3>( x, y, z ) 
{
    pBoundary = boundary;

    if ( pBoundary == INTERIOR ) 
        name = maxInteriorName++; 
    else 
        name = minBoundaryName--;

    pDoFDesc = NULL;
}

TVertex<2>::TVertex ( const TVector<2>& vector, uint vtype ) 
        : TVector<2>( vector ) 
{
    if ( vtype == 0 ) 
        name = maxInteriorName++; 
    else 
        name = minBoundaryName--;

    pDoFDesc = NULL;
}

TVertex<3>::TVertex ( const TVector<3> &vector, TBoundary *boundary )
    : TVector<3>( vector ) 
{
    pBoundary = boundary;

    if ( pBoundary == INTERIOR ) 
        name = maxInteriorName++; 
    else 
        name = minBoundaryName--;

    pDoFDesc = NULL;
}

//
// destructor
//
template <int N>
TVertex<N>::~TVertex ()
{
    if (pDoFDesc != NULL)
        delete pDoFDesc;
}

//
// get names
//
template <int N>
const uint 
TVertex<N>::getName () const
{ 
    return name; 
}

template <int N>
uint 
TVertex<N>::getMaxInteriorName ()
{ 
    return maxInteriorName; 
}

template <int N>
uint 
TVertex<N>::getMinBoundaryName ()
{ 
    return ULONG_MAX - minBoundaryName; 
}

template <int N>
void 
TVertex<N>::resetNames () 
{
    maxInteriorName = 0;
    minBoundaryName = ULONG_MAX;
}

//
// access boundary
//
template <int N>
TBoundary *
TVertex<N>::getBoundary () const
{
    return pBoundary;
}
    
//
// access DoFDesc
//
template <int N>
TDoFDesc *
TVertex<N>::getDoFDesc () const
{
    return pDoFDesc;
}

template <int N>
void
TVertex<N>::initDoFDesc ( int m )
{
    if (pDoFDesc == NULL)
        pDoFDesc = new TDoFDesc( m );
    else
        cerr << "TVertex::initDoFDesc (int) : already have DoFDesc" << endl;
}

//
// template instanciation
//
template class TVertex<2>;
template class TVertex<3>;
