//
//  Project   : libnet
//  File      : TGMRES.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for solver implementing GMRES
//
// arch-tag: 6a18666a-f20f-4daa-9927-6bafe3d3d660
//

extern "C" {
#include <math.h>
}

#include "TGMRES.hh"

//
// constructor and destructor
//
TGMRES::TGMRES( TSolver * prec,
                uint      amaxiter, 
                TFloat    aaccur, 
                uint      averb )
        : TPrecSolver ( prec, NULL, NULL, NULL, amaxiter, aaccur, averb) 
{
    setName( "GMRES" );

    _m = 10;
}

TGMRES::~TGMRES ()
{
}

//
// iteration
//
    
void
TGMRES::init ()
{
}

int
TGMRES::step ()
{
    // do nothing because all is done in solve
    return 0;
}

void
TGMRES::solve ()
{
    TBlockMatrix           * A[1];
    TBlockVector< TFloat > * RHS[1];
    TBlockVector< TFloat > * x[1];

    // simulate multigrid
    A[0]   = pA;
    RHS[0] = pRHS;
    x[0]   = pX;

    solve( 0, A, RHS, x );
}

void
TGMRES::solve ( int                     level,
                TBlockMatrix         ** A,
                TBlockVector<TFloat> ** RHS,
                TBlockVector<TFloat> ** x )
{
    uint                     it;
    int                      j, i, l, dim;
    TFloat                   h, norm, c, s, giv_r, giv_h;
    TFloat                   norm_0, norm_old;
    TFloat                   y[_m], e1[_m+1];
    TBlockVector< TFloat >   d, v_new, p, r;
    TFloat                   vec_size;

    TBlockMatrix           * A_l = A[level];
    TBlockVector< TFloat > * b_l = RHS[level];
    TBlockVector< TFloat > * x_l = x[level];

    TFloat                   aH[_m+1][_m];
    TFloat                   aC[_m+1], aS[_m+1];

    //
    // check given values
    //

    if ((A_l == NULL) || (b_l == NULL) || (x_l == NULL))
    {
        cout << "TGMRES::solve : got NULL arguments" << endl;
        return;
    }// if
    
    //
    // initialize variables
    //

    _v.setSize( _m + 1 );

    x_l->clone( d );
    x_l->clone( p );
    
    // get size of vectors to adjust norms (weightedEuclidNorm)
    vec_size = sqrt( TFloat( x_l->getSize() ) );

    pPrecond->setA( A_l );
    pPrecond->setX( x_l ); // necessary for Schur preconditioner

    //
    // begin of GMRES-Iteration
    //
    
    if (verbosity & PRINT_SOLVERNAME)
        cout << "TGMRES::solve : running (" << x_l->getSize() << " unknowns)" << endl;

    // r = b - Ax
    r  = * b_l;
    r -= (*A_l) * (*x_l);
    
    // v_0 = r / || r ||
    norm_0 = norm = r.euclidNorm();
    
    //
    // iteration itself
    //
    
    it = 0;
    
    while ( TRUE )
    {
        //
        // preprocess, set vecsizes etc.
        //
        
        // v_0 = r / || r ||
        if (norm != 0)
            r *= TFloat(1) / norm;
        else
        {
            cout << "TGMRES::step : residual norm = 0" << endl;
            break;
        }// if
        
        _v[0] = r;
        
        // e1 = || r || (1,0,0,...)^T
        for (i = 1; i <= _m; i++)
            e1[i] = 0;
        e1[0] = norm;
        
        norm_old = norm;

        //
        // iterate in krylov-subspace
        //
        for ( j = 0; j < _m; j++ )
        {
            // preconditioning
            if (isPrecSecondNF())
                p = (*A_l) * ((TSecondNF*)pPrecond)->WInvers( _v[j] );
            else
            {
                p = 0;
                pPrecond->step( level, A, & _v[j], & p );
                p = (*A_l) * p;
            }// else

            v_new = p;
            
            // adjust H and calc v'
            for ( i = 0; i <= j; i++ )
            {
                h = p * _v[i];
                aH[i][j] = h;
                
                v_new.daxpy( _v[i], -h );
            }// for
            
            norm     = v_new.euclidNorm();
            _v[j+1]  = v_new;
            _v[j+1] *= TFloat(1) / norm;
            aH[j+1][j] = norm;
            
            //
            // apply old givens rotations to new column
            //
            for ( i = 0; i < j; i++ )
            {
                giv_r = aH[i][j];
                giv_h = aH[i+1][j];
                
                aH[i][j]   = aS[i]*giv_r + aC[i]*giv_h;
                aH[i+1][j] = aC[i]*giv_r - aS[i]*giv_h;
            }// for
            
            //
            // do givens rotation for new element
            //
            
            giv_r = aH[j][j];
            giv_h = aH[j+1][j];
            norm  = sqrt( giv_r*giv_r + giv_h*giv_h );
            
            if (giv_r > TFloat(0))
            {
                c = giv_h / norm;
                s = giv_r / norm;
                aH[j][j] = norm;
            }// if
            else
            {
                c = - giv_h / norm;
                s = - giv_r / norm;
                aH[j][j] = - norm;
            }// else
            
            // aH[j+1][j] = 0;
            
            // apply givens rotation with RHS (new pos is 0 in e1)
            giv_r   = e1[j];
            e1[j]   = s*giv_r;
            e1[j+1] = c*giv_r;
            
            // save coeff. for future rotations
            aC[j] = c;
            aS[j] = s;

            norm = fabs( e1[j+1] );
                
            it++;
            
            //
            // print statistics
            //
            
            printStep( it, norm_old / vec_size, norm / vec_size );
            
            norm_old = norm;
            
            //
            // test stop criteria
            //

            if (testStop( it, norm_0 / vec_size, norm / vec_size ))
            {
                j++;
                break;
            }// if
        }// for

        // set reached dimension in krylov-subspace
        dim = j;
        
        //
        // solve min_y || beta e^1 - H y ||
        //
        
        for ( i = dim-1; i >= 0; i-- )
        {
            s = e1[i];
            
            for ( l = i+1; l < dim; l++ )
                s -= aH[i][l] * y[l];
            
            y[i] = s / aH[i][i];
        }// for
        
        //
        // adjust x
        //

        d = 0;
        for ( i = 0; i < dim; i++ )
            d.daxpy( _v[i], y[i] );
        
        if (isPrecSecondNF())
            *x_l += ((TSecondNF*)pPrecond)->WInvers( d );
        else
        {
            p = 0;
            pPrecond->step( level, A, & d, & p );
            *x_l += p;
        }// else

        //
        // recalc residual
        //
        
        // r = b - Ax
        r  = * b_l;
        r -= (*A_l) * (*x_l);

        //
        // test stop criteria
        //

        if (testStop( it, norm_0 / vec_size, norm / vec_size ))
            break;
    }// while

    //
    // print statistics
    //

    printResult( it, norm_0 / vec_size, norm / vec_size );
}

//
// parameter settings
//

void
TGMRES::setRestart ( int m )
{
    _m = m;
}
