//
//  Project   : libnet
//  File      : TSOR.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for a SOR-based solver in second normal form
//
// arch-tag: fba3aedf-e689-4ab3-a3e2-977962496b0e
//

extern "C" {
#include <math.h>
}

#include "TSOR.hh"

////////////////////////////////////
//
// constructor
//
TSOR::TSOR ( TBlockMatrix         * pA, 
             TBlockVector<TFloat> * pRHS, 
             TBlockVector<TFloat> * pX,  
             uint                   maxiter, 
             TFloat                 accur, 
             uint                   verb,
             TFloat                 damp )
        : TSecondNF( pA, pRHS, pX, maxiter, accur, verb, damp )
{
    setName( "FSOR" );

    _omega = 1.0;
    method = FORWARD_SOR;

    buildOmega  = FALSE;
    bInit       = TRUE;
    norm_old    = 0;
    vecX_old    = 0;
    update_step = 0;
}

////////////////////////////////////
//
// methods for solving
//

//
// implement W^-1 * v
//
TBlockVector<TFloat>
TSOR::WInvers ( const TBlockVector<TFloat> & v )
{
    TBlockVector<TFloat>  tmpVec;
	int     i, j, n, size;
    TFloat  omega_inv = 1.0 / omega();

    tmpVec = v;

    // dim of matrix (we assume rows == columns)
    n = pA->getNoOfRows();

    //
    // calc W^-1 * v for each block. All blocks below
    // diag are handled via block*vector
    //
    if (method == FORWARD_SOR)
    {
        for ( i = 0; i < n; i++ )
        {
            for ( j = 0; j < i; j++ )
            {
                if ((*pA)[i][j].getNoOfRows() > 0)
                    tmpVec[i] -= (*pA)[i][j] * tmpVec[j];
            }// for
            
            tmpVec[i] = (*pA)[i][i].invLowerMult( tmpVec[i], omega_inv );
        }// for
    }// else if
    else if (method == BACKWARD_SOR)
    {
        for ( i = n; i > 0; i-- )
        {
            for ( j = n; j > i; j-- )
            {
                if ((*pA)[i-1][j-1].getNoOfRows() > 0)
                    tmpVec[i-1] -= (*pA)[i-1][j-1] * tmpVec[j-1];
            }// for
            
            tmpVec[i-1] = (*pA)[i-1][i-1].invUpperMult( tmpVec[i-1], omega_inv );
        }// for
    }// else if
    else if (method == SYMMETRIC_SOR)
    {
        // forward sor
        for ( i = 0; i < n; i++ )
        {
            for ( j = 0; j < i; j++ )
            {
                if ((*pA)[i][j].getNoOfRows() > 0)
                    tmpVec[i] -= (*pA)[i][j] * tmpVec[j];
            }// for
            
            tmpVec[i] = (*pA)[i][i].invLowerMult( tmpVec[i], omega_inv );
        }// for
        
        // mult mith inverse diag
        for (i = 0; i < n; i++)
        {
            // for all elements in a vectorblock
            size = (*pA)[i][i].getNoOfRows();
            
            for ( j = 0; j < size; j++ )
                tmpVec[i][j] *= (*pA)[i][i][j].getValue( 0 );
        }// for
        
        // apply omega
        tmpVec *= (2.0 - omega()) / omega();

        // backward SOR
        for ( i = n; i > 0; i-- )
        {
            for ( j = n; j > i; j-- )
            {
                if ((*pA)[i-1][j-1].getNoOfRows() > 0)
                    tmpVec[i-1] -= (*pA)[i-1][j-1] * tmpVec[j-1];
            }// for
            
            tmpVec[i-1] = (*pA)[i-1][i-1].invUpperMult( tmpVec[i-1], omega_inv );
        }// for
    }// else if
    else
        cout << "TSOR::WInvers (...) : wrong method" << endl;

    return tmpVec;
}

//
// access SOR-damp factor
//

const TFloat &
TSOR::omega () const
{
    return _omega;
}

void
TSOR::setOmega ( const TFloat & o )
{
    _omega = o;
}

//
// set SOR method
//

void
TSOR::setMethod ( const uint m )
{
    method = m;

    switch (method)
    {
        case FORWARD_SOR   :  setName( "FSOR" ); break;
        case BACKWARD_SOR  :  setName( "BSOR" ); break;
        case SYMMETRIC_SOR :  setName( "SSOR" ); break;
    }// switch
}

//
// update (here : choice of optimal omega)
//
void
TSOR::update ( const TBlockVector< TFloat > & vecX )
{
    if (buildOmega)
    {
        TFloat                 norm_new, lambda, beta;
        TBlockVector< TFloat > tmp;

        norm_new = lambda = 0;

        if ( ! bInit )
        {
            tmp  = vecX;
            tmp -= vecX_old;

            norm_new = vecX.euclidNorm();
            
            if (norm_old != 0)
            {
                lambda = norm_new / norm_old;
                beta   = lambda + omega() - 1.0;
                if (beta < 0)
                    beta *= -1;
                beta  /= omega();
                beta  /= sqrt(lambda);
                cout << "TSOR::update (...) : lambda = " << lambda
                     << "  beta = " << beta
                     << "  omega = " << 2.0 / (1.0 + (sqrt( 1.0 - (beta*beta))))
                     << endl;
            }// if
        }// if
        else
            bInit = FALSE;

        norm_old = norm_new;
        vecX_old = vecX;
    }// if
}

//
// several spezializations of SOR
//
TFSOR::TFSOR ( TBlockMatrix         * pA, 
               TBlockVector<TFloat> * pRHS, 
               TBlockVector<TFloat> * pX,  
               uint                   maxiter, 
               TFloat                 accur, 
               uint                   verb,
               TFloat                 damp )
        : TSOR( pA, pRHS, pX, maxiter, accur, verb, damp )
{
    setMethod( FORWARD_SOR );
}

TBSOR::TBSOR ( TBlockMatrix         * pA, 
               TBlockVector<TFloat> * pRHS, 
               TBlockVector<TFloat> * pX,  
               uint                   maxiter, 
               TFloat                 accur, 
               uint                   verb,
               TFloat                 damp )
        : TSOR( pA, pRHS, pX, maxiter, accur, verb, damp )
{
    setMethod( BACKWARD_SOR );
}

TSSOR::TSSOR ( TBlockMatrix         * pA, 
               TBlockVector<TFloat> * pRHS, 
               TBlockVector<TFloat> * pX,  
               uint                   maxiter, 
               TFloat                 accur, 
               uint                   verb,
               TFloat                 damp )
        : TSOR( pA, pRHS, pX, maxiter, accur, verb, damp )
{
    setMethod( SYMMETRIC_SOR );
}
