//
//  Project  : libnet
//  File     : bicgStab.cc
//  Author   : Sabine Gutsch / Thomas Probst
//  Purpose  : bicg-STAB solver
//
// arch-tag: a9ed3389-d3c5-4c20-8950-d21f88e13819
//

extern "C" {
#include <math.h>
}

#include "TTerminal.hh"

#include "bicgStab.hh"

//
// constructor
//
TBicgStab::TBicgStab( TSolver  * prec,
                      uint       amaxiter, 
                      TFloat     aaccur, 
                      uint       averb ) 
        : TPrecSolver ( prec, NULL, NULL, NULL, amaxiter, aaccur, averb) 
{
    setName( "BiCG-STAB" ); 
}

//
// iteration
//

void 
TBicgStab::init ()
{
}

int
TBicgStab::step ()
{
    // do nothing because all is done in solve
    return 0;
}

void
TBicgStab::solve ()
{
    TBlockMatrix           * A[1];
    TBlockVector< TFloat > * RHS[1];
    TBlockVector< TFloat > * x[1];

    // simulate multigrid
    A[0]   = pA;
    RHS[0] = pRHS;
    x[0]   = pX;

    solve( 0, A, RHS, x );
}

void 
TBicgStab::solve ( int                     level,
                   TBlockMatrix         ** A,
                   TBlockVector<TFloat> ** RHS,
                   TBlockVector<TFloat> ** x )
{
#define EPSILON  1e-20

    uint                 it;
    TFloat               rho, h, rhoNew, alpha, beta, omega;
    TBlockVector<TFloat> r, r0, p, y, v, t;
    TFloat               norm, norm_old, norm_0;
    uint                 nob;
    
    TBlockMatrix           * A_l = A[level];
    TBlockVector< TFloat > * b_l = RHS[level];
    TBlockVector< TFloat > * x_l = x[level];

    //
    // check given values
    //

    if ((A_l == NULL) || (b_l == NULL) || (x_l == NULL))
    {
        cout << "TBiCGStab::solve : got NULL arguments" << endl;
        return;
    }// if
    
    //
    // initialize variables
    //

    pPrecond->setA( A_l );
    pPrecond->setX( x_l ); // necessary for Schur preconditioner
    
    x_l->clone( p );
    x_l->clone( v );
    x_l->clone( y );

    pPrecond->setA( A_l );

    //
    // begin iteration
    //
    
    if (verbosity & PRINT_SOLVERNAME)
        cout << "TBiCGStab::solve () : running (" << x_l->getSize() << " unknowns)" << endl;

    // r = b - Ax
    r  = (*b_l);
    r -= (*A_l) * (*x_l); 

    norm = norm_old = norm_0 = r.weightedEuclidNorm();
    
    r0 = r;

    p = 0;
    v = 0;

    rho = alpha = omega = 1;

    //
    // iteration
    //
    
    it = 0;
    
    while ( TRUE )
    {
        //
        // preprocess, set vecsizes etc.
        //
        
        rhoNew = r0 * r;

        // look if breakdown occured
        if ((fabs(rho) < EPSILON) || (fabs(omega) < EPSILON))
        {
            cout << "TBicgStab::solve : breakdown occured (rho = "
                 << rho << " , omega = " << omega << ")" << endl;
            break;
        }// if
        
        beta = ( rhoNew / rho ) * ( alpha / omega );
        
        // p = r + beta (p - omega*v)
        p.daxpy( v, -omega );
        p *= beta;
        p.daxpy( r, 1.0 );

        // preconditioning
        if (isPrecSecondNF())
            y = ((TSecondNF*)pPrecond)->WInvers( p );
        else
        {
            y = 0;
            pPrecond->step( level, A, & p, & y );
        }// else
        
        v = (*A_l) * y;
        h = ( r0 * v );
        
        // look if breakdown occured
        if (fabs(h) < EPSILON)
        {
            cout << "TBicgStab::solve : breakdown occured (h = " << h << ")" << endl;
            break;
        }// if
        
        alpha  = rhoNew / h;
        
        // apply first correction to x
        // x <- x + alpha y
        x_l->daxpy( y, alpha );

        it++;
        
        // s = r - alpha*v ( r is no longer used, so recycle it )
        r.daxpy( v, -alpha );
        
        // r is already our residuum
        norm = r.weightedEuclidNorm();

        //
        // print statistics
        //

        printStep( it, norm_old, norm );
            
        norm_old = norm;

        //
        // test stop criteria
        //

        if (testStop( it, norm_0, norm ))
            break;

        // preconditioning
        if (isPrecSecondNF())
            y = ((TSecondNF*)pPrecond)->WInvers( r );
        else
        {
            y = 0;
            pPrecond->step( level, A, & r, & y );
        }// else
        
        t      = (*A_l) * y;
        omega  = ( t * r ) / ( t * t );
        
        // apply second correction to x
        // x <- x + omega y
        x_l->daxpy( y, omega );
        
        it++;
        
        // pressure correction
        nob = r.getNoOfBlocks();
        if (nob > 1) 
        {
            TVectorBlock<TFloat> vecOne( r[nob-1] );
            vecOne.set( 1.0 );
            TFloat factor = (vecOne * (*x_l)[nob-1]) / (vecOne * vecOne);
            vecOne.set( factor );
            (*x_l)[nob-1] -= vecOne;
        }// if
        
        // r = s - omega*t (now reset r)
        r.daxpy( t, -omega );
        
        rho = rhoNew;

        //
        // calc residual norm
        //
        
        // r is already our residuum
        norm = r.weightedEuclidNorm();

        //
        // print statistics
        //

        printStep( it, norm_old, norm );

        //
        // test stop criteria
        //

        if (testStop( it, norm_0, norm ))
            break;
    }// while

    //
    // print statistics
    //

    printResult( it, norm_0, norm );
}
