//
//  Project  : libnet
//  File     : nonlinearSolver.cc
//  Author   : Sabine Gutsch 
//  Purpose  : basis class for nonlinear solver
//
// arch-tag: 231d9020-13cd-4022-ae94-3c96c15e0928
//

#include "T2DPlanarGraph.hh"
#include "criterion.hh"

#include "nonlinearSolver.hh"

//
// constructor
//
TNonLinearSolver::TNonLinearSolver( TBasisEqn<2> *aEqn, TBasisFE<2> *aFE, 
									T2DNetHierarchy *aNetHier, 
									T2DStokesHierarchy *aStokesHier ) 
{
	eqn         = aEqn;
	fe          = aFE; 
	netHier     = aNetHier;
	stokesHier  = aStokesHier;
	accuracy    = 1e-8;
	maxit       = 2;
	noOfVectors = aNetHier->getNoOfNets()-1;
	vecPerm     = new (TVectorBlock<uint>*)[noOfVectors];
	vecInvPerm  = new (TVectorBlock<uint>*)[noOfVectors];
    cyclicBlocks = new (TVectorBlock<uint>*)[noOfVectors];

	for (uint i = 0; i < noOfVectors; i++)
	{
		vecPerm[i] = NULL;
		vecInvPerm[i] = NULL;
        cyclicBlocks[i] = NULL;
	}
	arrVectors = new TDynArray< TVector<2> * >;
	T2DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
        
	T2DNet *net = iterNet.last();
	net = iterNet.prev();
	fe->getCoord(net, *arrVectors ); 

	ordering = NONE;
}

//
// destructor
//
TNonLinearSolver::~TNonLinearSolver ( ) 
{
	if (arrVectors) {
        delete (arrVectors); 
        arrVectors = NULL;
	}

	for (uint i = 0; i < noOfVectors; i++)
	{
		if (vecPerm[i]) delete (vecPerm[i]);
		vecPerm[i] = NULL;
		if (vecInvPerm[i]) delete (vecInvPerm[i]);
		vecInvPerm[i] = NULL;
		if (cyclicBlocks[i]) delete (cyclicBlocks[i]);
		cyclicBlocks[i] = NULL;
	}
}

//
// get and set name
//
char*
TNonLinearSolver::getName () { return (name); }

void 
TNonLinearSolver::setName ( const char *n )  
{ 
	name = new char[strlen(n)+1]; 
	strcpy(name, n);
}

// 
// update hierarchy (first re-establish original ordering, then
// update (convective part of) hierarchy, determine new ordering 
// and reorder)
//
void
TNonLinearSolver::updateHierarchy( TMultigrid *mgSolver )
{
	cout << "update hierarchy\n";
	uint i;
	uint low  = mgSolver->getLowLevel();
	uint high = mgSolver->getHighLevel();

	// regain original ordering
	if ( vecPerm[low] != NULL ) 
		for ( i = low; i <= high; i++ ) 
		{
			mgSolver->getRHSHierarchy()[i]->permute( vecInvPerm[i] );
			mgSolver->getXHierarchy()[i]->permute( vecInvPerm[i] );
			mgSolver->getMatrixHierarchy()[i]->permute(vecInvPerm[i]);
			if ( i > low ) {
				(mgSolver->getProlongation()[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1] );
				(mgSolver->getRestriction()[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1] );
			}
		}

	// update pX on coarser levels
	mgSolver->updateXHierarchy(); 

	// update convective part (and rhs)
	stokesHier->updateHierarchy( eqn, fe, netHier, 
								 mgSolver->getMatrixHierarchy(), 
								 mgSolver->getRHSHierarchy(), 
								 mgSolver->getXHierarchy() );

	// compute new ordering
	TBlockInfo blockinfo;
    blockinfo.setSortMethod( SORT_OFD );

    TVectorBlock<uint>  *arrFVSSize = new TVectorBlock<uint>();

	if ( (ordering != NONE) && (vecInvPerm[high] == NULL) ) 
	    for ( i = 0; i <= high; i++ ) { 
            vecInvPerm[i] = new TVectorBlock<uint>;
			vecPerm[i] = new TVectorBlock<uint>;
            cyclicBlocks[i] = new TVectorBlock<uint>;
		}// for
       
	if ( ordering & CONCENTRIC ) 
	{
	    blockinfo.setCycleName("psfiles/cycle.ps"); 
	    blockinfo.setGraphName("psfiles/graph.ps");
		blockinfo.buildCYCIndepPerm( vecInvPerm, cyclicBlocks, 
								*fe, netHier->getList(), 
								mgSolver->getMatrixHierarchy(),  
								1.0/3.0, E3 );
        for (i = 0; i <= high; i++)
            cout << " Number of cycles in net " << i+1 << " :  " 
                 << cyclicBlocks[i]->getSize()/2 << endl;
	}// if

	if ( ordering == PLANAR_FVS ) 
	{
		blockinfo.buildPlanarFVSIndepPerm( vecInvPerm, arrFVSSize,
                                           *fe, netHier->getList(),
								     mgSolver->getMatrixHierarchy(),  
								     1.0/3.0, E3 );
        for (i = 0; i <= high; i++) 
            cout << " Number of FVS nodes in net " << i+1 << " :  " 
				 << (*arrFVSSize)[i] << endl;
	}// if

	if ( ordering == HEURISTIC_FVS ) 
	{
		blockinfo.buildHeuristFVSIndepPerm( vecInvPerm, arrFVSSize,
                                           *fe, netHier->getList(),
								     mgSolver->getMatrixHierarchy(),  
								     1.0/3.0, E3 );
        for (i = 0; i <= high; i++) 
            cout << " Number of FVS nodes in net " << i+1 << " :  " 
				 << (*arrFVSSize)[i] << endl;
	}// if

	// get inverse permutation
	if ( ordering != NONE )	{
		cout << "building inverse permutations" << endl;
	    for ( i = low; i <= high; i++ )  
			invertPerm( (*vecPerm[i]), *vecInvPerm[i] );
	}// if

	// draw current velocity
	T2DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
	//	T2DNet *net = iterNet.last();
	//	net = iterNet.prev();
	//	TDrawApp app;
	//	app.draw( *arrVectors, 
	//			  *(mgSolver->getXHierarchy()[mgSolver->getHighLevel()]), 
	//			  *net, "psfiles/vecfield.ps");

	// reorder
	if ( vecPerm[low] != NULL ) 
		for ( i = low; 
			  i <= high; i++ ) 
		{
			mgSolver->getRHSHierarchy()[i]->permute( vecPerm[i] );
			mgSolver->getXHierarchy()[i]->permute( vecPerm[i] );
			mgSolver->getMatrixHierarchy()[i]->permute(vecPerm[i]);
			if ( i > 1 ) {
				(mgSolver->getProlongation()[i-1]->pBlockProlong)[0]->permute( vecPerm[i], vecPerm[i-1]);
				(mgSolver->getRestriction()[i-1]->pBlockProlong)[0]->permute( vecPerm[i], vecPerm[i-1]);
			}
		}
	cout << "update hierarchy done\n";
}


//
// solve method
//
TBlockVector<TFloat> 
TNonLinearSolver::solve( TMultigrid *mgSolver )
{
	uint i = 0; 
	TFloat initialDefectNorm = (*(mgSolver->getA()) * *(mgSolver->getX()) - *(mgSolver->getRHS()) ).weightedEuclidNorm();
	TFloat newDefectNorm = initialDefectNorm;
	TFloat oldDefectNorm = initialDefectNorm;

	TBlockVector<TFloat> vecX( *(mgSolver->getX()) );

	while ( (i++ < maxit) && (accuracy < newDefectNorm) ) {

		if ( (vecX - *(mgSolver->getX())).weightedEuclidNorm(0) > 0.01
			 && (newDefectNorm < oldDefectNorm) )
		{
			cout <<  "change X " << (vecX - *(mgSolver->getX())).weightedEuclidNorm(0) << " " << oldDefectNorm << " " << newDefectNorm << endl; 
			vecX =  *(mgSolver->getX());
			updateHierarchy( mgSolver );
		}

		oldDefectNorm = newDefectNorm;
		newDefectNorm = step ( mgSolver );

	}

	cout << "\nAverage convergence rate of " << name << ": " 
		 << pow( (newDefectNorm / initialDefectNorm), 1.0 / (i-1) ) << endl;

	// regain original ordering
	if ( vecPerm[mgSolver->getLowLevel()] != NULL ) 
		for ( i = mgSolver->getLowLevel(); 
			  i <= mgSolver->getHighLevel(); i++ ) 
		{
			mgSolver->getRHSHierarchy()[i]->permute( vecInvPerm[i] );
			mgSolver->getXHierarchy()[i]->permute( vecInvPerm[i] );
			mgSolver->getMatrixHierarchy()[i]->permute(vecInvPerm[i]);
			if ( i > 1 )
				(mgSolver->getProlongation()[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1] );
		}
	return ( *(mgSolver->getX()) );
}
	
