//
//  Project   : libnet
//  File      : secondNF.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : (classical) linear solvers in second normal form
//
// arch-tag: f49a22ed-757f-441a-82e0-e58cbdeff9e4
//

#include "secondNF.hh"

// constructor
TSecondNF::TSecondNF (TBlockMatrix         * paA,
                      TBlockVector<TFloat> * paRHS, 
                      TBlockVector<TFloat> * paX,
                      uint                   amaxiter, 
                      TFloat                 aaccur, 
                      uint                   averb,
                      TFloat                 aDamp )
	: TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb)
{ 
	dampFactor = aDamp; 
}

//
// destructor
//
TSecondNF::~TSecondNF ()
{}

// one iteration step
int 
TSecondNF::step () 
{ 
	r  = (*pA) * (*pX);           // r  := Ax - rhs
	r -= (*pRHS);

	r  = WInvers ( r );           // r  := W^{-1} r
    if (dampFactor != 1.0)
        r *= dampFactor;

	(*pX) -= r;                   // x  := x - dampFactor * Wr

	// pressure correction
	uint nob = r.getNoOfBlocks();
    if (nob > 1) 
    {
	    TVectorBlock<TFloat> vecOne( r[nob-1] );
        TFloat               factor;
        
        vecOne = 1.0;
        
        factor = (vecOne * (*pX)[nob-1]) / (vecOne * vecOne);

        vecOne = factor;
        (*pX)[nob-1] -= vecOne;
	}// if

    // update internal variables
    // update( *pX );
    
	return RETURN_CONTINUE;
}

//
// set and get damping factor
//
void 
TSecondNF::setDampFactor ( const TFloat & aDamp )
{
    dampFactor = aDamp;
}

const TFloat &
TSecondNF::getDampFactor () const
{
    return dampFactor;
}

//
// update internal variables if nessecary
//
void
TSecondNF::update ( const TBlockVector< TFloat > & )
{
    // nothing to be done
}

//////////////////////////////////////////////////
//
// class for Jacobi iteration
//

TJacobi::TJacobi ( TBlockMatrix         * paA, 
				   TBlockVector<TFloat> * paRHS, 
				   TBlockVector<TFloat> * paX,  
				   uint                   amaxiter, 
				   TFloat                 aaccur, 
				   uint                   averb,
                   TFloat                 aDamp ) 
	: TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("Jac"); 
}

TBlockVector<TFloat> 
TJacobi::WInvers ( const TBlockVector<TFloat> &v )
{ 
	return pA->invDiagMult( v );
}

//////////////////////////////////////////////////
//
// class for Richardson iteration
//

TRichardson::TRichardson ( TBlockMatrix         * paA, 
                           TBlockVector<TFloat> * paRHS, 
                           TBlockVector<TFloat> * paX,  
                           uint                   amaxiter, 
                           TFloat                 aaccur, 
                           uint                   averb,
                           TFloat                 aDamp ) 
	: TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("Rich"); 
}

TBlockVector<TFloat> 
TRichardson::WInvers ( const TBlockVector<TFloat> &v )
{ 
	return v;
}


//////////////////////////////////////////////////
//
// class for forward Gauss-Seidel iteration
//

TForwardGaussSeidel::TForwardGaussSeidel ( TBlockMatrix         * paA, 
                                           TBlockVector<TFloat> * paRHS, 
                                           TBlockVector<TFloat> * paX,  
                                           uint                   amaxiter, 
                                           TFloat                 aaccur, 
                                           uint                   averb,
                                           TFloat                 aDamp ) 
	: TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("FGS"); 
}

TBlockVector<TFloat> 
TForwardGaussSeidel::WInvers ( const TBlockVector<TFloat> &v )
{ 
	return pA->invLowerMult( v );
}

//////////////////////////////////////////////////
//
// class for backward Gauss-Seidel iteration
//

TBackwardGaussSeidel::TBackwardGaussSeidel ( TBlockMatrix         * paA, 
                                             TBlockVector<TFloat> * paRHS, 
                                             TBlockVector<TFloat> * paX,  
                                             uint                   amaxiter, 
                                             TFloat                 aaccur, 
                                             uint                   averb,
                                             TFloat                 aDamp ) 
	: TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("BGS"); 
}

TBlockVector<TFloat> 
TBackwardGaussSeidel::WInvers ( const TBlockVector<TFloat> &v )
{ 
	return pA->invUpperMult( v );
}

//////////////////////////////////////////////////
//
// class for symmetric Gauss-Seidel iteration
//

TSymmetricGaussSeidel::TSymmetricGaussSeidel ( TBlockMatrix         * paA, 
                                               TBlockVector<TFloat> * paRHS, 
                                               TBlockVector<TFloat> * paX,  
                                               uint                   amaxiter, 
                                               TFloat                 aaccur, 
                                               uint                   averb,
                                               TFloat                 aDamp ) 
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{
	setName("SGS"); 
}

TBlockVector<TFloat> 
TSymmetricGaussSeidel::WInvers ( const TBlockVector<TFloat> &v )
{
    return pA->invUpperMult( pA->diagMult( pA->invLowerMult( v )));
}

//////////////////////////////////////////////////
//
// class for ILU iteration
//

TILU::TILU ( TBlockMatrix         * paA, 
		     TBlockVector<TFloat> * paRHS,
			 TBlockVector<TFloat> * paX,
             uint                   amaxiter, 
             TFloat                 aaccur, 
             uint                   averb,
             TFloat                 aDamp ) 
	: TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{
	setName("ILU"); 
}

TBlockVector<TFloat> 
TILU::WInvers ( const TBlockVector<TFloat> & v )
{ 
    if ( pA->hasLUdecomp() == FALSE) 
        pA->calculateILUdecomp();

	return pA->solveLUdecomp( v );
}
