//
//  Project   : libnet
//  File      : twogrid.cc
//  Author    : Sabine Gutsch
//  Purpose   : twogrid method
//  Date      : 12.06.1997
//  Changes   : 
//  Problems  : 
//
// arch-tag: 897db19d-3916-47cb-8eb6-c48ab5fe6268
//

#include "twogrid.hh"

/////////////////////////////////////////////////////////
//
// class containing all the twogrid parameters
//

//
// constructor
//
TtwogridInfo::TtwogridInfo ( uint alow, uint ahigh, uint aRecCall,
                             TSolver* apPreSmoother, TSolver* apPostSmoother,
                             TSolver* apCoarseSolver) 
{ 
    low  = alow;
    high = ahigh; 
    noOfLevel = high - low + 1;
    pPreSmoother = apPreSmoother;
    pPostSmoother = apPostSmoother;
    pCoarseSolver = apCoarseSolver;
    preSmoothingSteps = 1;
    postSmoothingSteps = 1;
    recursiveCalls = aRecCall;
    CGCdamping = 1.0;
}

//
// get lowest/highest level
//
uint 
TtwogridInfo::getHighLevel () { return (high); }

uint 
TtwogridInfo::getLowLevel () { return (low); }

uint 
TtwogridInfo::getNoOfLevel () { return (noOfLevel); }

//
// set smoother		
//
void 
TtwogridInfo::setSmoother ( TSolver* apSmoother ) 
{
    pPreSmoother = apSmoother;
    pPostSmoother = apSmoother;
}

// set coarse grid solver
void 
TtwogridInfo::setCoarseSolver ( TSolver* apCoarseSolver ) 
{
    pCoarseSolver = apCoarseSolver;
}

/////////////////////////////////////////////////////////
//
// class for a twogrid solver
// 

// 
// constructor
// 
Ttwogrid::Ttwogrid ( TBlockMatrix **apA, TBlockVector<TFloat> **apRHS, 
                     TBlockVector<TFloat> **apX, 
                     TProlongation **apRestrict, 
                     TProlongation **apProlong, 
                     uint alow, uint ahigh, uint aRecCall,  
                     TSolver* apPreSmoother, TSolver* apPostSmoother,
                     TSolver* apCoarseSolver, 
                     uint amaxiter,
                     TFloat aaccur,
                     uint averb )
    : TSolver ( apA[ahigh-1], apRHS[ahigh-1], apX[ahigh-1], 
                amaxiter, aaccur, averb ),
      TtwogridInfo( alow, ahigh, aRecCall, apPreSmoother, 
                    apPostSmoother, apCoarseSolver )
{ 
    if (pPreSmoother != NULL)
    {
        TString newName;
        
        newName  = "TwoG ";
        newName += "(";
        newName += pPreSmoother->getName();
        newName += ")";
        setName( newName ); 
    }// if
    else
        setName( "TwoG" ); 
    
    
    pAH     = apA;
    pXH     = apX; 
    pRHSH   = apRHS; 
    pProlong = apProlong;
    pRestrict = apRestrict;

    pA = apA[ahigh];
    pX = apX[ahigh];
    pRHS = apRHS[ahigh];
}

//
// one twogrid step on level 'level'
//
void 
Ttwogrid::step ( uint level ) 
{

    if (level == low) {
        //        dp = (*pRHSH[level]) - (*pAH[level]) * (*pXH[level]); 
        //  TFloat en = dp.euclidNorm();cout<<" dn before cgs: "<<en<<" ";
        pCoarseSolver->solve( pAH[level], pRHSH[level], pXH[level] );
        //dp = (*pRHSH[level]) - (*pAH[level]) * (*pXH[level]); 
        //  en = dp.euclidNorm();cout<<" dn after cgs: "<<en<<endl;
 }else 
    {
        // presmoothing
        pPreSmoother->setMaxit( preSmoothingSteps );
        pPreSmoother->solve( pAH[level], pRHSH[level], pXH[level] );

        // calculate defect on fine grid
        d = (*pRHSH[level]) - (*pAH[level]) * (*pXH[level]);

        // restrict defect to coarse grid
        *pRHSH[level-1] = pRestrict[level-1]->restrict( d );

        // project *pRHSH[level-1] onto range of *pAH[level-1]
       
       

        pXH[level-1]->set( 0.0 );
        // coarse grid solution
        
        dp = (*pRHSH[level-1]);
        TFloat defect_norm =dp.weightedEuclidNorm( )/ sqrt(dp.getSize());
        TFloat defect_norm_old = defect_norm;

        // control number of inner calls:
        // usually by defect reduction by factor relaccrate,
        // but for safety by a maximal iteration number maxinnercalls
        // Since the coarse grid solver is assumed to be 'exact', only one step is meaningful
        // regardless of the actual accuracy.

        uint innercalls=0;
        uint maxinnercalls;
        if (level==low+1)
            maxinnercalls=1;
        else
            maxinnercalls=30;

        
        TFloat relaccrate;
        if ( level==high)
            relaccrate = 1e-12;
        else
            relaccrate= 1e-13;


        while (( defect_norm/defect_norm_old >relaccrate ) && (innercalls <maxinnercalls)){
            innercalls++;
            step( level-1 );
            dp = (*pRHSH[level-1]) - (*pAH[level-1]) * (*pXH[level-1]);
            TFloat en = dp.euclidNorm();
            defect_norm= en/sqrt(dp.getSize());
        }
        if (level==high)
            cout<< "innercalls on level:"<<level<<" : " <<innercalls<<endl;
        
        // coarse grid correction
        (*pXH[level]) += pProlong[level-1]->prolong( (*pXH[level-1]) );

        // postsmoothing
        pPostSmoother->setMaxit( postSmoothingSteps );
        pPostSmoother->solve( pAH[level], pRHSH[level], pXH[level] );
               

    }// else
     
} 

//
// one twogrid step on highest level
//
int 
Ttwogrid::step () 
{ 
    step( high ); 
    return (RETURN_CONTINUE);
}

//
// update pX on all levels by restricting fine grid pX to coarser levels
//
void 
Ttwogrid::updateXHierarchy()
{
    for (uint i = high; i > low; i--)
        *pXH[i-1] = pRestrict[i-1]->restrict( *pXH[i] );
}

//
// add/subtract twople of identity to all levels
//
void 
Ttwogrid::addIdentity( const TFloat factor )
{
    for (uint i = high; i > low; i--)
        pAH[i]->addIdentity( factor );
}

void 
Ttwogrid::subtractIdentity( const TFloat factor )
{
    for (uint i = high; i > low; i--)
        pAH[i]->subtractIdentity( factor );
}

//
// set smoother		
//
void 
Ttwogrid::setSmoother ( TSolver* apSmoother ) 
{
    pPreSmoother = apSmoother;
    pPostSmoother = apSmoother;
    if (pPreSmoother != NULL)
    {
        TString newName;
        
        newName  = "TwoG ";
        newName += "(";
        newName += pPreSmoother->getName();
        newName += ")";
        setName( newName ); 
    }// if
}
