//
//  Project :  libvisual
//  File    :  T2DFlowView.cc
//  Purpose :  class for visualization of a 2D flow
//  Author  :  Ronald Kriemann
//
// arch-tag: dc041003-772c-4650-ba97-a3945975b94f
//

#include "basisEqn.hh"
#include "TRandom.hh"

#include "T2DFlowView.hh"

///////////////////////////////////////////
//
// constructors and destructors
//

T2DFlowView::T2DFlowView ( T2DBasisEqn * pEqn,
                           TFloat stepX, TFloat stepY )
{
    stepWidthX = stepX;
    stepWidthY = stepY;
    
    buildFlowArray( pEqn );
}

T2DFlowView::~T2DFlowView ()
{
}

///////////////////////////////////////////
//
// visualization methods
//

void
T2DFlowView::draw ( TPrinter & p )
{
    if (arrPoints.getSize() == 0)
        return;

    int         i;
    int         noOfPoints = arrPoints.getSize();
    TVector<2>  orig, dest, b;
    TFloat      maxFlow, minDist;
    TFloat      norm, factor, len;

    //
    // look for maximal flow
    //
    maxFlow = 0;
    for (i = 0; i < noOfPoints; i++)
    {
        if ((norm = arrFlow[i].euclidNorm()) > maxFlow)
            maxFlow = norm;
    }// for
    
    // calc factor for downsizing flow
    if (stepWidthX <= stepWidthY)
        minDist = stepWidthX;
    else
        minDist = stepWidthY;
    
    factor  = minDist * 1.3;
    factor /= maxFlow;

    p.setRGBColor( COLOR_BLACK );

    //
    // draw flow
    //
    for (i = 0; i < noOfPoints; i++)
    {
        dest = orig = arrPoints[i];
        b = arrFlow[i];

        norm = b.euclidNorm();
        if (norm == 0)
            continue;

        len  = norm;
        b   *= 1 / len;
        len *= factor;
        b   *= len;
        
        b *= 0.45;
        orig -= b;
        dest += b;
        
//        p.setHSVColor( 300 - (300 * norm/maxFlow), 1, 1 );

        // draw arrow, but center at origin
        p.drawArrow( orig[X], orig[Y], dest[X], dest[Y] );
    }// for
}

//
// encapsulate building of point/flow array
//
void
T2DFlowView::buildFlowArray ( T2DBasisEqn * pEqn )
{
    int            i;
    TVector<2>     b, min, max, loc;
    P2DVectorFunc  convX;
    P2DVectorFunc  convY;
    
    if (pEqn == NULL)
    {
        cout << "T2DFlowView::buildFlowArray (...) : got NULL equation" << endl;
        return;
    }// if

    convX = pEqn->getConvb0();
    convY = pEqn->getConvb1();
    
    arrPoints.setGrowBy( 16 );
    arrFlow.setGrowBy( 16 );

    // to be done : get min/max from domain
    min[X] = min[Y] = 0;
    max[X] = max[Y] = 1;

    i = 0;
    loc[X] = min[X] + stepWidthX;
    while (loc[X] + 0.001 < max[X])
    {
        loc[Y] = min[Y] + stepWidthY;
        while (loc[Y] + 0.001 < max[Y])
        {
            // to be done : check if in domain
            
            arrPoints.setSize( i + 1 );
            arrFlow.setSize( i + 1 );

            arrPoints[i][X] = loc[X];
            arrPoints[i][Y] = loc[Y];

            b[X] = convX( arrPoints[i] );
            b[Y] = convY( arrPoints[i] );
                
            arrFlow[i] = b;

            i++;

            loc[Y] += stepWidthY;
        }// while

        loc[X] += stepWidthX;
    }// while
}
