//
//  Project :  libvisual
//  File    :  T3DFlowView.cc
//  Purpose :  class for visualization of a 3D flow
//  Author  :  Ronald Kriemann
//
// arch-tag: 32f77a88-f68d-4e33-aceb-e1d316e6bf40
//

#include "TConvEqn.hh"
#include "TRandom.hh"

#include "T3DFlowView.hh"

///////////////////////////////////////////
//
// constructors and destructors
//

T3DFlowView::T3DFlowView ( TConvEqn<3> * pEqn,
                           TFloat stepX, TFloat stepY, TFloat stepZ )
{
    stepWidthX = stepX;
    stepWidthY = stepY;
    stepWidthZ = stepZ;
    
    buildFlowArray( pEqn );
}

T3DFlowView::~T3DFlowView ()
{
}

///////////////////////////////////////////
//
// visualization methods
//

void
T3DFlowView::draw3D ( T3DPrinter & p )
{
    if (arrPoints.getSize() == 0)
        return;

    int         i;
    int         noOfPoints = arrPoints.getSize();
    TColorType  oldColor = p.color();
    TVector<3>  orig, dest, b;
    TFloat      maxFlow, minDist = 0.0;
    TFloat      norm, factor, len;

    //
    // look for maximal flow
    //
    maxFlow = 0;
    for (i = 0; i < noOfPoints; i++)
    {
        if ((norm = arrFlow[i].euclidNorm()) > maxFlow)
            maxFlow = norm;
    }// for
    
    // calc factor for downsizing flow
    if      ((stepWidthX <= stepWidthY) && (stepWidthX <= stepWidthZ))
        minDist = stepWidthX;
    else if ((stepWidthY <= stepWidthX) && (stepWidthY <= stepWidthZ))
        minDist = stepWidthY;
    else if ((stepWidthZ <= stepWidthX) && (stepWidthZ <= stepWidthY))
        minDist = stepWidthZ;
    else
        cout << "T3DFlowView::draw3D (...) : error while calc. minimal distance" << endl;
    
    factor  = minDist * 1.3;
    factor /= maxFlow;

    //
    // draw flow
    //
    for (i = 0; i < noOfPoints; i++)
    {
        dest = orig = arrPoints[i];
        b = arrFlow[i];

        norm = b.euclidNorm();
        if (norm == 0)
            continue;

        len  = norm;
        b   *= 1 / len;
        len *= factor;
        b   *= len;
        
        b *= 0.5;
        orig -= b;
        dest += b;
        
        p.setHSVColor( 300 - (300 * norm/maxFlow), 1, 1 );
        p.drawArrow( & orig, & dest, 0.05 * len );
    }// for

    p.drawCoordSystem();
    p.setColor( oldColor );
}

//
// encapsulate building of point/flow array
//
void
T3DFlowView::buildFlowArray ( TConvEqn<3> * pEqn )
{
#define EPS  0.0001
    
    int         i;
    TVector<3>  b, min, max, loc;
    
    if (pEqn == NULL)
    {
        cout << "T3DFlowView::buildFlowArray (...) : got NULL equation" << endl;
        return;
    }// if

    arrPoints.setGrowBy( 16 );
    arrFlow.setGrowBy( 16 );

    // to be done : get min/max from domain
    min[X] = min[Y] = min[Z] = 0;
    max[X] = max[Y] = max[Z] = 1;

    i = 0;
    loc[X] = min[X] + stepWidthX;
    while (loc[X] + 0.001 < max[X])
    {
        loc[Y] = min[Y] + stepWidthY;
        while (loc[Y] + 0.001 < max[Y])
        {
            loc[Z] = min[Z] + stepWidthZ;

            // to be done : check if in domain
            
            while (loc[Z] + 0.001 < max[Z])
            {
                arrPoints.setSize( i + 1 );
                arrFlow.setSize( i + 1 );

                arrPoints[i][X] = loc[X];
                arrPoints[i][Y] = loc[Y];
                arrPoints[i][Z] = loc[Z];
                
                pEqn->evalConv( & arrPoints[i], & b );
                
                arrFlow[i] = b;

                i++;

                loc[Z] += stepWidthZ;
            }// while

            loc[Y] += stepWidthY;
        }// while

        loc[X] += stepWidthX;
    }// while
}
