//
//  Project :  libvisual
//  File    :  T3DGraphView.cc
//  Purpose :  class for visualization of a 3D graph
//  Author  :  Ronald Kriemann
//
// arch-tag: e2c0d5a9-002d-4abc-a715-d0c658387e25
//

#include "TRandom.hh"

#include "TPermutation.hh"

#include "graphtool.hh"
#include "fvs.hh"
#include "planarFVS.hh"
#include "T3DCellEdgeGraph.hh"
#include "T3DCellEdgeNode.hh"
#include "T3DSurfaceGraph.hh"
#include "T3DGraph.hh"
#include "T3DPlanarNode.hh"
#include "T3DMarkableNode.hh"

#include "T3DGraphView.hh"

///////////////////////////////////////////
//
// constructors and destructors
//

T3DGraphView::T3DGraphView ( T3DCellEdgeGraph * pGraph )
{
    reset();
    pCellEdgeGraph = pGraph;
}

T3DGraphView::T3DGraphView ( T3DSurfaceGraph * pGraph )
{
    reset();
    pSurfaceGraph  = pGraph;
}

T3DGraphView::T3DGraphView ( T3DGraph * pGraph )
{
    reset();
    pNetGraph      = pGraph;
}

T3DGraphView::T3DGraphView ( TList< T3DCellEdgeGraph > * fsList )
{
    reset();
    pFSList        = fsList;
}

T3DGraphView::T3DGraphView ( TList< T3DSurfaceGraph >  * sgList )
{
    reset();
    pSGList = sgList;
}

T3DGraphView::T3DGraphView ( TList< TGraph > * listCycles )
{
    reset();
    pListCycles    = listCycles;
}

T3DGraphView::~T3DGraphView ()
{
}

///////////////////////////////////////////
//
// access member variables
//

void
T3DGraphView::setPermutation ( TPermutation * pPerm )
{
    pPermutation = pPerm;
}
    
///////////////////////////////////////////
//
// visualization methods
//

void
T3DGraphView::draw3D ( T3DPrinter & p )
{
    if (pCellEdgeGraph != NULL)
        drawCellEdgeGraph( p );

    if (pSurfaceGraph != NULL)
        drawSurfaceGraph( p );

    if (pNetGraph != NULL)
        drawNetGraph( p );

    if (pFSList != NULL)
        drawFlowSurfaces( p );

    if (pSGList != NULL)
        drawSurfaceGraphs( p );

    if (pListCycles != NULL)
        drawPlanarCycles( p );
}

//
// reset all member variables
//
void
T3DGraphView::reset ()
{
    pCellEdgeGraph = NULL;
    pSurfaceGraph  = NULL;
    pNetGraph      = NULL;
    pFSList        = NULL;
    pSGList        = NULL;
    pListCycles    = NULL;

    pPermutation = NULL;
}
    
//
// encapsulate drawing for each graphtype
//
void
T3DGraphView::drawCellEdgeGraph ( T3DPrinter & p )
{
    uint                    i, j;
    uint                    noOfNodes;
    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode        *pNode;
    T3DFrontier            *pFace;
    TColorType              oldColor = p.color();
    

    if (pCellEdgeGraph == NULL)
        return;
    
    iterNodes.setList( pCellEdgeGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        p.setRGBColor( COLOR_GRAY );
        p.setFilter( 0 );

        p.drawLine( pNode->cellEdge()->getNode( 0 ),
                    pNode->cellEdge()->getNode( 2 ) );

        // draw fs-face and alternative fs-face
        for (j = 0; j < 1; j++)
        {
            if ((pFace = pNode->fsFace( j )) != NULL)
            {
                noOfNodes = pFace->getNoOfNodes();
                
                p.setRGBColor( COLOR_RED );
                p.setFilter( 0.6 );
                p.beginPolygon( FILLED );
                
                for (i = 0; i < noOfNodes; i++)
                {
                    // only every second node is a corner
                    if ((i % 2) == 0)
                        p.addNode( pFace->getNode( i ) );
                }// for
                
                p.endPolygon();
            }// if
        }// for
    }// while

	// finally draw coord system
	p.drawCoordSystem();

    p.setColor( oldColor );
}

void
T3DGraphView::drawSurfaceGraph ( T3DPrinter & p )
{
    if (pSurfaceGraph == NULL)
        return;

    TListIterator< TNode >  iterNodes;
    T3DPlanarNode          *pNode, *pAdjNode;
    TListIterator< TEdge >  iterEdges;
    TEdge                  *pEdge;
    TVector<3>              vecOrig;
    char                    szString[64];
    TColorType              oldColor = p.color();
    TFontType               oldFont = p.font();

    iterNodes.setList( pSurfaceGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DPlanarNode*) iterNodes++) != NULL)
    {
        switch (pNode->mark())
        {
            case C        : p.setRGBColor( COLOR_RED );    break;
            case F        : p.setRGBColor( COLOR_BLUE );   break;
            case L        : p.setRGBColor( COLOR_GREEN );  break;
            case FVS_NODE : p.setRGBColor( COLOR_YELLOW ); break;
            default       : p.setRGBColor( COLOR_BLACK );  break;
        }// switch;
        p.drawSphere( pNode->vector(), 0.01 );

        p.setRGBColor( COLOR_BLACK );
        p.setFont( 0.02 );

        if (pPermutation != NULL)
            sprintf( szString, "%d", (*pPermutation)[ pNode->getName() ] );
        else
            sprintf( szString, "%d", pNode->getName() );
        
        vecOrig = * pNode->vector();
        vecOrig[X] += 0.01;
        vecOrig[Y] += 0.015;
        vecOrig[Z] += 0.0;
        p.drawText( & vecOrig, szString );
            
        p.setRGBColor( COLOR_GRAY );
        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdjNode = (T3DPlanarNode*) pEdge->getHead();

            p.drawArrow( pNode->vector(), pAdjNode->vector(), 0.0025 );
        }// while
    }// while

    p.drawCoordSystem();

    p.setColor( oldColor );
    p.setFont( oldFont );
}

void
T3DGraphView::drawNetGraph ( T3DPrinter & p )
{
    if (pNetGraph == NULL)
        return;

    TListIterator< TNode >  iterNodes;
    T3DMarkableNode        *pNode, *pAdjNode;
    TListIterator< TEdge >  iterEdges;
    TEdge                  *pEdge;
    TVector<3>              vecOrig;
    char                    szString[64];
    TColorType              oldColor = p.color();
    TFontType               oldFont = p.font();

    iterNodes.setList( pNetGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DMarkableNode*) iterNodes++) != NULL)
    {
        switch (pNode->mark())
        {
            case C        : p.setRGBColor( COLOR_RED );    break;
            case F        : p.setRGBColor( COLOR_BLUE );   break;
            case L        : p.setRGBColor( COLOR_GREEN );  break;
            case FVS_NODE : p.setRGBColor( COLOR_YELLOW ); break;
            default       : p.setRGBColor( COLOR_BLACK );  break;
        }// switch;
        p.drawSphere( pNode->vector(), 0.01 );

        p.setRGBColor( COLOR_BLACK );
        p.setFont( 0.02 );

        if (pPermutation != NULL)
            sprintf( szString, "%d", (*pPermutation)[ pNode->getName() ] );
        else
            sprintf( szString, "%d", pNode->getName() );
        
        vecOrig = * pNode->vector();
        vecOrig[X] += 0.01;
        vecOrig[Y] += 0.015;
        vecOrig[Z] += 0.0;
        p.drawText( & vecOrig, szString );
            
        p.setRGBColor( COLOR_GRAY );
        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdjNode = (T3DMarkableNode*) pEdge->getHead();

            p.drawArrow( pNode->vector(), pAdjNode->vector(), 0.0025 );
        }// while
    }// while

    p.drawCoordSystem();

    p.setColor( oldColor );
    p.setFont( oldFont );
}

void
T3DGraphView::drawFlowSurfaces ( T3DPrinter & p )
{
    uint                               i, j;
    uint                               noOfNodes;
    TListIterator< T3DCellEdgeGraph >  iterFS;
    T3DCellEdgeGraph                  *pFlowSurface;
    TListIterator< TNode >             iterNodes;
    T3DCellEdgeNode                   *pNode;
    T3DFrontier                       *pFace;
    TFloat                             h, s, v;
    TFloat                             colMax, colMin, colStep;
    TColorType                         oldColor = p.color();
    

    if (pFSList == NULL)
        return;

    colMax = 0.8;
    colMin = 0.4;
    colStep = (colMax - colMin) / TFloat(pFSList->getSize());
    
    h = -1;
    s = 1;
    v = colMin;
        
    iterFS.setList( *pFSList );
    iterFS.first();
    while ((pFlowSurface = iterFS++) != NULL)
    {
//         v += 2 * colStep;
//         if (v > colMax)
//             v = colMin;
//         if (v < colMin)
//             v = colMin;
        
        iterNodes.setList( pFlowSurface->getNodeList() );
        iterNodes.first();
        while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
        {
            p.setRGBColor( COLOR_GRAY );
            p.setFilter( 0 );
            
            p.drawLine( pNode->cellEdge()->getNode( 0 ),
                        pNode->cellEdge()->getNode( 2 ) );
            
//             p.setHSVColor( h, s, v );
//             p.setHSVColor( h, s, v, AMBIENT );
            p.setRGBColor( 1, 0, 0, PIGMENT );
            p.setFilter( 0.6 );
            
            // draw fs-face and alternative fs-face
            for (j = 0; j < 2; j++)
            {
                if ((pFace = pNode->fsFace( j )) != NULL)
                {
                    p.beginPolygon( FILLED );
                
                    noOfNodes = pFace->getNoOfNodes();
                    
                    for (i = 0; i < noOfNodes; i++)
                    {
                        // only every second node is a corner
                        if ((i % 2) == 0)
                            p.addNode( pFace->getNode( i ) );
                    }// for 
                    
                    p.endPolygon();
                }// if  
            }// for 

            p.setRGBColor( COLOR_BLACK, AMBIENT );
        }// while   
    }// while

    // finally draw coord system
    p.drawCoordSystem();

    p.setColor( oldColor );
}

void
T3DGraphView::drawSurfaceGraphs ( T3DPrinter & p )
{
    if (pSGList == NULL)
        return;

    TListIterator< T3DSurfaceGraph >   iterGraph;
    T3DSurfaceGraph                  * pSGraph;
    
    TListIterator< TNode >  iterNodes;
    T3DPlanarNode          *pNode, *pAdjNode;
    TListIterator< TEdge >  iterEdges;
    TEdge                  *pEdge;
    TVector<3>              vecOrig;
    char                    szString[64];
    TColorType              oldColor = p.color();
    TFontType               oldFont = p.font();

    iterGraph.setList( * pSGList );
    iterGraph.first();
    while ((pSGraph = iterGraph++) != NULL)
    {
        iterNodes.setList( pSGraph->getNodeList() );
        iterNodes.first();
        while ((pNode = (T3DPlanarNode*) iterNodes++) != NULL)
        {
            switch (pNode->mark())
            {
                case C        : p.setRGBColor( COLOR_RED );    break;
                case F        : p.setRGBColor( COLOR_BLUE );   break;
                case L        : p.setRGBColor( COLOR_GREEN );  break;
                case FVS_NODE : p.setRGBColor( COLOR_YELLOW ); break;
                default       : p.setRGBColor( COLOR_BLACK );  break;
            }// switch;
            p.drawSphere( pNode->vector(), 0.01 );
            
            p.setRGBColor( COLOR_BLACK );
            p.setFont( 0.02 );
            
            if (pPermutation != NULL)
                sprintf( szString, "%d", (*pPermutation)[ pNode->getName() ] );
            else
                sprintf( szString, "%d", pNode->getName() );
            
            vecOrig = * pNode->vector();
            vecOrig[X] += 0.01;
            vecOrig[Y] += 0.015;
            vecOrig[Z] += 0.0;
            p.drawText( & vecOrig, szString );
            
            p.setRGBColor( COLOR_GRAY );
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pAdjNode = (T3DPlanarNode*) pEdge->getHead();
                
                p.drawArrow( pNode->vector(), pAdjNode->vector(), 0.0025 );
            }// while
        }// while
    }// while

    p.drawCoordSystem();

    p.setColor( oldColor );
    p.setFont( oldFont );
}

void
T3DGraphView::drawPlanarCycles ( T3DPrinter & p )
{
    if (pListCycles == NULL)
        return;

    TListIterator< TGraph >   iterCycles;
    TGraph                  * pCycle;
    TListIterator< TNode >    iterNodes;
    T3DPlanarNode           * pNode, * pFirstNode, * pLastNode;
    TVector<3>                vecOrig;
    char                      szString[64];
    TColorType                oldColor = p.color();
    TFontType                 oldFont = p.font();

    iterCycles.setList( * pListCycles );
    iterCycles.first();
    while ((pCycle = iterCycles++) != NULL)
    {
        iterNodes.setList( pCycle->getNodeList() );
        pFirstNode = pLastNode = (T3DPlanarNode*) iterNodes.first();
        
        while ((pNode = (T3DPlanarNode*) iterNodes++) != NULL)
        {
            p.setRGBColor( COLOR_RED );
            p.drawSphere( pNode->vector(), 0.01 );
            
            p.setRGBColor( COLOR_BLACK );
            p.setFont( 0.015 );
            
            if (pPermutation != NULL)
                sprintf( szString, "%d", (*pPermutation)[ pNode->getName() ] );
            else
                sprintf( szString, "%d", pNode->getName() );
            
            vecOrig = * pNode->vector();
            vecOrig[X] += 0.01;
            vecOrig[Y] += 0.015;
            vecOrig[Z] += 0.0;
            p.drawText( & vecOrig, szString );
            
            if (pNode != pLastNode)
            {
                p.setRGBColor( COLOR_GRAY );
                p.drawArrow( pNode->vector(), pLastNode->vector(), 0.0025 );
            }// if

            pLastNode = pNode;
        }// while

        if (pLastNode != pFirstNode)
        {
            p.setRGBColor( COLOR_GRAY );
            p.drawArrow( pFirstNode->vector(), pLastNode->vector(), 0.0025 );
        }// if
    }// while

    p.drawCoordSystem();

    p.setColor( oldColor );
    p.setFont( oldFont );
}
