//
//  Project :  libvisual
//  File    :  TVRML.cc
//  Purpose :  class for a VRML interface
//  Author  :  Ronald Kriemann
//
// arch-tag: 00bcbde5-4048-4c38-8465-3c527a53f5dc
//

#include <string.h>

#include "defines.hh"

#include "TVRML.hh"

// use simple drawing (cubes for spheres etc)
bool TVRML::bSimpleDraw = TRUE;

//
// constructor
//
TVRML::TVRML ( const char *filename )
        : T3DPrinter( filename )
{}

//
// destructor
//
TVRML::~TVRML ()
{}

//
// change between drawing styles
//
void
TVRML::setSimpleDraw ( bool b )
{
    bSimpleDraw = b;
}
    
//
// begin (end) printing
//
void 
TVRML::begin ()
{
    T3DPrinter::begin();

    if (state() != UNDEF)
    {
        indentWrite( "#VRML V1.0 ascii\n" );
        beginSection();

        // set initial camera
        //        beginNode( "OrthographicCamera" );
        //        indentWrite( "position       0 0 1\n" );
        //        indentWrite( "orientation    0 0 1  0\n" );
        //        indentWrite( "focalDistance  5\n" );
        //        indentWrite( "height         2\n" );
        //        endNode();
    }// else
}

void 
TVRML::end ()
{
    if (state() != UNDEF)
        endSection();
    
    T3DPrinter::end();
}

//
// begin (end) new node section 
// (if name == NULL begin Separator)
//
void 
TVRML::beginSection ( const char *name )
{
    char  szString[256];

    if (state() != UNDEF)
    {
        if (name != NULL)
            sprintf( szString, "%s {", name );
        else
            strcpy( szString, "Separator {" );

        indentWrite( szString );
        indentUp();
    }// if
}

void 
TVRML::endSection ()
{
    if ((state() != UNDEF) && (indentLevel() > 0))
    {
        indentDown();
        indentWrite( "}" );
    }// if
}

//
// misc. draw commands
//

void 
TVRML::drawText ( TVector<3> *orig, const char *text )
{
    TVector<3>  locOrig;
    char        szString[256];

    if (orig == NULL)
    {
        cerr << "TVRML::drawText (...) : gave NULL node" << endl;
        return;
    }// if

    T3DPrinter::trans( & (locOrig = *orig) );
    
    beginSection();

    writeColor();
    writeFont();

    beginSection( "Translation" );
    sprintf( szString, "translation %.4g %.4g %.4g",
             double( locOrig[X] ),
             double( locOrig[Y] ),
             double( locOrig[Z] ) );
    indentWrite( szString );
    endSection();
    
    beginSection( "AsciiText" );

    sprintf( szString, "string        \"%s\"", text );
    indentWrite( szString );

    if (font().width >= 0)
    {
        sprintf( szString, "width         %.4g", double( font().width ) );
        indentWrite( szString );
    }// if
    
    switch (font().justification)
    {
        case JUST_LEFT   : indentWrite( "justification LEFT" ); break;
        case JUST_RIGHT  : indentWrite( "justification RIGHT" ); break;
        case JUST_CENTER : indentWrite( "justification CENTER" ); break;
        default          : break;
    }// switch
    
    if (font().spacing >= 0)
    {
        sprintf( szString, "spacing       \"%.4g\"", double( font().spacing ) );
        indentWrite( szString );
    }// if
    
    endSection();
    endSection();
}

void
TVRML::drawLine ( TVector<3> *orig, TVector<3> *dest )
{
    TVector<3>  locOrig, locDest;
    char        szString[256];
    TColorType  oldColor = color();

    if ((orig == NULL) || (dest == NULL))
        return;

    if ( *orig == *dest )
        return;
    
    T3DPrinter::trans( & (locOrig = *orig) );
    T3DPrinter::trans( & (locDest = *dest) );
    
    setFilter( 0 );
    
    beginSection();

    writeColor();
    
    beginSection( "Coordinate3" );
    indentWrite( "point [" );
    indentUp();
    
    sprintf( szString, "%.4g %.4g %.4g,", 
             double( locOrig[X] ),
             double( locOrig[Y] ),
             double( locOrig[Z] ) );
    indentWrite( szString );

    sprintf( szString, "%.4g %.4g %.4g,", 
             double( locDest[X] ),
             double( locDest[Y] ),
             double( locDest[Z] ) );
    indentWrite( szString );

    indentDown();
    indentWrite( "]" );
    endSection();
    
    beginSection( "IndexedLineSet" );
    indentWrite( "coordIndex [ 0, 1, -1 ]" );
    endSection();

    endSection();

    setColor( oldColor );
}
    
void
TVRML::drawCylinder ( TVector<3> *orig, TVector<3> *dest,
                      TFloat radius )
{
    TVector<3>  locOrig, locDest;
    TVector<3>  rotAxis, trans;
    TFloat      angle, len;
    char        szString[256];

    if ((orig == NULL) || (dest == NULL))
    {
        cerr << "TVRML::drawCylinder (...) : gave NULL args" << endl;
        return;
    }// if

    if ( *orig == *dest )
        return;
    
    T3DPrinter::trans( & (locOrig = *orig) );
    T3DPrinter::trans( & (locDest = *dest) );
    
    calcTrans( & locOrig, & locDest, & trans, & rotAxis, & angle );

    len = (locDest - locOrig).euclidNorm();
    
    //
    // write the node
    //
    beginSection();

    writeColor();
    
    beginSection( "Translation" );
    sprintf( szString, "translation %.4g %.4g %.4g",
             double( trans[X] ),
             double( trans[Y] ),
             double( trans[Z] ) );
    indentWrite( szString );
    endSection();
    
    beginSection( "Rotation" );
    sprintf( szString, "rotation %.4g %.4g %.4g %.4g",
             double( rotAxis[X] ),
             double( rotAxis[Y] ),
             double( rotAxis[Z] ),
             double( angle ) );
    indentWrite( szString );
    endSection();

    if (bSimpleDraw)
    {
        beginSection( "Cube" );
        sprintf( szString, "width  %.4g", double( 2.0*radius ) );
        indentWrite( szString );
        sprintf( szString, "height %.4g", double( len ) );
        indentWrite( szString );
        sprintf( szString, "depth  %.4g", double( 2.0*radius ) );
        indentWrite( szString );
        endSection();
    }// if
    else
    {
        beginSection( "Cylinder" );
        sprintf( szString, "radius %.4g", double( radius ) );
        indentWrite( szString );
        sprintf( szString, "height %.4g", double( len ) );
        indentWrite( szString );
        endSection();
    }// else

    endSection();
}

void
TVRML::drawCone ( TVector<3> *orig, TFloat origRad,
                  TVector<3> *dest, TFloat destRad )
{
    TVector<3>  locOrig, locDest;
    TVector<3>  rotAxis, trans, temp;
    TFloat      angle, len;
    char        szString[256];

    if ((orig == NULL) || (dest == NULL))
    {
        cerr << "TVRML::drawCone (...) : gave NULL args" << endl;
        return;
    }// if

    if ( *orig == *dest )
        return;
    
    T3DPrinter::trans( & (locOrig = *orig) );
    T3DPrinter::trans( & (locDest = *dest) );
    
    calcTrans( & locOrig, & locDest, & trans, & rotAxis, & angle );

    len = (locDest - locOrig).euclidNorm();
    
    //
    // write the node
    //
    beginSection();

    writeColor();

    beginSection( "Translation" );
    sprintf( szString, "translation %.4g %.4g %.4g",
             double( trans[X] ),
             double( trans[Y] ),
             double( trans[Z] ) );
    indentWrite( szString );
    endSection();
    
    beginSection( "Rotation" );
    sprintf( szString, "rotation %.4g %.4g %.4g %.4g",
             double( rotAxis[X] ),
             double( rotAxis[Y] ),
             double( rotAxis[Z] ),
             double( angle ) );
    indentWrite( szString );
    endSection();

    if (bSimpleDraw)
    {
        len /= 2.0;
    
        beginSection( "Coordinate3" );
        
        indentWrite( "point [" );
        indentUp();

        T3DPrinter::trans( & (temp = TVector<3>( -origRad, -origRad, -len )) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( temp[X] ),
                 double( temp[Y] ),
                 double( temp[Z] ) );
        indentWrite( szString );
        T3DPrinter::trans( & (temp = TVector<3>( origRad, -origRad, -len )) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( temp[X] ),
                 double( temp[Y] ),
                 double( temp[Z] ) );
        indentWrite( szString );
        T3DPrinter::trans( & (temp = TVector<3>( origRad, origRad, -len )) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( temp[X] ),
                 double( temp[Y] ),
                 double( temp[Z] ) );
        indentWrite( szString );
        T3DPrinter::trans( & (temp = TVector<3>( -origRad, origRad, -len )) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( temp[X] ),
                 double( temp[Y] ),
                 double( temp[Z] ) );
        indentWrite( szString );
        T3DPrinter::trans( & (temp = TVector<3>( 0, 0, len )) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( temp[X] ),
                 double( temp[Y] ),
                 double( temp[Z] ) );
        indentWrite( szString );
        
        indentDown();
        indentWrite( "]" );
        
        endSection();
        
        beginSection( "IndexedFaceSet" );
        
        indentWrite( "coordIndex [" );
        indentUp();
        
        indentWrite( "0, 1, 4, 0, -1," );
        indentWrite( "1, 2, 4, 1, -1," );
        indentWrite( "2, 3, 4, 2, -1," );
        indentWrite( "3, 0, 4, 3, -1," );
        indentWrite( "3, 2, 1, 0, 3, -1," );
        
        indentDown();
        indentWrite( "]" );
        
        endSection();
    }// if
    else
    {
        beginSection( "Cone" );
        sprintf( szString, "bottomRadius %.4g", double( origRad ) );
        indentWrite( szString );
        sprintf( szString, "height %.4g", double( len ) );
        indentWrite( szString );
        endSection();
    }// else

    endSection();
}

void
TVRML::drawSphere ( TVector<3> * orig, TFloat radius )
{
    TVector<3>  locOrig;
    char  szString[256];

    if (orig == NULL)
    {
        cerr << "TVRML::drawSphere (...) : gave NULL args" << endl;
        return;
    }// if

    T3DPrinter::trans( & (locOrig = *orig) );
    
    beginSection();

    writeColor();

    beginSection( "Translation" );
    sprintf( szString, "translation %.4g %.4g %.4g",
             double( locOrig[X] ),
             double( locOrig[Y] ),
             double( locOrig[Z] ) );
    indentWrite( szString );
    endSection();

    if (bSimpleDraw)
    {
        beginSection( "Cube" );
        sprintf( szString, "width  %.4g", double( 2.0*radius ) );
        indentWrite( szString );
        sprintf( szString, "height %.4g", double( 2.0*radius ) );
        indentWrite( szString );
        sprintf( szString, "depth  %.4g", double( 2.0*radius ) );
        indentWrite( szString );
        endSection();
    }// if
    else
    {
        beginSection( "Sphere" );
        sprintf( szString, "radius %.4g", double( radius ) );
        indentWrite( szString );
        endSection();
    }// else

    endSection();
}
    
void 
TVRML::drawPolygon ()
{
    TVector<3>                   locVec;
    uint                         i, noOfNodes;
    TListIterator< TVector<3> >  iterNodes( polygonData().listNodes );
    TVector<3>                  *v;
    char                         szString[256];

    
    noOfNodes = polygonData().listNodes.getSize();
    
    if (noOfNodes == 0)
        return;

    beginSection();

    writeColor();
    
    beginSection( "Coordinate3" );
        
    indentWrite( "point [" );
    indentUp();

    iterNodes.first();
    while ((v = iterNodes++) != NULL)
    {
        T3DPrinter::trans( & (locVec = *v) );
        sprintf( szString, "%.4g %.4g %.4g,",
                 double( locVec[X] ),
                 double( locVec[Y] ),
                 double( locVec[Z] ) );
        indentWrite( szString );
    }// while
    
    indentDown();
    indentWrite( "]" );

    endSection();

    //
    // write connection between nodes
    //
    
    beginSection( "IndexedFaceSet" );

    indentWrite( "coordIndex [" );
    indentUp();

    strcpy( szString, "" );
    
    for (i = 0; i < noOfNodes; i++)
    {
        sprintf( szString, "%s %d,", szString, i );
    }// for

    // check if we have to set the last edge
    // (connecting last with first node)
    if (iterNodes.first() == iterNodes.last())
        strcat( szString, " -1," );
    else
        strcat( szString, " 0, -1," );
    indentWrite( szString );
    
    indentDown();
    indentWrite( "]" );

    endSection();

    endSection();
}

//
// misc. style commands
//

//
// internal transformation
//
TFloat 
TVRML::transZ ( TVector<3> * v ) const
{
    return - T3DPrinter::transZ( v );
}

//
// write color/font def to file
//
void
TVRML::writeColor ()
{
    char  szString[256];
    
    beginSection( "Material" );

    sprintf( szString, "diffuseColor %.4g %.4g %.4g",
             double( color().r ), double( color().g ), double( color().b ) );
    indentWrite( szString );
    
    if ((color().amb_r != 0) || (color().amb_g != 0) || (color().amb_b != 0))
    {
        sprintf( szString, "ambientColor %.4g %.4g %.4g",
                 double( color().amb_r ), double( color().amb_g ), double( color().amb_b ) );
        indentWrite( szString );
    }// if
    
    if (color().filter > 0)
    {
        sprintf( szString, "transparency %.4g", double( color().filter ) );
        indentWrite( szString );
    }// if
    
    endSection();
}

void
TVRML::writeFont ()
{
    char  szString[256];

    beginSection( "FontStyle" );

    sprintf( szString, "size   %.2g", double( font().size ) );
    indentWrite( szString );

    switch (font().font)
    {
        case FONT_SERIF      : indentWrite( "family SERIF" ); break;
        case FONT_SANS       : indentWrite( "family SANS" ); break;
        case FONT_TYPEWRITER : indentWrite( "family TYPEWRITER" ); break;
        default              : break;
    }// switch

    switch (font().style)
    {
        case STYLE_NONE                  : indentWrite( "style  NONE" ); break;
        case STYLE_BOLD                  : indentWrite( "style  BOLD" ); break;
        case STYLE_ITALIC                : indentWrite( "style  ITALIC" ); break;
        case (STYLE_BOLD | STYLE_ITALIC) : indentWrite( "style  BOLD | ITALIC" ); break;
        default                          : break;
    }// switch
    
    endSection();
}

//
// calc translation/rotation for given orig/dest
//
void
TVRML::calcTrans ( TVector<3> * pOrig, TVector<3> * pDest,
                   TVector<3> * pTrans, TVector<3> * pRotAxis,
                   TFloat * pAngle )
{
    TVector<3>  rotAxis, dir, locOrig, locDest, translation;
    TFloat      angle, len;

    if ((pOrig == NULL) || (pDest == NULL) ||
        (pTrans == NULL) || (pRotAxis == NULL) || (pAngle == NULL))
    {
        cerr << "TVRML::calcTrans (...) : gave NULL node" << endl;
        return;
    }// if
    
    locOrig = *pOrig;
    locDest = *pDest;
    
    // calc rotation axis with vector-product
    dir  = locDest;
    dir -= locOrig;
    
    rotAxis[X] = dir[Z];
    rotAxis[Y] = 0;
    rotAxis[Z] = -dir[X];

    // adjust if undefined rotaxis
    if (rotAxis.euclidNorm() == 0)
        rotAxis[X] = 1;
    
    len = dir.euclidNorm();
    if (len != 0)
        angle = dir[Y] / len;
    else
        angle = 1;

    if (angle > 1)
        angle = 1;
    if (angle < -1)
        angle = -1;
    
    angle = acos(angle);

    // calc translation
    translation  = dir;
    translation *= 0.5;
    translation += locOrig;
    
    *pTrans   = translation;
    *pRotAxis = rotAxis;
    *pAngle   = angle;
}
