//
//  Project :  libvisual
//  File    :  drawapp.cc
//  Purpose :  class for a x-application (using qt)
//  Author  :  Ronald Kriemann / Sabine Gutsch
//
// arch-tag: a731f907-8663-4a25-9a37-7c35abab6600
//

#include "basegraph.hh"

#include "printer.hh"

#include "vectorfield.hh"
#include "netview.hh"
#include "graphview.hh"
#include "matrixview.hh"
//#include "streakline.hh"
#include "TConvView.hh"
#include "TVRML.hh"
#include "TPOVRay.hh"
#include "T3DGraphView.hh"
#include "T3DFlowView.hh"
#include "T2DFlowView.hh"
#include "view.hh"

#include "drawapp.hh"

// directory to store all files in
#define DATADIR "files"

//
// constructor
//
TDrawApp::TDrawApp ()
{}

////////////////////////////////////////////
//
// draw a vector field
//
void 
TDrawApp::draw ( TDynArray< TVector<2> * >& vertices,
                 TBlockVector<TFloat>& solution,
                 const T2DNet& net,
                 const char filename[],
                 bool scale )
{
    TVectorField view( vertices, solution, net, scale );

    drawPS( & view, filename );
}

////////////////////////////////////////////
//
// draw a net
//
void 
TDrawApp::draw ( T2DNet& net,
                 const char filename[] )
{
    TNetView view( net );

    drawPS( & view, filename );
}

//
// draw a net where the nodenames are permuted
//
void 
TDrawApp::draw ( T2DNet& net, TVectorBlock<uint>& permutation,
                 const char filename[] )
{
    TNetView view( net, permutation );

    drawPS( & view, filename );
}


void 
TDrawApp::drawNet ( T3DNet *net, const char *filename, int dev )
{
    TNetView  view( net );

    if (net == NULL)
    {
        cout << "TDrawApp::drawNet (...) : got NULL net" << endl;
        return;
    }// if
    
    draw( &view, filename, dev );
}

void
TDrawApp::drawMG ( T3DMeshGenerator * mg, const char *filename, int dev )
{
    TNetView  view( mg );

    if (mg == NULL)
    {
        cout << "TDrawApp::drawNet (...) : got NULL meshgen" << endl;
        return;
    }// if
    
    draw( &view, filename, dev );
}

////////////////////////////////////////////
//
// draw a blockmatrix
//
void 
TDrawApp::drawMatrix ( TBlockMatrix * pMatrix,
                       TFloat xi,
                       const char filename[] )
{
    drawMatrix( pMatrix, xi, FALSE, filename );
}

void 
TDrawApp::drawMatrix ( TMatrixBlock * pMatrix,
                       TFloat xi,
                       const char filename[] )
{
    drawMatrix( pMatrix, xi, bool(FALSE), filename );
}

void
TDrawApp::drawMatrix ( TMatrixBlock * pMatrix, TFloat xi,
                       TDynArray< uint > * pBlockStruct,
                       const char *filename )
{
    TMatrixView  view( pMatrix, xi, FALSE );

    view.setBlockStructure( pBlockStruct );

    drawPS( & view, filename );
}

void 
TDrawApp::drawMatrix ( TBlockMatrix * pMatrix, TFloat xi, bool smaller,
                       const char *filename )
{
    TMatrixView  view( pMatrix, xi, smaller );

    drawPS( & view, filename );
}

void
TDrawApp::drawMatrix ( TMatrixBlock * pMatrix, TFloat xi,
                       bool smaller, const char * filename )
{
    TMatrixView  view( pMatrix, xi, smaller );

    drawPS( & view, filename );
}

////////////////////////////////////////////
//
// draw a graph
//
void 
TDrawApp::draw ( TGraph& graph, const char filename[], bool drawDel )
{
    TGraphView  view( graph );

    view.setDrawDeleted( drawDel );
    
    drawPS( & view, filename );
}

void 
TDrawApp::draw ( TGraph& graph, TVectorBlock<uint>& permutation,
                 const char *filename, bool drawDel  )
{
    TGraphView  view( graph, permutation );

    view.setDrawDeleted( drawDel );

    drawPS( & view, filename );
}

void 
TDrawApp::draw ( TList<TGraph>& cycles, const char *filename )
{
    TGraphView  view( cycles );

    drawPS( & view, filename );
}

void 
TDrawApp::draw ( TList<TGraph>& cycles, TVectorBlock<uint>& permutation, 
                 const char *filename )
{
    TGraphView  view( cycles, permutation );

    drawPS( & view, filename );
}

void 
TDrawApp::draw ( TList<TNode>& listCycNodes, const char *filename )
{
    TGraphView  view( listCycNodes );

    drawPS( & view, filename );
}

void
TDrawApp::drawGraph ( T3DCellEdgeGraph * pGraph,
                      const char *filename, int dev )
{
    if (pGraph == NULL)
    {
        cout << "TDrawApp::drawGraph (...) : got NULL graph" << endl;
        return;
    }// if

    T3DGraphView   view( pGraph );

    draw( &view, filename, dev );
}
    
void
TDrawApp::drawGraph ( T3DSurfaceGraph * pGraph, const char *filename,
                      int dev )
{
    drawGraph( pGraph, NULL, filename, dev );
}

void
TDrawApp::drawGraph ( T3DSurfaceGraph * pGraph, TPermutation * pPerm,
                      const char *filename, int dev )
{
    if (pGraph == NULL)
    {
        cout << "TDrawApp::drawGraph (...) : got NULL graph" << endl;
        return;
    }// if

    T3DGraphView   view( pGraph );

    view.setPermutation( pPerm );
    
    draw( &view, filename, dev );
}

void
TDrawApp::drawGraph ( TList< T3DSurfaceGraph > * pList,
                      TPermutation * pPerm,
                      const char *filename, int dev )
{
    if (pList == NULL)
    {
        cout << "TDrawApp::drawGraph (...) : got NULL arg" << endl;
        return;
    }// if

    T3DGraphView   view( pList );

    view.setPermutation( pPerm );
    
    draw( &view, filename, dev );
}

void
TDrawApp::drawGraph ( T3DGraph * pGraph,
                      const char *filename, int dev )
{
    drawGraph( pGraph, NULL, filename, dev );
}

void
TDrawApp::drawGraph ( T3DGraph * pGraph, TPermutation * pPerm,
                      const char *filename, int dev )
{
    if (pGraph == NULL)
    {
        cout << "TDrawApp::drawGraph (...) : got NULL graph" << endl;
        return;
    }// if

    T3DGraphView   view( pGraph );

    view.setPermutation( pPerm );
    
    draw( &view, filename, dev );
}

void
TDrawApp::drawFlowSurfaces ( TList< T3DCellEdgeGraph > * pFSList,
                             const char *filename, int dev )
{
    T3DGraphView   view( pFSList );

    if (pFSList == NULL)
    {
        cout << "TDrawApp::drawFlowSurfaces (...) : got NULL flow surfaces" << endl;
        return;
    }// if

    draw( &view, filename, dev );
}

void
TDrawApp::drawCycles ( TList< TGraph > * pListCycles,
                       const char *filename,
                       int dev )
{
    T3DGraphView   view( pListCycles );

    if (pListCycles == NULL)
    {
        cout << "TDrawApp::drawCycles (...) : got NULL cyclelist" << endl;
        return;
    }// if

    draw( &view, filename, dev );
}


////////////////////////////////////////////
//
// draw flow
//

void
TDrawApp::drawFlow ( T2DBasisEqn * pEqn,
                     TFloat stepX, TFloat stepY,
                     const char * filename )
{
    T2DFlowView  view( pEqn, stepX, stepY );
    
    if (pEqn == NULL)
    {
        cout << "TDrawApp::drawFlow : got NULL equation" << endl;
        return;
    }// if

    drawPS( &view, filename );
}

void
TDrawApp::drawFlow ( TConvEqn<3> * pEqn,
                     TFloat stepX, TFloat stepY, TFloat stepZ,
                     const char * filename, int dev )
{
    T3DFlowView  view( pEqn, stepX, stepY, stepZ );

    if (pEqn == NULL)
    {
        cout << "TDrawApp::drawFlow (...) : got NULL equation" << endl;
        return;
    }// if

    draw( &view, filename, dev );
}

////////////////////////////////////////////
//
// draw streaklines
//
// void 
// TDrawApp::drawStreakLine ( T2DNet &net, TBlockVector<TFloat> &velocity,
//                            T2DBasisEqn &eqn,
//                            const char *filename )
// {
//     TStreakLineView  view( net, velocity, eqn );
    
//     view.buildTraces( 0.05, 200 );
//     drawPS( & view, filename );
// }

////////////////////////////////////////////
//
// draw convergence rates
//
void 
TDrawApp::drawConv ( TDynArray<TFloat> *parrConv, const char *filename )
{
    TConvView  view( parrConv );

    drawPS( & view, filename );
}

////////////////////////////////////////////
//
// misc. helpfull methods
//

//
// draw given view via postscript
//
void
TDrawApp::drawPS ( TView * pView, const char * filename )
{
    TPrinter  * pPrinter;
    char        szFilename[256];

    if ((pView == NULL) || (filename == NULL))
        return;
    
    sprintf( szFilename, "%s/%s.eps", DATADIR, filename );
    pPrinter = new TPrinter( szFilename );

    pPrinter->begin();
    pView->draw( *pPrinter );
    pPrinter->end();

    delete pPrinter;
}

//
// draw given view according to device
void
TDrawApp::draw ( TView * pView, const char * filename, int dev )
{
    if (dev & VRML)
        drawVRML( pView, filename );
    
    if (dev & POVRAY)
        drawPOVRay( pView, filename );
}
    
//
// draw given view via VRML
//
void
TDrawApp::drawVRML ( TView * pView, const char *filename )
{
    T3DPrinter  *pPrinter;
    char         szFilename[256];

    if ((pView == NULL) || (filename == NULL))
        return;
    
    sprintf( szFilename, "%s/%s.wrl", DATADIR, filename );
    pPrinter = new TVRML( szFilename );

    pPrinter->begin();
    pView->draw3D( *pPrinter );
    pPrinter->end();

    delete pPrinter;
}

//
// draw given view via POVRay
//
void
TDrawApp::drawPOVRay ( TView * pView, const char *filename )
{
    T3DPrinter  *pPrinter;
    char         szFilename[256];

    if ((pView == NULL) || (filename == NULL))
        return;
    
    sprintf( szFilename, "%s/%s.pov", DATADIR, filename );
    pPrinter = new TPOVRay( szFilename );

    pPrinter->begin();
    pView->draw3D( *pPrinter );
    pPrinter->end();

    delete pPrinter;
}
