//
//  Project :  libvisual
//  File    :  graphview.cc
//  Purpose :  class for visualization of a graph
//  Author  :  Ronald Kriemann / Sabine Gutsch
//
// arch-tag: 21b7288f-58b9-4440-9b6f-c6f20e5eb10f
//

#include <stdio.h>
#include <math.h>

#include "vector.hh"

#include "nodes.hh"
#include "T2DPlanarNode.hh"
#include "TPlanarCycle.hh"
#include "graphtool.hh"
#include "fvs.hh"

#include "graphview.hh"

// cos and sin of the angle for the arrow (20 deg)
#define COSANGLE  0.9396936208
#define SINANGLE  0.3420201433

// type of cycles to draw (min=0 or max=1)
#define TYPE  0

////////////////////////////////////////////////////
//
// some extra functions
//
uint 
pow_2(int bn) 
{
	// calculate power of 2:  2^(bn-1) mod ns
	int ns = 16;

	while (bn > ns)
		bn -= ns;

	return (uint)(floor(pow(2,bn-1)+0.5));
}

TFloat 
getLargeSector(uint bf)
{
	int    ns = 16, i;
	uint   j, k;
	int    c_sind = 1, c_len = 0;
	int    m_sind = 1, m_len = 0;
	TFloat rv;

	if (bf & pow_2(1))
		k = 1;
	else {
		k      = 0;
		c_sind = 1; 
		c_len  = 1;
	}// else

	for (i = 2; i <= 2*ns; i++)
	{
		if (bf & pow_2(i)) j = 1; 
		else j = 0;
		
		if ((k == 1) && (j == 0)) {
			// free sector starts;
			c_len  = 1; 
			c_sind = i; 
			k      = 0;
		}// if
		else
			if ((k == 0) && (j == 0)) {
				// free sector continues
				c_len++;
			} else
				if ((k == 0) && (j == 1)) {
					// free sector ends;
					k = 1;
					if (c_len > m_len) {
						m_len  = c_len;
						m_sind = c_sind;
					}// if
					c_len = 0;
				}// if
		//(else):  blocked sector continues
	}// fir

	if (c_len > m_len) {
		m_len  = c_len;
		m_sind = c_sind;
	}// if
	
	if (m_len > 0) {
		rv = m_sind + 0.5 * m_len;
		if (rv > ns) rv -= ns;
	}// if
	else 
		rv = 1e6;
	
	return rv;
}

////////////////////////////////////////////////////
//
// now the class
//

//
// constructor
//
TGraphView::TGraphView ( TGraph& graph )
{
    pGraph        = &graph;
    pPermuteBlock = NULL;
    plistCycNodes = NULL;
    plistCycle    = NULL;
    
    minDist = minimalDist( pGraph );
    
    if (minDist > 1e39)
        minDist = 0.5;
    
    bDrawDeleted = FALSE;
    
}

TGraphView::TGraphView ( TGraph& graph, TVectorBlock<uint>& permutation )
{
	pGraph        = &graph;
	pPermuteBlock = &permutation;
	plistCycNodes = NULL;
	plistCycle    = NULL;

    minDist = minimalDist( pGraph );
    
    if (minDist > 1e39)
        minDist = 0.5;

    bDrawDeleted = FALSE;
}

TGraphView::TGraphView ( TList<TGraph>& cycles )
{
	plistCycle    = &cycles;
	pGraph        = NULL;
	pPermuteBlock = NULL;

    minDist = 1e40;

	if (cycles.getSize() > 0)
	{
		TListIterator<TGraph>  iterCycles( cycles );
		TGraph                *pCycle;
        TFloat                 minimum;

		minDist = 1e40;

		while ((pCycle = iterCycles++) != NULL)
		{
            minimum = minimalDist( pCycle );
            
            if (minimum < minDist)
                minDist = minimum;
		}// while
	}// if
	else
		plistCycle = NULL;

    bDrawDeleted = FALSE;
}

TGraphView::TGraphView ( TList<TGraph>& cycles, TVectorBlock<uint>& permutation )
{
	plistCycle    = &cycles;
	pGraph        = NULL;
	plistCycNodes = NULL;
	pPermuteBlock = &permutation;

    minDist = 1e40;

	if (cycles.getSize() > 0)
	{
		TListIterator<TGraph>  iterCycles( cycles );
		TGraph                *pCycle;
        TFloat                 minimum;

		minDist = 1e40;

		while ((pCycle = iterCycles++) != NULL)
		{
            minimum = minimalDist( pCycle );
            
            if (minimum < minDist)
                minDist = minimum;
		}// while
	}// if
	else
		plistCycle = NULL;

    bDrawDeleted = FALSE;
}

TGraphView::TGraphView ( TList<TNode>& listCycNodes )
{
	plistCycNodes    = &listCycNodes;
	plistCycle       = NULL;
	pGraph           = NULL;
	pPermuteBlock    = NULL;

    minDist = 1e40;

	if (listCycNodes.getSize() > 0)
	{
		TListIterator<TNode>   iterCycles( listCycNodes );
		TCycleNode            *pCycNode;
		TGraph                *pCycle;
        TFloat                 minimum;

		minDist = 1e40;

		while ((pCycNode = (TCycleNode*) iterCycles++) != NULL)
		{
			pCycle = pCycNode->cycle();

            minimum = minimalDist( pCycle );
            
            if (minimum < minDist)
                minDist = minimum;
		}// while
	}// if
	else
		plistCycNodes = NULL;

    bDrawDeleted = FALSE;
}

//
// draw the view
//
void
TGraphView::draw ( TPrinter& p )
{
	if (pGraph != NULL)
		drawGraph( p );
	else if ((plistCycle != NULL) || (plistCycNodes != NULL))
		drawCycles( p );
}

//
// draw the graph
//
void 
TGraphView::drawGraph ( TPrinter& p )
{
	// fontsize with respect to the min node distance
	int  fntSize = int(fabs(p.getView().getScaleY() * minDist) /3.5);
	// distance between node and name
	TFloat  fntDist = 0, fntFactor = 0;
  
	// if too small, no names will be drawn
	bool bDrawNames;
	char szNodeName[18];
  
	if (fntSize > 12) fntSize = 12;
  
	if (fntSize < 4)
		bDrawNames = FALSE;
	else
    {
		bDrawNames = TRUE;
		fntFactor=  fabs(minDist /6.3);
		fntDist = fabs(minDist / 10.0);
		p.setFont( "Courier", fntSize );
    }// else
  
    bDrawNames = FALSE;

    uint   nodeName,SuccNodeName;
  
	TVector<2> vecNode, vecDir, vecSucc, offSet;
  
	TFloat  origX, origY, succX, succY, rX, rY;
	TFloat offX,offY,aalpha; 
	// angle in degrees 
	int alpha;
	// number of sectors (only even numbers allowed! )
	int noOfSectors = 16;
	int noOfSectors_2 = noOfSectors/2;
	// sectorsize in degree
	int  Secsize = (int)floor(360.0/noOfSectors+0.5);
	int blockNo1,blockNo2, i, noOfdigits; 
	TFloat ac,as;

	p.setRGBColor( COLOR_BLACK );
	p.setLineWidth( 1 );
  
	TListIterator<TNode>  iter( pGraph->getNodeList() );
	TListIterator<TEdge>  iterEdges;
	TEdge                *pEdge;
	T2DPlanarNode        *pNode, *pSuccNode;

	// determine the largest nodeName;
	iter.first();
	uint maxNodeName=0;
	while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
		nodeName = pNode->getName();
		maxNodeName=MAX(maxNodeName,nodeName);

		iterEdges.setList(pNode->getOutEdges());
		iterEdges.first();
		while ((pEdge = iterEdges++) != NULL)
			maxNodeName = MAX(maxNodeName,pEdge->getHead()->getName());
    }// while
	TVectorBlock<uint> *areaCoding = new TVectorBlock<uint>(0,maxNodeName+1);
  
	iter.first();
	while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
        if ( ! bDrawDeleted && pNode->isDeleted( TYPE ))
            continue;
        
		nodeName = pNode->getName();
		vecNode  = *(pNode->vector());
      
		// coord of the node
		origX  = vecNode[X];
		origY  = vecNode[Y];
      
		// draw the edges to all succ
		iterEdges.setList( pNode->getOutEdges() );
		iterEdges.first();
		while ((pEdge = iterEdges++) != NULL)
        {
			pSuccNode = (T2DPlanarNode*) pEdge->getNode();

            if ( ! bDrawDeleted && pSuccNode->isDeleted( TYPE ))
                continue;
        
			SuccNodeName=pSuccNode->getName();
			if (SuccNodeName!=nodeName)
            {
				vecSucc = *(pSuccNode->vector());
              
				// coord of the succ-node
				succX = vecSucc[X];
				succY = vecSucc[Y];
              
				// draw the edge
				p.drawLine( origX, origY, succX, succY );
              
				// now show direction
				vecDir = vecNode - vecSucc;
              
				vecDir *= 1.0 / vecDir.euclidNorm();
				vecDir *= minDist / 4.0;
              
				// the arrow on the edge (which shows the direction) 
				// starts not at the vertex coordinats of the vertex it points to, 
				// but on the border of the bullet which denots the vertex.
				// Therefore, an offSet to the vertex coordinates (vecSucc) has to be 
				// calculated, depending on the bullet's diameter  (minDist/8.5)
				// So, offSet= 0.5*minDist / 4.0* (1.0 / vecDir.euclidNorm()) * vecDir 
				// is o.k.
              
				offSet= 0.5*vecDir;
				offX=offSet[X]; offY= offSet[Y];
              
				// rotate dir-vector
				rX = vecDir[X] * COSANGLE - vecDir[Y] * SINANGLE;
				rY = vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
				p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );
              
				rX =   vecDir[X] * COSANGLE + vecDir[Y] * SINANGLE;
				rY = - vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
				p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );


				// code area

				// alpha is the angle in degrees (0..360) of the vecDir and
				// the x-axis in positiv direction.

				aalpha= atan2(-vecDir[Y],-vecDir[X]);
				if (aalpha>0)
					alpha = (int)floor(aalpha*180/PI+0.5);
				else
					alpha =360+ (int)floor(aalpha*180/PI-0.5); 
                
				// calculate, which sectors are to be blocked ( +- 20 degree 
				// around direction)
				blockNo1 = (int)floor((alpha-20.0)*1.0/Secsize)+1;
				blockNo2 = (int)floor((alpha+20.0)*1.0/Secsize)+1;
              
				// block sectors by bitcoding
				for (i= blockNo1; i <= blockNo2;i++)
					(*areaCoding)[nodeName] = ((*areaCoding)[nodeName] | pow_2(i));
          
				// block opposite direction in the sector of the  sucsessing node
				blockNo1+=noOfSectors_2;
				blockNo2+=noOfSectors_2;
             
				for (i= blockNo1; i <= blockNo2;i++)
					(*areaCoding)[SuccNodeName] = ((*areaCoding)[SuccNodeName] |
												   pow_2(i));
			}// if
        }// while
    }// while
  
	// now paint all nodes 
	// (therefore the mark is visible, we paint them over the edges)
	iter.first();
	while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
        if ( ! bDrawDeleted && pNode->isDeleted( TYPE ))
            continue;

		nodeName = pNode->getName();
		vecNode  = *(pNode->vector());
      
		// coord of the node
		origX  = vecNode[X];
		origY  = vecNode[Y];
      
		// mark the node
		switch (pNode->mark( TYPE ))
        {
            case C:          p.setRGBColor( COLOR_RED );   p.setLineWidth( 1 ); break;
            case F:          p.setRGBColor( COLOR_BLUE );  p.setLineWidth( 1 ); break;
            case L:          p.setRGBColor( COLOR_GREEN ); p.setLineWidth( 1 ); break;
            case FVS_NODE:   p.setRGBColor( COLOR_YELLOW ); p.setLineWidth( 2 ); break;
            default:         p.setRGBColor( COLOR_BLACK ); break;
        }// switch
		p.drawFilledArc( origX, origY, minDist/8.5, 0, 360 );
      
		// and the name
		if (bDrawNames == TRUE)
        { 
            p.setRGBColor( COLOR_BLACK );
            p.setLineWidth( 0 );
            
			// find the free largest sector for labeling;
			uint arc=(*areaCoding)[nodeName];
			TFloat fs=getLargeSector(arc);
			if (fs > 1e5) {
				cerr << "TGraphView::drawGraph( TPrinter& ) : "
					 << "no possible space for node "
					 << nodeName << endl;
				// take standardplace
				fs = 1;
			}// if

			ac   = cos((fs-1)*Secsize/360*2*PI); as=sin((fs-1)*Secsize/360*2*PI);
			offX = ac; 
			offY = as;
          
			// calculate number of digits of the label
			if (pPermuteBlock != NULL) {
				if(((*pPermuteBlock)[nodeName]) == 0)
					noOfdigits = 1; 
				else
					noOfdigits = (int)floor(log10((*pPermuteBlock)[nodeName]))+1;
			}// if
			else {
				if (nodeName == 0)
					noOfdigits = 1; 
				else
					noOfdigits = (int)floor(log10(nodeName))+1;
			}// else

			// calculate offsett to move the label
			offX = fntDist*ac;
			offY = fntDist*as;
			// move by width of the label, if label is left of the node
			if (offX < 0) offX -= (noOfdigits*fntFactor);
			// move by height of the label, if label is below the  node
			if (offY < 0) offY -= 1.3*fntFactor;
          
			if (pPermuteBlock != NULL)
				sprintf( szNodeName, "%d", (*pPermuteBlock)[nodeName] );
			else
				sprintf( szNodeName, "%d", nodeName );
			p.drawText( origX + offX, origY + offY,   szNodeName );
        }// if
    }// for

	delete areaCoding;
}

//
// draw the cycles
//
void 
TGraphView::drawCycles ( TPrinter& p )
{
	// fontsize with respect to the min node distance
	int  fntSize = int(fabs(p.getView().getScaleY() * minDist) / 3.5);
	// distance between node and name
	TFloat  fntDist = 0.0, fntFactor = 0.0;
	
	// if too small, no names will be drawn
	bool bDrawNames;
	char szNodeName[18];

	if (fntSize > 12) fntSize = 12;

	if (fntSize < 4)
		bDrawNames = FALSE;
	else
	{
		bDrawNames = TRUE;
		fntFactor  = fabs(minDist /6.3);
		fntDist    = fabs(minDist / 10.0);
		p.setFont( "Courier", fntSize );
	}// else

	uint   nodeName,SuccNodeName;;
	
	TVector<2> vecNode, vecDir, vecSucc, offSet;
	TFloat  origX, origY, succX, succY, rX, rY;
	
	TFloat offX,offY,aalpha; 
	// angle in degrees 
	int alpha;
	// number of sectors (only even numbers allowed! )
	int noOfSectors = 16;
	int noOfSectors_2 = noOfSectors/2;
	// sectorsize in degree
	int Secsize = (int)floor(360.0/noOfSectors+0.5);
 	int blockNo1,blockNo2, i, noOfdigits; 
	TFloat ac,as;
	
	p.setRGBColor( COLOR_BLACK );
	p.setLineWidth( 1 );

	TListIterator<TGraph> iterCycles;
	TGraph               *pCycle = NULL;

	TListIterator<TNode>  iterCycNode;
	TCycleNode           *pCycNode;

	TListIterator<TNode>  iter;
	T2DPlanarNode        *pNode, *pSuccNode, *pFirstNode;

	TListIterator<TEdge>  iterEdges;
	TEdge                *pEdge;

	uint maxNodeName=0;
 
	if (plistCycle != NULL)
	{
		iterCycles.setList( *plistCycle );
		iterCycles.first();
		while ((pCycle = iterCycles++) != NULL)
		{
			iter.setList( pCycle->getNodeList() );
			iter.first();
			while ((pNode = (T2DPlanarNode*) iter++) != NULL)
			{
				nodeName = pNode->getName();
				maxNodeName=MAX(maxNodeName,nodeName);
				
				iterEdges.setList(pNode->getOutEdges());
				iterEdges.first();
				while ((pEdge = iterEdges++) != NULL)
					maxNodeName = MAX(maxNodeName,pEdge->getHead()->getName());
			}// while
		}// while
	}// if
	else if (plistCycNodes != NULL)
	{
		iterCycNode.setList( *plistCycNodes );
		iterCycNode.first();
		while ((pCycNode = (TCycleNode*)iterCycNode++) != NULL)
		{
			pCycle = pCycNode->cycle();
			iter.setList( pCycle->getNodeList() );
			iter.first();
			while ((pNode = (T2DPlanarNode*) iter++) != NULL)
			{
				nodeName = pNode->getName();
				maxNodeName=MAX(maxNodeName,nodeName);
				
				iterEdges.setList(pNode->getOutEdges());
				iterEdges.first();
				while ((pEdge = iterEdges++) != NULL)
					maxNodeName = MAX(maxNodeName,pEdge->getHead()->getName());
			}// while
		}// while
	}// if

	TVectorBlock<uint> *areaCoding = new TVectorBlock<uint>(0,maxNodeName+1);

	// go over all cycles and draw them
	if (plistCycle != NULL)
		iterCycles.first();
	else if (plistCycNodes != NULL)
		iterCycNode.first();

	do 
	{
		if (plistCycle != NULL)
		{
			pCycle = iterCycles++;
		}// if
		else if (plistCycNodes != NULL)
		{
			pCycNode = (TCycleNode*) iterCycNode++;
			if (pCycNode != NULL)
				pCycle = pCycNode->cycle();
			else
				pCycle = NULL;
		}// if

		if (pCycle != NULL)
		{
			iter.setList( pCycle->getNodeList() );
			
			pNode = pFirstNode = (T2DPlanarNode*) iter.first();
			iter++;
			
			while ((pSuccNode = (T2DPlanarNode*) iter++) != NULL){
				
				vecNode = *(pNode->vector());
				vecSucc = *(pSuccNode->vector());
				SuccNodeName=pSuccNode->getName();
				nodeName = pNode->getName();
				
				// coord of the tail
				origX = vecSucc[X];
				origY = vecSucc[Y];
				
				// coord of the head
				succX = vecNode[X];
				succY = vecNode[Y];
				
				// draw the edge
				p.drawLine( origX, origY, succX, succY );
				
				// now show direction
				vecDir = vecSucc - vecNode;
				
				vecDir *= 1.0 / vecDir.euclidNorm();
				vecDir *= minDist / 4.0;
				
				// the arrow on the edge (which shows the direction) 
				// starts not at the vertex coordinats of the vertex it points to, 
				// but on the border of the bullet which denots the vertex.
				// Therefore, an offSet to the vertex coordinates (vecSucc) has to be 
				// calculated, depending on the bullet's diameter  (minDist/8.5)
				// So, offSet= 0.5*minDist / 4.0* (1.0 / vecDir.euclidNorm()) * vecDir 
				// is o.k.
				
				offSet= 0.5*vecDir;
				offX=offSet[X]; offY= offSet[Y];
				
				// rotate dir-vector
				rX = vecDir[X] * COSANGLE - vecDir[Y] * SINANGLE;
				rY = vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
				p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );
				
				rX =   vecDir[X] * COSANGLE + vecDir[Y] * SINANGLE;
				rY = - vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
				p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );
         
				// code area

				// alpha is the angle in degrees (0..360) of the vecDir and
				// the x-axis in positiv direction.

				aalpha = atan2(-vecDir[Y],-vecDir[X]);
				if (aalpha > 0)
					alpha = (int)floor(aalpha*180/PI+0.5);
				else
					alpha = 360+ (int)floor(aalpha*180/PI-0.5); 
                
				// calculate, which sectors are to be blocked ( +- 20 degree 
				// around direction)
				blockNo1 = (int)floor((alpha-20.0)*1.0/Secsize)+1;
				blockNo2 = (int)floor((alpha+20.0)*1.0/Secsize)+1;
              
				// block sectors of SuccNode by bitcoding
				for (i = blockNo1; i <= blockNo2; i++)
					(*areaCoding)[SuccNodeName] = ((*areaCoding)[SuccNodeName] | pow_2(i));
          
				// block opposite direction in the sector of the originating  node
				blockNo1 += noOfSectors_2;
				blockNo2 += noOfSectors_2;
             
				for (i = blockNo1; i <= blockNo2; i++)
					(*areaCoding)[nodeName] = ((*areaCoding)[nodeName] |  pow_2(i));
				pNode = pSuccNode;
			}// while

			// and from the last node to the first node
			vecNode = *(pNode->vector());
			vecSucc = *(pFirstNode->vector());
 
			SuccNodeName=pFirstNode->getName();
			nodeName = pNode->getName();
    
			// coord of the tail
			origX = vecSucc[X];
			origY = vecSucc[Y];
			
			// coord of the head
			succX = vecNode[X];
			succY = vecNode[Y];

			// draw the edge
			p.drawLine( origX, origY, succX, succY );

			// now show direction
			vecDir =  vecSucc- vecNode;

			vecDir *= 1.0 / vecDir.euclidNorm();
			vecDir *= minDist / 4.0;

			// the arrow on the edge (which shows the direction) 
			// starts not at the vertex coordinats of the vertex it points to, 
			// but on the border of the bullet which denots the vertex.
			// Therefore, an offSet to the vertex coordinates (vecSucc) has to 
			// be calculated, depending on the bullet's diameter  (minDist/8.5)
			// So, offSet= 0.5*minDist / 4.0* (1.0 / vecDir.euclidNorm()) * vecDir 
			// is o.k.
         
			offSet= 0.5*vecDir;
			offX=offSet[X]; offY= offSet[Y];
         
			// rotate dir-vector
			rX = vecDir[X] * COSANGLE - vecDir[Y] * SINANGLE;
			rY = vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
			p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );
         
			rX =   vecDir[X] * COSANGLE + vecDir[Y] * SINANGLE;
			rY = - vecDir[X] * SINANGLE + vecDir[Y] * COSANGLE;
			p.drawLine( succX+offX, succY+offY, succX + rX+offX, succY + rY+offY );
         
			// code area
         
			// alpha is the angle in degrees (0..360) of the vecDir and
			// the x-axis in positiv direction.
         
			aalpha = atan2(-vecDir[Y],-vecDir[X]);
			if (aalpha > 0)
				alpha = (int)floor(aalpha*180/PI+0.5);
			else
				alpha = 360+ (int)floor(aalpha*180/PI-0.5); 
         
			// calculate, which sectors are to be blocked ( +- 20 degree around direction)
			blockNo1 = (int)floor((alpha-20.0)*1.0/Secsize)+1;
			blockNo2 = (int)floor((alpha+20.0)*1.0/Secsize)+1;
         
			// block sectors of SuccNode by bitcoding
			for (i = blockNo1; i <= blockNo2; i++)
				(*areaCoding)[SuccNodeName] = ((*areaCoding)[SuccNodeName] | pow_2(i));
         
			// block opposite direction in the sector of the  originating node
			blockNo1 += noOfSectors_2;
			blockNo2 += noOfSectors_2;
         
			for (i = blockNo1; i <= blockNo2; i++)
				(*areaCoding)[nodeName] = ((*areaCoding)[nodeName] |  pow_2(i));
		}// if
	} while (pCycle != NULL);

	// now paint all nodes 
	// (therefore the mark is visible, we paint them over the edges)
	if (plistCycle != NULL)
		iterCycles.first();
	else if (plistCycNodes != NULL)
		iterCycNode.first();

	do 
	{
		if (plistCycle != NULL)
		{
			pCycle = iterCycles++;
		}// if
		else if (plistCycNodes != NULL)
		{
			pCycNode = (TCycleNode*) iterCycNode++;
			if (pCycNode != NULL)
				pCycle = pCycNode->cycle();
			else
				pCycle = NULL;
		}// if

		if (pCycle != NULL)
		{
			iter.setList( pCycle->getNodeList() );
			iter.first();
			while ((pNode = (T2DPlanarNode*) iter++) != NULL)
			{
				nodeName = pNode->getName();
				vecNode  = *(pNode->vector());

				// coord of the node
				origX  = vecNode[X];
				origY  = vecNode[Y];

				// mark the node
				switch (pNode->mark( TYPE ))
				{
                    case C:          p.setRGBColor( COLOR_RED );   p.setLineWidth( 1 ); break;
                    case F:          p.setRGBColor( COLOR_BLUE );  p.setLineWidth( 1 ); break;
                    case L:          p.setRGBColor( COLOR_GREEN ); p.setLineWidth( 1 ); break;
                    case FVS_NODE:   p.setRGBColor( COLOR_YELLOW ); p.setLineWidth( 2 ); break;
                    default:         p.setRGBColor( COLOR_BLACK ); break;
				}// switch
				p.drawFilledArc( origX, origY, minDist/8.5, 0, 360 );

				// and the name
				if (bDrawNames == TRUE)
				{ 
                    p.setRGBColor( COLOR_BLACK );
                    p.setLineWidth( 0 );
                    
					// find the free largest sector for labeling;
					uint  arc = (*areaCoding)[nodeName];
					TFloat fs  = getLargeSector(arc);
					if (fs > 1e5) {
						cerr << "TGraphView::drawCycles(TPrinter&) : "
							 << "Nn possible space for node "
							 << nodeName << endl;
						// take standardplace
						fs = 1;
					}// if

					ac   = cos((fs-1)*Secsize/360*2*PI); as=sin((fs-1)*Secsize/360*2*PI);
					offX = ac; 
					offY = as;
          
					// calculate number of digits of the label
					if (pPermuteBlock != NULL){
						if(((*pPermuteBlock)[nodeName]) == 0)
							noOfdigits = 1; 
						else
							noOfdigits = (int)floor(log10((*pPermuteBlock)[nodeName]))+1;
					}// if
					else {
						if (nodeName == 0)
							noOfdigits = 1; 
						else
							noOfdigits = (int)floor(log10(nodeName))+1;
					}// else

					// calculate offsett to move the label
					offX = fntDist*ac;
					offY = fntDist*as;
					// move by width of the label, if label is left of the node
					if (offX < 0) offX -= (noOfdigits*fntFactor);
					// move by height of the label, if label is below the  node
					if (offY < 0) offY -= 1.3*fntFactor;
          
					if (pPermuteBlock != NULL)
						sprintf( szNodeName, "%d", (*pPermuteBlock)[nodeName] );
					else
						sprintf( szNodeName, "%d", nodeName );
					p.drawText( origX + offX, origY + offY,   szNodeName );
				}// if
			}// while
		}// if
	} while (pCycle != NULL);

	delete areaCoding;
}

//
// set drawing of deleted nodes
//
void
TGraphView::setDrawDeleted ( bool b )
{
    bDrawDeleted = b;
}

//
// determine minimal distance between nodes in graph
//
TFloat
TGraphView::minimalDist ( TGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "TGraphView::minDist (...) : got NULL arg" << endl;
        return 0.0;
    }// if

	// iterate over graph to get max distance between two nodes
	TListIterator<TNode>   iterNodes;
	TListIterator<TEdge>   iterEdges;
    TEdge                * pEdge;
	T2DPlanarNode        * pNode, * pAdjNode;;
	T2DVector              vecNode, vecAdjNode;
    TFloat                 minimum;

	minimum = 1e40;

    iterNodes.setList( pGraph->getNodeList() );
	iterNodes.first();
	while ((pNode = (T2DPlanarNode*) iterNodes++) != NULL)
	{
		vecNode = *(pNode->vector());

        iterEdges.setList( pNode->getOutEdges() );
		iterEdges.first();
		while ((pEdge = iterEdges++) != NULL)
		{
            pAdjNode = (T2DPlanarNode*) pEdge->getHead();
            
			if (pNode->getName() != pAdjNode->getName())
            {
				vecAdjNode  = *(pAdjNode->vector());
				vecAdjNode -= vecNode;
				
				if (vecAdjNode.euclidNorm() < minimum)
					minimum = vecAdjNode.euclidNorm();
			}// if
		}// while

        iterEdges.setList( pNode->getInEdges() );
		iterEdges.first();
		while ((pEdge = iterEdges++) != NULL)
		{
            pAdjNode = (T2DPlanarNode*) pEdge->getTail();
            
			if (pNode->getName() != pAdjNode->getName())
            {
				vecAdjNode  = *(pAdjNode->vector());
				vecAdjNode -= vecNode;
				
				if (vecAdjNode.euclidNorm() < minimum)
					minimum = vecAdjNode.euclidNorm();
			}// if
		}// while
	}// while

    return minimum;
}
    
