//
//  Project :  libvisual
//  File    :  dnetview.cc
//  Purpose :  class for net-visualization
//  Author  :  Ronald Kriemann / Sabine Gutsch
//
// arch-tag: 01b69c6a-510a-4af7-b059-6dfd5010efac
//

extern "C" {
#include <stdio.h>
#include <math.h>
}

#include "T3DFrontier.hh"
#include "T3DCellEdge.hh"

#include "netview.hh"

typedef  TCell<2>  T2DCell;

// #define STRETCH_CELL

//
// constructor
//
TNetView::TNetView ( T2DNet& net )
{
    p3DNet   = NULL;
    p2DNet   = &net;
    pMeshGen = NULL;
    
    maxInteriorName = p2DNet->getNoOfInteriorNodes();
    pPermuteBlock = NULL;

    // get the min (max) distance between two vertices
    TListIterator<T2DCell>  iterCells( net.getList() );
    T2DCell                *pCell;
    TFloat                  distV0_V1, distV1_V2, distV2_V0;

    minDistNodes = 1.0;
    maxDistNodes = 0.0;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        distV0_V1 = ((*(pCell->getNode(0))) - (*(pCell->getNode(1)))).euclidNorm();
        distV1_V2 = ((*(pCell->getNode(1))) - (*(pCell->getNode(2)))).euclidNorm();
        distV2_V0 = ((*(pCell->getNode(2))) - (*(pCell->getNode(0)))).euclidNorm();
        
        if (minDistNodes > distV0_V1) minDistNodes = distV0_V1;
        if (minDistNodes > distV1_V2) minDistNodes = distV1_V2;
        if (minDistNodes > distV2_V0) minDistNodes = distV2_V0;

        if (maxDistNodes < distV0_V1) maxDistNodes = distV0_V1;
        if (maxDistNodes < distV1_V2) maxDistNodes = distV1_V2;
        if (maxDistNodes < distV2_V0) maxDistNodes = distV2_V0;
    }// while
}

TNetView::TNetView ( T2DNet& net, TVectorBlock<uint>& permutation )
{
    p3DNet = NULL;
    p2DNet = &net;
    maxInteriorName = p2DNet->getNoOfInteriorNodes();
    pPermuteBlock = &permutation;

    // get the min (max) distance between two vertices
    TListIterator<T2DCell>  iterCells( net.getList() );
    T2DCell                *pCell;
    TFloat                  distV0_V1, distV1_V2, distV2_V0;

    minDistNodes = 1.0;
    maxDistNodes = 0.0;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        distV0_V1 = ((*(pCell->getNode(0))) - (*(pCell->getNode(1)))).euclidNorm();
        distV1_V2 = ((*(pCell->getNode(1))) - (*(pCell->getNode(2)))).euclidNorm();
        distV2_V0 = ((*(pCell->getNode(2))) - (*(pCell->getNode(0)))).euclidNorm();
        
        if (minDistNodes > distV0_V1) minDistNodes = distV0_V1;
        if (minDistNodes > distV1_V2) minDistNodes = distV1_V2;
        if (minDistNodes > distV2_V0) minDistNodes = distV2_V0;

        if (maxDistNodes < distV0_V1) maxDistNodes = distV0_V1;
        if (maxDistNodes < distV1_V2) maxDistNodes = distV1_V2;
        if (maxDistNodes < distV2_V0) maxDistNodes = distV2_V0;
    }// while
}

TNetView::TNetView ( T3DNet * net )
{
    p2DNet   = NULL;
    p3DNet   = net;
    pMeshGen = NULL;
}

TNetView::TNetView ( T3DMeshGenerator * mg )
{
    p2DNet   = NULL;
    p3DNet   = NULL;
    pMeshGen = mg;
}

//
// draw the net
//
void 
TNetView::draw ( TPrinter& p )
{
    T2DVertex *pV0, *pV1, *pV2;
    uint     nodeName;
    uint     permuteSize;

    TListIterator<TCell<2> > iterCell( p2DNet->getList() );
    TCell<2>                *cell;

    // fontsize with respect to the min node distance
    int  fntSize;
    
    // distance between node and name
    TFloat  fntDist = 0.0;

    // if too small, no names will be drawn
    bool bDrawNames;
    char szNodeName[32];

    fntSize = int(fabs(p.getView().getScaleY() * minDistNodes) / 3.5);
    
    if (fntSize > 12)
        fntSize = 12;

    if (fntSize < 3)
        bDrawNames = FALSE;
    else
    {
        bDrawNames = TRUE;
        fntDist = fabs(minDistNodes / 25.0);
        p.setFont( "Courier", fntSize );
    }// else

//    bDrawNames = FALSE;
    
    p.setRGBColor( COLOR_BLACK );
    p.setLineWidth( 0 );

    permuteSize = pPermuteBlock != NULL ? pPermuteBlock->getSize() : 0;
    
    iterCell.first();
    while ((cell = iterCell++) != NULL)
    {
        pV0 = cell->getNode(0);
        pV1 = cell->getNode(1);
        pV2 = cell->getNode(2);

        // draw names of the nodes
        if (bDrawNames)
        {
            nodeName = pV0->getName();
            if (nodeName < BOUNDARY_VERTEX/2)
                if (pPermuteBlock != NULL)
                    sprintf( szNodeName, "%u", (*pPermuteBlock)[nodeName]);
                else
                    sprintf( szNodeName, "%u", nodeName);
            else
            {
                nodeName = BOUNDARY_VERTEX - nodeName + maxInteriorName;
                if ((pPermuteBlock != NULL) && (nodeName < permuteSize))
                    sprintf( szNodeName, "B(%u)", (*pPermuteBlock)[nodeName] );
                else
                    sprintf( szNodeName, "B(%u)", nodeName );
            }// else
            p.drawText( (*pV0)[X] + fntDist, (*pV0)[Y] + fntDist, szNodeName );

            nodeName = pV1->getName();
            if (nodeName < BOUNDARY_VERTEX/2)
                if (pPermuteBlock != NULL)
                    sprintf( szNodeName, "%u", (*pPermuteBlock)[nodeName]);
                else
                    sprintf( szNodeName, "%u", nodeName);
            else
            {
                nodeName = BOUNDARY_VERTEX - nodeName + maxInteriorName;
                if ((pPermuteBlock != NULL) && (nodeName < permuteSize))
                    sprintf( szNodeName, "B(%u)", (*pPermuteBlock)[nodeName] );
                else
                    sprintf( szNodeName, "B(%u)", nodeName );
            }// else
            p.drawText( (*pV1)[X] + fntDist, (*pV1)[Y] + fntDist, szNodeName );

            nodeName = pV2->getName();
            if (nodeName < BOUNDARY_VERTEX/2)
                if (pPermuteBlock != NULL)
                    sprintf( szNodeName, "%u", (*pPermuteBlock)[nodeName]);
                else
                    sprintf( szNodeName, "%u", nodeName);
            else
            {
                nodeName = BOUNDARY_VERTEX - nodeName + maxInteriorName;
                if ((pPermuteBlock != NULL) && (nodeName < permuteSize))
                    sprintf( szNodeName, "B(%u)", (*pPermuteBlock)[nodeName] );
                else
                    sprintf( szNodeName, "B(%u)", nodeName );
            }// else
            p.drawText( (*pV2)[X] + fntDist, (*pV2)[Y] + fntDist, szNodeName );
        }// if

        p.drawLine( (*pV0)[X], (*pV0)[Y], (*pV1)[X], (*pV1)[Y] );
        p.drawLine( (*pV1)[X], (*pV1)[Y], (*pV2)[X], (*pV2)[Y] );
        p.drawLine( (*pV2)[X], (*pV2)[Y], (*pV0)[X], (*pV0)[Y] );
    }// for
}

void
TNetView::draw3D ( T3DPrinter & p )
{
    if (p3DNet != NULL)
        drawNet( p );

    if (pMeshGen != NULL)
        drawMG( p );
}

void 
TNetView::drawNet ( T3DPrinter &p )
{
    uint                       i, j;
    TCell<3>                  *pCell;
    TListIterator<TCell<3> >   iterCells;
    TFrontier<3>              *pFrontier;
    TVertex<3>                *pVertex;
    uint                       noOfFrontiers, noOfVertices;
    TColorType                 oldColor = p.color();

#ifdef STRETCH_CELL
    TVector<3>  s, d, * tmpVec;
#endif
    
    if (p3DNet == NULL)
        return;

    // adjust color for net
    p.setRGBColor( COLOR_BLACK );

    iterCells.setList( p3DNet->getList() );
    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFrontiers = pCell->getNoOfFrontiers();

#ifdef STRETCH_CELL
        s  = 0;
        s += * pCell->getNode( 0 );
        s += * pCell->getNode( 1 );
        s += * pCell->getNode( 2 );
        s += * pCell->getNode( 3 );
        s *= 0.25;
#endif
        
        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = pCell->getFrontier( i );

            /*
              uint noOfEdges = ((T3DFrontier*)pFrontier)->getNoOfEdges();
              
              for ( j = 0; j < noOfEdges; j++ )
              {
              T3DCellEdge * pEdge = ((T3DFrontier*)pFrontier)->getEdge( j );
              
              if (pEdge->getBoundary() == INTERIOR)
              {
              p.setRGBColor( COLOR_RED );
              p.drawLine( pEdge->getNode( 0 ), pEdge->getNode( 2 ) );
              p.setRGBColor( COLOR_BLACK );
              }// if
              else
              p.drawLine( pEdge->getNode( 0 ), pEdge->getNode( 2 ) );
              }// for
              */
            
            // create new polyhedron
            p.beginPolygon();

            noOfVertices = pFrontier->getNoOfNodes();

            for (j = 0; j < noOfVertices; j++)
            {
                // add all vertices of frontier
                pVertex = pFrontier->getNode( pCell, j );
                if (pVertex != NULL)
                {
#ifdef STRETCH_CELL
                    tmpVec = new TVector<3>( *pVertex );

                    d  = 0;
                    d  = *tmpVec;
                    d -= s;
                    d *= 0.3;
                    d += s;
                    *tmpVec = d;

                    p.addNode( tmpVec );
#else
                    p.addNode( pVertex );
#endif
                }// if
            }// for

            // draw polyhedron
            p.endPolygon();
        }// for
    }// while

    p.setColor( oldColor );

    // finally draw coord system
    p.drawCoordSystem();
}

void
TNetView::drawMG ( T3DPrinter & p )
{
    TListIterator< T3DQuad >   iter;
    T3DQuad                  * pQuad;

    p.setRGBColor( COLOR_BLACK );

    iter.setList( pMeshGen->quads() );
    iter.first();
    while ((pQuad = iter++) != NULL)
    {
        p.drawLine( pQuad->vertex( 0 ), pQuad->vertex( 1 ) );
        p.drawLine( pQuad->vertex( 1 ), pQuad->vertex( 2 ) );
        p.drawLine( pQuad->vertex( 2 ), pQuad->vertex( 3 ) );
        p.drawLine( pQuad->vertex( 3 ), pQuad->vertex( 0 ) );
        
        p.drawLine( pQuad->vertex( 4 ), pQuad->vertex( 5 ) );
        p.drawLine( pQuad->vertex( 5 ), pQuad->vertex( 6 ) );
        p.drawLine( pQuad->vertex( 6 ), pQuad->vertex( 7 ) );
        p.drawLine( pQuad->vertex( 7 ), pQuad->vertex( 4 ) );
        
        p.drawLine( pQuad->vertex( 0 ), pQuad->vertex( 4 ) );
        p.drawLine( pQuad->vertex( 1 ), pQuad->vertex( 5 ) );
        p.drawLine( pQuad->vertex( 2 ), pQuad->vertex( 6 ) );
        p.drawLine( pQuad->vertex( 3 ), pQuad->vertex( 7 ) );
    }// while

    p.drawCoordSystem();
}
