//
//  Project :  libvisual
//  File    :  printer.cc
//  Purpose :  class for a ps printer driver
//  Author  :  Ronald Kriemann
//
// arch-tag: af272890-59da-4a6a-848d-60250fa328c7
//

extern "C" {
#include <math.h>
}

#include "printer.hh"

// size of normal postscript page
#define _SIZE_X_  595
#define _SIZE_Y_  842

// scaling for 1 mm
#define _MM_  2.83465 

//
// constructor
//
TPrinter::TPrinter ( const char filename[], uint paper )
{
	pFile = NULL;
	strcpy( szFilename, filename );

	switch (paper)
	{
        case A4:
        default:
            width  = 2100;
            heigth = 2970;
	}// switch

	bLeft = bRight = bTop = bBottom = 20;

	// set coord transformation
	view.setWorld( 0, 0, 1, 1 );
	if (_SIZE_X_ > _SIZE_Y_)
	{
		TFloat  diff = (bTop+bBottom)*_MM_;
		TFloat  min  = TFloat(_SIZE_X_ - _SIZE_Y_ - diff) / 2.0;
		view.setViewport( min+(bBottom*_MM_), bLeft*_MM_, 
						  _SIZE_Y_ - diff, _SIZE_Y_ - diff );
	}// if
	else
	{
		TFloat  diff = (bLeft+bRight)*_MM_;
		TFloat  min  = TFloat(_SIZE_Y_ - _SIZE_X_ - diff) / 2.0;
		view.setViewport( bBottom*_MM_, min+(bLeft*_MM_), 
						  _SIZE_X_ - diff, _SIZE_X_ - diff );
	}// if

    old_r  = old_g = old_b = -1;
    old_lw = -1;
    old_size = -1;
    strcpy( old_font, "" );
}

//
// destructor
//
TPrinter::~TPrinter ()
{
	if (pFile != NULL)
        fclose( pFile );
}

//
// return the viewport
//
TViewport & 
TPrinter::getView ()
{
	return view;
}

//
// begin (end) printing
//
void 
TPrinter::begin ()
{
	pFile = fopen( szFilename, "w" );
	if (pFile == NULL)
		cerr << "TPrinter::start () : can not open file (" << szFilename << ")" 
			 << endl;
	else
	{
		TFloat  minX, minY, maxX, maxY;
		char  szBB[64];

		view.getViewport( minX, minY, maxX, maxY );
		sprintf( szBB, "%%%%BoundingBox: %d %d %d %d \n", 
				 int(minX-bLeft), int(minY-bBottom), int(maxX+bRight), int(maxY+bTop) );

		// print preamble
		fputs( "%!PS-Adobe-2.0 EPSF-2.0 \n", pFile );
		fputs( "%%Title:  \n", pFile );
		fputs( "%%Creator: stokes \n", pFile );
        fputs( "%%Orientation: Portrait \n", pFile );
		fputs( szBB, pFile );
		fputs( "%%Pages: 0 \n", pFile );
        fputs( "%%BeginSetup \n", pFile );
        fputs( "%%EndSetup \n", pFile );
        fputs( "%%Magnification: 1.0000 \n", pFile );
		fputs( "%%EndComments \n", pFile );

        fputs( "/$F2psDict 200 dict def \n", pFile );
        fputs( "$F2psDict begin \n", pFile );
        fputs( "$F2psDict /mtrx matrix put \n", pFile );
        fputs( "end \n", pFile );
        fputs( "save \n", pFile );

		// print prolog
		fputs( "/D {bind def} bind def \n", pFile );
		fputs( "/SP {showpage} D \n", pFile );

		fputs( "/NP {newpath} D \n", pFile );
		fputs( "/MT {moveto} D \n", pFile );
		fputs( "/LT {lineto} D \n", pFile );
		fputs( "/DA {NP arc CP} D \n", pFile );
		fputs( "/CP {closepath} D \n", pFile );
		fputs( "/S {stroke} D \n", pFile );
		fputs( "/F {fill} D \n", pFile );
		fputs( "/DL {4 2 roll NP MT LT CP S} D \n", pFile );

		fputs( "/SW {setlinewidth} D \n", pFile );
		fputs( "/SRGB {setrgbcolor} D \n", pFile );

		fputs( "/FF {findfont} D \n", pFile );
		fputs( "/SF {setfont} D \n", pFile );
		fputs( "/SCF {scalefont} D \n", pFile );
		fputs( "/T {3 1 roll NP MT show} D \n", pFile );

        fputs( "/$F2psBegin {$F2psDict begin /$F2psEnteredState save def} def \n", pFile );
        fputs( "/$F2psEnd {$F2psEnteredState restore end} def \n", pFile );
		fputs( "%%EndProlog \n", pFile );
	}// else
}

void 
TPrinter::end ()
{
	if (pFile != NULL)
	{
		// show the printed page
		// fputs( "SP \n", pFile ); (not recommended in EPS)

        fputs( "$F2psBegin \n", pFile );
        fputs( "$F2psEnd \n", pFile );
        fputs( "restore \n", pFile );
        
		fclose( pFile );
	}
	pFile = NULL;
}

//
// misc drawing commands
//
void 
TPrinter::drawPoint( TFloat x, TFloat y )
{
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "%f %f %f %f DL \n", 
				 double( view.viewX(x)-0.5 ), double( view.viewY(y) ), 
				 double( view.viewX(x)+0.5 ), double( view.viewY(y) ) );
		fputs( string, pFile );
	}// if
}

void 
TPrinter::drawLine( TFloat x1, TFloat y1, TFloat x2, TFloat y2 )
{
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "%f %f %f %f DL \n", 
				 double( view.viewX(x1) ), double( view.viewY(y1) ), 
                 double( view.viewX(x2) ), double( view.viewY(y2) ) );
		fputs( string, pFile );
	}// if
}

void 
TPrinter::drawText( TFloat x, TFloat y, const char text[] )
{
	if (pFile != NULL)
	{
		char string[1024];
		sprintf( string, "%f %f (%s) T \n",
                 double( view.viewX(x) ), double( view.viewY(y) ), text );
		fputs( string, pFile );
	}// if
}

void 
TPrinter::drawArc( TFloat x, TFloat y, TFloat rad,
				   TFloat sang, TFloat eang )
{
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "%f %f %f %f %f DA S \n", 
				 double( view.viewX(x) ), double( view.viewY(y) ),
                 double( rad ), double( sang ), double( eang ) );
		fputs( string, pFile );
	}// if
}

void 
TPrinter::drawFilledArc( TFloat x, TFloat y, TFloat rad,
						 TFloat sang, TFloat eang )
{
	if (pFile != NULL)
	{
		char string[256];
		//		cout << "rad : " << view.getScaleX() * rad << endl;

		sprintf( string, "%f %f %f %f %f DA F \n", 
				 double( view.viewX(x) ), double( view.viewY(y) ),
                 double( view.getScaleX()*rad ), double( sang ), double( eang ) );
		fputs( string, pFile );
	}// if
}

void
TPrinter::drawArrow ( TFloat x1, TFloat y1,
                      TFloat x2, TFloat y2,
                      TFloat angle, TFloat lengthArrow )
{
    TFloat  u, v, su, sv, ru, rv;
    TFloat  c, s;

    // calc cos/sin of rotation angle
    c = cos( angle * M_PI / TFloat ( 180 ) );
    s = sin( angle * M_PI / TFloat ( 180 ) );
    
    // calc direction
    u = x2 - x1;
    v = y2 - y1;
        
    // the direction of the vector
    drawLine( x1, y1, x2, y2 );
		
    // show direction
    su = - (u * (1.0 - lengthArrow));
    sv = - (v * (1.0 - lengthArrow));
		
    ru = su * c - sv * s;
    rv = su * s + sv * c;
    drawLine( x1+u     , y1+v,
              x1+u + ru, y1+v + rv );
		
    ru =   su * c + sv * s;
    rv = - su * s + sv * c;
    drawLine( x1+u     , y1+v,
              x1+u + ru, y1+v + rv );
}

//
// misc design commands
//
void 
TPrinter::setLineWidth( TFloat width )
{
    if (width == old_lw)
        return;
    
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "%f SW \n", double( width ) );
		fputs( string, pFile );

        old_lw = width;
	}// if
}

void 
TPrinter::setRGBColor( TFloat r, TFloat g, TFloat b )
{
    if ((r == old_r) && (g == old_g) && (b == old_b))
        return;
    
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "%f %f %f SRGB \n", double( r ), double( g ), double( b ) );
		fputs( string, pFile );

        old_r = r;
        old_g = g;
        old_b = b;
	}// if
}

void 
TPrinter::setRGBColor( uint rgb )
{
	if (pFile != NULL)
	{
		TFloat r = TFloat((rgb & 0xFF0000) >> 16) / TFloat(0xFF),
            g = TFloat((rgb & 0x00FF00) >> 8) / TFloat(0xFF),
            b = TFloat( rgb & 0x0000FF) / TFloat(0xFF);

        if ((r == old_r) && (g == old_g) && (b == old_b))
            return;
    
		char string[256];
		sprintf( string, "%f %f %f SRGB \n", double( r ), double( g ), double( b ) );

		fputs( string, pFile );

        old_r = r;
        old_g = g;
        old_b = b;
	}// if
}

void
TPrinter::setHSVColor( TFloat h, TFloat s, TFloat v )
{
    int  loc_h, loc_s, loc_v;
    int  r = 0, g = 0, b = 0;

    if (h != -1)
        loc_h = (loc_h = (h > 360 ? 360 : int( h ))) < 0 ? 0 : loc_h;
    else
        loc_h = -1;
    
    loc_s = (loc_s = (s > 1 ? 255 : int( 255*s ))) < 0 ? 0 : loc_s;
    loc_v = (loc_v = (v > 1 ? 255 : int( 255*v ))) < 0 ? 0 : loc_v;

    if (( loc_s == 0 ) || ( loc_h == -1 ))
    {
        r = loc_v;
        g = loc_v;
        b = loc_v;
    }// if
    else
    {
        if ( uint( loc_h ) >= 360 )
            loc_h %= 360;
        
        uint f = loc_h % 60;
        loc_h /= 60;
        
        uint p = (uint)(2*loc_v*(255-loc_s)+255)/510;
        uint q, t;
        
        if ( loc_h & 1 )
        {
            q = (uint)(2*loc_v*(15300-loc_s*f)+15300)/30600;

            switch( loc_h )
            {
                case 1: r=(int)q;     g=(int)loc_v, b=(int)p; break;
                case 3: r=(int)p;     g=(int)q,     b=(int)loc_v; break;
                case 5: r=(int)loc_v; g=(int)p,     b=(int)q; break;
            }// switch
        }// if
        else
        {
            t = (uint)(2*loc_v*(15300-(loc_s*(60-f)))+15300)/30600;

            switch( loc_h )
            {
                case 0: r = (int)loc_v; g = (int)t,     b = (int)p;     break;
                case 2: r = (int)p;     g = (int)loc_v, b = (int)t;     break;
                case 4: r = (int)t;     g = (int)p,     b = (int)loc_v; break;
            }// switch
        }// else
    }// else

    setRGBColor( TFloat(r) / TFloat(255),
                 TFloat(g) / TFloat(255),
                 TFloat(b) / TFloat(255) );
}

void 
TPrinter::setFont( const char font[], int size )
{
    if ((size == old_size) && (strcmp( font, old_font ) == 0))
        return;
    
	if (pFile != NULL)
	{
		char string[256];
		sprintf( string, "/%s FF %d SCF SF \n", font, size );
		fputs( string, pFile );

        old_size = size;
        strcpy( old_font, font );
	}// if
}
